/* -----------------------------------------------------------------------
 *
 * Project:     UT-SNMP
 *
 * Filename:    rfc1157_snmp.h
 *
 * Date:        May 1994
 *
 * Purpose:     include file with constants for all the snmp modules.
 *
 * Author(s):   Vincent Berkhout, Frederic P. H. van Hengstum
 *
 * -----------------------------------------------------------------------
 */
#ifndef _rfc1157_snmp_h_
#define _rfc1157_snmp_h_
#include <snmpTypes.h>



#define COLDSTART 0
#define WARMSTART 1
#define LINKDOWN 2
#define LINKUP 3
#define AUTHENTICATIONFAILURE 4
#define EGPNEIGHBORLOSS 5
#define ENTERPRISESPECIFIC 6

typedef AsnInt Trap_PDUV1Int; /* INTEGER { COLDSTART(0), WARMSTART(1), LINKDOWN(2), LINKUP(3), AUTHENTICATIONFAILURE(4), EGPNEIGHBORLOSS(5), ENTERPRISESPECIFIC(6) }  */

#define BEncTrap_PDUV1IntContent BEncAsnIntContent

#define BDecTrap_PDUV1IntContent BDecAsnIntContent

#define PrintTrap_PDUV1Int PrintAsnInt

#define FreeTrap_PDUV1Int FreeAsnInt 





/*#define NOERROR 0*/
#define TOOBIG 1
#define NOSUCHNAME 2
#define BADVALUE 3
#define READONLY 4
#define GENERR 5

typedef AsnInt PDUV1Int; /* INTEGER { NOERROR(0), TOOBIG(1), NOSUCHNAME(2), BADVALUE(3), READONLY(4), GENERR(5) }  */

#define BEncPDUV1IntContent BEncAsnIntContent

#define BDecPDUV1IntContent BDecAsnIntContent

#define PrintPDUV1Int PrintAsnInt

#define FreePDUV1Int FreeAsnInt 





#define VERSION_1 0

typedef AsnInt MessageV1Int; /* INTEGER { VERSION_1(0) }  */

#define BEncMessageV1IntContent BEncAsnIntContent

#define BDecMessageV1IntContent BDecAsnIntContent

#define PrintMessageV1Int PrintAsnInt

#define FreeMessageV1Int FreeAsnInt 



typedef enum MessageV1ChoiceChoiceId
{
        MESSAGEV1CHOICE_GET_REQUEST = 0,
        MESSAGEV1CHOICE_GET_NEXT_REQUEST = 1,
        MESSAGEV1CHOICE_GET_RESPONSE = 2,
        MESSAGEV1CHOICE_SET_REQUEST = 3,
        MESSAGEV1CHOICE_TRAP = 4
} MessageV1ChoiceChoiceId;


typedef struct MessageV1Choice /* CHOICE */
{
    MessageV1ChoiceChoiceId  choiceId;
    union MessageV1ChoiceChoiceUnion
    {
    AsnOcts* get_request; /* [0] IMPLICIT OCTET STRING */
    AsnOcts* get_next_request; /* [1] IMPLICIT OCTET STRING */
    AsnOcts* get_response; /* [2] IMPLICIT OCTET STRING */
    AsnOcts* set_request; /* [3] IMPLICIT OCTET STRING */
    AsnOcts* trap; /* [4] IMPLICIT OCTET STRING */
    } a;
} MessageV1Choice;

AsnLen BEncMessageV1ChoiceContent PROTO((BUF_TYPE b, MessageV1Choice* v));

void BDecMessageV1ChoiceContent PROTO(( BUF_TYPE b, AsnTag tagId0, AsnLen elmtLen0, MessageV1Choice* v, AsnLen* bytesDecoded, ENV_TYPE env));


void PrintMessageV1Choice PROTO((FILE* f, MessageV1Choice* v, unsigned short int indent));
void FreeMessageV1Choice PROTO((MessageV1Choice* v));


typedef struct PDUV1 /* SEQUENCE */
{
    AsnInt request_id; /* INTEGER */
    PDUV1Int error_status; /* PDUV1Int */
    AsnInt error_index; /* INTEGER */
    VarBindList* variable_bindings; /* VarBindList */
} PDUV1;

AsnLen BEncPDUV1Content PROTO((BUF_TYPE b, PDUV1* v));

void BDecPDUV1Content PROTO(( BUF_TYPE b, AsnTag tagId0, AsnLen elmtLen0, PDUV1* v, AsnLen* bytesDecoded, ENV_TYPE env));


void PrintPDUV1 PROTO((FILE* f, PDUV1* v, unsigned short int indent));
void FreePDUV1 PROTO((PDUV1* v));





typedef struct Trap_PDUV1 /* [4] IMPLICIT SEQUENCE */
{
    AsnOid enterprise; /* OBJECT IDENTIFIER */
    IpAddress agent_addr; /* IpAddress */
    Trap_PDUV1Int generic_trap; /* Trap-PDUV1Int */
    AsnInt specific_trap; /* INTEGER */
    TimeTicks time_stamp; /* TimeTicks */
    VarBindList* variable_bindings; /* VarBindList */
} Trap_PDUV1;

AsnLen BEncTrap_PDUV1Content PROTO((BUF_TYPE b, Trap_PDUV1* v));

void BDecTrap_PDUV1Content PROTO(( BUF_TYPE b, AsnTag tagId0, AsnLen elmtLen0, Trap_PDUV1* v, AsnLen* bytesDecoded, ENV_TYPE env));


void PrintTrap_PDUV1 PROTO((FILE* f, Trap_PDUV1* v, unsigned short int indent));
void FreeTrap_PDUV1 PROTO((Trap_PDUV1* v));





typedef struct PDUV1 GetRequest_PDUV1; /* [0] IMPLICIT PDUV1 */

#define BEncGetRequest_PDUV1Content BEncPDUV1Content

#define BDecGetRequest_PDUV1Content BDecPDUV1Content

#define PrintGetRequest_PDUV1 PrintPDUV1

#define FreeGetRequest_PDUV1 FreePDUV1 




typedef struct PDUV1 GetNextRequest_PDUV1; /* [1] IMPLICIT PDUV1 */

#define BEncGetNextRequest_PDUV1Content BEncPDUV1Content

#define BDecGetNextRequest_PDUV1Content BDecPDUV1Content

#define PrintGetNextRequest_PDUV1 PrintPDUV1

#define FreeGetNextRequest_PDUV1 FreePDUV1 




typedef struct PDUV1 GetResponse_PDUV1; /* [2] IMPLICIT PDUV1 */

#define BEncGetResponse_PDUV1Content BEncPDUV1Content

#define BDecGetResponse_PDUV1Content BDecPDUV1Content

#define PrintGetResponse_PDUV1 PrintPDUV1

#define FreeGetResponse_PDUV1 FreePDUV1 




typedef struct PDUV1 SetRequest_PDUV1; /* [3] IMPLICIT PDUV1 */

#define BEncSetRequest_PDUV1Content BEncPDUV1Content

#define BDecSetRequest_PDUV1Content BDecPDUV1Content

#define PrintSetRequest_PDUV1 PrintPDUV1

#define FreeSetRequest_PDUV1 FreePDUV1 




typedef struct MessageV1 /* SEQUENCE */
{
    MessageV1Int version; /* MessageV1Int */
    AsnOcts community; /* OCTET STRING */
    struct MessageV1Choice* data; /* MessageV1Choice */
} MessageV1;

AsnLen BEncMessageV1 PROTO((BUF_TYPE b, MessageV1* v));

void BDecMessageV1 PROTO(( BUF_TYPE b, MessageV1* result, AsnLen* bytesDecoded, ENV_TYPE env));
AsnLen BEncMessageV1Content PROTO((BUF_TYPE b, MessageV1* v));

void BDecMessageV1Content PROTO(( BUF_TYPE b, AsnTag tagId0, AsnLen elmtLen0, MessageV1* v, AsnLen* bytesDecoded, ENV_TYPE env));


void PrintMessageV1 PROTO((FILE* f, MessageV1* v, unsigned short int indent));
void FreeMessageV1 PROTO((MessageV1* v));


typedef enum PDUsChoiceChoiceId
{
   PDUSV1_GET_REQUEST = 0,
   PDUSV1_GET_NEXT_REQUEST = 1,
   PDUSV1_GET_RESPONSE = 2,
   PDUSV1_SET_REQUEST = 3,
   PDUSV1_TRAP = 4
} PDUsChoiceChoiceId;

typedef struct PDUsV1 /* CHOICE */
{
    PDUsChoiceChoiceId choiceId;
    union PDUsV1ChoiceUnion
    {
    GetRequest_PDUV1* get_request; /* GetRequest-PDUV1 */
    GetNextRequest_PDUV1* get_next_request; /* GetNextRequest-PDUV1 */
    GetResponse_PDUV1* get_response; /* GetResponse-PDUV1 */
    SetRequest_PDUV1* set_request; /* SetRequest-PDUV1 */
    struct Trap_PDUV1* trap; /* Trap-PDUV1 */
    } a;
} PDUsV1;

AsnLen BEncPDUsV1 PROTO((BUF_TYPE b, PDUsV1* v));

void BDecPDUsV1 PROTO(( BUF_TYPE b, PDUsV1* result, AsnLen* bytesDecoded, ENV_TYPE env));
AsnLen BEncPDUsV1Content PROTO((BUF_TYPE b, PDUsV1* v));

void BDecPDUsV1Content PROTO(( BUF_TYPE b, AsnTag tagId0, AsnLen elmtLen0, PDUsV1* v, AsnLen* bytesDecoded, ENV_TYPE env));


void PrintPDUsV1 PROTO((FILE* f, PDUsV1* v, unsigned short int indent));
void FreePDUsV1 PROTO((PDUsV1* v));

#endif /* conditional include of rfc1157_snmp.h */
