/* -----------------------------------------------------------------------
 *
 * project:     UT-SNMP
 *
 * Filename:    TYPES.c
 *
 * Date:        June 1994
 *
 * Purpose:     Additional functions to fool around with variable bindings
 *				in a more convenient manner.
 *
 * Author(s):   Vincent Berkhout, Frederic P. H. van Hengstum, Rob Post
 *
 * -----------------------------------------------------------------------
 */

#include <snmpTypes.h>



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mOcNew
 *
 * Arguments:	l, length of string (IN).
 *
 * Return:	On success, reference to created variable, or NULL otherwise.
 *
 * Purpose:    	Allocate space for variable.
 *
 * Remarks:	The allocation of the string field is an desired 
 *		side-effect !
 *
 * -----------------------------------------------------------------------
 */
	AsnOcts *
mOcNew( unsigned int l )
{
	AsnOcts *p;
	p = (AsnOcts *) Asn1Alloc( sizeof( AsnOcts ) );
	p->octs = (unsigned char *) Asn1Alloc( p->octetLen = l );
	return( p );
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mOcPut
 *
 * Arguments:	p, reference to variable (IN).
 *		s, contence of string (IN).
 *		l, length of string (IN).
 *
 * Return:	Reference to variable.
 *
 * Purpose:    	Fill variable with string and length.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	AsnOcts *
mOcPut( AsnOcts *p, unsigned char *s, unsigned int l )
{
	memcpy( p->octs, s, p->octetLen = l );
	return( p );
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mOcGet
 *
 * Arguments:	p, reference to variable (IN).
 *		s, contence of string (OUT).
 *		l, length of string (OUT).
 *
 * Return:	Reference to variable.
 *
 * Purpose: 	Fill string and length with variable values.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	AsnOcts *
mOcGet( AsnOcts *p, unsigned char *s, unsigned int *l )
{
	memcpy( s, p->octs, *l = p->octetLen );
	return( p );
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mBiNew
 *
 * Arguments:	l, length of string (bits) (IN).
 *
 * Return:	On success, reference to created variable, or NULL otherwise.
 *
 * Purpose:    	Allocate space for variable.
 *
 * Remarks:	The allocation of the (bit)string filed is an desired 
 *		side-effect !
 *
 * -----------------------------------------------------------------------
 */
	AsnBits *
mBiNew( int l )
{
	AsnBits *p;
	p = (AsnBits *) Asn1Alloc( sizeof( AsnBits ) );
	p->bits = (char *) Asn1Alloc( p->bitLen = l );
	return( p );
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mBiPut
 *
 * Arguments:	p, reference to variable (IN).
 *		s, contence of string (bits) (IN).
 *		l, length of string (IN).
 *
 * Return:	Reference to variable.
 *
 * Purpose:    	Fill variable with string and length.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	AsnBits *
mBiPut( AsnBits *p, char *s, int l )
{
	memcpy( p->bits, s, p->bitLen = l );
	return( p );
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mBiGet
 *
 * Arguments:	p, reference to variable (IN).
 *		s, contence of string (OUT).
 *		l, length of string (OUT).
 *
 * Return:	Reference to variable. 
 *
 * Purpose:    	Fill variable with string (bits) and length.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	AsnBits *
mBiGet( AsnBits *p, char *s, int *l )
{
	memcpy( s, p->bits, *l = p->bitLen );
	return( p );
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mObNew
 *
 * Arguments:	none.	
 *
 * Return:	On success, reference to created variable, or NULL otherwise.
 *
 * Purpose:    	Allocate space for variable.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	ObjectSyntax *
mObNew( )
{
	return (ObjectSyntax *) Asn1Alloc( sizeof( ObjectSyntax ) );
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mObPut
 *
 * Arguments:	p, reference to variable (IN).
 *		i, variable type (IN).
 *		v, variable value (IN).
 *
 * Return:	Reference to variable.
 *
 * Purpose:    	Fill variable with type and value.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	ObjectSyntax *
mObPut( ObjectSyntax *p, ObjectSyntaxChoiceId i, AsnInt v )
{
	p->choiceId = i;
	p->a.integer_value = v;
	return p;
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mObGet
 *
 * Arguments:	p, reference to variable (IN).
 *		i, variable type (IN).
 *
 * Return:	Reference to variable value.
 *
 * Purpose:    	Fill type and value with varibale values.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	AsnInt 
mObGet( ObjectSyntax *p, ObjectSyntaxChoiceId *i) 
{
	*i = p->choiceId;
	return p->a.integer_value;
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mVcPut
 *
 * Arguments:	p, reference to variable (IN).
 *		i, variable type (IN).
 *		v, variable value (IN).
 *
 * Return:	Reference to variable.
 *
 * Purpose:    	Fill variable with type and value.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	VarBindChoice *
mVcPut( VarBindChoice *p, VarBindChoiceChoiceId i, ObjectSyntax *v )
{
	p->choiceId = i;
	p->a.value = v;
	return p;
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mVcGet
 *
 * Arguments:	p, reference to variable (IN).
 *		i, varible type (OUT)
 *
 * Return:	Reference to variable value.
 *
 * Purpose:    	Fill type with variable type.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	ObjectSyntax *
mVcGet( VarBindChoice *p, VarBindChoiceChoiceId *i ) 
{
	*i = p->choiceId;
	return p->a.value;
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mVcNew
 *
 * Arguments:	l, length of string (IN)
 *
 * Return:	On success, reference to created variable, or NULL otherwise.
 *
 * Purpose:    	Allocate space for variable.
 *
 * Remarks:	The allocation of the string filed is an desired side-
 *		effect !
 *
 * -----------------------------------------------------------------------
 */
	VarBind *
mVbNew( unsigned int l )
{
	VarBind *p;
	p = (VarBind *) Asn1Alloc( sizeof( VarBind ) );
	p->name.octs = (unsigned char *) Asn1Alloc( p->name.octetLen=l );
	return p;
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mVbPut
 *
 * Arguments:	p, reference to variable (IN).
 *		n, name of variable (IN).
 *		v, value of variable (IN).
 *
 * Return:	Reference to variable.
 *
 * Purpose:   	Fill variable with name and value.	 
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	VarBind *
mVbPut( VarBind *p, ObjectName *n, VarBindChoice *v )
{
	memcpy( p->name.octs, n->octs, p->name.octetLen=n->octetLen );
	p->varBindChoice = v;
	return p;
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mVbGet
 *
 * Arguments:	p, reference to variable (IN).
 *		n, name of variable (OUT).
 *
 * Return:	Reference to variable value.
 *
 * Purpose:    	Fill name with variable name.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	VarBindChoice *
mVbGet( VarBind *p, ObjectName *n )
{
	memcpy( n->octs, p->name.octs, n->octetLen=p->name.octetLen );
	return p->varBindChoice;
}

/* -----------------------------------------------------------------------
 */
/* -----------------------------------------------------------------------
 *
 * Project:	UT-SNMP
 *
 * Filename:    spmMisc.c
 *
 * Date:        April 1994.
 *
 * Purpose:     General SPM functions.
 *
 * Author(s):   J.W.Togtema, Frederic P. H. van Hengstum
 *
 * -----------------------------------------------------------------------
 */

#include <snmpTypes.h>
#include <rfc1157_snmp.h>
#include <spmMD5.h>

AsnInt max_bindings = 2147483647;

AsnOid iso			= { 1, "\1" };
AsnOid org			= { 1, "\53" };
AsnOid dod 			= { 2, "\53\6" };
AsnOid internet 		= { 3, "\53\6\1" };
AsnOid directory 		= { 4, "\53\6\1\1" };
AsnOid mgmt 			= { 4, "\53\6\1\2" };
AsnOid mib2 			= { 5, "\53\6\1\2\1" };
AsnOid asnsystem 		= { 6, "\53\6\1\2\1\1" };
AsnOid sysDescr 		= { 7, "\53\6\1\2\1\1\1" };
AsnOid sysObjectID 		= { 7, "\53\6\1\2\1\1\2" };
AsnOid sysUpTime 		= { 7, "\53\6\1\2\1\1\3" };
AsnOid sysContact 		= { 7, "\53\6\1\2\1\1\4" };
AsnOid sysName 			= { 7, "\53\6\1\2\1\1\5" };
AsnOid sysLocation 		= { 7, "\53\6\1\2\1\1\6" };
AsnOid sysServices	 	= { 7, "\53\6\1\2\1\1\7" };
AsnOid interfaces 		= { 6, "\53\6\1\2\1\2" };
AsnOid at 			= { 6, "\53\6\1\2\1\3" };
AsnOid ip 			= { 6, "\53\6\1\2\1\4" };
AsnOid icmp 			= { 6, "\53\6\1\2\1\5" };
AsnOid tcp 			= { 6, "\53\6\1\2\1\6" };
AsnOid udp 			= { 6, "\53\6\1\2\1\7" };
AsnOid egp 			= { 6, "\53\6\1\2\1\10" };
AsnOid snmp 			= { 6, "\53\6\1\2\1\13" };
AsnOid experimental 		= { 4, "\53\6\1\3" };
AsnOid private 			= { 4, "\53\6\1\4" };
AsnOid enterprises 		= { 5, "\53\6\1\4\1" };
AsnOid utV1domain 		= { 6, "\53\6\1\4\1\7" }; 
AsnOid snmpV2 			= { 4, "\53\6\1\6" };
AsnOid snmpDomains 		= { 5, "\53\6\1\6\1" };
AsnOid snmpUDPDomain 		= { 6, "\53\6\1\6\1\1" }; 
AsnOid snmpCLNSDomain 		= { 6, "\53\6\1\6\1\2" }; 
AsnOid snmpCONSDomain 		= { 6, "\53\6\1\6\1\3" }; 
AsnOid snmpDDPDomain		= { 6, "\53\6\1\6\1\4" }; 
AsnOid snmpIPXDomain		= { 6, "\53\6\1\6\1\5" }; 
AsnOid snmpTCPDomain		= { 6, "\53\6\1\6\1\6" }; /*!! attention: no mib2 */
AsnOid snmpProxys		= { 5, "\53\6\1\6\2" };
AsnOid rfc1157Proxy		= { 6, "\53\6\1\6\2\1" };
AsnOid rfc1157Domain		= { 7, "\53\6\1\6\2\1\1" };
AsnOid rfc1157noAuth		= { 7, "\53\6\1\6\2\1\2" };
AsnOid snmpModules		= { 5, "\53\6\1\6\3" };
AsnOid snmpMIB			= { 6, "\53\6\1\6\3\1" }; 
AsnOid snmpMIBObjects		= { 7, "\53\6\1\6\3\1\1" }; 
AsnOid snmpStats		= { 8, "\53\6\1\6\3\1\1\1" }; 
AsnOid snmpV1			= { 8, "\53\6\1\6\3\1\1\2" }; 
AsnOid snmpOR			= { 8, "\53\6\1\6\3\1\1\3" }; 
AsnOid snmpTrap			= { 8, "\53\6\1\6\3\1\1\4" }; 
AsnOid snmpTrapOID		= { 9, "\53\6\1\6\3\1\1\4\1" }; 
AsnOid snmpTrapTable		= { 9, "\53\6\1\6\3\1\1\4\2" }; 
AsnOid snmpTrapEntry		= { 10, "\53\6\1\6\3\1\1\4\2\1" }; 
AsnOid snmpTrapNumbers		= { 11, "\53\6\1\6\3\1\1\4\2\1\1" }; 
AsnOid snmpTrapEnterprise	= { 9, "\53\6\1\6\3\1\1\4\3" }; 
AsnOid snmpTraps		= { 8, "\53\6\1\6\3\1\1\5" }; 
AsnOid coldStart		= { 9, "\53\6\1\6\3\1\1\5\1" }; 
AsnOid warmStart		= { 9, "\53\6\1\6\3\1\1\5\2" }; 
AsnOid linkDown			= { 9, "\53\6\1\6\3\1\1\5\3" }; 
AsnOid linkUp			= { 9, "\53\6\1\6\3\1\1\5\4" }; 
AsnOid authenticationFailure    = { 9, "\53\6\1\6\3\1\1\5\5" }; 
AsnOid egpNeighborLoss  	= { 9, "\53\6\1\6\3\1\1\5\6" }; 
AsnOid snmpSet			= { 8, "\53\6\1\6\3\1\1\6" };
AsnOid snmpM2M 							 = {  6, "\53\6\1\6\3\2" };
AsnOid snmpM2MObjects					 = {  7, "\53\6\1\6\3\2\1" };
AsnOid snmpAlarm						 = {  8, "\53\6\1\6\3\2\1\1" };
AsnOid snmpAlarmNextIndex				 = {  9, "\53\6\1\6\3\2\1\1\1" };
AsnOid snmpAlarmTable					 = {  9, "\53\6\1\6\3\2\1\1\2" };
AsnOid snmpAlarmEntry					 = { 10, "\53\6\1\6\3\2\1\1\2\1" };
AsnOid snmpAlarmIndex					 = { 11, "\53\6\1\6\3\2\1\1\2\1\1" };
AsnOid snmpAlarmVariable				 = { 11, "\53\6\1\6\3\2\1\1\2\1\2" };
AsnOid snmpAlarmInterval				 = { 11, "\53\6\1\6\3\2\1\1\2\1\3" };
AsnOid snmpAlarmSampleType				 = { 11, "\53\6\1\6\3\2\1\1\2\1\4" };
AsnOid snmpAlarmValue					 = { 11, "\53\6\1\6\3\2\1\1\2\1\5" };
AsnOid snmpAlarmStartupAlarm			 = { 11, "\53\6\1\6\3\2\1\1\2\1\6" };
AsnOid snmpAlarmRisingThreshold			 = { 11, "\53\6\1\6\3\2\1\1\2\1\7" };
AsnOid snmpAlarmFallingThreshold		 = { 11, "\53\6\1\6\3\2\1\1\2\1\10" };
AsnOid snmpAlarmRisingEventIndex		 = { 11, "\53\6\1\6\3\2\1\1\2\1\11" };
AsnOid snmpAlarmFallingEventIndex		 = { 11, "\53\6\1\6\3\2\1\1\2\1\12" };
AsnOid snmpAlarmUnavailableEventIndex	 = { 11, "\53\6\1\6\3\2\1\1\2\1\13" };
AsnOid snmpAlarmStatus					 = { 11, "\53\6\1\6\3\2\1\1\2\1\14" };
AsnOid snmpAlarmNotifications			 = {  9, "\53\6\1\6\3\2\1\1\3" };
AsnOid snmpRisingAlarm					 = { 10, "\53\6\1\6\3\2\1\1\3\1" };
AsnOid snmpFallingAlarm					 = { 10, "\53\6\1\6\3\2\1\1\3\2" };
AsnOid snmpObjectUnavailableAlarm		 = { 10, "\53\6\1\6\3\2\1\1\3\3" };
AsnOid snmpEvent						 = {  8, "\53\6\1\6\3\2\1\2" };
AsnOid snmpEventNextIndex				 = {  9, "\53\6\1\6\3\2\1\2\1" };
AsnOid snmpEventTable					 = {  9, "\53\6\1\6\3\2\1\2\2" };
AsnOid snmpEventEntry					 = { 10, "\53\6\1\6\3\2\1\2\2\1" };
AsnOid snmpEventIndex					 = { 11, "\53\6\1\6\3\2\1\2\2\1\1" };
AsnOid snmpEventID						 = { 11, "\53\6\1\6\3\2\1\2\2\1\2" };
AsnOid snmpEventDescription				 = { 11, "\53\6\1\6\3\2\1\2\2\1\3" };
AsnOid snmpEventEvents					 = { 11, "\53\6\1\6\3\2\1\2\2\1\4" };
AsnOid snmpEventLastTimeSent			 = { 11, "\53\6\1\6\3\2\1\2\2\1\5" };
AsnOid snmpEventStatus					 = { 11, "\53\6\1\6\3\2\1\2\2\1\6" };
AsnOid snmpEventNotifyMinInterval		 = {  9, "\53\6\1\6\3\2\1\2\3" };
AsnOid snmpEventNotifyMaxRetransmissions = {  9, "\53\6\1\6\3\2\1\2\4" };
AsnOid snmpEventNotifyTable				 = {  9, "\53\6\1\6\3\2\1\2\5" };
AsnOid snmpEventNotifyEntry				 = { 10, "\53\6\1\6\3\2\1\2\5\1" };
AsnOid snmpEventNotifyIntervalRequested  = { 11, "\53\6\1\6\3\2\1\2\5\1\1" };
AsnOid snmpEventNotifyRetransmissionsRequested\
										 = { 11, "\53\6\1\6\3\2\1\2\5\1\2" };
AsnOid snmpEventNotifyLifetime			 = { 10, "\53\6\1\6\3\2\1\2\5\1\3" };
AsnOid snmpEventNotifyStatus			 = { 10, "\53\6\1\6\3\2\1\2\5\1\4" };
AsnOid partyMIB			= { 6, "\53\6\1\6\3\3" }; 
AsnOid partyAdmin		= { 7, "\53\6\1\6\3\3\1" }; 
AsnOid partyProtocols		= { 8, "\53\6\1\6\3\3\1\1" }; 
AsnOid noAuth			= { 9, "\53\6\1\6\3\3\1\1\1" }; 
AsnOid noPriv			= { 9, "\53\6\1\6\3\3\1\1\2" }; 
AsnOid desPrivProtocol		= { 9, "\53\6\1\6\3\3\1\1\3" }; 
AsnOid v2md5AuthProtocol	= { 9, "\53\6\1\6\3\3\1\1\4" }; 
AsnOid currentTime		= { 9, "\53\6\1\6\3\3\1\2\1" }; 
AsnOid restartTime		= { 9, "\53\6\1\6\3\3\1\2\2" }; 
AsnOid cacheTime		= { 9, "\53\6\1\6\3\3\1\2\3" }; 
AsnOid initialPartyId		= { 8, "\53\6\1\6\3\3\1\3" }; 
AsnOid initialContextId		= { 8, "\53\6\1\6\3\3\1\4" };
AsnOid partyMIBObjects		= { 7, "\53\6\1\6\3\3\2" }; 
AsnOid snmpParties		= { 8, "\53\6\1\6\3\3\2\1" }; 
AsnOid partyTable		= { 9, "\53\6\1\6\3\3\2\1\1" }; 
AsnOid snmpContexts		= { 8, "\53\6\1\6\3\3\2\2" }; 
AsnOid contextTable		= { 9, "\53\6\1\6\3\3\2\2\1" }; 
AsnOid snmpAccess		= { 8, "\53\6\1\6\3\3\2\3" }; 
AsnOid aclTable			= { 9, "\53\6\1\6\3\3\2\3\1" }; 
AsnOid snmpViews		= { 8, "\53\6\1\6\3\3\2\4" }; 
AsnOid viewTable		= { 9, "\53\6\1\6\3\3\2\4\1" }; 

/* 
 * -----------------------------------------------------------------------
 *
 * Function:    DomainConvert
 *
 * Arguments:   o, domain type identifier (IN).
 *		domain, domain type identifier (OUT).
 *
 * Return:	STATUS, TRUE on success, and FALSE otherwise.
 *
 * Purpose:     This function ...
 *
 * Remarks:	-'\001' substracted for offset in enumerated type (TDOMAIN)
 *
 * -----------------------------------------------------------------------
 */
	STATUS
DomainConvert( AsnOid *o, AsnInt *domain )
{
	*domain = (int) ((o->octs[o->octetLen-1])-'\001') ;
	return OK ;
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:    ASCIItoOID
 *
 * Arguments:   oid, converted object identifier (IN/OUT).
 *
 * Return:	STATUS, TRUE on success, and FALSE otherwise.
 *
 * Purpose:     This function converts an oid from UT to SNACC format.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	 STATUS
OIDtoAsnOid( AsnOid *oid )
{
	register short i = 0;
	AsnOid *d;

	if (oid->octetLen == 0)
		return(OK);
	d = pNewOid(2*oid->octetLen) ;
	if( !((oid->octs[0] == 0) || (oid->octs[0] == 1) ||
	     (oid->octs[0] == 2)) )
		return( OK );
	if( ((oid->octs[0] == 0) || (oid->octs[0] == 1)) &&
	    (oid->octs[1] >= 40) )
		return( OK );

	d->octetLen=0;
	d->octs[d->octetLen++] = (oid->octs[0] * 40);
	if (oid->octetLen > 1)
		d->octs[0] += oid->octs[1];
	for( i = 2; i < oid->octetLen; i++ ) {
		if( ( oid->octs[i] >> 7 ) != 0 ) {
			d->octs[d->octetLen++] = 0x81;
			d->octs[d->octetLen++] = oid->octs[i] & 0x7F;
		} else
			d->octs[d->octetLen++] = oid->octs[i];
	}
	
	oid->octetLen = d->octetLen;
	memcpy( oid->octs, d->octs, d->octetLen);

	pFreeOid( d );
	return OK;
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:    ASCIItoOID
 *
 * Arguments:   oid, object identifier (IN).
 *		str, converted object identifier (OUT).
 *
 * Return:	STATUS, TRUE on success, and FALSE otherwise.
 *
 * Purpose:     This function converts object identifier from internal -
 *		to external format.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS
ASCIItoOID( char *str, AsnOid *oid )
{
	char *tmpstr;
	int index, len, i;
	char string[BUFSIZ];

	len = strlen( str );		
	if( len <= 0 )
		return( FAILED );
	strcpy( string, str );
	string[ len ] = '\0'; 
	for( i = 0; i < len; i++)
		if (!(isdigit(string[i]) || (string[i] == '.')))
			return( FAILED);
	index = 0;
	tmpstr = (char *) strtok( (char *) string, (char *) ". " );
   	while( tmpstr != NULL )
   	{
     		oid->octs[index] = atoi( tmpstr );
     		index++;
     		tmpstr = (char *) strtok( (char *) NULL,  (char *) ". " );
   	}
  	oid->octetLen = index;
	return( OIDtoAsnOid( oid ) );
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:    OIDtoASCII
 *
 * Arguments:   oid, object identifier (IN).
 *		str, converted object identifier (OUT).
 *
 * Return:	STATUS, TRUE on success, and FALSE otherwise.
 *
 * Purpose:     This function converts object identifiers from internal -
 *		to external format.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS
OIDtoASCII( AsnOid *eoid, char *str )
{
    	int i;
    	int arcNum;
    	int firstArcNum;
    	int secondArcNum;
   	unsigned char ts[BUFSIZ];

	*str = '\0';

    	if (eoid == NULL)
	return;

	for (arcNum = 0, i=0; (i < eoid->octetLen) && (eoid->octs[i] & 0x80);i++)
		arcNum = (arcNum << 7) + (eoid->octs[i] & 0x7f);

	arcNum = (arcNum << 7) + (eoid->octs[i] & 0x7f);
	i++;

	firstArcNum = arcNum / 40;
	if (firstArcNum > 2 )
	   	firstArcNum = 2;

	secondArcNum = arcNum - (firstArcNum * 40);
	sprintf( ts, "%d.", firstArcNum);
	strcat( str, ts );
	sprintf( ts, "%d", secondArcNum);
	strcat( str, ts );
	for ( ; i < eoid->octetLen; )
	{
		for (arcNum = 0 ; (i < eoid->octetLen) && (eoid->octs[i] & 0x80);i++)
			arcNum = (arcNum << 7) + (eoid->octs[i] & 0x7f);
		arcNum = (arcNum << 7) + (eoid->octs[i] & 0x7f);
		i++;

		sprintf( ts, ".%d", arcNum);
		strcat( str, ts );
	}
	return OK ;
}

/*
 * -----------------------------------------------------------------------
 *
 * Function:    CMPOID
 *
 * Arguments:   AsnOid  *o1     (in)
 *              AsnOid  *o2     (in)
 *
 * Return:	COMPARETYPE
 *
 * Purpose:     Compare two object identifiers lexicographically.
 *              LESS     <->     o1 < o2
 *              EQUAL    <->     o1 = o2
 *              GREATER  <->     o1 > o2
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
        COMPARETYPE
CMPOID( AsnOid *o1, AsnOid *o2)
{
        int     diff, res;

        diff = o1->octetLen - o2->octetLen;
        if (o1->octetLen < o2->octetLen)
                res = memcmp( (char *)o1->octs, (char *)o2->octs, o1->octetLen);
        else
                res = memcmp( (char *)o1->octs, (char *)o2->octs, o2->octetLen);
        if ((diff != 0) && (res == 0))
                res = diff;
	if (res < 0)
		return(LESS);
	else if (res > 0)
		return(GREATER);
	else	return(EQUAL);
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:    CmpOID
 *
 * Arguments:   o1, object identifier 1 (IN).
 *		od2, object identifier 2 (IN).
 *
 * Return:	STATUS, TRUE on success, and FALSE otherwise.
 *
 * purpose:     This function compare both identifiers.
 *
 * Remarks: 	This routine _just_ compares, and _not_ orders, both oid's.
 *
 * -----------------------------------------------------------------------
 */
	STATUS
CmpOID( AsnOid *o1, AsnOid *o2 )
{
	return( AsnOctsEquiv( o1, o2) );
}


/*  
 * -----------------------------------------------------------------------
 *
 * Function:    printArgumentList 
 *
 * Arguments:	variable argument list
 *		( format, value, value, value, etc ) 
 *
 * Return:	void
 *
 * Purpose:     print a variable argument list for a function/procedure
 *
 * Remarks:	could be optimized ??? why ???
 *
 * -----------------------------------------------------------------------
 */
	void 
printArgumentList( va_alist ) va_dcl
{
	va_list args ;
	char	*format ;

	va_start(args);
	format = va_arg(args, char *) ;
	(void) vfprintf(stderr, format, args);
	(void) fprintf( stderr, "\n" ) ; 
	va_end(args) ;
}

/*  
 * --------------------------------------------------------------------
 *
 * Function:    NewOct 		(volatile memory)
 *
 * Arguments:	size, size of octet string (IN).
 *
 * Return:	AsnOcts * 
 *
 * Purpose:     This function allocates space for variable of AsnOcts type.
 *
 * Remarks:	USES NIBBLEMEM MEMORY !
 * 		The octetLen field is filled with zero.
 *
 * --------------------------------------------------------------------
 */
	AsnOcts*
NewOct(UAsnInt size)
{
        char    *octets;
        AsnOcts *octetstr;

        if ((octetstr = (AsnOcts *)Asn1Alloc(sizeof(AsnOcts))) == NULL)
                return(NULL);
        if ((octets = (char *)Asn1Alloc(size)) == NULL)
                return(NULL);
        octetstr->octs = octets;
        return(octetstr);
}


/* ^L
 * --------------------------------------------------------------------
 *
 * Function:  pNewOct		(permanent memory) 
 *
 * Arguments:	size, size of octet string (IN).
 *
 * Return:	AsnOcts * 
 *
 * Purpose:     This function mallocs space for variable of AsnOcts type.
 *
 * Remarks:	USES SYSTEM MEMORY !
 * 		The octetLen field is filled with zero.
 *
 * --------------------------------------------------------------------
 */
       AsnOcts*
pNewOct( UAsnInt size )
{ 
        char    *octets;
        AsnOcts *octetstr;

        if ((octetstr = (AsnOcts *)malloc(sizeof(AsnOcts))) == NULL)
                return(NULL);
        if ((octets = (char *)malloc(size)) == NULL)
                return(NULL);
		memset( octets, 0x00, size ) ;
        octetstr->octs = octets;
        return(octetstr);
}


/* 
 * --------------------------------------------------------------------
 *
 * Function: 	FreeOct		(volatile memory) 
 *
 * Arguments	octs, reference to superfluous space.(IN)
 *
 * Returns:	none
 *
 * Prupose:	Does not free the space (not yet provided by SNACC).
 *
 * Remarks:	Only the pointer (octs) will be set to NULL.
 * 		The octs will be freed during a shutdownNibbleMem
 *
 * --------------------------------------------------------------------
 */
	void
FreeOct(AsnOcts *octs)
{
	octs = (AsnOcts *) NULL;
}

/* 
 * --------------------------------------------------------------------
 *
 * Function: 	pFreeOct	(permanent memory)
 *
 * Arguments:	octs, reference to superfluous space.(IN)
 *
 * Returns:	none
 *
 * Prupose:	This uses permanent system memory
 *
 * Remarks:	The pointer (octs) will be set to NULL.
 *
 * --------------------------------------------------------------------
 */
	void
pFreeOct(AsnOcts *octs)
{
	(void) free(octs->octs);
	(void) free(octs);
	octs = (AsnOcts *) NULL;
}



/* 
 * --------------------------------------------------------------------
 *
 * Function:	pNewVbl
 *
 * Arguments:	
 *
 * Return:	VarBindList*
 *
 * Purpose:	Allocates a pointer to VarBindList. The fields
 *		first, curr and last are NULL, count is 0
 *		and datasize is (sizeof(char *)
 *
 * Remarks: 	To fill the fields of VarBindList used the 
 *		function Vb or Vbp.
 * !!!!!!!!!!!!! Function is obsolete: 
 * 		NewVbl() is now defined as AsnListNew !
 * 
 * ---------------------------------------------------------------------
	VarBindList*
NewVbl( )
{
	VarBindList 	*vbl;

	vbl = (VarBindList *) AsnListNew(sizeof(char *));
	if (vbl == NULL)
		return( (VarBindList *) NULL);
	return(vbl);
}
*/

/* 
 * --------------------------------------------------------------------
 *
 * Function:	pFreeVbl
 *
 * Arguments:	vbl, the pointer to the varbindlist (IN)  
 *
 * Returns:	none
 *
 * Purpose:	frees the VarBindlist pointer and all its fields inside. 
 *
 * Remarks:	The vbl pointer will be set to NULL.	
 *
 * --------------------------------------------------------------------
 */
	void
pFreeVbl(VarBindList *vbl)
{
  FreeVarBindList(vbl); /* uses the real free */
  vbl = (VarBindList *) NULL;
}


/* 
 * --------------------------------------------------------------------
 *
 * Function:	NewVb    
 *
 * Arguments:   vbl, the pointer to the varbindlist(IN), 
 *		objectNameSize, size  of the objectName inside the varbind(IN)
 *		varbindChoice, choice of varbind (5 different choices)(IN)
 *		syntaxChoice, choice of object (12 different choices) (IN)
 *		valueSize, size of value (only for AsnOcts & AsnBits) (IN)
 *
 * Returns:    	VarBind* 
 *
 * Purpose:     Appends a variable binding to vbl and allocates
 *		space for this structure. 
 *
 * Remarks:	There is no free routine for a variable binding
 *		The freeing is done with the function _Vbl.
 *		The octetLen (or bitLen) fields are always set
 *		to zero.
 *		All of the choiceId fields are filled with the input
 *		values. 
 *
 * --------------------------------------------------------------------
 */
	VarBind *
NewVb(VarBindList *vbl, UAsnInt objectNameSize,
   VarBindChoiceChoiceId varbindChoice, 
   ObjectSyntaxChoiceId syntaxChoice,
   UAsnInt valueSize)
{
  VarBind	**vb;

  vb = (VarBind**) AsnListAppend(vbl);
  (*vb) = (VarBind *) Asn1Alloc (sizeof(VarBind));
  if ((*vb) == NULL)
	return( (VarBind *) NULL);
  (*vb)->name.octs = (unsigned char *) Asn1Alloc ( objectNameSize );
  if ((*vb)->name.octs == NULL)
	return( (VarBind *) NULL);
  (*vb)->name.octetLen = 0;

  (*vb)->varBindChoice = (VarBindChoice *) Asn1Alloc ( sizeof (VarBindChoice));
  if ((*vb)->varBindChoice == NULL)
	return( (VarBind *) NULL);
  (*vb)->varBindChoice->choiceId = varbindChoice;

  switch(varbindChoice)
  {
    case(VARBINDCHOICE_UNSPECIFIED):
  (*vb)->varBindChoice->a.unSpecified = (AsnNull) NULL;
    break;
    case(VARBINDCHOICE_NOSUCHOBJECT):
  (*vb)->varBindChoice->a.noSuchObject = (AsnNull) NULL;
    break;
    case(VARBINDCHOICE_NOSUCHINSTANCE):
  (*vb)->varBindChoice->a.noSuchInstance = (AsnNull) NULL;
    break;
    case(VARBINDCHOICE_ENDOFMIBVIEW):
  (*vb)->varBindChoice->a.endOfMibView = (AsnNull) NULL;
    break;
    case(VARBINDCHOICE_VALUE):
  (*vb)->varBindChoice->a.value = (ObjectSyntax *) Asn1Alloc(sizeof(ObjectSyntax));
  if ((*vb)->varBindChoice->a.value == NULL)
	return( (VarBind *) NULL);
  (*vb)->varBindChoice->a.value->choiceId = syntaxChoice;
  switch(syntaxChoice)
  {
    case(OBJECTSYNTAX_STRING_VALUE):
    case(OBJECTSYNTAX_OBJECTID_VALUE):
    case(OBJECTSYNTAX_IPADDRESS_VALUE):
    case(OBJECTSYNTAX_ARBITRARY_VALUE):
    case(OBJECTSYNTAX_NSAPADDRESS_VALUE):
  (*vb)->varBindChoice->a.value->a.string_value = 
			(AsnOcts*) Asn1Alloc(sizeof(AsnOcts));
  if ((*vb)->varBindChoice->a.value->a.string_value == NULL)
	return( (VarBind *) NULL);
  (*vb)->varBindChoice->a.value->a.string_value->octs = 
			(unsigned char *) Asn1Alloc( valueSize);
  if ((*vb)->varBindChoice->a.value->a.string_value->octs == NULL)
	return( (VarBind *) NULL);
  (*vb)->varBindChoice->a.value->a.string_value->octetLen = 0;
    break;
    case(OBJECTSYNTAX_BIT_VALUE):
  (*vb)->varBindChoice->a.value->a.bit_value = 
			(AsnBits*) Asn1Alloc(sizeof(AsnBits));
  if ((*vb)->varBindChoice->a.value->a.bit_value == NULL)
	return( (VarBind *) NULL);
  (*vb)->varBindChoice->a.value->a.bit_value->bits = 
			(unsigned char *) Asn1Alloc( valueSize);
  if ((*vb)->varBindChoice->a.value->a.bit_value->bits == NULL)
	return( (VarBind *) NULL);
  (*vb)->varBindChoice->a.value->a.bit_value->bitLen = 0;
    break;
    default:
    break;
   }
   }

  return( (*vb) );
}


/* 
 * -------------------------------------------------------------
 *
 * Function:	NewVbp
 *
 * Arguments:   vbl, the pointer to the varbindlist(IN), 
 *		objectNameSize, size  of the objectName inside the varbind(IN)
 *		varbindChoice, choice of varbind (5 different choices)(IN)
 *		syntaxChoice, choice of object (12 different choices) (IN)
 *		valueSize, size of value (only for AsnOcts & AsnBits) (IN)
 *
 * Returns:    	VarBind* 
 *
 * Purpose:     Prepends a variable binding to vbl and allocates
 *		space for this structure. 
 *
 * Remarks:  	The octetLen (or bitLen) fields are always set
 *		to zero.
 *		All of the choiceId fields are filled with the input
 *		values. 
 *
 * ------------------------------------------------------------
 */
	VarBind*
NewVbp(VarBindList *vbl, UAsnInt objectNameSize,
   VarBindChoiceChoiceId varbindChoice, 
   ObjectSyntaxChoiceId syntaxChoice,
   UAsnInt valueSize)

{
  VarBind	**vb;

  vb = (VarBind **) varBindListPrepend(vbl);
  (*vb) = (VarBind *) malloc(sizeof(VarBind));
  if ((*vb) == NULL)
	return( (VarBind *) NULL);
  (*vb)->name.octs = (unsigned char *) malloc( objectNameSize );
  if ((*vb)->name.octs == NULL)
	return( (VarBind *) NULL);
  (*vb)->name.octetLen = 0;

  (*vb)->varBindChoice = (VarBindChoice *) malloc( sizeof (VarBindChoice));
  if ((*vb)->varBindChoice == NULL)
	return( (VarBind *) NULL);
  (*vb)->varBindChoice->choiceId = varbindChoice;

  switch(varbindChoice)
  {
    case(VARBINDCHOICE_UNSPECIFIED):
  (*vb)->varBindChoice->a.unSpecified = (AsnNull) NULL;
    break;
    case(VARBINDCHOICE_NOSUCHOBJECT):
  (*vb)->varBindChoice->a.noSuchObject = (AsnNull) NULL;
    break;
    case(VARBINDCHOICE_NOSUCHINSTANCE):
  (*vb)->varBindChoice->a.noSuchInstance = (AsnNull) NULL;
    break;
    case(VARBINDCHOICE_ENDOFMIBVIEW):
  (*vb)->varBindChoice->a.endOfMibView = (AsnNull) NULL;
    break;
    case(VARBINDCHOICE_VALUE):
  (*vb)->varBindChoice->a.value = (ObjectSyntax *) malloc(sizeof(ObjectSyntax));
  if ((*vb)->varBindChoice->a.value == NULL)
	return( (VarBind *) NULL);
  (*vb)->varBindChoice->a.value->choiceId = syntaxChoice;
  switch(syntaxChoice)
  {
    case(OBJECTSYNTAX_STRING_VALUE):
    case(OBJECTSYNTAX_OBJECTID_VALUE):
    case(OBJECTSYNTAX_IPADDRESS_VALUE):
    case(OBJECTSYNTAX_ARBITRARY_VALUE):
    case(OBJECTSYNTAX_NSAPADDRESS_VALUE):
  (*vb)->varBindChoice->a.value->a.string_value = 
			(AsnOcts*) malloc(sizeof(AsnOcts));
  if ((*vb)->varBindChoice->a.value->a.string_value == NULL)
	return( (VarBind *) NULL);
  (*vb)->varBindChoice->a.value->a.string_value->octs = 
			(unsigned char *) malloc( valueSize);
  if ((*vb)->varBindChoice->a.value->a.string_value->octs == NULL)
	return( (VarBind *) NULL);
  (*vb)->varBindChoice->a.value->a.string_value->octetLen = 0;
    break;
    case(OBJECTSYNTAX_BIT_VALUE):
  (*vb)->varBindChoice->a.value->a.bit_value = (AsnBits*) malloc(sizeof(AsnBits));
  if ((*vb)->varBindChoice->a.value->a.bit_value == NULL)
	return( (VarBind *) NULL);
  (*vb)->varBindChoice->a.value->a.bit_value->bits = (unsigned char *) malloc( valueSize);
  if ((*vb)->varBindChoice->a.value->a.bit_value->bits == NULL)
	return( (VarBind *) NULL);
  (*vb)->varBindChoice->a.value->a.bit_value->bitLen = 0;
    break;
    default:
    break;
   }
   }

  return( (*vb) );
}



/* 
 * -------------------------------------------------------------
 *
 * Function:	AllocPrivMsg
 *
 * Arguments:	privDstSize, size of the privDst oid (IN).
 *		privDataSize, size of the privData string (IN).   
 *
 * Returns:     SnmpPrivMsg *
 *
 * Purpose:	Allocates space for SnmpPrivMsg. This structure
 *		contains two fields(privDst and privData). Space for 
 *		these fields will also be allocated.
 *
 * Remarks:	The octetLen fields will be set to zero.		
 *		
 *
 * ------------------------------------------------------------
 */
	SnmpPrivMsg*
AllocPrivMsg(UAsnInt privDstSize, UAsnInt privDataSize)
{
  SnmpPrivMsg	*privMsg;

  privMsg = (SnmpPrivMsg *) Asn1Alloc(sizeof(SnmpPrivMsg));
  if (privMsg == NULL)
	return( (SnmpPrivMsg *) NULL);
  privMsg->privDst.octs = Asn1Alloc(privDstSize);
  if (privMsg->privDst.octs == NULL)
	return( (SnmpPrivMsg *) NULL);
  privMsg->privDst.octetLen = 0;
  privMsg->privData.octs = Asn1Alloc(privDataSize);
  if (privMsg->privData.octs == NULL)
	return( (SnmpPrivMsg *) NULL);
  privMsg->privData.octetLen = 0;

  return ( privMsg );
}


/* 
 * -------------------------------------------------------------
 *
 * Function:	AllocAuthMsg
 *
 * Arguments:	authDataSize, size of the authData string (IN).
 *		choice, choice between auth & noauth (IN).
 *
 * Returns:     SnmpAuthMsg* 
 *
 * Purpose:	Allocates space for SnmpAuthMsg. This structure
 *		contains two fields (noAuth & authData or
 *		v2md5AuthProtocol & authData), the space of
 *		these fields will also be allocated.
 *
 * Remarks:	If the choice is noAuth then the noAuth field
 *		will be filled with NULL and the octetLen will
 *		be set to zero.
 *		If the choice is v2md5AuthProtocol then the octetLen field
 *		of  the digest will be set to DIGESTSIZE.
 *		The octetLen field of authData will be set to zero.		
 *		
 *
 * ------------------------------------------------------------
 */

	SnmpAuthMsg*
AllocAuthMsg(UAsnInt authDataSize, SnmpAuthMsgChoiceChoiceId choice)
{
  SnmpAuthMsg	*authMsg;

  authMsg = (SnmpAuthMsg *) Asn1Alloc(sizeof(SnmpAuthMsg));
  if (authMsg == NULL)
	return( (SnmpAuthMsg *) NULL);
  authMsg->authInfo = (SnmpAuthMsgChoice *)
		      Asn1Alloc(sizeof(SnmpAuthMsgChoice));
  if (authMsg->authInfo == NULL)
	return( (SnmpAuthMsg *) NULL);

  authMsg->authInfo->choiceId = choice;
  switch(choice)
  {
    case(SNMPAUTHMSGCHOICE_NOAUTH):
  authMsg->authInfo->a.noAuth = (AsnOcts *) Asn1Alloc ( sizeof( AsnOcts ));
  if (authMsg->authInfo->a.noAuth == NULL)
	return( (SnmpAuthMsg *) NULL);
  authMsg->authInfo->a.noAuth->octs = (unsigned char *) Asn1Alloc(1);
  if (authMsg->authInfo->a.noAuth->octs == NULL)
	return( (SnmpAuthMsg *) NULL);
  authMsg->authInfo->a.noAuth->octetLen = 0;
  authMsg->authInfo->a.noAuth->octs = (unsigned char *) NULL;
    break;
    case(SNMPAUTHMSGCHOICE_V2MD5AUTHPROTOCOL):
  authMsg->authInfo->a.v2md5AuthProtocol = 
	(SnmpAuthMsgChoiceSeq*) Asn1Alloc(sizeof(SnmpAuthMsgChoiceSeq));
  if (authMsg->authInfo->a.v2md5AuthProtocol == NULL)
	return( (SnmpAuthMsg *) NULL);
  authMsg->authInfo->a.v2md5AuthProtocol->authDigest.octs = 
	(unsigned char *) Asn1Alloc(DIGESTSIZE);
  if (authMsg->authInfo->a.v2md5AuthProtocol->authDigest.octs == NULL)
	return( (SnmpAuthMsg *) NULL);
  authMsg->authInfo->a.v2md5AuthProtocol->authDigest.octetLen = DIGESTSIZE;
    break;
  }

  authMsg->authData.octs = (unsigned char *)
			   Asn1Alloc( authDataSize );
  if ( authMsg->authData.octs == NULL)
	return( (SnmpAuthMsg *) NULL);
  authMsg->authData.octetLen = 0;

  return( authMsg );
}


/* 
 * -------------------------------------------------------------
 *
 * Function:	AllocMgmtCom
 *
 * Arguments:	tag, choice between the PDUs (IN).
 *		operationSize, size of the 'operation' field(IN).
 *		srcPartySize, size of the srcParty field (IN).
 *		dstPartySize, size of the dstParty field (IN).
 *		contextSize, size of the context field (IN).
 *
 * Returns:     SnmpMgmtCom* 
 *
 * Purpose:	Allocates space for SnmpMgmtCom. This structure
 *		contains four fields (dstParty, srcParty, context and
 *		operation), for these fields space will be allocated.
 *
 * Remarks:	The choiceId field will be set to 'tag'. The octetLen
 *		fields of the four fields will be set to zero.
 *		
 *
 * ------------------------------------------------------------
 */
	SnmpMgmtCom*
AllocMgmtCom(PDUsChoiceId tag,UAsnInt operationSize,
	     UAsnInt srcPartySize, UAsnInt dstPartySize,
	     UAsnInt contextSize)
{
  SnmpMgmtCom	*mgmtCom;

  mgmtCom = (SnmpMgmtCom *) Asn1Alloc(sizeof(SnmpMgmtCom));
  if ( mgmtCom == NULL)
	return( (SnmpMgmtCom *) NULL);
  mgmtCom->operation = (SnmpMgmtComChoice *) 
		       Asn1Alloc(sizeof(SnmpMgmtComChoice));
  if ( mgmtCom->operation == NULL)
	return( (SnmpMgmtCom *) NULL);

  mgmtCom->operation->choiceId = tag;
  mgmtCom->operation->a.get_request = (AsnOcts *) Asn1Alloc(sizeof(AsnOcts));
  if ( mgmtCom->operation->a.get_request == NULL)
	return( (SnmpMgmtCom *) NULL);
  
  mgmtCom->operation->a.get_request->octs = (unsigned char *)
					    Asn1Alloc(operationSize);
  if ( mgmtCom->operation->a.get_request->octs == NULL)
	return( (SnmpMgmtCom *) NULL);
  mgmtCom->operation->a.get_request->octetLen = 0;

  mgmtCom->srcParty.octs = (unsigned char *) 
			   Asn1Alloc( srcPartySize );
  if ( mgmtCom->srcParty.octs == NULL)
	return( (SnmpMgmtCom *) NULL);
  mgmtCom->srcParty.octetLen = 0;
  mgmtCom->dstParty.octs = (unsigned char *)
			   Asn1Alloc( dstPartySize );
  if ( mgmtCom->dstParty.octs == NULL)
	return( (SnmpMgmtCom *) NULL);
  mgmtCom->dstParty.octetLen = 0;
  mgmtCom->context.octs = (unsigned char *)
			  Asn1Alloc( contextSize );
  if ( mgmtCom->context.octs == NULL)
	return( (SnmpMgmtCom *) NULL);
  mgmtCom->context.octetLen = 0;

  return( mgmtCom );
}


/* 
 * -------------------------------------------------------------
 *
 * Function:	AllocPDUs
 *
 * Arguments:	tag, choice between the PDUs (IN).
 *
 * Returns:     PDUs* 
 *
 * Purpose:	Allocates space for PDUs. This structure
 *		contains the variable_bindings field for which
 *		space will be allocated.
 *
 * Remarks:	The choiceId field will be set to 'tag'.
 *		For the variable_bindings field only space for
 *		the pointer will be allocated. To create a 
 *		VarBindList use Vbl.
 *
 * ------------------------------------------------------------
 */
	PDUs*
AllocPDUs(PDUsChoiceId tag)
{
 PDUs	*pduMsg;

 pduMsg = (PDUs *) Asn1Alloc(sizeof(PDUs));
  if ( pduMsg == NULL)
	return( (PDUs*) NULL);
 pduMsg->choiceId = tag;

 if (tag == PDUS_GET_BULK_REQUEST)
   {
      pduMsg->a.get_bulk_request = (GetBulkRequest_PDU*) 
				    Asn1Alloc(sizeof(GetBulkRequest_PDU));
      if ( pduMsg->a.get_bulk_request == NULL)
	return( (PDUs*) NULL);
      pduMsg->a.get_bulk_request->variable_bindings = 
			    (VarBindList*) Asn1Alloc(sizeof(VarBindList));
      if ( pduMsg->a.get_bulk_request->variable_bindings == NULL)
	return( (PDUs*) NULL);
    }
  else
    {
      pduMsg->a.get_request = (GetRequest_PDU*)
			      Asn1Alloc(sizeof(GetRequest_PDU));
      if ( pduMsg->a.get_request == NULL)
	return( (PDUs*) NULL);
      pduMsg->a.get_request->variable_bindings = 
		  (VarBindList*) Asn1Alloc(sizeof(VarBindList));
      if ( pduMsg->a.get_request->variable_bindings == NULL)
	return( (PDUs*) NULL);
    }
  return( pduMsg );
}


/* 
 * ----------------------------------------------------------------------
 *
 * Function:	AllocMessage 
 *
 * Arguments:	tag, choice between the PDUs (IN).
 *		messageSize, size of the data field(IN).
 *		comminitySize, size of the community field (IN).
 *
 * Returns:     MessageV1* 
 *
 * Purpose:	Allocates space for MessageV1. This structure
 *		contains three fields (version, community and data),
 *		for these fields space will be allocated.
 *
 * Remarks:	The choiceId field of the data field  will be set to 'tag'.
 *		The octetLen field of community and data will be set to zero.
 *		
 *
 * ---------------------------------------------------------------------
 */

	MessageV1*
AllocMessage( PDUsChoiceId tag, UAsnInt messageSize, UAsnInt communitySize)

{
 MessageV1	*message;

 message = (MessageV1 *) Asn1Alloc(sizeof(MessageV1));
 if (message == NULL)
	return( (MessageV1*) NULL);
 
 message->community.octs = (unsigned char *) Asn1Alloc (communitySize);
 if (message->community.octs == NULL)
	return( (MessageV1*) NULL);
 message->community.octetLen = 0;
 
 message->data = (MessageV1Choice* ) Asn1Alloc(sizeof(MessageV1Choice));
 if (message->data == NULL)
	return( (MessageV1*) NULL);
 
 message->data->choiceId = tag;

 message->data->a.get_request = (AsnOcts*) Asn1Alloc (sizeof(AsnOcts));
 if (message->data->a.get_request == NULL)
	return( (MessageV1*) NULL);
 message->data->a.get_request->octs = (unsigned char *) Asn1Alloc(messageSize);
 if (message->data->a.get_request->octs == NULL)
	return( (MessageV1*) NULL);

 message->data->a.get_request->octetLen = 0;


 return( message );
}


/* 
 * -------------------------------------------------------------
 *
 * Function:	AllocPDUsV1
 *
 * Arguments:	tag, choice between the PDUs (IN).
 *		enterpriseSize, only with trap-pdu  (IN).
 *		agent_addrSize, only with trap-pdu (IN).
 *
 * Returns:     PDUsV1* 
 *
 * Purpose:	Allocates space for PDUs. This structure
 *		contains the variable_bindings field for which
 *		space will be allocated. For a trap-pdu the enterprise
 *		and agent_addr will also be allocated.
 *
 * Remarks:	The choiceId field will be set to 'tag'.
 *		For the variable_bindings field only space for
 *		the pointer will be allocated. To create a 
 *		VarBindList use Vbl.
 *
 * ------------------------------------------------------------
 */
	PDUsV1*
AllocPDUsV1( PDUsChoiceId tag, UAsnInt enterpriseSize, UAsnInt agent_addrSize)
{
 PDUsV1		*pduMsg;

 pduMsg = (PDUsV1 *) Asn1Alloc(sizeof(PDUsV1));
 if (pduMsg == NULL)
	return( (PDUsV1*) NULL);
 pduMsg->choiceId = tag;

 if (tag == PDUS_TRAP)
 {
   pduMsg->a.trap = (Trap_PDUV1*) Asn1Alloc(sizeof(Trap_PDUV1));
   if (pduMsg->a.trap == NULL)
	return( (PDUsV1*) NULL);
   pduMsg->a.trap->enterprise.octs = (unsigned char *) Asn1Alloc(enterpriseSize);
   if (pduMsg->a.trap->enterprise.octs == NULL)
	return( (PDUsV1*) NULL);
   pduMsg->a.trap->enterprise.octetLen = 0;
   pduMsg->a.trap->agent_addr.octs = (unsigned char *) Asn1Alloc(agent_addrSize);
   if (pduMsg->a.trap->agent_addr.octs == NULL)
	return( (PDUsV1*) NULL);
   pduMsg->a.trap->agent_addr.octetLen = 0;
   pduMsg->a.trap->variable_bindings = (VarBindList*) Asn1Alloc(sizeof(VarBindList));
   if (pduMsg->a.trap->variable_bindings == NULL)
	return( (PDUsV1*) NULL);
 }
 else
 {
   pduMsg->a.get_request = (GetRequest_PDUV1*) Asn1Alloc(sizeof(GetRequest_PDUV1));
   if (pduMsg->a.get_request == NULL)
	return( (PDUsV1*) NULL);
   pduMsg->a.get_request->variable_bindings = (VarBindList*) Asn1Alloc(sizeof(VarBindList));
   if (pduMsg->a.get_request->variable_bindings == NULL)
	return( (PDUsV1*) NULL);
  }
  return( pduMsg );
}




/* 
 * -------------------------------------------------
 * Function:	CalcLength
 * 
 * Arguments:	value, ... (IN)
 *
 * Purpose:	calculate length of value. This function is needed for
 *		the calculation of the number of octets needed for
 *		the timestamps in BerEncodeAuthMsg. This value is
 *		needed to calculate the authDigestOffset.
 *
 * Returns:	length, ... (OUT)
 * -----------------------------------------------------
 */

	AsnInt
 CalcLength (AsnInt value)
{
 unsigned long int	mask;
 int			len;

 mask = (0x7f80 << ((sizeof(AsnInt) - 2) * 8));
 if (((long int) value) < 0)
 {
	len = sizeof(UAsnInt) + 1;
 }
 else
 {
	for (len = sizeof(UAsnInt); len > 1; --len)
	{
	    if ( ( value & mask) == 0)
		mask >>=8;
	    else
		break;
	}
 }
 return( len );
}


/* 
 * --------------------------------------------------------------------
 *
 * Function:    DynToStatVbl
 *
 * Arguments:	VarBindList	*vbl, Dynamic Varbindlist (in)
 *		chare		*staticvbl, Static Varbindlist (in/out)
 *
 * Returns:     OK or FAILED
 *
 * Purpose:     Copies the dynamic data into a static piece of memory
 *              The available memory size is give by svbl->ocetLen (in). 
 *              
 *
 * Remarks:     Should check for available space left in octetstring!
 *             
 * 
 * ---------------------------------------------------------------------
 */
	STATUS
DynToStatVbl ( VarBindList *vbl, char *staticvbl)
{
	VarBind	*vb ;
	char    *svbl; /* svbl is the pointer from the static list */
	int	memLen, byteLen ;

	svbl = staticvbl;

	/* write the total number of varbinds in new array 
	*/
	StatBufCopy( svbl, &vbl->count, sizeof(int) );

	FOR_EACH_LIST_ELMT( vb,  vbl) {

		/*	copy name length  of object name and the name itself */
		StatBufCopy( svbl, &vb->name.octetLen, sizeof(AsnInt) );
		StatBufCopy( svbl, vb->name.octs, vb->name.octetLen );

		/*	copy choice of varbind */ 
		StatBufCopy( svbl, &vb->varBindChoice->choiceId, sizeof(VarBindChoiceChoiceId) );

		switch (vb->varBindChoice->choiceId) {
		case (VARBINDCHOICE_UNSPECIFIED) :  break ;
		case (VARBINDCHOICE_NOSUCHOBJECT) :  break ;
		case (VARBINDCHOICE_NOSUCHINSTANCE) :  break ;
		case (VARBINDCHOICE_ENDOFMIBVIEW) :  break ;
		case (VARBINDCHOICE_VALUE) : 
			StatBufCopy( svbl, &vb->varBindChoice->a.value->choiceId, 
					sizeof(ObjectSyntaxChoiceId) );
			switch (vb->varBindChoice->a.value->choiceId) {

			/*
			 *  	COPYING OF ASNINTs
			 */
			case (OBJECTSYNTAX_INTEGER_VALUE) : 
				StatBufCopy( svbl, &vb->varBindChoice->a.value->a.integer_value, 
					sizeof(AsnInt) ); break;
			case (OBJECTSYNTAX_COUNTER_VALUE) :
				StatBufCopy( svbl, &vb->varBindChoice->a.value->a.counter_value, 
					sizeof(UAsnInt) ); break;
			case (OBJECTSYNTAX_GAUGE_VALUE) :
				StatBufCopy( svbl, &vb->varBindChoice->a.value->a.gauge_value, 
					sizeof(UAsnInt) ); break;
			case (OBJECTSYNTAX_TIMETICKS_VALUE) :
				StatBufCopy( svbl, &vb->varBindChoice->a.value->a.timeticks_value, 
					sizeof(UAsnInt) ); break;
			case (OBJECTSYNTAX_BIG_COUNTER_VALUE) :
				StatBufCopy( svbl, &vb->varBindChoice->a.value->a.big_counter_value, 
					sizeof(UAsnInt) ); break;
			case (OBJECTSYNTAX_UNSIGNED_INTEGER_VALUE) :				
				StatBufCopy( svbl, &vb->varBindChoice->a.value->a.unsigned_integer_value, 
					sizeof(UAsnInt) ); break;

			/*
			 *	COPYING OF ASNOCTs
			 */
			case(OBJECTSYNTAX_STRING_VALUE):
			case(OBJECTSYNTAX_OBJECTID_VALUE):
			case(OBJECTSYNTAX_IPADDRESS_VALUE):
			case(OBJECTSYNTAX_ARBITRARY_VALUE):
			case(OBJECTSYNTAX_NSAPADDRESS_VALUE):
				StatBufCopy( svbl, &vb->varBindChoice->a.value->a.string_value->octetLen, 
					sizeof(AsnInt) ) ; 
				StatBufCopy( svbl, vb->varBindChoice->a.value->a.string_value->octs, 
					vb->varBindChoice->a.value->a.string_value->octetLen); break;

			/*
			 * 	COPYING OF ASNBITs
			 */
			case(OBJECTSYNTAX_BIT_VALUE):
 				if (vb->varBindChoice->a.value->a.bit_value->bitLen == 0)
   					byteLen = 0;
				else
					byteLen = ((vb->varBindChoice->a.value->a.bit_value->bitLen-1) / 8) + 1;
				StatBufCopy( svbl, &vb->varBindChoice->a.value->a.bit_value->bitLen, 
				 	sizeof(int) ); 
				StatBufCopy( svbl, vb->varBindChoice->a.value->a.bit_value->bits, 
				 	byteLen ); break;
	
			default:
				return FAILED ;
			} 	/* end switch */
		        break;
		default:
			return FAILED ;
		} 	/* end switch */

	}	/* end foreach */

	return OK; 

} /* end function */


/* 
 * --------------------------------------------------------------------
 *
 * Function:    StatToDynVbl
 *
 * Arguments:	AsnOcts		*staticvbl, Dynamic Varbindlist (in)
 *		VarBindList	*vbl, Static Varbindlist (in/out)
 *
 * Returns:     OK or FAILED
 *
 * Purpose:     Copy a static varbinding list to a dynamic var bindings list 
 *              The available memory size is give by svbl->ocetLen (in). 
 *              
 *
 * Remarks:     Should check for available space left in octetstring!
 *             
 * 
 * ---------------------------------------------------------------------
 */
	STATUS
StatToDynVbl ( char *staticvbl, VarBindList *vbl)
{
	VarBind	*vb ;
	char	*cp ; 		/* the static array, used as current pointer */
	int	i, vbNr;	/* number of variables */
	int 	tempLen ;
	AsnInt	len ;
	AsnOid	*objectName ;
	VarBindChoiceChoiceId varBindChoice; 
	ObjectSyntaxChoiceId syntaxChoice=0; 
	UAsnInt	valueSize, bitLen; 


/*
 *	first save the data on the static memory, so it is possible to return 
 * 	the static varbind list identically
 */ 


	cp = staticvbl ; 
	
	/* 	Get the number of available var binds 
	 * 	vbNr was originally vbl->count, but NewVb will increase the vbl->count !! 
	 */
	DynBufCopy( &vbNr, cp , sizeof(int) );

	for (i=0; i < vbNr; i++) {

		valueSize = 0 ;

		/*	copy name length  of object name and the name itself */
		DynBufCopy( &len , cp, sizeof(AsnInt) );	
		objectName = pNewOid (len) ;
		objectName->octetLen = len ;		
		DynBufCopy( objectName->octs, cp, objectName->octetLen );

		/*	copy choice of varbind */ 
		DynBufCopy( &varBindChoice, cp, sizeof(VarBindChoiceChoiceId) );

		switch (varBindChoice) {
		case (VARBINDCHOICE_UNSPECIFIED) : break ;
		case (VARBINDCHOICE_NOSUCHOBJECT) :  break ;
		case (VARBINDCHOICE_NOSUCHINSTANCE) : break ;
		case (VARBINDCHOICE_ENDOFMIBVIEW) : break ;

		case (VARBINDCHOICE_VALUE) : 

			DynBufCopy( &syntaxChoice, cp, sizeof(ObjectSyntaxChoiceId) );
			switch (syntaxChoice) {

			case (OBJECTSYNTAX_INTEGER_VALUE) : 
			case (OBJECTSYNTAX_COUNTER_VALUE) :
			case (OBJECTSYNTAX_GAUGE_VALUE) :
			case (OBJECTSYNTAX_TIMETICKS_VALUE) :
			case (OBJECTSYNTAX_BIG_COUNTER_VALUE) :
			case (OBJECTSYNTAX_UNSIGNED_INTEGER_VALUE) :
				valueSize = 0 ; 				
				break;

			case(OBJECTSYNTAX_STRING_VALUE):
			case(OBJECTSYNTAX_OBJECTID_VALUE):
			case(OBJECTSYNTAX_IPADDRESS_VALUE):
			case(OBJECTSYNTAX_ARBITRARY_VALUE):
			case(OBJECTSYNTAX_NSAPADDRESS_VALUE):
				DynBufCopy( &valueSize, cp, sizeof(UAsnInt) ) ; 
				break;

			case(OBJECTSYNTAX_BIT_VALUE):
				DynBufCopy( &bitLen, cp, sizeof(UAsnInt) ) ; 
 				if (bitLen == 0)
   					valueSize = 0;
				else
					valueSize = ((bitLen-1) / 8) + 1;
				break ;
	
			default: 
				return FAILED ;
			} 	/* end switch */
			break;
		default: 
			return FAILED ;
		} 	/* end switch */
		
		vb = NewVbp ( vbl, objectName->octetLen, varBindChoice, syntaxChoice, valueSize ) ;

		(void) memcpy( vb->name.octs, objectName->octs, objectName->octetLen ) ;
		vb->name.octetLen= objectName->octetLen ;

		/*  varBindChoice & syntaxChoice are already set in NewVbl */

		if ( varBindChoice == VARBINDCHOICE_VALUE ) {
			switch (syntaxChoice) {

			case (OBJECTSYNTAX_INTEGER_VALUE) : 
				DynBufCopy( &vb->varBindChoice->a.value->a.integer_value, 
					cp, sizeof(AsnInt) ) ; break;
			case (OBJECTSYNTAX_COUNTER_VALUE) :
				DynBufCopy( &vb->varBindChoice->a.value->a.counter_value, 
					cp, sizeof(AsnInt) ) ; break;
			case (OBJECTSYNTAX_GAUGE_VALUE) :
				DynBufCopy( &vb->varBindChoice->a.value->a.gauge_value, 
					cp, sizeof(AsnInt) ) ; break;
			case (OBJECTSYNTAX_TIMETICKS_VALUE) :
				DynBufCopy( &vb->varBindChoice->a.value->a.timeticks_value, 
					cp, sizeof(AsnInt) ) ; break;
			case (OBJECTSYNTAX_BIG_COUNTER_VALUE) :
				DynBufCopy( &vb->varBindChoice->a.value->a.big_counter_value, 
					cp, sizeof(AsnInt) ) ; break;
			case (OBJECTSYNTAX_UNSIGNED_INTEGER_VALUE) :				
				DynBufCopy( &vb->varBindChoice->a.value->a.unsigned_integer_value, 
					cp, sizeof(AsnInt) ) ; break;

			case(OBJECTSYNTAX_STRING_VALUE):
			case(OBJECTSYNTAX_OBJECTID_VALUE):
			case(OBJECTSYNTAX_IPADDRESS_VALUE):
			case(OBJECTSYNTAX_ARBITRARY_VALUE):
			case(OBJECTSYNTAX_NSAPADDRESS_VALUE):
				DynBufCopy( vb->varBindChoice->a.value->a.string_value->octs, 
					cp, valueSize ) ; 
				vb->varBindChoice->a.value->a.string_value->octetLen = valueSize;
				break;

			case(OBJECTSYNTAX_BIT_VALUE):
				DynBufCopy( &valueSize, cp, valueSize ) ; 
				DynBufCopy( vb->varBindChoice->a.value->a.bit_value->bits, 
					cp, valueSize ) ; 
				vb->varBindChoice->a.value->a.bit_value->bitLen = bitLen ;
				break ;
			
			default:	
				return FAILED ;
			} 	/* end switch */

		}	/* end if */

	}	/* end foreach */

	return OK; 

} /* end function */



	STATUS
V1TrapToVarbindlist( VarBindList **varbindlist,  AsnOid *enterprise, 
		AsnInt *generic_trap, AsnInt *specific_trap, UAsnInt *time_stamp)
{
	UAsnInt	vbSize ;
	VarBind	*newVb ;
	int	dummy=0 ;
	char	kar ;

	vbSize = 0 ;
	if ( *generic_trap == ENTERPRISESPECIFIC ) {
		vbSize = enterprise->octetLen + 1 + sizeof(AsnInt) ;
		newVb = NewVbp( *varbindlist, 9, VARBINDCHOICE_VALUE, 
			OBJECTSYNTAX_OBJECTID_VALUE, vbSize) ;
		newVb->name.octetLen = 9 ;
		memcpy( newVb->name.octs, snmpTrapOID.octs, 8 ) ;
		memcpy( newVb->name.octs+8, &dummy, 1 ) ;
		newVb->varBindChoice->a.value->a.objectID_value->octetLen = vbSize ;
		memcpy(newVb->varBindChoice->a.value->a.objectID_value->octs, enterprise->octs, 
			enterprise->octetLen ) ;
		memcpy(newVb->varBindChoice->a.value->a.objectID_value->octs+enterprise->octetLen, 
			&dummy, 1 ) ;
		memcpy( newVb->varBindChoice->a.value->a.objectID_value->octs+enterprise->octetLen+1, 
			specific_trap, sizeof(AsnInt) ) ;
	} else {
		newVb = NewVbp( *varbindlist, 9, VARBINDCHOICE_VALUE, 
			VARBINDCHOICE_UNSPECIFIED, 9) ;
		newVb->name.octetLen = 9 ;
		memcpy( newVb->name.octs, snmpTrapOID.octs, 8 ) ;
		memcpy( newVb->name.octs+8, &dummy, 1 ) ;
		memcpy(newVb->varBindChoice->a.value->a.objectID_value->octs, snmpTraps.octs, 
			snmpTraps.octetLen ) ;
		kar  = (char) (*generic_trap)+1 ;
		memcpy(newVb->varBindChoice->a.value->a.objectID_value->octs+8, &kar, 1) ;
	} 
		
	newVb = NewVbp( *varbindlist, 8, VARBINDCHOICE_VALUE, 
			OBJECTSYNTAX_UNSIGNED_INTEGER_VALUE, 0) ;
	newVb->name.octetLen = 8 ;
	memcpy( newVb->name.octs, sysUpTime.octs, 7 ) ;
	memcpy( newVb->name.octs+7, &dummy, 1 ) ;
	newVb->varBindChoice->a.value->a.unsigned_integer_value = *time_stamp ;


	newVb = NewVb( *varbindlist, 10, VARBINDCHOICE_VALUE, 
			OBJECTSYNTAX_OBJECTID_VALUE, enterprise->octetLen) ;
	newVb->name.octetLen = 10 ;
	memcpy( newVb->name.octs, snmpTrapEnterprise.octs, 9) ;
	memcpy( newVb->name.octs+9, &dummy, 1 ) ;
	memcpy( newVb->varBindChoice->a.value->a.objectID_value->octs, enterprise->octs, 
			enterprise->octetLen ) ;
	
	return OK ;
	
} 	/* V1TrapToVarbindlist */


	STATUS
VarbindlistToV1Trap( VarBindList **varbindlist,  AsnOid *enterprise, 
		AsnInt *generic_trap, AsnInt *specific_trap, UAsnInt *time_stamp)
{
	VarBind		*newVb ;

	newVb = FIRST_LIST_ELMT( *varbindlist ) ;
	*time_stamp = newVb->varBindChoice->a.value->a.unsigned_integer_value ;

	AsnListRemove (*varbindlist) ;

	newVb = FIRST_LIST_ELMT( *varbindlist) ;
	*generic_trap = (AsnInt) *(newVb->varBindChoice->a.value->a.objectID_value->octs+8);

	/*!! enterprise specific trap NOT YET supported */

	AsnListRemove (*varbindlist) ;

	newVb = LAST_LIST_ELMT( *varbindlist) ;
	memcpy( enterprise->octs, newVb->varBindChoice->a.value->a.objectID_value->octs, 
		newVb->varBindChoice->a.value->a.objectID_value->octetLen ) ;

	(*varbindlist)->curr = (*varbindlist)->last;
	AsnListRemove (*varbindlist) ;
}

/*
 * -------------------------------------------------------------------
 *
 * Function:    CopyVbl
 *
 * Arguments:   vbl1 (in)
 *              vbl2 (out)
 *
 * Purpose:     copy varbindlist vbl1 to varbindlist vbl2
 *
 * Author:      K. Flapper, Rob Post
 *
 * ------------------------------------------------------------------- */

	void 
CopyVbl( VarBindList *vbl1, VarBindList *vbl2)
{
    char    buf[10000]; /* correct size? */

    (void) DynToStatVbl( vbl1, buf);
    (void) StatToDynVbl( buf, vbl2);
}


/* ---------------------- end of file TYPES.c ----------------------
 */

