/* -----------------------------------------------------------------------------
 * project:     UT-SNMP
 * file:	snmpInitPerkins.c 
 * date:        21 September 1994  
 * purpose:     initialisation *.perk-files, containing SNMPv2 administrative
 *		configuration information as proposed by David Perkins 
 * author(s):	Martijn Visser 
 * -----------------------------------------------------------------------------
 */

#define MAIN

/* -----------------------------------------------------------------------
 * include files
 * -----------------------------------------------------------------------
 */
#include "snmpInitPerkins.h"      /* New header files */
#include <snmpInitAgentUser.h>    /* User interface functions */
#include <snmpInitAgentSys.h>     /* System functions */

/* -----------------------------------------------------------------------
 * typedefs
 * -----------------------------------------------------------------------
 */
typedef enum {
   guest = 1,
   operator,
   root
} AuthLevel;

typedef enum {
   public = 1,		/* noAuth/noPriv */
   authenticated,	/* auth/noPriv */
   secret		/* auth/priv */
} CommType;


/* -----------------------------------------------------------------------
 * Function : InitViewRecord
 * Arguments: view initial pointer to viewentry (IN/OUT)
 * Purpose  : This function initialises a view record.
 * Remarks  :
 * -----------------------------------------------------------------------
 */
void InitViewRecord(viewEntry *view)
{
  view->viewIndex = 1;          /* First record in view.conf file */
  strcpy(view->viewSubtree,"\0"); /* Variable empty */
  strcpy(view->viewMask,"\0");
  strcpy(view->viewType,"\0");
}

/* -----------------------------------------------------------------------
 * Function:  	InitAclRecord
 * Arguments: 	acl, initial pointer to aclentry (IN/OUT)
 * Purpose:   	This function initialises an acl record.
 * Remarks:
 * -----------------------------------------------------------------------
 */
void InitAclRecord(aclEntry *acl)
{
  strcpy(acl->aclName,"\0");
  strcpy(acl->aclPrivileges,"\0");
}

/* -----------------------------------------------------------------------
 * Function :  	InitContextRecord
 * Arguments: 	context, initial pointer to contextentry (IN/OUT) 
 * Purpose  :  	This function initialises an context record.
 * Remarks  :
 * -----------------------------------------------------------------------
 */
void InitContextRecord(contextEntry *context)
{
  strcpy(context->contextName,"\0");
  strcpy(context->contextIdentity,InitialContextId);
  context->contextIndex = 1;
  context->contextLocal = 0;
  context->contextViewIndex = 1;
  strcpy(context->contextLocalEntity,"Null");
  strcpy(context->contextLocalTime,"currentTime");
  strcpy(context->contextProxyDstParty,"0");
  strcpy(context->contextProxySrcParty,"0");
  strcpy(context->contextProxyContext,".0.0");
}
  
/* -----------------------------------------------------------------------
 * Function:  	InitPartyRecord
 * Arguments: 	party, initial pointer to partyentry (IN/OUT) 
 * Purpose:   	This function initialises an party record.
 * Remarks:
 * -----------------------------------------------------------------------
 */
void InitPartyRecord(partyEntry *party)
{
   strcpy(party->partyName,"\0");  /* Initialisation party record */
   strcpy(party->partyIdentity,"\0");
   strcpy(party->IPaddress,"\0");
   strcpy(party->partyTAddress,"\0");
   strcpy(party->partyAuthProtocol,"\0");
   strcpy(party->partyAuthPrivate,"\0");
   strcpy(party->partyAuthPublic,"\0");
   strcpy(party->partyPrivProtocol,"\0");
   strcpy(party->partyPrivPrivate,"\0");
   strcpy(party->partyPrivProtocol,"\0");
   strcpy(party->partyPrivPublic,"\0");
   strcpy(party->partyMaxMessageSize,"0");      /* Default */
   strcpy(party->partyTDomain,"snmpUDPDomain"); /* Default */
   strcpy(party->partyAuthClock,"00000000");    /* Default */
   strcpy(party->partyAuthLifetime,"300");      /* Default */
   strcpy(party->partyLocal,"\0");	        /* Default */	
}

/* -----------------------------------------------------------------------
 * function:	addHeader
 * arguments:	pOutFile (IN) the stream to which the a header is added
 * 		szFilename (IN) name of the file containing the header
 * returns:	status, 0 on succes and -1 when failed
 * purpose:	this functions adds a header to a stream
 * remarks:	-
 * -----------------------------------------------------------------------
 */
addHeader(FILE *pOutFile, char *szFilename)
{
    char szPathname[FILENAMELEN];
    char szBuf[MAXRECSTR];
    FILE * pInFile;
        
    if (getenv("SNMP_PARTYMIB_PATH") !=  (char *)NULL)
        strcpy(szPathname,getenv("SNMP_PARTYMIB_INITPATH"));
    strcat(szPathname,szFilename);
    
    if ((pInFile = fopen(szPathname,"rt")) == (FILE *)NULL) {
        fprintf(stderr,"%s, %d: could not open \"%s\" for read...\n",__FILE__,__LINE__,szFilename);
        return(-1);
    }
    
    while (fgets(szBuf,MAXRECSTR,pInFile) != (char *)NULL) {
        fputs(szBuf,pOutFile);
    }
    fputc('\n',pOutFile);
    return 0;
}

/* -----------------------------------------------------------------------
 * function:	getNextIndex
 * arguments:	pFile (IN) pointer to agent-file
 *		pNextPartyIndex (IN) pointer to last used index
 * returns:	-
 * purpose:	search in an agent-file for the first unused indices
 * remarks:	-
 * -----------------------------------------------------------------------
 */
void getNextIndex(FILE *pFile, int *pNextPartyIndex)
{
    char szLine[MAXRECSTR], szField[MAXRECSTR];
    int nIndex;
    
    *pNextPartyIndex = 0;
    while (fgets(szLine,MAXRECSTR,pFile) != (char *)NULL) {
        sscanf(szLine,"%s = %d",szField,&nIndex);
        if ((strcmp(szField,"partyIndex") == 0) && (nIndex > *pNextPartyIndex))
            *pNextPartyIndex = nIndex;
    }
}
        

/* -----------------------------------------------------------------------
 * function:	buildAclNames
 * arguments:	nComm (IN) communication-type
 *		nAuth (IN) authorization-level
 *	        szManagerName (IN) hostname of manager
 *              agentAclName, managerAclName (OUT)
 * returns:	-
 * purpose:	builds names for ACL-pair depending on communication-type
 *		and authorization-level
 * remarks:	-
 * -----------------------------------------------------------------------
 */
void buildAclNames(CommType nComm, AuthLevel nAuth, char *szManagerName,
    char *agentAclName, char *managerAclName)
{
    switch (nComm) {
        case public:	    strcpy(agentAclName,"anonymousToAgent");
                            strcpy(managerAclName,"agentToAnonymous");
                            break;
        case authenticated:
        case secret:        strcpy(agentAclName,szManagerName);
                            strcpy(managerAclName,"agentTo");
                            strcat(agentAclName,"ToAnonymous");
                            strcat(managerAclName,szManagerName);
                            break;
    }
    switch (nComm) {
        case public:        strcat(agentAclName,"-public");
                            strcat(managerAclName,"-public");
                            break;
        case authenticated: strcat(agentAclName,"-authenticated");
                            strcat(managerAclName,"-authenticated");
                            break;
        case secret:        strcat(agentAclName,"-secret");
                            strcat(managerAclName,"-secret");
                            break;
    }
    switch (nAuth) {
        case guest:    strcat(agentAclName,"-guest");
                       strcat(managerAclName,"-guest");
                       break;
        case operator: strcat(agentAclName,"-operator");
                       strcat(managerAclName,"-operator");
                       break;
        case root:     strcat(agentAclName,"-root");
                       strcat(managerAclName,"-root");
                       break;
    }
}


/* -----------------------------------------------------------------------
 * function:	writeView
 * arguments:	pFile (IN) pointer to stream to which the viewEntry is written
 *		view (IN) the viewEntry
 * returns:	-
 * purpose:	writes viewEntry to file
 * remarks:	-
 * -----------------------------------------------------------------------
 */
void writeView(FILE *pFile, viewEntry view)
{
    fprintf(pFile,"%%view <%s>\n",view.viewName);
    fprintf(pFile,"    viewIndex = %d\n",view.viewIndex);
    fprintf(pFile,"    viewSubtree = %s\n",view.viewSubtree);
    if (strcmp(view.viewType,"included") != 0)
        fprintf(pFile,"    viewType = %s\n",view.viewType);
}

/* -----------------------------------------------------------------------
 * function:	writeContext
 * arguments:	pFile (IN) pointer to stream to which the contextEntry is written
 *		context (IN) the contextEntry
 * returns:	-
 * purpose:	writes contextEntry to file
 * remarks:	-
 * -----------------------------------------------------------------------
 */
void writeContext(FILE *pFile, contextEntry context)
{
    fprintf(pFile,"%%context <%s>\n",context.contextName);
    fprintf(pFile,"    contextIdentity = %s.%d\n",context.contextIdentity,context.contextIndex);
    fprintf(pFile,"    contextIndex = %d\n",context.contextIndex);
    fprintf(pFile,"    contextViewIndex = %d\n",context.contextViewIndex);
}

/* -----------------------------------------------------------------------
 * function:	writeParty
 * arguments:	pFile (IN) pointer to stream to which the partyEntry is written
 *		party (IN) the partyEntry
 * returns:	-
 * purpose:	writes partyEntry to file
 * remarks:	-
 * -----------------------------------------------------------------------
 */
void writeParty(FILE *pFile, partyEntry party)
{
    fprintf(pFile,"%%party <%s>\n",party.partyName);
    fprintf(pFile,"    partyIdentity = %s.%d",party.partyIdentity,party.partyType);
    if (party.partyType > 2)
        fprintf(pFile,".%s",party.IPaddress);
    fputc('\n',pFile); 
    fprintf(pFile,"    partyIndex = %d\n",party.partyIndex);
    fprintf(pFile,"    partyTDomain = %s\n",party.partyTDomain);
    fputs("    partyTAddress = ",pFile);
    if (strcmp(party.partyName,"public-manager") == 0)
        fputs("0.0.0.0",pFile);
    else
        fprintf(pFile,"%s",party.IPaddress);
    fprintf(pFile," %s\n",party.partyTAddress);
    if (strcmp(party.partyMaxMessageSize,"484") != 0)
        fprintf(pFile,"    partyMaxMessageSize = %s\n",party.partyMaxMessageSize);
    if (strcmp(party.partyLocal,"true") != 0)
        fprintf(pFile,"    partyLocal = %s\n",party.partyLocal);
    fprintf(pFile,"    partyAuthProtocol = %s\n",party.partyAuthProtocol);
    if (strcmp(party.partyPrivProtocol,"noPriv") != 0)
        fprintf(pFile,"    partyPrivProtocol = %s\n",party.partyPrivProtocol);
    fprintf( stdout, "p\b" ) ;
    usleep( 130000 ) ;
	fflush( stdout ) ;
}

/* -----------------------------------------------------------------------
 * function:	writeAcl
 * arguments:	pFile (IN) pointer to stream to which the aclEntry is written
 *		acl (IN) the aclEntry
 * returns:	-
 * purpose:	writes aclEntry to file
 * remarks:	-
 * -----------------------------------------------------------------------
 */
void writeAcl(FILE *pFile, aclEntry acl)
{
    fprintf(pFile,"%%acl <%s>\n",acl.aclName);
    fprintf(pFile,"    aclTarget = %d\n",acl.aclTarget);
    fprintf(pFile,"    aclSubject = %d\n",acl.aclSubject);
    fprintf(pFile,"    aclResources = %d\n",acl.aclResources);
    fprintf(pFile,"    aclPrivileges = %s\n",acl.aclPrivileges);
	fprintf( stdout, "a\b" ) ;
	usleep( 130000 ) ;
	fflush( stdout ) ;
}

/* -----------------------------------------------------------------------
 * function:	makeViews
 * arguments:	pFile (IN) pointer to stream to wich the views are written
 *		view (IN) a default viewEntry
 * returns:	-
 * purpose:	generates the 3 initial view-families
 * remarks:	-
 * -----------------------------------------------------------------------
 */
void makeViews(FILE *pFile, viewEntry view)
{
    for (view.viewIndex = 1; view.viewIndex < 4; view.viewIndex++) {
        switch (view.viewIndex) {
            case 1: strcpy(view.viewName,"system");
                    strcpy(view.viewSubtree,"system");
                    strcpy(view.viewType,"included");
                    writeView(pFile,view);
                    strcpy(view.viewName,"snmpStats");
                    strcpy(view.viewSubtree,"snmpStats");
                    writeView(pFile,view);
                    strcpy(view.viewName,"snmpParties");
                    strcpy(view.viewSubtree,"snmpParties");
                    writeView(pFile,view);
                    break;
            case 2: strcpy(view.viewName,"internet");
                    strcpy(view.viewSubtree,"internet");
                    strcpy(view.viewType,"included");
                    writeView(pFile,view);
                    strcpy(view.viewName,"excludedSNMPContexts");
                    strcpy(view.viewSubtree,"snmpContexts");
                    strcpy(view.viewType,"excluded");
                    writeView(pFile,view);
                    strcpy(view.viewName,"excludedSNMPAccess");
                    strcpy(view.viewSubtree,"snmpAccess");
                    writeView(pFile,view);
                    strcpy(view.viewName,"excludedSNMPViews");
                    strcpy(view.viewSubtree,"snmpViews");
                    writeView(pFile,view);
                    break;
            case 3: strcpy(view.viewName,"internet");
                    strcpy(view.viewSubtree,"internet");
                    strcpy(view.viewType,"included");
                    writeView(pFile,view);
                    break;
            }
    }
    fputc('\n',pFile);
	fprintf( stdout, "v\b" ) ;
	usleep( 130000 ) ;
	fflush( stdout ) ;
}

/* -----------------------------------------------------------------------
 * function:	makeContexts
 * arguments:	pFile (IN) pointer to file to wich the contexts are written
 * 		context (IN) a default contextEntry
 * returns:	-
 * purpose:	generates the 3 initial contexts
 * remarks:	-
 * -----------------------------------------------------------------------
 */
void makeContexts(FILE *pFile, contextEntry context)
{
    AuthLevel nAuth;
    for (nAuth = 1; nAuth < 4; nAuth++) {
        switch (nAuth) {
            case guest:    strcpy(context.contextName,"guest");
                           break;
            case operator: strcpy(context.contextName,"operator");
                           break;
            case root:     strcpy(context.contextName,"root");
                           break;
        }
        context.contextIndex = nAuth;
        context.contextViewIndex = nAuth;
        writeContext(pFile,context);
    }
    fputc('\n',pFile);
	fprintf( stdout, "c\b" ) ;
	usleep( 130000 ) ;
	fflush( stdout ) ;
}     

/* -----------------------------------------------------------------------
 * function:	makePartiesNAcls
 * arguments:	pFile (IN) pointer to stream to which the parties & ACL's are written
 * 		partyAgent, partyManager (IN) default partyEntries
 *		aclAgent, aclManager (IN) default aclEntries
 *  		nextPartyIndex (IN) last used indices            
 *		szManagerName (IN) hostname of manager
 * returns:	-
 * purpose:	generates the 3 initial party- & 5 ACL-pairs
 * remarks:	-
 * -----------------------------------------------------------------------
 */
void makePartiesNAcls(FILE *pFile, partyEntry partyAgent, partyEntry partyManager,
    aclEntry aclAgent, aclEntry aclManager, int nextPartyIndex, char *szManagerName)
{
    CommType nComm;
    
    partyManager.partyIndex = (partyAgent.partyIndex = nextPartyIndex+1)+1;
    for (nComm = nextPartyIndex?authenticated:public; nComm < 4; nComm++) {
        switch (nComm) {
            case public:        fputs("%manager <_anonymous_>\n",pFile);
            			strcpy(partyAgent.partyName,"public-");
                                strcpy(partyAgent.partyAuthProtocol,"noAuth");
                                strcpy(partyAgent.partyPrivProtocol,"noPriv");
                                partyAgent.partyType = 1;                 
                                break;
            case authenticated: fprintf(pFile,"%%manager <%s>\n",szManagerName);
                                strcpy(partyAgent.partyName,"authenticated-");
                                strcpy(partyAgent.partyAuthProtocol,"v2md5AuthProtocol");
                                strcpy(partyAgent.partyPrivProtocol,"noPriv");
                                partyAgent.partyType = 13;
                                break;
            case secret:        strcpy(partyAgent.partyName,"secret-");
                                strcpy(partyAgent.partyAuthProtocol,"v2md5AuthProtocol");
                                strcpy(partyAgent.partyPrivProtocol,"desPrivProtocol");
                                partyAgent.partyType = 15;
                                break;
        }
        strcpy(partyManager.partyName,partyAgent.partyName);
        strcpy(partyManager.partyAuthProtocol,partyAgent.partyAuthProtocol);
        strcpy(partyManager.partyPrivProtocol,partyAgent.partyPrivProtocol);
        
        partyManager.partyType = partyAgent.partyType+1;
        
        if (strcmp(partyAgent.partyName,"public-") == 0) {
            strcat(partyAgent.partyName,"agent");
            strcat(partyManager.partyName,"manager");
        } else {
            strcat(partyAgent.partyName,"agent-");
            strcat(partyAgent.partyName,szManagerName);
            strcat(partyManager.partyName,szManagerName);
            strcat(partyManager.partyName,"-agent");
        }
        
        strcpy(partyAgent.partyLocal,"true");
        if (partyManager.IPaddress == partyAgent.IPaddress)
            strcpy(partyManager.partyLocal,"true");
        else
            strcpy(partyManager.partyLocal,"false");
	
	writeParty(pFile,partyAgent);
        writeParty(pFile,partyManager);
        
        partyAgent.partyIndex += 2;
        partyManager.partyIndex += 2;
        
        switch (nComm) {
            case public:        buildAclNames(nComm,guest,szManagerName,
                                    aclAgent.aclName,aclManager.aclName);
                                aclAgent.aclTarget = aclManager.aclSubject = nextPartyIndex+1;
                                aclAgent.aclSubject = aclManager.aclTarget = nextPartyIndex+2;
                                aclAgent.aclResources = aclManager.aclResources = 1;
                                strcpy(aclAgent.aclPrivileges,"gnb");
                                strcpy(aclManager.aclPrivileges,"r");
                                writeAcl(pFile,aclAgent); 
                                writeAcl(pFile,aclManager);
                                break;
            case authenticated: buildAclNames(nComm,operator,szManagerName,
                                    aclAgent.aclName,aclManager.aclName);
                                aclAgent.aclTarget = aclManager.aclSubject = nextPartyIndex+3;
                                aclAgent.aclSubject = aclManager.aclTarget = nextPartyIndex+4;
                                aclAgent.aclResources = aclManager.aclResources = 2;
                                strcpy(aclAgent.aclPrivileges,"gnb");
                                strcpy(aclManager.aclPrivileges,"ru");
                                writeAcl(pFile,aclAgent); 
                                writeAcl(pFile,aclManager);
                                buildAclNames(nComm,root,szManagerName,
                                    aclAgent.aclName,aclManager.aclName);
                                aclAgent.aclResources = aclManager.aclResources = 3;
                                strcpy(aclAgent.aclPrivileges,"gnbs");
                                strcpy(aclManager.aclPrivileges,"ru");
                                writeAcl(pFile,aclAgent); 
                                writeAcl(pFile,aclManager);
                                break;
            case secret:        buildAclNames(nComm,operator,szManagerName,
                                    aclAgent.aclName,aclManager.aclName);
                                aclAgent.aclTarget = aclManager.aclSubject = nextPartyIndex+5;
                                aclAgent.aclSubject = aclManager.aclTarget = nextPartyIndex+6;
                                aclAgent.aclResources = aclManager.aclResources = 2;
                                strcpy(aclAgent.aclPrivileges,"gnb");
                                strcpy(aclManager.aclPrivileges,"ru");
                                writeAcl(pFile,aclAgent);
                                writeAcl(pFile,aclManager);
                                buildAclNames(nComm,root,szManagerName,
                                    aclAgent.aclName,aclManager.aclName);
                                aclAgent.aclResources = aclManager.aclResources = 3;
                                strcpy(aclAgent.aclPrivileges,"gnbs");
                                strcpy(aclManager.aclPrivileges,"ru");
                                writeAcl(pFile,aclAgent); 
                                writeAcl(pFile,aclManager);
                                break;
        }
        fputc('\n',pFile);
    }
}

/* -----------------------------------------------------------------------
 * function:	main 
 * arguments:	-
 * returns:	-
 * purpose:	initialisation *.perk-files
 * remarks:	- 
 * -----------------------------------------------------------------------
 */
void main()
{
    FILE *pAgentFile, *pManagerFile;
    char szManagerName[ARLEN], szAgentName[ARLEN];
    char szAgentFilename[FILENAMELEN] = "\0", szManagerFilename[FILENAMELEN] = "\0";
    
    partyEntry partyAgent, partyManager;  
    contextEntry context;
    viewEntry view;
    aclEntry aclAgent, aclManager;
    
    int nextPartyIndex;
   
    InitContextRecord(&context);
    InitViewRecord(&view);
    InitAclRecord(&aclAgent);
    InitAclRecord(&aclManager);
    InitPartyRecord(&partyAgent);
    InitPartyRecord(&partyManager);
    
    /* user input */
    UINintro( 	szManagerName, partyManager.IPaddress, szAgentName,
				partyAgent.IPaddress, partyAgent.partyMaxMessageSize,
				partyManager.partyTAddress, partyAgent.partyTAddress ) ;
    
	fprintf( stdout, "\n     Status: [ ] (creating acl-, view-, party- or contextrecord)" );
	fprintf( stdout, "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b" );
	fflush( stdout ) ;
    strcpy(partyManager.partyMaxMessageSize,partyAgent.partyMaxMessageSize);
    
    /* initialPartyId '.' <agent IP address> */   
    strcpy(partyAgent.partyIdentity,InitialPartyId);
    strcat(partyAgent.partyIdentity,partyAgent.IPaddress);
    strcpy(partyManager.partyIdentity,InitialPartyId);
    strcat(partyManager.partyIdentity,partyAgent.IPaddress);
    
    strcat(context.contextIdentity,partyAgent.IPaddress);
    
    /* building filenames */
    if (getenv("SNMP_PARTYMIB_PATH") != (char *)NULL) {
        strcat(szAgentFilename,getenv("SNMP_PARTYMIB_PATH"));
        strcat(szManagerFilename,getenv("SNMP_PARTYMIB_PATH"));
    }
    strcat(szAgentFilename,szAgentName);
    strcat(szAgentFilename,"_agent.perkins.conf");
    strcat(szManagerFilename,szAgentName);
    strcat(szManagerFilename,"_manager_");
    strcat(szManagerFilename,szManagerName);
    strcat(szManagerFilename,".perkins.conf");
    
    /* opening files */
    if ((pManagerFile = fopen(szManagerFilename,"wt")) == (FILE *)NULL) {
        fprintf(stderr,"%s, %d: could not create \"%s\"...\n",__FILE__,__LINE__,szAgentFilename);
        exit(-1);
    }
      
    if ((pAgentFile = fopen(szAgentFilename,"rt")) != (FILE *)NULL) {    
        
        /* add manager to existing agent-file */
        getNextIndex(pAgentFile,&nextPartyIndex);
        fclose(pAgentFile);
        pAgentFile = fopen(szAgentFilename,"at");
        makePartiesNAcls(pAgentFile,partyAgent,partyManager,aclAgent,aclManager,
            nextPartyIndex,szManagerName);
    } else 
        if ((pAgentFile = fopen(szAgentFilename,"wt")) != (FILE *)NULL) {
    
            /* create new agent-file */
            addHeader(pAgentFile,AgentHeader);
            nextPartyIndex = 0;
            fprintf(pAgentFile,"%%global <%s>\n",szAgentName);
            makeViews(pAgentFile,view);
            makeContexts(pAgentFile,context);
            makePartiesNAcls(pAgentFile,partyAgent,partyManager,aclAgent,aclManager,
                nextPartyIndex,szManagerName);
        } else {            
            fprintf(stderr,"%s, %d: could not create \"%s\"...\n",__FILE__,__LINE__,szAgentFilename);
            exit(-1);
        }
    /* creating manager-file */
    addHeader(pManagerFile,AgentHeader);
    nextPartyIndex = 0;
    fprintf(pManagerFile,"%%global <%s>\n",szAgentName);
    makeViews(pManagerFile,view);
    makeContexts(pManagerFile,context);
    makePartiesNAcls(pManagerFile,partyAgent,partyManager,aclAgent,aclManager,
        nextPartyIndex,szManagerName);
    
	fprintf( stdout, 
"\n\n************************************************************************\n\n" ) ;
    fclose(pAgentFile);
    fclose(pManagerFile);
}
