 /* -----------------------------------------------------------------------
 *
 * Project:	UT-SNMP
 *
 * Filename:    spmPerkins.c
 *
 * Date:        September 1994
 *
 * Purpose:     This file contains the primairy partymib functions.
 *
 * Remarks:	This file is an update of spmPatyMIB.c
 *		New is the format of the initialisation-files. The new
 *		format was propused by Perkins. 
 *
 * Author( s):  (Koert Flapper) Erwin Schrijver
 *
 * -----------------------------------------------------------------------
 */

#include "spmPerkins.h"

int	psize, pradix, csize, cradix;
FoidT	*pFoid;

/*
 * -----------------------------------------------------------------------
 * Function:	SpmInitPartyMIB
 * Arguments:	none.
 * Purpose:     Initializes the Party MIB.		
 * Remarks:
 * -----------------------------------------------------------------------
 */
	STATUS
SpmInitPartyMIB () 
{
	if (MibInit( partymibtab, partymsize, partytabtab, partytsize) == OK)
		return(PartyMibOpen());
	return(ERROR);
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:    OpenPartyFile
 *
 * Arguments:   filname, filename .
 *
 * Purpose:     check if filename exists, and if it an regular file.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS
OpenPartyFile( char *filname, char *type, FILE **fp )
{
	struct stat buf;

	if( stat( filname, &buf ) == 0 ) {
		if( S_ISREG( buf.st_mode ) ) {
			*fp = fopen( filname, type );
			return(OK);
		}
	}
	return( FAILED );
}


				
/* 
 * -----------------------------------------------------------------------
 *
 * Function:	PartyMibOpen
 *
 * Arguments:	none.
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     Initializes the Party MIB.		
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS
PartyMibOpen()
{
	FILE	*pfFile;
	char	szLine[BUFSIZE], szEntry[BUFSIZE], szFile[MAXFILENAME];
	char	*cp;
	AsnInt	blank; 	
	
	strncpy( szFile, getenv( "SNMP_PARTYMIB_PATH" ), MAXFILENAME-1);
	strncat( szFile, CONFFILE, MAXFILENAME - strlen( szFile ) -1 );
	
	if (OpenPartyFile( szFile, "r", &pfFile ) != OK){
		Error( "Unable to open perkins.conf" );
		return ( ERROR );
	}
	
	if (foidOpen( &pFoid ) == ERROR)
		return ( ERROR );
		
	while( fgets( szLine, BUFSIZE, pfFile )) {
	
		while( 1 ) {
			if (strlen( szLine ) > 250 ){
				Error( "Line longer than 250 bytes" );
				fclose( pfFile );
				return ( ERROR );
			}
			
			if( (szLine[0] == '#') || (szLine[0] == ';') ||
			    (!strncmp( szLine, "//", 2 )) )
			    	break;
		
			blank = TRUE;
			for( cp = szLine ; *cp ; cp++ )
				if (!isspace( *cp )) {
					blank = FALSE;
					break;
				}
			if (blank) break;
			
			if (sscanf( szLine, "%%%s", szEntry ) != 1)
				break;
				
			if (!strncmp( szLine, "%view", 4 )) {
				if (LoadViewSection( pfFile, szLine ) != OK) {
					fclose ( pfFile );
					return ( ERROR );
				}
			}
			else if (!strncmp( szLine, "%acl", 3 )) {
				if (LoadAclSection( pfFile, szLine ) != OK) {
					fclose( pfFile );
					return ( ERROR );
				}
			}
			else if (!strncmp( szLine, "%party", 5 )) {
				if (LoadPartySection( pfFile, szLine ) != OK) {
					fclose( pfFile );
					return ( ERROR );
				}
			}
			else if (!strncmp( szLine, "%context", 7 )) {
				if (LoadContextSection( pfFile, szLine ) != OK) {
					fclose( pfFile );
					return ( ERROR );
				}
			}
			else break;
		}
	}
	fclose( pfFile );

	PrintPartyMIB();
	SavePartyMIB();
	
	foidClose( &pFoid );
	
	return ( OK );	
}	



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	PrintPartyMIB
 *
 * Arguments:	none.
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     Prints the Party MIB.		
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS
PrintPartyMIB()
{
	PartyEntry	*pp;
	ContextEntry	*cp;
	AclEntry	*ap;
	ViewEntry	*vp;

	fprintf(stderr, "The partytable looks like this\n");
	fprintf(stderr, "Ind\t Name\t\t\t   Local TAddress\t\tID\n");
	pp = PartyMIB.parties;
	while (pp) {
		PrintPartyEntry( pp);
		pp = pp->next;
		}
	fprintf(stderr, "\n");

	fprintf(stderr, "The contexttable looks like this\n");
	fprintf(stderr, "Ind\t Name\t\t\t  Local ID\n");
	cp = PartyMIB.contexts;
	while (cp) { 
		PrintContextEntry( cp);
		cp = cp->next;
		}
	fprintf(stderr, "\n");

	fprintf(stderr, "The acltable looks like this\n");
	fprintf(stderr, "TInd\t SInd\t RInd\t Privileges\n");
	ap = PartyMIB.acls;
	while( ap) {
		PrintAclEntry( ap);
	      	ap = ap->next;
	}
	fprintf(stderr, "\n");


	fprintf(stderr, "The viewtable looks like this\n");
	fprintf(stderr, "Ind\t Subtree \t\t     Type Mask\n");
	vp = PartyMIB.views;
	while( vp) { 
		PrintViewEntry( vp);
	      	vp = vp->next;
	}
	fprintf(stderr, "\n");
	return(OK);
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	SavePartyMIB
 *
 * Arguments:	none.
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     Saves the Party MIB.		
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS
SavePartyMIB()
{
	FILE 		*pfSave;
	char		szFile[MAXFILENAME];
	PartyEntry	*pp;
	ContextEntry	*cp;
	AclEntry	*ap;
	ViewEntry	*vp;

	strncpy( szFile, getenv( "SNMP_PARTYMIB_PATH" ), MAXFILENAME-1);
	strncat( szFile, CONFFILE, MAXFILENAME - strlen( szFile ) -1 );
	
	if( (pfSave = fopen( szFile, "w" )) == NULL){
		Error( "Unable to open perkins.temp" );
		return ( ERROR );
	}	

	vp = PartyMIB.views;
	while( vp) { 
		SaveViewEntry( vp, pfSave);
	      	vp = vp->next;
	}

	cp = PartyMIB.contexts;
	while (cp) { 
		SaveContextEntry( cp, pfSave );
		cp = cp->next;
		}

	pp = PartyMIB.parties;
	while (pp) {
		SavePartyEntry( pp, pfSave );
		pp = pp->next;
		}

	ap = PartyMIB.acls;
	while( ap) {
		SaveAclEntry( ap, pfSave );
	      	ap = ap->next;
	}

	fclose( pfSave );
	return(OK);
}

/* 
 * -----------------------------------------------------------------------
 *
 * Function:	LoadPartySection
 *
 * Arguments:	*pfFile, pointer to input-stream (IN).
 *		*szLine, string containing first line of party-section
 *
 * Returns:	STATUS, which is OK on success
 *
 * Purpose:     This function loads the context section.		
 *
 * Remarks:	
 *
 * -----------------------------------------------------------------------
 */
	STATUS
LoadPartySection( FILE *pfFile, char *szLine )
{
	PartyEntry	*party, pTemp;
	char    	szEntry[BUFSIZE], szValue[BUFSIZE],
			hostname[MAXHOSTNAMELEN];
	char		*cp;
    	AsnInt 		blank;
    	UAsnInt		tmpint;
    	STATUS		nResult;
        u_long 		myaddr, addr, byte;
        u_short		port;
   	struct hostent  *he=NULL;
   	
	psize = 101;
	pradix = 21;
	phashtab = (struct Hashnode **)malloc(psize *sizeof(struct Hashnode *));
	hashclear(phashtab, psize);
	
	pTemp.partyIdentity		= NULL;
	pTemp.partyIndex		= 0;
	pTemp.partyName			= NULL;
	pTemp.partyTDomain		= NULL;
	pTemp.partyTAddress		= NULL;
	pTemp.partyMaxMessageSize	= 484;
	pTemp.partyLocal		= UNDEF;
	pTemp.partyAuthProtocol		= NULL;
	pTemp.partyAuthClock		= 0;
	pTemp.partyAuthPrivate		= NULL;
	pTemp.partyAuthPublic		= NULL;
	pTemp.partyAuthLifetime		= 0;
	pTemp.partyPrivProtocol		= NULL;
	pTemp.partyPrivPrivate		= NULL;
	pTemp.partyPrivPublic		= NULL;
	pTemp.partyCloneFrom		= NULL;
	pTemp.partyStorageType		= Volatile;
	pTemp.partyStatus		= active;
	
	if (sscanf( szLine+1, "%*s <%s>", szEntry ) == 1){
		pTemp.partyName = (char *) malloc( strlen( szEntry ) );
		strncpy( pTemp.partyName, szEntry, strlen( szEntry )-1 );
		strncpy( pTemp.partyName + strlen( szEntry )-1, "\0", 1);
	}
		
    	while (fgets(szLine, BUFSIZE, pfFile)) {
  		if (szLine[0] == '%') break;  	
		
		if (strlen( szLine ) > 250 ){
			Error( "Line longer than 250 bytes" );
			FreeParty( &pTemp );
			return ( ERROR );
		}
		
		if( (szLine[0] == '#') || (szLine[0] == ';') ||
		    (!strncmp( szLine, "//", 2 )) )
		    	continue;
		
		blank = TRUE;
		for( cp = szLine ; *cp ; cp++ )
			if (!isspace( *cp )) {
				blank = FALSE;
				break;
			}
		if (blank) continue;
			
        	if (sscanf(szLine, " %s = %s", szEntry, szValue ) != 2) {
            		Error( "Bad party parse" );
			FreeParty( &pTemp );
			return ERROR ;
		}
			
		if (!strcmp( szEntry, "partyIdentity" ))
			nResult = LoadPartyEntry( PARTYIDENTITY, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyIndex" ))
			nResult = LoadPartyEntry( PARTYINDEX, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyTDomain" ))
			nResult = LoadPartyEntry( PARTYTDOMAIN, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyTAddress" )) 
			nResult = LoadPartyEntry( PARTYTADDRESS, szLine, &pTemp );
		else if (!strcmp( szEntry, "partyMaxMessageSize" ))
			nResult = LoadPartyEntry( PARTYMAXMESSAGESIZE, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyLocal" ))
			nResult = LoadPartyEntry( PARTYLOCAL, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyAuthProtocol" ))
			nResult = LoadPartyEntry( PARTYAUTHPROTOCOL, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyAuthClock" ))
			nResult = LoadPartyEntry( PARTYAUTHCLOCK, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyAuthPrivate" ))
			nResult = LoadPartyEntry( PARTYAUTHPRIVATE, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyAuthPublic" ))
			nResult = LoadPartyEntry( PARTYAUTHPUBLIC, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyAuthLifetime" ))
			nResult = LoadPartyEntry( PARTYAUTHLIFETIME, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyPrivProtocol" ))
			nResult = LoadPartyEntry( PARTYPRIVPROTOCOL, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyPrivPrivate" ))
			nResult = LoadPartyEntry( PARTYPRIVPRIVATE, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyPrivPublic" ))
			nResult = LoadPartyEntry( PARTYPRIVPUBLIC, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyStorageType" ))
			nResult = LoadPartyEntry( PARTYSTORAGETYPE, szValue, &pTemp );
		else if (!strcmp( szEntry, "partyStatus" ))
			nResult = LoadPartyEntry( PARTYSTATUS, szValue, &pTemp );
		else {
			Error( "Bad party parse" );
			FreeParty( &pTemp );
			return ( ERROR );
		}
			
		if (nResult != OK) {
			FreeParty( &pTemp );
			return ( nResult );
		}
	}
		
	if( (pTemp.partyIdentity == NULL) || (pTemp.partyIndex == 0) ||
	    (pTemp.partyTDomain == NULL) || (pTemp.partyTAddress == NULL) ||
	    (pTemp.partyAuthProtocol == NULL) ){
		Error( "Party-section should contain partyIdentity, partyIndex, partyTDomain,\n"
		       "partyTAddress and partyAuthProtocol" );
		FreeParty( &pTemp );
		return ( ERROR );
	}

	if ( gethostname(hostname, MAXHOSTNAMELEN) == -1) {
		Error( "Could not get hostname" );
	  	return ERROR ;
	}
	
	if (GetPartyIndex( pTemp.partyIdentity, &tmpint ) == OK) {
		Error( "Party-identity already in use" );
		FreeParty( &pTemp );
		return ( ERROR );
	}
	
	if (CreatePartyEntry( pTemp.partyIdentity, &party ) != OK) {
		Error( "Failed to create new party" );
		FreeParty( &pTemp );
		return ( ERROR );
	}
	
	party->partyName		= pTemp.partyName;
	party->partyIdentity		= pTemp.partyIdentity;
	party->partyIndex		= pTemp.partyIndex;
	party->partyTDomain		= pTemp.partyTDomain;
	party->partyTAddress		= pTemp.partyTAddress;
	party->partyMaxMessageSize 	= pTemp.partyMaxMessageSize;
	party->partyLocal		= pTemp.partyLocal;
	party->partyAuthProtocol	= pTemp.partyAuthProtocol;
	party->partyAuthClock		= pTemp.partyAuthClock;
	party->partyAuthPrivate		= pTemp.partyAuthPrivate;
	party->partyAuthPublic		= pTemp.partyAuthPublic;
	party->partyAuthLifetime	= pTemp.partyAuthLifetime;
	party->partyPrivProtocol	= pTemp.partyPrivProtocol;
	party->partyPrivPrivate		= pTemp.partyPrivPrivate;
	party->partyPrivPublic		= pTemp.partyPrivPublic;
	party->partyCloneFrom		= pTemp.partyCloneFrom;
	party->partyStorageType		= pTemp.partyStorageType;
	party->partyStatus		= pTemp.partyStatus;
					
	if (party->partyPrivProtocol == NULL) {
		party->partyPrivProtocol = pNewOid( noPriv.octetLen);
		party->partyPrivProtocol->octetLen = noPriv.octetLen;
		memcpy( party->partyPrivProtocol->octs, noPriv.octs,
					party->partyPrivProtocol->octetLen );
	}

	he = gethostbyname(hostname);
	if (he != (struct hostent *)0)
		myaddr = inet_addr( inet_ntoa( *((struct in_addr *)(he->h_addr)) ) );
	
	myaddr = htonl( myaddr );
	if (party->partyLocal == UNDEF) {
		if( ( ( memcmp( party->partyTDomain->octs, snmpUDPDomain.octs, 
				      snmpUDPDomain.octetLen ) == OK ) || 
		( memcmp( party->partyTDomain->octs, snmpTCPDomain.octs, 
				      snmpUDPDomain.octetLen ) == OK ) )  && 
		( !memcmp( &myaddr, party->partyTAddress->octs, 4 ) ) )
				party->partyLocal = true;
		    	else
				party->partyLocal = false;
	}
		  	
    	return(OK);
}

/* 
 * -----------------------------------------------------------------------
 *
 * Function:	LoadPartyEntry
 *
 * Arguments:	nEntry,	number to object to be added (IN).
 *		*szValue, string containing value belonging to object (IN).
 *		*cTemp,	pointer to temporary-party entry.
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     This funtcion translates szValue in the appropriate field.		
 *
 * Remarks:	none.
 *
 * -----------------------------------------------------------------------
 */
	STATUS
LoadPartyEntry( AsnInt nEntry, char *szValue, PartyEntry *pTemp )
{	
	char	*cp, *ucp;
	char	szAddress[BUFSIZE], szPort[BUFSIZE];
	AsnOcts	*octFriendly, *octOID;
	u_long	addr, byte;
	u_short	port;
	AsnInt	nonhex, nDesc, nError;
		
	switch (nEntry) {
		case PARTYIDENTITY :
			octFriendly 		= pNewOct( BUFSIZE );
			octFriendly->octetLen 	= BUFSIZE-1;
			octOID			= pNewOct( BUFSIZE );
			octOID->octetLen 	= BUFSIZE-1;
			switch (foidConv( pFoid, szValue, *octFriendly,
					  *octOID, &nDesc, &nError )) {
				case ERROR :
					FreeOct( octFriendly );
					FreeOct( octOID );
					return ( ERROR );
					break;
				case FAILED :
					if (nError == ILLEGAL) {
						Error( "Bad party object ID" );
						FreeOct( octFriendly );
						FreeOct( octOID );
						return ( ERROR );
					}
					break;
				default :
					break;
			}
		   	pTemp->partyIdentity = pNewOid(strlen(octOID->octs)) ; 
			if( ASCIItoOID(octOID->octs, pTemp->partyIdentity) != OK ) {
				Error( "Bad party object ID" );
				FreeOid( pTemp->partyIdentity);
				FreeOct( octFriendly );
				FreeOct( octOID );
				pTemp->partyIdentity = NULL;
				return ERROR ;
		   	}
			FreeOct( octFriendly );
			FreeOct( octOID );
		    	break;
		case PARTYINDEX :
			for( cp = szValue; *cp; cp++ )
                		if (!isdigit(*cp)) {
                    			Error( "Party-index not an integer" );
					return ERROR ;
				}
            		pTemp->partyIndex = atoi( szValue );
            		break;
            	case PARTYTDOMAIN :
			if( ! (int) strcasecmp( szValue, "snmpUdpDomain") ) {
				pTemp->partyTDomain = pNewOid( snmpUDPDomain.octetLen);
				pTemp->partyTDomain->octetLen = snmpUDPDomain.octetLen;
				memcpy( pTemp->partyTDomain->octs, snmpUDPDomain.octs,
					pTemp->partyTDomain->octetLen);
			} else if (!strcasecmp( szValue, "snmpTcpDomain") ) {
				pTemp->partyTDomain = pNewOid( snmpTCPDomain.octetLen );
				pTemp->partyTDomain->octetLen = snmpTCPDomain.octetLen;
				memcpy( pTemp->partyTDomain->octs, snmpTCPDomain.octs,
					pTemp->partyTDomain->octetLen);
			} else if (!strcasecmp( szValue, "rfc1157Domain") ) {
				pTemp->partyTDomain = pNewOid( rfc1157Domain.octetLen);
				pTemp->partyTDomain->octetLen = rfc1157Domain.octetLen;
				memcpy( pTemp->partyTDomain->octs, rfc1157Domain.octs,
					pTemp->partyTDomain->octetLen);
			} else	{
				Error( "Bad domain type" );
				return ERROR ;
			}
			break;
		case PARTYTADDRESS :
			if (sscanf( szValue, " %*s = %s %s", szAddress, szPort ) != 2)
				return ( ERROR );
			pTemp->partyTAddress = pNewOct(6);
			pTemp->partyTAddress->octetLen = 6;
		    	if( (addr = inet_addr( szAddress )) == -1 ) {
				Error( "Bad IP address" );
				FreeOct( pTemp->partyTAddress );
				pTemp->partyTAddress = NULL;
				return ( ERROR );
			}
			addr = htonl(addr);
			memcpy( pTemp->partyTAddress->octs, &addr, 4 );
		    	for( cp = szPort; *cp; cp++ )
				if( !isdigit( *cp ) ) {
					Error( "Not a port number" );
					return ERROR ;
				}
			port = htons( atoi( szPort ));
			memcpy( pTemp->partyTAddress->octs + 4, &port, 2 );
		    	break;
		case PARTYMAXMESSAGESIZE :
		    	for( cp = szValue; *cp; cp++ )
				if( !isdigit(*cp) ) {
					Error( "Bad Max Message Size value" );
					return ERROR ;
				}
		    	pTemp->partyMaxMessageSize = atoi( szValue );
		    	if( ( pTemp->partyMaxMessageSize < MINMAXMS ) ||
		    	    ( pTemp->partyMaxMessageSize > MAXMAXMS ) ) {
				Error( "Bad Max Message Size value" );
				return ERROR ;
			}
			break;
		case PARTYLOCAL :
			if ( !strncmp( szValue, "true", 4 ))
				pTemp->partyLocal = true;
			else if (!strncmp( szValue, "false", 5 ))
				pTemp->partyLocal = false;
			else {
				Error( "Bad partyLocal value" );
				return ( ERROR );
			}
			break;
		case PARTYAUTHPROTOCOL :		
		    	if( !strcasecmp( szValue, "noAuth") ) {
				pTemp->partyAuthProtocol = pNewOid( noAuth.octetLen);
				pTemp->partyAuthProtocol->octetLen = noAuth.octetLen;
				memcpy( pTemp->partyAuthProtocol->octs, 
					noAuth.octs,
					pTemp->partyAuthProtocol->octetLen );
		    	} else if( !strcasecmp(szValue, "v2md5AuthProtocol")) {
				pTemp->partyAuthProtocol = pNewOid( v2md5AuthProtocol.octetLen);
				pTemp->partyAuthProtocol->octetLen = v2md5AuthProtocol.octetLen;
				memcpy( pTemp->partyAuthProtocol->octs, 
					v2md5AuthProtocol.octs,
					pTemp->partyAuthProtocol->octetLen );
		    	} else if( !strcasecmp( szValue, "rfc1157noAuth") ) {
				pTemp->partyAuthProtocol = pNewOid( rfc1157noAuth.octetLen);
				pTemp->partyAuthProtocol->octetLen = rfc1157noAuth.octetLen;
				memcpy( pTemp->partyAuthProtocol->octs, 
					rfc1157noAuth.octs,
					pTemp->partyAuthProtocol->octetLen );
			} else	{
				Error( "Bad authentication protocol type" );
				return ERROR ;
			}
			break;
		case PARTYAUTHCLOCK :
		    	for( cp = szValue; *cp; cp++ ) 
				if( !isxdigit(*cp) ) {
			    		Error( "Bad clock value" );
					return ERROR ;
				}
		    	if( sscanf( szValue, "%x", &pTemp->partyAuthClock ) != 1 ) {
				Error( "Bad clock value" );
				return ERROR ;
			}
		    	break;
		case PARTYAUTHPRIVATE :
			if (CMPOID( pTemp->partyAuthProtocol, &rfc1157noAuth) == EQUAL) {
				pTemp->partyAuthPrivate= pNewOct(strlen(szValue));
				pTemp->partyAuthPrivate->octetLen = strlen(szValue);
				memcpy(pTemp->partyAuthPrivate->octs, szValue,
					pTemp->partyAuthPrivate->octetLen);
			} else {
		    		if( strlen(szValue) != 32 ) {
					Error( "Bad private authentication key value" );
					return ERROR ;
				}
			    	for(cp = szValue; *cp; cp++)
					if (!isxdigit(*cp)) {
			    			Error( "Bad private authentication key value" );
						return ERROR ;
					}
		    		ucp = ( char *) ((pTemp->partyAuthPrivate=(AsnOcts *)pNewOct(KEYLEN))->octs);
		    		for( cp = szValue; *cp; cp += 2, ucp++ ) {
					if( sscanf( cp, "%2x", &byte ) != 1 ) {
			    			Error( "Bad private authentication key value" );
						return ERROR ;
					}
					*ucp = byte;
			     	}
				pTemp->partyAuthPrivate->octetLen = (AsnInt) ucp - (AsnInt)
					pTemp->partyAuthPrivate->octs;
			}
			break;
		case PARTYAUTHPUBLIC :
			if (CMPOID( pTemp->partyAuthProtocol, &rfc1157noAuth) != EQUAL) {		
				nonhex = 0;
				for( cp = szValue; ((*cp) && (!(nonhex))); cp++ ) {
					if( !isxdigit( *cp ) )
				    		nonhex = 1;
			    	}
		    		if( nonhex ) {
			    		Error( "Bad public authentication key value" );
					return ERROR ;
			    	} else {
					if( (strlen(szValue) % 2) || 
					    (strlen(szValue) > 32) ) {
			    			Error( "Bad public authentication key value" );
						return ERROR ;
					}
					ucp = (char *) ((pTemp->partyAuthPublic=
					      (AsnOcts *)pNewOct(KEYLEN))->octs);
					for( cp = szValue; *cp; cp += 2, ucp++ ) {
			    			if( sscanf( cp, "%2x", &byte) != 1 ) {
							Error( "Bad public authentication key value" );
							return ERROR ;
						}
			    			*ucp = byte;
					}
					pTemp->partyAuthPublic->octetLen = (AsnInt) ucp - (AsnInt)
						pTemp->partyAuthPublic->octs;
				}
			}
		    	break;
		case PARTYAUTHLIFETIME :		
		    	for( cp = szValue; *cp; cp++ )
				if( !isdigit( *cp ) ) {
			    		Error( "Bad lifetime value" );
					return ERROR ;
				}
		    	pTemp->partyAuthLifetime = atoi( szValue );
		    	if( ( pTemp->partyAuthLifetime < 0 ) || 
		    	    ( pTemp->partyAuthLifetime > MAXLIFETIME) ) {
				Error("Bad auth lifetime value" );
				return ERROR ;
			}
			break;
		case PARTYPRIVPROTOCOL :
		    	if( !strcasecmp( szValue, "noPriv" ) ) {
				pTemp->partyPrivProtocol = pNewOid( noPriv.octetLen);
				pTemp->partyPrivProtocol->octetLen = noPriv.octetLen;
				memcpy( pTemp->partyPrivProtocol->octs, noPriv.octs,
					pTemp->partyPrivProtocol->octetLen );
		    	} else 	if( !strcasecmp( szValue, "desPrivProtocol" ) ) {
				pTemp->partyPrivProtocol = pNewOid( desPrivProtocol.octetLen);
				pTemp->partyPrivProtocol->octetLen = desPrivProtocol.octetLen;
				memcpy( pTemp->partyPrivProtocol->octs, desPrivProtocol.octs,
					pTemp->partyPrivProtocol->octetLen );
		    	} else {
				Error( "Bad privacy protocol type" );
				return ERROR ;
			}
		    	break;
		case PARTYPRIVPRIVATE :
		    	if( strlen(szValue) != 32 ) {
				Error( "Bad private privacy key value" );
				return ERROR ;
			}
			for( cp = szValue; *cp; cp++ ) {
				if( !isxdigit(*cp) ) {
			    		Error( "Bad private privacy key value" );
					return ERROR ;
				}
		    	}
		    	ucp = ( char * ) (pTemp->partyPrivPrivate= pNewOct(KEYLEN) )->octs;
		    	for( cp = szValue; *cp; cp += 2, ucp++) {
				if( sscanf(cp, "%2x", &byte ) != 1 ) {
			    		Error( "Bad private privacy key value" );
					return ERROR ;
				}
				*ucp = byte;
		    	}
			pTemp->partyPrivPrivate->octetLen = (AsnInt) ucp -
				(AsnInt)pTemp->partyPrivPrivate->octs;
			break;
		case PARTYPRIVPUBLIC :
		    	nonhex = 0;	
		    	for(cp = szValue; ((*cp) && (!(nonhex))); cp++) {
				if( !isxdigit( *cp ) )
			    		nonhex = 1;
		    	}
		    	if( nonhex ) {
				Error( "Bad public privacy key value" );
				return ERROR ;
			} else {
		        	if( (strlen(szValue) % 2) || 
		        	    (strlen(szValue) > 32) ) {
			    		Error( "Bad public privacy key value" );
					return ERROR ;
				}
				ucp = (char *) ((pTemp->partyPrivPublic=(AsnOcts *)pNewOct(KEYLEN))->octs) ;
				for( cp = szValue; *cp; cp += 2, ucp++ ) {
			    		if (sscanf(cp, "%2x", &byte) != 1) {
						Error( "Bad public privacy key value" );
						return ERROR ;
					}
				    	*ucp = byte;
				}
       		         	pTemp->partyPrivPublic->octetLen = (AsnInt)ucp -
					(AsnInt)pTemp->partyPrivPublic->octs;
	    		}
			break;
		case PARTYSTORAGETYPE :
			/* value is ignored */
			break;
		case PARTYSTATUS :
			if (!strncmp( szValue, "active", 6 ))
				pTemp->partyStatus = active;
			else if (!strncmp( szValue, "notInService", 7 ))
				pTemp->partyStatus = notReady;
			else {
				Error( "Bad party status value" );
				return ( ERROR );
			}
			break;
		default:
			Error( "unknown state" );
			return ERROR ;
    	}
    	return( OK );
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	LoadContextSection
 *
 * Arguments:	*pfFile, pointer to input-stream (IN).
 *		*szLine, string containing first line of context-section
 *
 * Returns:	STATUS, which is OK on success
 *
 * Purpose:     This function loads the context section.		
 *
 * Remarks:	Problems on deciding upon value of contextLocal
 *
 * -----------------------------------------------------------------------
 */
	STATUS
LoadContextSection( FILE *pfFile, char *szLine )
{
	ContextEntry	*context, cTemp;
	char    	szEntry[BUFSIZE], szValue[BUFSIZE],
			hostname[MAXHOSTNAMELEN];
	char		*cp;
    	AsnInt 		blank;
    	UAsnInt		tmpint;
    	STATUS		nResult;
        u_long 		myaddr, addr;
   	struct hostent  *he;
	
	csize = 101;
	cradix = 13;
	chashtab = (struct Hashnode **)malloc( csize * sizeof( struct Hashnode *));
	hashclear( chashtab, csize );
		
	cTemp.contextIndex 		= 0;
	cTemp.contextName 		= NULL;
	cTemp.contextIdentity		= NULL;
	cTemp.contextLocal		= UNDEF;
	cTemp.contextViewIndex 		= 0;
	cTemp.contextLocalEntity 	= NULL;
	cTemp.contextLocalTime		= NULL;
	cTemp.contextProxyDstParty 	= NULL;
	cTemp.contextProxySrcParty 	= NULL;
	cTemp.contextProxyContext 	= NULL;
	cTemp.contextStorageType 	= Volatile;
	cTemp.contextStatus		= active;
	
	if (sscanf( szLine+1, "%*s <%s>", szEntry ) == 1){
		cTemp.contextName = (char *) malloc( strlen( szEntry ) );
		strncpy( cTemp.contextName, szEntry, strlen( szEntry )-1 );
		strncpy( cTemp.contextName + strlen( szEntry )-1, "\0", 1);
	}
		
    	while (fgets(szLine, BUFSIZE, pfFile)) {
    		if (szLine[0] == '%') break;
    		
		if (strlen( szLine ) > 250 ){
			Error( "Line longer than 250 bytes" );
			FreeContext( &cTemp );
			return ( ERROR );
		}
			
		if( (szLine[0] == '#') || (szLine[0] == ';') ||
		    (!strncmp( szLine, "//", 2 )) ) 
		    	continue;
		
		blank = TRUE;
		for( cp = szLine ; *cp ; cp++ )
			if (!isspace( *cp )) {
				blank = FALSE;
				break;
			}
		if (blank) continue;
			
       		if (sscanf(szLine, " %s = %s", szEntry, szValue ) != 2) {
       			Error( "Bad context parse" );
            		FreeContext( &cTemp );
			return ERROR ;
		}
			
		if (!strcmp( szEntry, "contextIdentity"))
			nResult = LoadContextEntry( CONTEXTIDENTITY, szValue, &cTemp );
		else if (!strcmp( szEntry, "contextIndex" ))
			nResult = LoadContextEntry( CONTEXTINDEX, szValue, &cTemp );
		else if (!strcmp( szEntry, "contextLocal" ))
			nResult = LoadContextEntry( CONTEXTLOCAL, szValue, &cTemp );
		else if (!strcmp( szEntry, "contextViewIndex" ))
			nResult = LoadContextEntry( CONTEXTVIEWINDEX, szValue, &cTemp );
		else if (!strcmp( szEntry, "contextLocalEntity" ))
			nResult = LoadContextEntry( CONTEXTLOCALENTITY, szValue, &cTemp );
		else if (!strcmp( szEntry, "contextLocalTime" ))
			nResult = LoadContextEntry( CONTEXTLOCALTIME, szValue, &cTemp );
		else if (!strcmp( szEntry, "contextProxyDstParty" ))
			nResult = LoadContextEntry( CONTEXTPROXYDSTPARTY, szValue, &cTemp );
		else if (!strcmp( szEntry, "contextProxySrcParty" ))
			nResult = LoadContextEntry( CONTEXTPROXYSRCPARTY, szValue, &cTemp );
		else if (!strcmp( szEntry, "contextProxyContext" ))
			nResult = LoadContextEntry( CONTEXTPROXYCONTEXT, szValue, &cTemp );
		else if (!strcmp( szEntry, "contextStorageType" ))
			nResult = LoadContextEntry( CONTEXTSTORAGETYPE, szValue, &cTemp );
		else if (!strcmp( szEntry, "contextStatus" ))
			nResult = LoadContextEntry( CONTEXTSTATUS, szValue, &cTemp );
		else {
			Error( "Bad context parse" );
			FreeContext( &cTemp );
			return ( ERROR );
		}
		
		if (nResult != OK) {
			FreeContext( &cTemp );
			return ( nResult );
		}
	}
		
	if( (cTemp.contextIdentity == NULL) || (cTemp.contextIndex ==0) ||
			(cTemp.contextViewIndex == 0)){
		Error( "Context-section should contain contextIdentity, contextIndex and contextViewIndex" );
		FreeContext( &cTemp );
		return ( ERROR );
	}
		
	if (GetContextIndex( cTemp.contextIdentity, &tmpint ) == OK) {
		Error( "ContextIdentity already in use" );
		FreeContext( &cTemp );
		return ERROR;
	}
		
	if (CreateContextEntry( cTemp.contextIdentity, &context) != OK) {
		Error( "Failed to create new context" );
		FreeContext( &cTemp );
		return(FAILED);
	}
 	
 	if (cTemp.contextLocalTime == NULL) LoadContextEntry( CONTEXTLOCALTIME, "currentTime", &cTemp );
 	
 	context->contextIndex		= cTemp.contextIndex;
 	context->contextName 		= cTemp.contextName;
 	context->contextViewIndex 	= cTemp.contextViewIndex;
 	context->contextLocalEntity 	= cTemp.contextLocalEntity;
 	context->contextLocalTime 	= cTemp.contextLocalTime;
 	context->contextProxyDstParty 	= cTemp.contextProxyDstParty;
 	context->contextProxySrcParty 	= cTemp.contextProxySrcParty;
 	context->contextProxyContext 	= cTemp.contextProxyContext;
 	context->contextStorageType 	= cTemp.contextStorageType;
 	context->contextStatus 		= cTemp.contextStatus;
 	
 	if (cTemp.contextLocal != UNDEF)
 		context->contextLocal = cTemp.contextLocal;
 	else {
 		if (context->contextViewIndex == 0)
 			context->contextLocal = FALSE;
		else {
			if ( gethostname(hostname, MAXHOSTNAMELEN) == -1) {
				Error( "Couldn't get hostname");
				return ERROR ;
			}
			(void) OIDtoASCII( context->contextIdentity, szEntry);
			(void) sscanf( szEntry, "1.3.6.1.6.3.3.1.4.%s", szValue);
			szValue[strlen( szValue ) - 2] = '\0';
			addr = inet_addr( szValue );
			he = gethostbyname(hostname);
			if ( he != (struct hostent *) 0)
				myaddr = inet_addr( 
					inet_ntoa(*((struct in_addr *)(he->h_addr))));
			if( !memcmp( &myaddr, &addr, (size_t) 4) )
			  			context->contextLocal = true;
			else context->contextLocal = false;
		}
	}
 	if (context->contextViewIndex > 0 )
		context->contextStatus = active;
	else
		context->contextStatus = notReady;
    	
    	return(OK);
}

/* 
 * -----------------------------------------------------------------------
 *
 * Function:	LoadContextEntry
 *
 * Arguments:	nEntry,	number to object to be added (IN).
 *		*szValue, string containing value belonging to object (IN).
 *		*cTemp,	pointer to temporary-context entry.
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     This funtcion translates szValue in the appropriate field.		
 *
 * Remarks:	none.
 *
 * -----------------------------------------------------------------------
 */
	STATUS
LoadContextEntry( AsnInt nEntry, char *szValue, ContextEntry *cTemp )
{	
	char	*cp;
	AsnOcts	*octFriendly, *octOID;
	AsnInt	nDesc, nError;
		
	switch (nEntry) {
		case CONTEXTIDENTITY :
			octFriendly 		= pNewOct( BUFSIZE );
			octFriendly->octetLen	= BUFSIZE - 1;
			octOID			= pNewOct( BUFSIZE );
			octOID->octetLen	= BUFSIZE - 1;
			switch (foidConv( pFoid, szValue, *octFriendly,
					  *octOID, &nDesc, &nError )) {
				case ERROR :
					FreeOct( octFriendly );
					FreeOct( octOID );
					return ( ERROR );
					break;
				case FAILED :
					if (nError == ILLEGAL) {
						FreeOct( octFriendly );
						FreeOct( octOID );
						Error( "Bad context ID" );
						return ( ERROR );
					}
					break;
				default :
					break;
			}
			cTemp->contextIdentity = pNewOid( strlen( octOID->octs ) );
            		if( ASCIItoOID( octOID->octs, cTemp->contextIdentity ) != 0) {
                		Error( "Bad context object ID" );
				FreeOid(cTemp->contextIdentity);
				FreeOct( octFriendly );
				FreeOct( octOID );
				return ERROR ;
			}
			break;
		case CONTEXTINDEX :
			for( cp = szValue; *cp; cp++ )
                		if (!isdigit(*cp)) {
                    			Error( "Context-index not an integer" );
					return ERROR ;
				}
            		cTemp->contextIndex = atoi( szValue );
            		break;
           	case CONTEXTLOCAL :
			if ( !strncmp( szValue, "true", 4 ))
				cTemp->contextLocal = true;
			else if (!strncmp( szValue, "false", 5 ))
				cTemp->contextLocal = false;
			else {
				Error( "Bad contextLocal value" );
				return ( ERROR );
			}
			break;
		case CONTEXTVIEWINDEX :
            		for( cp = szValue; *cp; cp++ )
                		if (!isdigit(*cp)) {
                    			Error( "View index not an integer" );
					return ERROR ;
				}
            		cTemp->contextViewIndex = atoi( szValue );
            		break;
             	case CONTEXTLOCALENTITY :
			cTemp->contextLocalEntity=pNewOct(strlen(szValue));
	  		cTemp->contextLocalEntity->octetLen = strlen(szValue);
			memcpy(cTemp->contextLocalEntity->octs, szValue, 
					cTemp->contextLocalEntity->octetLen);
			break;
		case CONTEXTLOCALTIME :
			if( !strcasecmp( szValue, "currentTime") ) {
				cTemp->contextLocalTime = pNewOid( currentTime.octetLen);
                		cTemp->contextLocalTime->octetLen = currentTime.octetLen;
				memcpy( cTemp->contextLocalTime->octs, currentTime.octs,
					cTemp->contextLocalTime->octetLen);
            		} else if( !strcasecmp(szValue, "restartTime") ) {
				cTemp->contextLocalTime = pNewOid( restartTime.octetLen);
                		cTemp->contextLocalTime->octetLen = restartTime.octetLen;
				memcpy( cTemp->contextLocalTime->octs, restartTime.octs,
					cTemp->contextLocalTime->octetLen);
            		} else if( !strcasecmp(szValue, "cacheTime")) {
				cTemp->contextLocalTime = pNewOid( cacheTime.octetLen);
                		cTemp->contextLocalTime->octetLen = cacheTime.octetLen;
				memcpy( cTemp->contextLocalTime->octs, cacheTime.octs,
					cTemp->contextLocalTime->octetLen);
	    		} else {
                		Error( "Bad local time value" );
				return ERROR ;
			}
            		break;
            	case CONTEXTPROXYDSTPARTY :
            		/* Ignore */
            		break;
            	case CONTEXTPROXYSRCPARTY :
			/* Ignore */
			break;
		case CONTEXTPROXYCONTEXT :
			/* Ignore */
			break;
		case CONTEXTSTORAGETYPE :
			/* value is ignored */
			break;
		case CONTEXTSTATUS :
			if (!strncmp( szValue, "active", 6 ))
				cTemp->contextStatus = active;
			else if (!strncmp( szValue, "notInService", 7 ))
				cTemp->contextStatus = notReady;
			else {
				Error( "Bad context status value" );
				return ( ERROR );
			}
			break;
		default:
            		Error("unknown context-entry" );
			return ERROR ;
    	}
	FreeOct( octFriendly );
	FreeOct( octOID );
    	return(OK);
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	LoadAclSection
 *
 * Arguments:	*pfFile, pointer to input-stream (IN).
 *		*szLine, string containing first line of acl-section (IN/OUT).
 *
 * Returns:	STATUS, which is OK on success
 *
 * Purpose:     This function loads an acl section.		
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS	
LoadAclSection( FILE *pfFile, char *szLine )
{
	AclEntry	*acl, aTemp;
	char    	szEntry[BUFSIZE], szValue[BUFSIZE];
	char		*cp;
    	AsnInt 		blank;
    	STATUS		nResult;
	
	aTemp.aclTarget 	= 0;
	aTemp.aclSubject 	= 0;
	aTemp.aclResources 	= 0;
	aTemp.aclPrivileges 	= 0;
	aTemp.aclStorageType 	= Volatile;
	aTemp.aclStatus 	= active;

    	while (fgets(szLine, BUFSIZE, pfFile)) {
    		if (szLine[0] == '%') break;
    		
		if (strlen( szLine ) > 250 ){
			Error( "Line longer than 250 bytes" );
			return ( ERROR );
		}
		
		if( (szLine[0] == '#') || (szLine[0] == ';') ||
		    (!strncmp( szLine, "//", 2 )) ) 
			continue;
		
		blank = TRUE;
		for( cp = szLine ; *cp ; cp++ )
			if (!isspace( *cp )) {
				blank = FALSE;
				break;
			}
		if (blank) continue;
				
        	if (sscanf(szLine, " %s = %s", szEntry, szValue ) != 2) {
           		Error( "Bad acl parse" );
			return ERROR ;
		}
			
		if (!strcmp( szEntry, "aclTarget" ))
			nResult = LoadAclEntry( ACLTARGET, szValue, &aTemp );
		else if (!strcmp( szEntry, "aclSubject" ))
			nResult = LoadAclEntry( ACLSUBJECT, szValue, &aTemp );
		else if (!strcmp( szEntry, "aclResources" ))
			nResult = LoadAclEntry( ACLRESOURCES, szValue, &aTemp );
		else if (!strcmp( szEntry, "aclPrivileges" ))
			nResult = LoadAclEntry( ACLPRIVILIGES, szValue, &aTemp );
		else if (!strcmp( szEntry, "aclStorageType" ))
			nResult = LoadAclEntry( ACLSTORAGETYPE, szValue, &aTemp );
		else if (!strcmp( szEntry, "aclStatus" ))
			nResult = LoadAclEntry( ACLSTATUS, szValue, &aTemp );
		else {
			Error( "Bad acl parse" );
			return ( ERROR );
		}
			
		if (nResult != OK) return ( nResult );
	}
		
	if( (aTemp.aclTarget    == 0) || (aTemp.aclSubject    == 0) ||
	    (aTemp.aclResources == 0) || (aTemp.aclPrivileges == 0)   ){
		Error( "Acl-section should contain aclTarget, aclSubject, aclResources and aclPrivileges" );
		return ( ERROR );
	}
		
	if (GetAclEntry( aTemp.aclTarget, aTemp.aclSubject, aTemp.aclResources, &acl ) == OK) {
		Error( "An acl-entry with these parameters already exists");
		return ERROR;
	}
		
	if (CreateAclEntry( aTemp.aclTarget, aTemp.aclSubject, aTemp.aclResources, &acl ) != OK) {
		Error( "Failed to create new acl-entry" );
		return(FAILED);
	}
 	acl->aclPrivileges	= aTemp.aclPrivileges;
   	acl->aclStorageType 	= aTemp.aclStorageType;
        acl->aclStatus 		= aTemp.aclStatus;
    	
    	return(OK);
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	LoadAclEntry
 *
 * Arguments:	nEntry,	number to object to be added (IN).
 *		*szValue, string containing value belonging to object (IN).
 *		*aTemp,	pointer to temporary-acl entry.
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     This funtcion translates szValue in the appropriate field.		
 *
 * Remarks:	none.
 *
 * -----------------------------------------------------------------------
 */
	STATUS
LoadAclEntry( AsnInt nEntry, char *szValue, AclEntry *aTemp )
{
	char	*cp;
	
	switch (nEntry) {
		case ACLTARGET :
			for (cp = szValue ; *cp ; cp++ )
				if (!isdigit( *cp )) {
					Error( "Acl-target index not an integer" );
					return ( ERROR );
				}
			aTemp->aclTarget = atoi( szValue );
			break;
		case ACLSUBJECT :
			for (cp = szValue ; *cp ; cp++ )
				if (!isdigit( *cp )) {
					Error( "Acl-subject index not an integer" );
					return ( ERROR );
				}
			aTemp->aclSubject = atoi( szValue );
			break;
		case ACLRESOURCES :
			for (cp = szValue ; *cp ; cp++ )
				if (!isdigit( *cp )) {
					Error( "Acl-resources index not an integer" );
					return ( ERROR );
				}
			aTemp->aclResources = atoi( szValue );
			break;
		case ACLPRIVILIGES :
			aTemp->aclPrivileges = 0;
        		for(cp = szValue; *cp; cp++){
            			switch(*cp){
              				case 'g': case 'G':
                				aTemp->aclPrivileges |= ACLPRIVILEGESGET; 
                				break;
              				case 'n': case 'N':
                				aTemp->aclPrivileges |= ACLPRIVILEGESGETNEXT; 
                				break;
              				case 'r': case 'R':
                				aTemp->aclPrivileges |= ACLPRIVILEGESGETRESPONSE; 
                				break;
              				case 's': case 'S':
                				aTemp->aclPrivileges |= ACLPRIVILEGESSET; 
                				break;
              				case 't': case 'T':
                				aTemp->aclPrivileges |= ACLPRIVILEGESTRAP; 
                				break;
              				case 'b': case 'B':
                				aTemp->aclPrivileges |= ACLPRIVILEGESBULK; 
                				break;
              				case 'i': case 'I':
                				aTemp->aclPrivileges |= ACLPRIVILEGESINFORM; 
                				break;
              				case 'u': case 'U':
                				aTemp->aclPrivileges |= ACLPRIVILEGESTRAP2; 
                				break;
              				default:
                				Error( "Bad privileges code" ); 
						return ERROR ;
            			}
            		}
        		break;			
		case ACLSTORAGETYPE :
			/* value is ignored */
			break;
		case ACLSTATUS :
			if (!strncmp( szValue, "active", 6 ))
				aTemp->aclStatus = active;
			else if (!strncmp( szValue, "notInService", 7 ))
				aTemp->aclStatus = notReady;
			else {
				Error( "Bad acl status value" );
				return ( ERROR );
			}
			break;
		default:
			Error( "Bad entry in section acl" );
			return ( ERROR );
	}
	return ( OK );
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	LoadViewSection
 *
 * Arguments:	*pfFile, pointer to input-stream (IN).
 *		*szLine, ifrst line of the section (IN/OUT).
 *
 * Returns:	STATUS, which is OK on success
 *
 * Purpose:     This function  loads one view-definition.		
 *
 * Remarks:	
 *
 * -----------------------------------------------------------------------
 */
	STATUS	
LoadViewSection( FILE *pfFile, char *szLine )
{
	ViewEntry	*view, vTemp;
	char    	szEntry[BUFSIZE], szValue[BUFSIZE];
	char		*cp;
    	AsnInt 		blank;
    	STATUS		nResult;
	
	vTemp.viewIndex 	= 0;
	vTemp.viewSubtree 	= NULL;
	vTemp.viewMask 		= NULL;
	vTemp.viewType 		= VIEWINCLUDED;
	vTemp.viewStorageType 	= Volatile;
	vTemp.viewStatus 	= active;

    	while (fgets(szLine, BUFSIZE, pfFile)) {
    		if (szLine[0] == '%') break;
    		
		if (strlen( szLine ) > 250 ){
			Error( "Line longer than 250 bytes" );
			FreeView( &vTemp );
			return ( ERROR );
		}
			
		if( (szLine[0] == '#') || (szLine[0] == ';') ||
		    (!strncmp( szLine, "//", 2 )) ) 
			continue;
		
		blank = TRUE;
		for( cp = szLine ; *cp ; cp++ )
			if (!isspace( *cp )) {
				blank = FALSE;
				break;
			}
		if (blank) continue;
				
       		if (sscanf(szLine, " %s = %s", szEntry, szValue ) != 2) {
            		Error( "Bad view parse" );
            		FreeView( &vTemp );
			return ERROR ;
		}
			
		if (!strcmp( szEntry, "viewIndex" ))
			nResult = LoadViewEntry( VIEWINDEX, szValue, &vTemp );
		else if (!strcmp( szEntry, "viewSubtree" ))
			nResult = LoadViewEntry( VIEWSUBTREE, szValue, &vTemp );
		else if (!strcmp( szEntry, "viewMask" ))
			nResult = LoadViewEntry( VIEWMASK, szValue, &vTemp );
		else if (!strcmp( szEntry, "viewType" ))
			nResult = LoadViewEntry( VIEWTYPE, szValue, &vTemp );
		else if (!strcmp( szEntry, "viewStorageType" ))
			nResult = LoadViewEntry( VIEWSTORAGETYPE, szValue, &vTemp );
		else if (!strcmp( szEntry, "viewStatus" ))
			nResult = LoadViewEntry( VIEWSTATUS, szValue, &vTemp );
		else {
			Error( "Bad view parse" );
			FreeView( &vTemp );
			return ( ERROR );				
		}
		
		if (nResult != OK) {
			FreeView( &vTemp );
			return ( nResult );
		}
	}
		
	if( (vTemp.viewIndex == 0) || (vTemp.viewSubtree == NULL)){
		Error( "View-section should contain viewIndex and viewSubtree" );
		FreeView( &vTemp );
		return ( ERROR );
	}
		
	if (ViewEntryExists( vTemp.viewIndex, vTemp.viewSubtree) == OK) {
		Error( "A view-entry with these parameters already exists");
		FreeView( &vTemp );
		return ERROR;
	}
		
	if (CreateViewEntry( vTemp.viewIndex, vTemp.viewSubtree, &view) != OK) {
		Error( "Failed to create new view-entry" );
		FreeView( &vTemp );
		return(FAILED);
	}
 	
 	view->viewMask 		= vTemp.viewMask;
 	view->viewType 		= vTemp.viewType;
   	view->viewStorageType 	= vTemp.viewStorageType;
        view->viewStatus 	= vTemp.viewStatus;
    	
    	return(OK);
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	LoadViewEntry
 *
 * Arguments:	nEntry,	number to object to be added (IN).
 *		*szValue, string containing value belonging to object (IN).
 *		*vTemp,	pointer to temporary-view entry.
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     This function translates szValue in the appropriate field.		
 *
 * Remarks:	none.
 *
 * -----------------------------------------------------------------------
 */
	STATUS
LoadViewEntry( AsnInt nEntry, char *szValue, ViewEntry *vTemp )
{
	char	*cp, *ucp;
	u_long	byte;
	AsnInt	nonhex = FALSE, nError, nDesc;
	AsnOcts	*octFriendly, *octOID;
	
	octFriendly 		= pNewOct( BUFSIZE );
	octFriendly->octetLen	= BUFSIZE - 1;
	octOID			= pNewOct( BUFSIZE );
	octOID->octetLen	= BUFSIZE - 1;
	
	switch (nEntry) {
		case VIEWINDEX :
			for (cp = szValue ; *cp ; cp++ )
				if (!isdigit( *cp )) {
					Error( "View index not an integer" );
					return ( ERROR );
				}
			vTemp->viewIndex = atoi( szValue );
			break;
		case VIEWSUBTREE :
			switch (foidConv( pFoid, szValue, *octFriendly,
					  *octOID, &nDesc, &nError )) {
				case ERROR :
					FreeOct( octFriendly );
					FreeOct( octOID );
					return ( ERROR );
					break;
				case FAILED :
					if (nError == ILLEGAL) {
						FreeOct( octFriendly );
						FreeOct( octOID );
						Error( "Bad subtree" );
						return ( ERROR );
					}
					break;
				default :
					break;
			}
			vTemp->viewSubtree = pNewOid( strlen( octOID->octs ));
			if (ASCIItoOID( octOID->octs, vTemp->viewSubtree ) != 0) {
				Error( "Bad subtree object ID" );
				FreeOid( vTemp->viewSubtree );
				vTemp->viewSubtree = NULL;
				FreeOct( octFriendly );
				FreeOct( octOID );
				return ( ERROR );
			}
			FreeOct( octFriendly );
			FreeOct( octOID );
			break;
		case VIEWMASK :
			for( cp = szValue ; *cp ; cp++ )
				if (!isxdigit( *cp )) nonhex = TRUE;
			if (nonhex) {
				Error( "Bad view mask value" );
				return ( ERROR );
			}
			if (strlen( szValue ) > 32) {
				Error( "Bad view mask value" );
				return ( ERROR );
			}
			ucp = (char *)((vTemp->viewMask = pNewOct(4))->octs);
			for( cp = szValue ; *cp ; cp += 2, ucp++ ) {
				if (sscanf( cp, "%2x", &byte ) != 1){
					Error( "Bad view mask value" );
					FreeOct( vTemp->viewMask );
					vTemp->viewMask = NULL;
					return ( ERROR );
				}
				*ucp = byte;
			}
			vTemp->viewMask->octetLen = (AsnInt) ucp - (AsnInt)
							vTemp->viewMask->octs;
			break;
		case VIEWTYPE :
			if (!strncmp( szValue, "include", 7 ))
				vTemp->viewType = VIEWINCLUDED;
			else if (!strncmp( szValue, "exclude", 7 ))
				vTemp->viewType = VIEWEXCLUDED;
			else {
				Error( "Bad view type value" );
				return ( ERROR );
			}
			break;
		case VIEWSTORAGETYPE :
			/* value is ignored */
			break;
		case VIEWSTATUS :
			if (!strncmp( szValue, "active", 6 ))
				vTemp->viewStatus = active;
			else if (!strncmp( szValue, "notInService", 7 ))
				vTemp->viewStatus = notReady;
			else {
				Error( "Bad view status value" );
				return ( ERROR );
			}
			break;
		default:
			Error( "Bad entry in section view" );
			return ( ERROR );
	}
	return ( OK );
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	CreatePartyEntry
 *
 * Arguments:	partyidentity,	name of the party to be created (IN).
 *		newparty, to new created party-entry (OUT).
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     This funtcion creates a new party entry of partyIdentity.		
 *
 * Remarks:	none.
 *
 * -----------------------------------------------------------------------
 */
	STATUS
CreatePartyEntry( AsnOid *partyidentity, PartyEntry **newparty )
{ 
	PartyEntry	*party, *ppp, *pp;
	struct Hashnode	*hashnode;
	int		loc;

	if ((hashnode = ( struct Hashnode *)malloc(sizeof(struct Hashnode))) 
			== (struct Hashnode *)NULL)
		{ Error("No memory available for new party hashnode");
		  return(FAILED);
		}
	if ((party = ( PartyEntry *)malloc(sizeof(PartyEntry))) 
			== (PartyEntry *)NULL)
		{ Error("No memory available for new party");
		  free( hashnode);
		  return(FAILED);
		}
	memset((char *)party, 0, (size_t) sizeof(PartyEntry));
	party->partyIdentity = partyidentity;

	/* Insertion in linked list indexed by partyidentity */
        ppp = pp = PartyMIB.parties;
        while (pp) {
                if (CMPOID( partyidentity, pp->partyIdentity) == LESS)
			break;
		ppp = pp;
                pp = pp->next;
        }
        if (pp == PartyMIB.parties) {
                party->next = PartyMIB.parties;
                PartyMIB.parties = party;
        } else {
                party->next = ppp->next;
                ppp->next = party;
        }

	/* Insertion in hash-table */
	hashnode->pentry = ( PartyEntry *)party;
	loc = hashoid( party->partyIdentity, psize, pradix);
	hashnode->pnext = phashtab[loc];
	phashtab[loc] = hashnode;
	*newparty = party;
	return(OK);
 }

/* 
 * -----------------------------------------------------------------------
 *
 * Function:	FreeParty
 *
 * Arguments:	PartyEntry	*pe, pointer to the party to be removed (IN).
 *
 * Returns:	STATUS, which is OK on success
 *
 * Purpose:     This function frees the memory allocated to the party entry.
 *
 * Remarks:	
 *
 * -----------------------------------------------------------------------
 */
void FreeParty( PartyEntry *pe)
{
	if (pe->partyName)
		free(pe->partyName);
	if (pe->partyIdentity)
		FreeOid( pe->partyIdentity);
	if (pe->partyTDomain)
		FreeOid( pe->partyTDomain);
	if (pe->partyTAddress)
		FreeOct( pe->partyTAddress);
	if (pe->partyAuthProtocol)
		FreeOid( pe->partyAuthProtocol);
	if (pe->partyAuthPrivate)
		FreeOct( pe->partyAuthPrivate);
	if (pe->partyAuthPublic)
		FreeOct( pe->partyAuthPublic);
	if (pe->partyPrivProtocol)
		FreeOid( pe->partyPrivProtocol);
	if (pe->partyPrivPrivate)
		FreeOct( pe->partyPrivPrivate);
	if (pe->partyPrivPublic)
		FreeOct( pe->partyPrivPublic);
	if ( pe->partyCloneFrom)
		FreeOid( pe->partyCloneFrom);
	free(pe);
	pe = NULL;
}

/* 
 * -----------------------------------------------------------------------
 *
 * Function:	RemovePartyEntry
 *
 * Arguments:	partyidentity, name of the party to be removed (IN).
 *
 * Returns:	STATUS, which is OK on success
 *
 * Purpose:     This function removes the party entry with partyIdentity.
 *
 * Remarks:	
 *
 * -----------------------------------------------------------------------
 */
	STATUS
RemovePartyEntry( AsnOid *partyid )
{
	int	loc;
	COMPARETYPE	res;
	struct Hashnode	*phn, *lphn;
	PartyEntry	*pp, *ppp;

	loc = hashoid(partyid, psize, pradix);
	lphn = NULL;
	phn = phashtab[loc];
	while (phn != NULL)
		if ( CMPOID( partyid, 
		  ((PartyEntry *)phn->pentry)->partyIdentity) == EQUAL)
			break;
		else {
			lphn = phn;
			phn = phn->pnext;
			}
	if ( phn == NULL )
		return(FAILED);
	if ( lphn == NULL )
		phashtab[loc] = NULL;
	else 
		lphn->pnext = phn->pnext;

	ppp = pp = PartyMIB.parties;
	while (pp) {
		if ( (res = CMPOID( partyid, pp->partyIdentity)) == GREATER)
			return(FAILED);
		else	if ( res == EQUAL)
				break;
		ppp = pp;
		pp = pp->next;
	}

	FreeParty(phn->pentry);
	free(phn);
	return(OK);
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	PrintPartyEntry
 *
 * Arguments:	none.
 *
 * Returns:	STATUS, which is OK on success
 *
 * Purpose:     This function prints a PartyEntry.		
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS	
PrintPartyEntry( PartyEntry *pe )
{
	char	oid[256];
	u_char	inetaddr[4];
	short	portnr;

	fprintf(stderr, "%-5u", pe->partyIndex);
	fprintf(stderr, "%-30s ",pe->partyName);
	if (pe->partyLocal == true)
	    fprintf(stderr, "True  ");
	else if (pe->partyLocal == false)
		fprintf(stderr, "False ");
	    else fprintf(stderr, "?     ");
	memcpy( inetaddr, pe->partyTAddress->octs, (size_t) 4);
	memcpy( &portnr, pe->partyTAddress->octs+4, (size_t) 2);
	(void) fprintf(stderr, "%u.%u.%u.%u:%u\t", inetaddr[0], 
			inetaddr[1], inetaddr[2], inetaddr[3], portnr);
	OIDtoASCII( pe->partyIdentity, oid);
	fprintf(stderr, "%s\n",oid);
	return(OK);
}

/* 
 * -----------------------------------------------------------------------
 *
 * Function:	SavePartyEntry
 *
 * Arguments:	pe	pointer to partyEntry (IN)
 *		pfFile 	pointer to input-stream (IN)
 *
 * Returns:	STATUS, which is OK on success
 *
 * Purpose:     This function saves a PartyEntry.		
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS	
SavePartyEntry( PartyEntry *pe, FILE *pfFile )
{
	char	oid[256];
	AsnInt	*hex, nLoop;
	u_char	inetaddr[4];
	short	portnr;

	if (pe->partyName)
		fprintf( pfFile, "%%party <%s>\n", pe->partyName );
	else
		fprintf( pfFile, "%%party\n" );
		
	OIDtoASCII( pe->partyIdentity, oid);
	fprintf( pfFile, "    partyIdentity = %s\n",oid);
	
	fprintf( pfFile, "    partyIndex = %u\n", pe->partyIndex);
	
	if (!memcmp( pe->partyTDomain->octs, snmpUDPDomain.octs,
		     snmpUDPDomain.octetLen ))
		fprintf( pfFile, "    partyTDomain = snmpUDPDomain\n" );
	else if (!memcmp( pe->partyTDomain->octs, snmpTCPDomain.octs,
			  snmpTCPDomain.octetLen ))
		fprintf( pfFile, "    partyTDomain = snmpTCPDomain\n" );
	else if (!memcmp( pe->partyTDomain->octs, rfc1157Domain.octs,
			  rfc1157Domain.octetLen ))
		fprintf( pfFile, "    partyTDomain = rfc1157Domain\n" );

	memcpy( inetaddr, pe->partyTAddress->octs,   (size_t) 4);
	memcpy( &portnr,  pe->partyTAddress->octs+4, (size_t) 2);
	fprintf( pfFile, "    partyTAddress = %u.%u.%u.%u %u\n", inetaddr[0], 
			inetaddr[1], inetaddr[2], inetaddr[3], portnr);

	fprintf( pfFile, "    partyMaxMessageSize = %u\n",
						pe->partyMaxMessageSize );
	
	if (pe->partyLocal == true)
	    fprintf( pfFile, "    partyLocal = true\n");
	else if (pe->partyLocal == false)
		fprintf( pfFile, "    partyLocal = false\n");

	if (!memcmp( pe->partyAuthProtocol->octs, noAuth.octs, noAuth.octetLen ))
		fprintf( pfFile, "    partyAuthProtocol = noAuth\n" );
	else if (!memcmp( pe->partyAuthProtocol->octs, v2md5AuthProtocol.octs,
			  v2md5AuthProtocol.octetLen ))
		fprintf( pfFile, "    partyAuthProtocol = v2md5AuthProtocol\n" ); 
	else if (!memcmp( pe->partyAuthProtocol->octs, rfc1157noAuth.octs,
			  rfc1157noAuth.octetLen ))
		fprintf( pfFile, "    partyAuthProtocol = rfc1157noAuth\n" );	 
	else	{
		Error( "Bad authentication protocol type" );
		return ERROR ;
	}

	if (pe->partyAuthClock)
		fprintf( pfFile, "    partyAuthClock = %u\n", pe->partyAuthClock );
	
	if (pe->partyAuthPrivate) {
		if (!memcmp( pe->partyAuthProtocol->octs, rfc1157noAuth.octs,
			     rfc1157noAuth.octetLen )) 
			fprintf( pfFile, "    partyAuthPrivate = %s\n", 
					pe->partyAuthProtocol->octs );
		else {
			memset( oid, 0, 256 );
			for( nLoop = 0, hex = (AsnInt *)pe->partyAuthPrivate->octs ; 
		     	     nLoop < pe->partyAuthPrivate->octetLen ;
		     	     nLoop+=sizeof(AsnInt), hex++ ) 
		     		sprintf( oid+(2*nLoop), "%x", *hex );
		     	sprintf( oid+(2*pe->partyAuthPrivate->octetLen), "\0" );
			fprintf( pfFile, "    partyAuthPrivate = %s\n", oid );
		}
	}
		     	
	if( (memcmp( pe->partyAuthProtocol->octs, rfc1157noAuth.octs,
		     rfc1157noAuth.octetLen )) &&
	    (pe->partyAuthPublic)) {
		memset( oid, 0, 256 );
		for( nLoop = 0, hex = (AsnInt *)pe->partyAuthPublic->octs ; 
		     nLoop < pe->partyAuthPublic->octetLen ;
		     nLoop+=sizeof(AsnInt), hex++ ) 
		     	sprintf( oid+(2*nLoop), "%x", *hex );
		sprintf( oid+(2*pe->partyAuthPublic->octetLen), "\0" );
		fprintf( pfFile, "    partyAuthPublic = %s\n", oid );
	}
		
	if (pe->partyAuthLifetime)
		fprintf( pfFile, "    partyAuthLifeTime = %u\n", pe->partyAuthLifetime );
	
	if (pe->partyPrivProtocol) {
		if (!memcmp( pe->partyPrivProtocol->octs, noPriv.octs,
			     noPriv.octetLen ))
			fprintf( pfFile, "    partyPrivProtocol = noPriv\n" );
		else if (!memcmp( pe->partyPrivProtocol->octs, 
				  desPrivProtocol.octs, desPrivProtocol.octetLen ))
			fprintf( pfFile, "    partyPrivProtocol = desPrivProtocol\n" ); 
		else	{
			Error( "Bad privacy protocol type" );
			return ERROR ;
		}
	}
	
	if (pe->partyPrivPrivate) {
		memset( oid, 0, 256 );
		for( nLoop = 0, hex = (AsnInt *) pe->partyPrivPrivate->octs ; 
	     	     nLoop < pe->partyPrivPrivate->octetLen ;
	     	     nLoop+=sizeof(AsnInt), hex++ )
	     	     	sprintf( oid+(2*nLoop), "%x", *hex ); 
	     	sprintf( oid+(2*pe->partyPrivPrivate->octetLen), "\0" );
		fprintf( pfFile, "    partyPrivPrivate = %s\n", oid );
	}
			     	
	if (pe->partyPrivPublic) {
		memset( oid, 0, 256 );
		for( nLoop = 0, hex = (AsnInt *) pe->partyPrivPublic->octs ; 
		     nLoop < pe->partyPrivPublic->octetLen ;
		     nLoop+=sizeof(AsnInt), hex++ ) 
		     	sprintf( oid+(2*nLoop), "%x", *hex );
		sprintf( oid+(2*pe->partyPrivPublic->octetLen), "\0" );
		fprintf( pfFile, "    partyPrivPublic = %s\n", oid );
	}
		
	if (pe->partyStatus == active )
		fprintf( pfFile, "    partyStatus = active\n" );
	else if ( pe->partyStatus == notReady )
		fprintf( pfFile, "    partyStatus = NotInService\n" );
				
	return(OK);
}



/* 
 * -------------------------------------------------------------------
 *
 * Function:    GetPartyByName
 *
 * Arguments:   partyName	*char		(in)
 *		partyId		*UAsnInt	(out) 
 *
 * Returns:     STATUS  which is OK on succes
 *
 * Purpose:     Retrieves partyid using the partyname
 *
 * -------------------------------------------------------------------
 */
	STATUS	
GetPartyByName( char *name, UAsnInt *index )
{ 
	PartyEntry	*pp;

	pp = PartyMIB.parties;
	while (pp) {
		if ( strcasecmp( pp->partyName, name) == 0 ) {
			*index = (UAsnInt) pp;
			return(OK);
		}
		pp = pp->next;
	}
	return(FAILED);
}	

/* 
 * -------------------------------------------------------------------
 *
 * Function:    GetPartyByIndex
 *
 * Arguments:   partyIndex	AsnInt		(in)
 *		partyP		*UAsnInt	(out) 
 *
 * Returns:     STATUS  which is OK on succes
 *
 * Purpose:     Retrieves partyid using the partyindex
 *
 * -------------------------------------------------------------------
 */
	STATUS	
GetPartyByIndex( AsnInt partyIndex, UAsnInt *partyP )
{ 
	PartyEntry	*pp;

	pp = PartyMIB.parties;
	while (pp) {
		if (pp->partyIndex == partyIndex ) {
			*partyP = (UAsnInt) pp;
			return(OK);
		}
		pp = pp->next;
	}
	return(FAILED);
}	

/* 
 * -------------------------------------------------------------------
 *
 * Function:    GetPartyIndex
 *
 * Arguments:   partyId		OID	(in)
 *		partyIndex	AsnInt	(out) 
 *
 * Returns:     STATUS  which is OK on succes
 *
 * Purpose:     Retrieves the party's index identified 
 * 		by  party identifier 
 *
 * Remarks:	none
 *
 *-------------------------------------------------------------------
 */
	STATUS	
GetPartyIndex( AsnOid *partyid, AsnInt *index )
{ 
	int	loc;
	struct Hashnode	*phn;

	loc = hashoid( partyid, psize, pradix);
	phn = phashtab[loc];
	while (phn != NULL) {
		if ( CMPOID( partyid, 
		  (((PartyEntry *)phn->pentry)->partyIdentity)) == EQUAL) {
		    *index = (int) phn->pentry;
		    return(OK);
		    }
		else phn = phn->pnext;
	}
	return(FAILED);
}		

/* 
 * -----------------------------------------------------------------------
 *
 * Function:	CreateContextEntry
 *
 * Arguments:	contextidentity, reference to context (IN)
 *		newcontext, created context (OUT).
 *
 * Returns:	STATUS
 *
 * Purpose:     This function creates a new context entry with 
 *		referenced by contextidentity;
 *
 * Remarks:	Because of the proposal of Perkins, this routine is altered.
 *		The contextIndex is not jet assigned.
 *
 * -----------------------------------------------------------------------
 */
	STATUS
CreateContextEntry( AsnOid *contextidentity, ContextEntry **newcontext )
{ 
	ContextEntry	*context, *pcp, *cp;
	struct Hashnode	*hashnode;
	int		loc;
	
	if ((hashnode = ( struct Hashnode *)malloc(sizeof(struct Hashnode))) 
				== (struct Hashnode *)NULL)
		{ fprintf(stderr, "No memory available for new context hashnode\n");
		  return(FAILED);
		}
	if ((context = ( ContextEntry *)malloc(sizeof(ContextEntry))) 
				== (ContextEntry *)NULL)
		{ fprintf(stderr, "No memory available for new context\n");
		  free( hashnode);
		  return(FAILED);
		}
	memset((char *)context, 0, (size_t) sizeof(ContextEntry));
	context->contextIdentity = contextidentity;

	/* First insertion in linked list */
        pcp = cp = PartyMIB.contexts;
        while (cp) {
                if (CMPOID( contextidentity, cp->contextIdentity) == LESS)
                        break;
                pcp = cp;
                cp = cp->next;
        }
        if (cp == PartyMIB.contexts) {
                context->next = PartyMIB.contexts;
                PartyMIB.contexts = context;
        } else {
                context->next = pcp->next;
                pcp->next = context;
        }

	/* Insertion in hash-table */
	hashnode->pentry = ( ContextEntry *)context;
	loc = hashoid( context->contextIdentity, csize, cradix);
	hashnode->pnext = chashtab[loc];
	chashtab[loc] = hashnode;
	*newcontext = context;
	return(OK);
}

/* 
 * -----------------------------------------------------------------------
 *
 * Function:	FreeContext
 *
 * Arguments:	ContextEntry	*ce, pointer to the context to be removed (IN).
 *
 * Returns:	STATUS, which is OK on success
 *
 * Purpose:     This function frees the memory allocated to the context entry.
 *
 * Remarks:	
 *
 * -----------------------------------------------------------------------
 */
void FreeContext( ContextEntry *ce)
{
	if (ce->contextName)
		free(ce->contextName);
	if (ce->contextIdentity)
		FreeOid( ce->contextIdentity);
	if (ce->contextLocalEntity)
		FreeOct( ce->contextLocalEntity);
	if (ce->contextProxyDstParty)
		FreeOid( ce->contextProxyDstParty);
	if (ce->contextProxySrcParty)
		FreeOid( ce->contextProxySrcParty);
	if (ce->contextProxyContext)
		FreeOid( ce->contextProxyContext);
	free(ce);
	ce = NULL;
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	RemoveContextEntry
 *
 * Arguments:	contextidentity, reference to context (IN).
 *
 * Returns:	STATUS 
 *
 * Purpose:     This function removes the referenced context entry.
 *
 * Remarks:	none
 *
 * -----------------------------------------------------------------------
 */
	STATUS
RemoveContextEntry( AsnOid *contextid )
{ 
	AsnInt	loc;
	struct Hashnode	*phn, *lphn;
	ContextEntry	*cp, *pcp;
	COMPARETYPE		res;

	loc = hashoid(contextid, csize, cradix);
	lphn = NULL;
	phn = chashtab[loc];
	while (phn != NULL) {
		if ( CMPOID( contextid, (((ContextEntry *)
			phn->pentry)->contextIdentity)) == EQUAL) break;
		lphn = phn;
		phn = phn->pnext;
	}
	if ( phn == NULL )
		return(FAILED);
	if ( lphn == NULL )
		chashtab[loc] = NULL;
	else 
		lphn->pnext = phn->pnext;

	pcp = cp = PartyMIB.contexts;
	while (cp) {
		if ((res = CMPOID( contextid, cp->contextIdentity)) == GREATER)
                        return(FAILED);
                if ( res == EQUAL)
                        break;
		pcp = cp;
		cp = cp->next;
	}
	if (!cp)
		return(FAILED);
	if ( pcp == PartyMIB.contexts)
		PartyMIB.contexts = cp;
	else
		pcp->next = cp->next;
	FreeContext(phn->pentry);
	free(phn);
	return(OK);
}

/* 
 * -----------------------------------------------------------------------
 *
 * Function:	PrintContextEntry
 *
 * Arguments:	ctable, reference to CONTEXTTABLE (IN)
 *
 * Returns:	STATUS 
 *
 * Purpose:     This function prints a ContextEntry.		
 *
 * Remarks:	none
 *
 * -----------------------------------------------------------------------
 */
	STATUS	
PrintContextEntry( ContextEntry *ce )
{ 
	char	oid[256];

	fprintf(stderr, "%-5u", ce->contextIndex);
	fprintf(stderr, "%-30s",ce->contextName);
	if (ce->contextLocal == true)
	    fprintf(stderr, "True  ");
	else if (ce->contextLocal == false)
		fprintf(stderr, "False ");
	    else fprintf(stderr, "?     ");
	OIDtoASCII( ce->contextIdentity, oid);
	fprintf(stderr, "%s\n",oid); 
	return(OK);
}

/* 
 * -----------------------------------------------------------------------
 *
 * Function:	SaveContextEntry
 *
 * Arguments:	ctable, reference to CONTEXTTABLE (IN)
 *		pfFile, pointer to output stream (IN)
 *
 * Returns:	STATUS 
 *
 * Purpose:     This function saves a ContextEntry.		
 *
 * Remarks:	none
 *
 * -----------------------------------------------------------------------
 */
	STATUS	
SaveContextEntry( ContextEntry *ce, FILE *pfFile )
{ 
	char	oid[256];

	if (ce->contextName == NULL)
		fprintf( pfFile, "%%context\n" );
	else
		fprintf( pfFile, "%%context <%s>\n", ce->contextName );
		
	OIDtoASCII( ce->contextIdentity, oid);
	fprintf( pfFile, "    contextIdentity = %s\n",oid);
	 
	fprintf( pfFile, "    contextIndex = %u\n", ce->contextIndex);
	
	if (ce->contextLocal == true)
		fprintf( pfFile, "    contextLocal = true\n");
	else if (ce->contextLocal == false)
		fprintf( pfFile, "    contextLocal = false\n");

	fprintf( pfFile, "    contextViewIndex = %u\n", ce->contextViewIndex );
		
	if (ce->contextLocalEntity)
		fprintf( pfFile, "    contextLocalEntity = %s\n", 
			ce->contextLocalEntity->octs );
			
	if (!memcmp( ce->contextLocalTime->octs, currentTime.octs,
		     currentTime.octetLen ))
		fprintf( pfFile, "    contextLocalTime = currentTime\n" );
	else if (!memcmp( ce->contextLocalTime->octs, restartTime.octs,
			  restartTime.octetLen ))
		fprintf( pfFile, "    contextLocalTime = restartTime\n" );
	else if (!memcmp( ce->contextLocalTime->octs, cacheTime.octs,
			  cacheTime.octetLen ))
		fprintf( pfFile, "    contextLocalTime = cacheTime\n" );
		
	if (ce->contextProxyDstParty) {
		OIDtoASCII( ce->contextProxyDstParty, oid);
		fprintf( pfFile, "    contextProxyDstParty = %s\n",oid);
	}
	
	if (ce->contextProxySrcParty) { 
		OIDtoASCII( ce->contextProxySrcParty, oid);
		fprintf( pfFile, "    contextProxySrcParty = %s\n",oid);
	}
	
	if (ce->contextProxyContext) { 
		OIDtoASCII( ce->contextProxyContext, oid);
		fprintf( pfFile, "    contextProxyContext = %s\n",oid);
	}
	
	if ( ce->contextStatus == active )
		fprintf( pfFile, "    contextStatus = active\n" );
	else if ( ce->contextStatus == notReady )
		fprintf( pfFile, "    contextStatus = NotInService\n" );
			
	return(OK);
}



/* 
 * -------------------------------------------------------------------
 *
 * Function:    GetContextByIndex
 *
 * Arguments:   contextIndex	AsnInt		(in)
 *		contextP	*UAsnInt	(out) 
 *
 * Returns:     STATUS
 *
 * Purpose:     This function maps the contextindex upon contexid.
 *
 * Remarks:	none.
 *
 * -------------------------------------------------------------------
 */
	STATUS
GetContextByIndex( AsnInt contextIndex, UAsnInt *contextP )
{ 
	ContextEntry	*cp;

	cp = PartyMIB.contexts;
	while (cp) {
		if (cp->contextIndex == contextIndex) { 
			*contextP = (UAsnInt)cp;
			return(OK);
		}
		cp = cp->next;
	}
	return(FAILED);
}

/* 
 * -------------------------------------------------------------------
 *
 * Function:    GetContextByName
 *
 * Arguments:   contextName	*char		(in)
 *		contextId	AsnOid	(out) 
 *
 * Returns:     STATUS
 *
 * Purpose:     This function maps the contextname upon contexid.
 *
 * Remarks:	none.
 *
 * -------------------------------------------------------------------
 */
	STATUS
GetContextByName( char *name, UAsnInt *index )
{ 
	ContextEntry	*cp;

	cp = PartyMIB.contexts;
	while (cp) {
		if (strcasecmp(cp->contextName, name) == 0) { 
			*index = (UAsnInt)cp;
			return(OK);
		}
		cp = cp->next;
	}
	return(FAILED);
}
/* 
 * -------------------------------------------------------------------
 *
 * Function:    GetContextIndex
 *
 * Arguments:   contextId		UAsnInt	(in)
 *		contextIndex	AsnInt	(out) 
 *
 * Returns:     STATUS  which is OK on succes
 *
 * Purpose:     Retrieves the context's index identified 
 *              by  context identifier 
 *
 * -------------------------------------------------------------------
 */
	STATUS
GetContextIndex( AsnInt contextid, UAsnInt *index )
{ 
	int	loc;
	struct Hashnode	*phn;

	loc = hashoid( contextid, csize, cradix);
	phn = chashtab[loc];
	while (phn != NULL) {
		if ( CMPOID( (AsnOid *)contextid, 
			(((ContextEntry *)phn->pentry)->contextIdentity))
			== EQUAL) {
		    *index = (int)phn->pentry;
		    return(OK);
		    }
		else phn = phn->pnext;
	}
	return(FAILED);
}		


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	GetAclEntry
 *
 * Arguments:	target, subject, name	AsnInteger
 *			indices of the acl to be found (IN).
 *		acl	*AclEntry, poAsnInter to acl-entry if found, 
 *			NULL otherwise (OUT)
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     Searches for a specified acl in the acl database.		
 *
 * Remarks:	
 *
 * -----------------------------------------------------------------------
 */
	STATUS
GetAclEntry( UAsnInt target, UAsnInt subject, UAsnInt resources,
	     AclEntry **acl )
{ 
	AclEntry	*ap;

	ap = PartyMIB.acls;
	while (ap) {
		if ( ap->aclTarget < target ) {
			ap = ap->next;
			continue;
		} else if ( ap->aclTarget != target )
				return(FAILED);
		if ( ap->aclSubject < subject ) {
			ap = ap->next;
			continue;
		} else if ( ap->aclSubject != subject )
				return(FAILED);
			
		if ( ap->aclResources < resources) {
			ap = ap->next;
			continue;
		} else if ( ap->aclResources != resources)
				return(FAILED);
	}
	if (!ap)
		return(FAILED);
	*acl = ap;
	return(OK);
}		



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	CreateAclEntry
 *
 * Arguments:	target, subject, name	AsnInteger
 *			indices of the acl to be found (IN).
 *		newacl	poAsnInter to new created acl-entry.
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     Creates a new acl entry with aclIdentity = aclidentity.		
 *
 * Remarks:	
 *
 * -----------------------------------------------------------------------
 */
	STATUS
CreateAclEntry( UAsnInt target, UAsnInt subject, UAsnInt resources,
		AclEntry **newacl )
{
	AclEntry	*acl, *ap, *pap;
	
	if ((acl = ( AclEntry *)malloc(sizeof(AclEntry))) == (AclEntry *)NULL)
		{ fprintf(stderr, "No memory available for new acl\n");
		  return(FAILED);
		}
	memset((char *)acl, 0, (size_t)sizeof(AclEntry));
	GetPartyByIndex( target, &target );
	GetPartyByIndex( subject, &subject );
	GetContextByIndex( resources, &resources );
	acl->aclTarget 		= target;
	acl->aclSubject 	= subject;
	acl->aclResources 	= resources;

	/* First insertion of new acl entry */
	pap = ap = PartyMIB.acls;
	while (ap) {
		if ( ((PartyEntry *)ap->aclTarget)->partyIndex >
			 ((PartyEntry *)target)->partyIndex) break;
		else	if ( ((PartyEntry *)ap->aclTarget)->partyIndex ==  
			     ((PartyEntry *)target)->partyIndex)
				if ( ((PartyEntry *)ap->aclSubject)->partyIndex 
					>  ((PartyEntry *)subject)->partyIndex)
					break;
				else	if ( ((PartyEntry *)ap->aclSubject)->partyIndex 
					  ==  ((PartyEntry *)subject)->partyIndex)
						if ( ((ContextEntry *)ap->aclResources)->contextIndex 
						   > ((ContextEntry *)resources)->contextIndex )
							break;
		pap = ap;
		ap = ap->next;
	}
	if (ap == PartyMIB.acls) { 
		acl->next = PartyMIB.acls;
		PartyMIB.acls = acl;
	} else {
		acl->next = pap->next;
		pap->next = acl;
	}
	*newacl = acl;
	return(OK);
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	RemoveAclEntry
 *
 * Arguments:	target, subject, name	AsnInteger
 *			indices of the acl to be found (IN).
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     Removes the acl entry with aclIdentity = aclidentity.		
 *
 * Remarks:	
 *
 * -----------------------------------------------------------------------
 */
	STATUS
RemoveAclEntry( UAsnInt target, UAsnInt subject, UAsnInt resources )
{
	AclEntry	*ap, *pap;

	pap = ap = PartyMIB.acls;
	while (ap) {
		if ( ap->aclTarget < target ) {
			pap = ap;
			ap = ap->next;
			continue;
		} else if ( ap->aclTarget != target )
				return(FAILED);
		if ( ap->aclSubject < subject ) {
			pap = ap;
			ap = ap->next;
			continue;
		} else if ( ap->aclSubject != subject )
				return(FAILED);
			
		if ( ap->aclResources < resources) {
			pap = ap;
			ap = ap->next;
			continue;
		} else if ( ap->aclResources != resources)
				return(FAILED);
	}
	if (ap == PartyMIB.acls)
		PartyMIB.acls = ap->next;
	else
		pap->next = ap->next;
	free(ap);
	return(OK);
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	PrintAclEntry
 *
 * Arguments:	acle	*AclEntry	(IN)
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     Prints an AclEntry.		
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS	
PrintAclEntry( AclEntry	*acle )
{
	fprintf(stderr, "%-12u", ((PartyEntry *)acle->aclTarget)->partyIndex);
	fprintf(stderr, "%-12u", ((PartyEntry *)acle->aclSubject)->partyIndex);
	fprintf(stderr, "%-12u", 
			((ContextEntry *)acle->aclResources)->contextIndex); 
	fprintf(stderr, "%5d\n", acle->aclPrivileges);
	return(OK);
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	SaveAclEntry
 *
 * Arguments:	acle	*AclEntry	(IN)
 *		pfFile	*FILE		(IN)
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     Saves an AclEntry.		
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS	
SaveAclEntry( AclEntry	*acle, FILE *pfFile )
{
	char 	szPriv[BUFSIZE] = "";
	AsnInt	nPriv;
	
	fprintf( pfFile, "%%acl\n" );
	
	fprintf( pfFile, "    aclTarget = %u\n", 
			((PartyEntry *)acle->aclTarget)->partyIndex);
	fprintf( pfFile, "    aclSubject = %u\n", 
			((PartyEntry *)acle->aclSubject)->partyIndex);
	fprintf( pfFile, "    aclResources = %u\n", 
			((ContextEntry *)acle->aclResources)->contextIndex);
			 
	fprintf( pfFile, "    aclPrivileges = " );
	
	nPriv = acle->aclPrivileges;
	if (nPriv & ACLPRIVILEGESGET) 
		fprintf( pfFile, "g" );
	if (nPriv & ACLPRIVILEGESGETNEXT) 
		fprintf( pfFile, "n" );
	if (nPriv & ACLPRIVILEGESGETRESPONSE) 
		fprintf( pfFile, "r" );
	if (nPriv & ACLPRIVILEGESSET) 
		fprintf( pfFile, "s" );
	if (nPriv & ACLPRIVILEGESTRAP)
		fprintf( pfFile, "t" );
	if (nPriv & ACLPRIVILEGESBULK) 
		fprintf( pfFile, "b" );
	if (nPriv & ACLPRIVILEGESINFORM) 
		fprintf( pfFile, "i" );
	if (nPriv & ACLPRIVILEGESTRAP2) 
		fprintf( pfFile, "u" );
	fprintf( pfFile, "\n" );
	
	if ( acle->aclStatus == active )
		fprintf( pfFile, "    aclStatus = active\n" );
	else if ( acle->aclStatus == notReady )
		fprintf( pfFile, "    aclStatus = NotInService\n" );
	return(OK);
}

/* 
 * -----------------------------------------------------------------------
 * 
 * Function:	ViewEntryExists
 *
 * Arguments:	index,		AsnInt, index used in conf-file. 
 *		*subtree,	poAsnInter to poAsnInter of new created 
 *				view-entry.
 *
 * Returns:	STATUS which is OK if a view-entry with these parameters 
 *			exists
 *
 * Purpose:     Check to see if a view-entry with these parameters already 
 *		exists in the partyMIB.		
 *
 * Remarks:	
 *
 * -----------------------------------------------------------------------
 */
	STATUS
ViewEntryExists( UAsnInt index, AsnOid *subtree)
{
	ViewEntry	*vp;

	vp = PartyMIB.views;
	while (vp) { 
		if ((vp->viewIndex == index) && 
		    (CMPOID( vp->viewSubtree, subtree) == EQUAL))
			return(OK);
		else	vp = vp->next;
	}
	return(FAILED);
}


/* 
 * -----------------------------------------------------------------------
 * 
 * Function:	CreateViewEntry
 *
 * Arguments:	index,		UAsnInt, index used in conf-file. 
 *		*subtree	AsnOid identifying the subtree (IN)
 *		**newview,	pointer to pointer of new created view-entry.
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     Creates a new view entry with the specified subtree and the 
 *		new index and places it at the correct position in the 
 *		viewtable.		
 *
 * Remarks:	
 *
 * -----------------------------------------------------------------------
 */
	STATUS
CreateViewEntry( AsnInt	index, AsnOid *subtree, ViewEntry **newview )
{
	ViewEntry	*view, *pvp, *pv;

	if ((view = ( ViewEntry *)malloc(sizeof(ViewEntry))) 
				== (ViewEntry *)NULL)
		  return(FAILED);
	memset((char *)view, 0, (size_t)sizeof(ViewEntry));

	pvp = pv = PartyMIB.views;
	while (pv) { 
		if (pv->viewIndex > index)
			break;
		else	if (pv->viewIndex == index)
				if (CMPOID( subtree, pv->viewSubtree) == LESS)
					break;
		pvp = pv;
		pv = pv->next;
	}

	view->viewIndex = index;
	view->viewSubtree = subtree;
	if (pv == PartyMIB.views) {
		view->next = PartyMIB.views;
		PartyMIB.views = view;
	} else {
		view->next = pvp->next;
		pvp->next = view;
	}
	*newview = view;
	return(OK);
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	FreeView
 *
 * Arguments:	ViewEntry	*ve, pointer to the view to be removed (IN).
 *
 * Returns:	STATUS, which is OK on success
 *
 * Purpose:     This function frees the memory allocated to the view entry.
 *
 * Remarks:	
 *
 * -----------------------------------------------------------------------
 */
void FreeView( ViewEntry *ve)
{
	if (ve->viewSubtree)
		FreeOid( ve->viewSubtree);
	if (ve->viewMask)
		FreeOct( ve->viewMask);
	free(ve);
	ve = NULL;
}


/*  
 * -----------------------------------------------------------------------
 *
 * Function:	RemoveViewEntry
 *
 * Arguments:	index,	poAsnInter to index of the view to be created (IN).
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     Removes the view entries with viewIndex = index.		
 *
 * Remarks:	
 *
 * -----------------------------------------------------------------------
 */
	STATUS
RemoveViewEntry( AsnInt index )
{
	ViewEntry	*vp, *pvp;
	AsnInt		found = 0;

	vp = PartyMIB.views;
	pvp = vp;
	while (vp)
	    { if (vp->viewIndex == index)
		{ if ( vp == PartyMIB.views)
		      PartyMIB.views = vp->next;
		  else pvp->next = vp->next;
		  FreeView( vp);
		  found = 1;
		}
	      else pvp = vp;
	      vp = vp->next;
	    }
	if (found)
	      return(OK);
	else return(FAILED);
}

/* 
 * -----------------------------------------------------------------------
 *
 * Function:	PrintViewEntry
 *
 * Arguments:	vtable	*VIEWTABLE	(IN)
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     Prints a ViewEntry.		
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS
PrintViewEntry( ViewEntry *vp )
{
	char	oid[256];

	fprintf(stderr, "%-5u", vp->viewIndex);
	OIDtoASCII( vp->viewSubtree, oid);
	fprintf(stderr, "%-25s", oid);
	if ( vp->viewType == VIEWINCLUDED)
	    fprintf(stderr, "incl \n");
	else if ( vp->viewType == VIEWEXCLUDED)
	    fprintf(stderr, "excl \n");
	    else fprintf(stderr, "?    \n");
	return(OK);
}

/* 
 * -----------------------------------------------------------------------
 *
 * Function:	SaveViewEntry
 *
 * Arguments:	vtable	*VIEWTABLE	(IN)
 *		pfFile	*FILE		(IN)
 *
 * Returns:	STATUS which is OK on success
 *
 * Purpose:     Saves a ViewEntry.		
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS
SaveViewEntry( ViewEntry *vp, FILE *pfFile )
{
	char	oid[256];
	AsnInt	nLoop;
	AsnInt	*hex, nDesc, nError;
	AsnOcts	*octFriendly, *octOID;
	
	octFriendly 		= pNewOct( BUFSIZE );
	octFriendly->octetLen	= BUFSIZE - 1;
	octOID			= pNewOct( BUFSIZE );
	octOID->octetLen	= BUFSIZE - 1;
 
	fprintf( pfFile, "%%view\n" );
	
	fprintf( pfFile, "    viewIndex = %u\n", vp->viewIndex);
	
	OIDtoASCII( vp->viewSubtree, oid);
	switch (foidConv( pFoid, oid, *octFriendly, *octOID, &nDesc, &nError )) {
		case OK:
			for( nLoop = strlen( octFriendly->octs )-1;
			     (nLoop > 0) && octFriendly->octs[nLoop] != '.' ;
			     nLoop-- );
			fprintf( pfFile, "    viewSubtree = %s\n",
				(nLoop) ? octFriendly->octs+nLoop+1 :
					  octFriendly->octs );
			break;
		case FAILED:
		case ERROR:
		default:
			fprintf( pfFile, "    viewSubtree = %s\n", oid );
			break;
	}
	
	if ( vp->viewMask != NULL ) {
		memset( oid, 0, 256 );
		for ( nLoop = 0, hex = (AsnInt *) vp->viewMask->octs ; 
		      nLoop < vp->viewMask->octetLen ;
		      nLoop+=sizeof(AsnInt), hex++ ) 
			sprintf( oid+(2*nLoop), "%x", *hex );
		sprintf( oid+(2*vp->viewMask->octetLen), "\0" );
		fprintf( pfFile, "    viewMask = %s\n", oid );
	} 
	
	if ( vp->viewType == VIEWINCLUDED)
		fprintf( pfFile, "    viewType = include\n");
	else if ( vp->viewType == VIEWEXCLUDED)
		fprintf( pfFile, "    viewType = exclude\n");
		
	if ( vp->viewStatus == active )
		fprintf( pfFile, "    viewStatus = active\n" );
	else if ( vp->viewStatus == notReady )
		fprintf( pfFile, "    viewStatus = NotInService\n" );
		
	return(OK);
}
