/* -----------------------------------------------------------------------
 *
 * Project:	UT-SNMP
 *
 * Filename:	spmTIM.c
 *
 * Date:	May 1994
 *
 * Purpose:	File with major function for the Transportportlayer
 *
 * Author(s):	Vincent G.M. Berkhout, Frederic P. H. van Hengstum
 *
 * -----------------------------------------------------------------------
 */

#include <spmTIM.h>


/*
 *	local variables
 */
static	int	addressSize = 6 ;

static  BOOLEAN transportDomainTable[DOMAINTOTAL] ;
static  TDOMAIN previousDomain ;

static  listenTable             *CLListenTable ;
static  listenTable             *COListenTable ;
static  activeConnectionsTable  *activeConnections ;

/*
 * --------------------------------------------------------------
 * function: 	SpmInitTim
 * Arguments:	none.
 * Purpose:	initialize Transport module.
 * Remarks:
 * -----------------------------------------------------------------------
 */
	STATUS
SpmInitTim( )
{
	int	i;

/*
 *	initialize the TransportDomainTable with FALSE and table pointers to NULL
 */
	for (i = 0 ; i < DOMAINTOTAL; i++ ) 
		transportDomainTable[i] = FALSE ;	
	CLListenTable = NULL ;		
	COListenTable = NULL ;
	activeConnections = NULL ;
	return OK ;

}	/* end of function TransportInitialize */

/* -----------------------------------------------------------------------
 * Function:	TransportOpen	  
 * Arguments:	Transport domain 	(in)
 * 		Transport address 	(in)
 * 		MaxMessageSize 		(in) (not implemented yet)
 * Purpose:	Open a transport connection on a specified domain 
 * Remarks:	
 * -----------------------------------------------------------------------
 */
	STATUS
TransportOpen(transportDomain, transportAddress)
TDOMAIN		transportDomain;	/* in */
TADDRESS	*transportAddress;	/* in */
{
	switch( transportDomain )
	{
	case UDPDOMAIN :
	case CLNSDOMAIN :
	case DDPDOMAIN :
	case IPXDOMAIN : {
		THANDLE		dataHandle ;
		listenTable	*openCurrent = NULL,
		 		*newEntry = NULL ;
		
		openCurrent = CLListenTable ;

		while ( openCurrent != NULL ) 
		{
			if ( memcmp( (void *) openCurrent->transportAddress, 
				(const void *) transportAddress, addressSize) == 0 ) 
			{
				(openCurrent->numberOfOpens ) ++ ;
				return OK ;
			} else 	openCurrent = openCurrent->next ;

		}	/* end while */

		if (TransPacketOpen( transportDomain, transportAddress, &dataHandle) < 0)
			return ERROR ;
		else {
		        newEntry = (listenTable *) malloc( sizeof( listenTable ) );
			newEntry->transportAddress = (char *) malloc ( addressSize ) ;
			(void) memcpy( (void *) newEntry->transportAddress, 
					(const void *) transportAddress, addressSize) ; 
			newEntry->handle = dataHandle ;
			newEntry->numberOfOpens = 1 ;
			newEntry->next = CLListenTable ;
			CLListenTable = newEntry ;

			transportDomainTable[ transportDomain ] = TRUE ;

			return OK ;

		} /* endif */

		break ;		/* never reached */

	}	/* end case UDPDOMAIN */

	case TCPDOMAIN :
	case CONSDOMAIN : {
		THANDLE		connectHandle;
		listenTable	*current = NULL,
				*newEntry = NULL ;

		current = COListenTable ;

		while ( current != NULL )
			if ( memcmp( current->transportAddress, 
			   ( const void *) transportAddress, addressSize)  == 0 ) 
			{
				current->numberOfOpens++ ;
				transportDomainTable[ transportDomain ] = TRUE ;
				return OK ;
			} else 	current = current->next ;


		if (TransPassiveOpen( transportDomain, transportAddress, &connectHandle) < 0)
			return ERROR ;

	        newEntry = ( listenTable *) malloc( sizeof( listenTable ) ) ;
		newEntry->transportAddress = (char *) malloc ( addressSize ) ;

		(void) memcpy( (void *) newEntry->transportAddress, 
				(const void *) transportAddress, addressSize) ; 
		newEntry->handle = connectHandle ;
		newEntry->numberOfOpens = 1 ;
		newEntry->next = COListenTable ;
		COListenTable = newEntry ;

		transportDomainTable[ transportDomain ] = TRUE ;

		return OK ;

	}	/* end case stream oriented */

	default:

	} 	/* end switch */

	return ERROR ;	/* never reached */

}	/* end function TransportOpen */ 

/* -----------------------------------------------------------------------
 * Function:	TransportSend	  
 * Arguments:	transportDomain,		(in)
 * 		localTAddress, 			(in)
 *		remoteTAddress, 		(in)
 * 		message				(in)
 * Purpose:	write message to datahandle on specified domain and address	
 * Remarks:	-
 * -----------------------------------------------------------------------
 */
	STATUS
TransportSend( transportDomain, localTAddress, remoteTAddress, tmessage )
TDOMAIN		transportDomain;	/* in */
TADDRESS	*localTAddress,		/* in */
		*remoteTAddress; 	/* in */
AsnOcts		*tmessage; 		/* in */
	
{
	TMESSAGE message;
	unsigned short portl, portr;

	message.length = tmessage->octetLen;
	memcpy( message.buf, tmessage->octs, message.length );

	memcpy( &portl, localTAddress+4, 2 );
	memcpy( &portr, remoteTAddress+4, 2 );
		
	TRACE( SNMP_DEBUG_TIM, "%u.%u.%u.%u:%d--%d bytes-->%u.%u.%u.%u:%d on %d domain",
			localTAddress[0], localTAddress[1], localTAddress[2],
			localTAddress[3], portl, message.length,
			remoteTAddress[0], remoteTAddress[1], remoteTAddress[2],
			remoteTAddress[3], portr, transportDomain );

	switch( transportDomain )
	{
	case UDPDOMAIN :
	case CLNSDOMAIN :
	case DDPDOMAIN :
	case IPXDOMAIN : {
		THANDLE		dataHandle;
		listenTable	*current = NULL ;
 
		current = CLListenTable ;

		dataHandle = -1 ;
/*
 *	find the dataHandle corresponding to the given local address
 */ 
		while ( current != NULL ) {
			if ( memcmp( current->transportAddress, localTAddress, addressSize) == 0 ) {
				dataHandle = current->handle ;
				break ;
			} else	current = current->next ;
		}	/* end while */
/*
 *	no datahandle found !
 */ 
		if (dataHandle < 1) {
			if ( TransportOpen(transportDomain, localTAddress) != OK ) 
				return FAILED;
			else {   /* and find the new dataHandle */
        		while ( current != NULL ) {
            		if ( memcmp( current->transportAddress, localTAddress, 
							addressSize) == 0 ) {
                			dataHandle = current->handle ;
                			break ;
            		} else  current = current->next ;
        		}   /* end while */
			} /* open call */
		} /* no dataHandle */ 

		if ( dataHandle < 0 ) {
/*			TRACE( SNMP_DEBUG_TIM, "No dataHandle found to send message on") ; */
			return FAILED ;
		} 	/* endif */
/*
 *	datahandle is available->send the data 
 */ 
		if ( TransPacketSend(transportDomain, &dataHandle, remoteTAddress, &message) < 0 )
			return FAILED ;

		return OK ;

	}	/* end case UDPDOMAIN */

	case TCPDOMAIN :
	case CONSDOMAIN :  {
		THANDLE		 	dataHandle;
		activeConnectionsTable	*current = NULL , 
					*newEntry = NULL ;

		current = activeConnections ;

/*
 *	look for an active connection modified looks only for remote address!
 */ 
		while ( (current != NULL) && 
			   (memcmp( current->remoteTAddress,remoteTAddress,addressSize ) != 0) )
					current = current->next ;
/*
 *	if the connection is not found then open a new connection and update the table.
 */ 
		if ( current == NULL )
			if (TransActiveOpen( transportDomain, remoteTAddress, &dataHandle) < 0)
				return ERROR ;		/* error message is already on lower level */

			else {
				newEntry = ( activeConnectionsTable *) 
					malloc( sizeof( activeConnectionsTable ) );
				newEntry->localTAddress = (char *) malloc ( addressSize ) ;
				newEntry->remoteTAddress = (char *) malloc ( addressSize ) ;
				newEntry->lastUpdate = (struct timeval *) malloc(sizeof(struct  timeval)) ;			

				(void) memcpy( (void *) newEntry->localTAddress, 
					(const void *) localTAddress, addressSize) ; 
				(void) memcpy( (void *) newEntry->remoteTAddress, 
					(const void *) remoteTAddress, addressSize) ; 
				newEntry->handle = dataHandle ; 
				gettimeofday( (newEntry->lastUpdate), (struct timezone *) NULL );
				newEntry->next = activeConnections ;
				activeConnections = current = newEntry ;

			} 	/* endif transport active open */

/*
 *	There is a connection available, now send the data
 */ 
		dataHandle = current->handle ;
		if (TransStreamSend( transportDomain, &dataHandle, &message) >= 0) {
               		gettimeofday( ( current->lastUpdate), (struct timezone *) NULL );
			return OK ;
		} 	/* endif transStreamSend */

		return FAILED ;

	}	/* end case TCPDOMAIN */
	default: 

	} 	/* end switch */

	return ERROR ;

}	/* end function TransportSend */


/* -----------------------------------------------------------------------
 * Function:	TransportReceive	  
 * Arguments:	transportDomain,	(out)
 * 		localTAddress, 		(out)
 * 		remoteTAddress,		(out)
 * 		message			(out) 
 * Purpose:	receive a message from the different domains 	
 * Remarks:	
 * -----------------------------------------------------------------------
 */
	STATUS
TransportReceive( transportDomain, localTAddress, remoteTAddress, message)
TDOMAIN		*transportDomain;	/* out */
TADDRESS 	*localTAddress,		/* out */
		*remoteTAddress; 	/* out */
TMESSAGE	*message; 		/* out */

{
	THANDLE	allHandles[MAXFILES],		/* all handles on which to perform a select */
		activeHandles[MAXFILES];	/* all active handles */
	int	i; 
	int 	domainCounter ;
	unsigned short portl, portr;

	*transportDomain = previousDomain ;
/*
 *	check al domains until a message found or no more domains available.
 */ 
	
for ( domainCounter = 1; domainCounter <= DOMAINTOTAL ; domainCounter++ )
{
	*transportDomain  = (previousDomain + domainCounter) % ( DOMAINTOTAL) ;
	if (transportDomainTable[ *transportDomain  ] == FALSE )
		continue ;	/* skip disabled transportdomains */

	switch( *transportDomain )
	{	
	case UDPDOMAIN :
	case CLNSDOMAIN :
	case DDPDOMAIN :
	case IPXDOMAIN : 
	{
		int		j ;
		THANDLE 	dataHandle; 
		listenTable	*clCurrent = NULL ;
		THANDLE		allClHandles[MAXFILES],
				activeClHandles[MAXFILES] ;
		
		clCurrent = CLListenTable ;
/*
 *	fill the polling set with all possible handles
 */ 
		memset (allClHandles, -1, sizeof ( allClHandles ) ) ;
		memset (activeClHandles, -1, sizeof ( activeClHandles ) ) ;

		i = 0 ;
		while ( clCurrent != NULL ) {
			allClHandles[i++] = clCurrent->handle ;
			clCurrent = clCurrent->next ;
		}	/* end while */

/*
 *	perform the "poll" (look for activity on the data handles)
 */ 
		if (TransCheckInputs( *transportDomain, allClHandles, activeClHandles ) < 0 ) 
			return ERROR ;
		
/*
 *	now process all active handles until a message is received
 */ 

		j = 0 ; 		/* reset i */

		while (	activeClHandles[j] != -1 ) {
/*
 *	receive the data from the active handle
 */ 
			if (TransPacketReceive( *transportDomain, &activeClHandles[j] , remoteTAddress, message ) < 0 )
				return ERROR ; 
/*
 *	if the message is not empty look for the local transport address
 */ 
			if ( message->length > 0 ) {

				clCurrent = CLListenTable ;
				while ( clCurrent->handle != activeClHandles[j] ) 
					clCurrent = clCurrent->next ;

				if (clCurrent != NULL)
					(void) memcpy( (void *) localTAddress, 
					(const void *) clCurrent->transportAddress, addressSize) ;

				memcpy( &portl, localTAddress+4, 2 );
				memcpy( &portr, remoteTAddress+4, 2 );
				TRACE( SNMP_DEBUG_TIM,  "%u.%u.%u.%u:%d<--%d bytes--%u.%u.%u.%u:%d (CL)",
					localTAddress[0], localTAddress[1], 
					localTAddress[2], localTAddress[3], portl, 
					message->length, remoteTAddress[0], 
					remoteTAddress[1], remoteTAddress[2], 
					remoteTAddress[3], portr ); 

				previousDomain = *transportDomain ;

				return OK ;

			} 	/* endif */

			j++ ;

		}	/* end while */

		break ;

	}	/* end case connection less communication */


	case CONSDOMAIN :
	case TCPDOMAIN: {
		THANDLE		allHandles[MAXFILES],	/* all handles on which to perform a select */
				activeHandles[MAXFILES],/* all active handles */
				dataHandle ;
		int		i,j ;			/* indication that a message is received */
		TADDRESS	returnAddress[addressSize] ;


		listenTable		*currentListen = NULL ;
		activeConnectionsTable	*current = NULL,
					*previous = NULL, 
					*newEntry = NULL ;


		current = previous = activeConnections ;

/*
 *	fille the array to perform the poll with all available handles
 */ 
		memset (allHandles, -1, sizeof ( allHandles ) ) ;
		memset (activeHandles, -1, sizeof ( activeHandles ) ) ;

		i = 0 ;
		while ( current != NULL ) 
		{
			allHandles[i++] = current->handle ;
			current = current->next ;
		} 	/* end while allHandles */

		if (TransCheckInputs( *transportDomain, allHandles, activeHandles ) < 0 )
			return ERROR ;

/*
 *	now process all active handles
 */ 
		i = 0 ; 
		while (	activeHandles[i] != -1 ) 
		{
			if (TransStreamReceive( *transportDomain, &activeHandles[i] , message ) < 0 )
				return ERROR ;

			current = activeConnections ;
			while ( ( current != NULL) && ( current->handle != activeHandles[i] ) ) {
				previous = current ;
				current = current->next ;
			}	/* endif */

			if ( current != NULL ) {
				if ( message->length > 0 ) {	
					gettimeofday( ( current->lastUpdate), 
						(struct timezone *) NULL );
					(void) memcpy( (void *) localTAddress, 
						(const void *) current->localTAddress, addressSize) ;
					(void) memcpy( (void *) remoteTAddress, 
						(const void *) current->remoteTAddress, addressSize) ;

					memcpy( &portl, localTAddress+4, 2 );
					memcpy( &portr, remoteTAddress+4, 2 );
					TRACE( SNMP_DEBUG_TIM,  "%u.%u.%u.%u:%d<--%d bytes--%u.%u.%u.%u:%d (CO)", localTAddress[0], localTAddress[1], localTAddress[2], localTAddress[3], portl, message->length, remoteTAddress[0], remoteTAddress[1], remoteTAddress[2], remoteTAddress[3], portr ); 

					previousDomain = *transportDomain ;

					return OK ;
				} else {
					if (current == activeConnections )
						activeConnections = activeConnections->next;
					else 	previous->next = current->next ;
					free (current) ;
				}	/* endif */
			} else return ERROR ;

			i++ ;

		}	/* end while */
/*
 *	The connection setup and data transfer phase are now finished
 * 	There might be disconnection requests
 */ 
		currentListen = COListenTable ;
		memset( allHandles, -1, sizeof(allHandles) ) ;
		memset( activeHandles, -1, sizeof(activeHandles) ) ;

		i = 0 ;
		while ( currentListen != NULL ) {
			allHandles[i++] = currentListen->handle ;
			currentListen = currentListen->next ;
		} 	/* end while */ 

/*
 *	perform the "poll" (look for activity on the connection handles)
 */ 
		if (TransCheckInputs( *transportDomain, allHandles, activeHandles ) < 0 )
			return ERROR ;

/*
 *	If there are connect Handle active accept the new connections
 */ 
		i = 0 ; 
		while (activeHandles[i] != -1 ) 
		{

			if ( TransAcceptConnection( *transportDomain, &activeHandles[i], 
							returnAddress, &dataHandle ) < 0 ) 
				return ERROR ;

			newEntry = ( activeConnectionsTable *) malloc( sizeof( activeConnectionsTable ) );
			newEntry->localTAddress = (char *) malloc( addressSize ) ;
			newEntry->remoteTAddress = (char *) malloc( addressSize ) ;
			newEntry->lastUpdate = (struct timeval *) malloc(sizeof(struct timeval));			
/*
 *	find the localTAddress corresponding with the connectHandle 
 */			
			currentListen = COListenTable ;
			j = 0 ;	

			while ( currentListen != NULL ) 
			{
				if ( currentListen->handle == activeHandles[j] ) 
				{
					(void) memcpy( (void *)newEntry->localTAddress, 
						(const void *) localTAddress, addressSize) ; 
					break ;
				} else	currentListen = currentListen->next ;

			}	/* end while */

			(void) memcpy( (void *) newEntry->remoteTAddress, 
				(const void *) returnAddress, addressSize) ; 
			newEntry->handle = dataHandle ; 
			gettimeofday( (newEntry->lastUpdate), (struct timezone *) NULL ) ;
			newEntry->next = activeConnections ;
			activeConnections = newEntry ;
			i++ ;
		} 	/* end while */
		
		break ;

	}	/* end case TCPDOMAIN */

	default: {
		return FAILED ;
	}

	} 	/* end switch */

}	/* end for */

/*
 *	no message received and all domains polled
 */ 
return ERROR ; 

}	/* end function TransportReceiveData */


/* -----------------------------------------------------------------------
 * Function:	TransportClose
 * Arguments:	transportDomain,	(in)
 * 		localTAddress		(in) 
 * Purpose:	close the handle (and associated handles!) on the given domain
 * Remarks:	-
 * -----------------------------------------------------------------------
 */
	STATUS
TransportClose( transportDomain, localTAddress)
TDOMAIN		transportDomain ;	/* in */
TADDRESS	*localTAddress ;	/* in */
{

	switch( transportDomain )
	{
	case UDPDOMAIN :
	case CLNSDOMAIN :
	case DDPDOMAIN :
	case IPXDOMAIN : 
	{
		listenTable	*closeCurrent,
				*previous ;

		closeCurrent = previous = CLListenTable ;

/*
 *	look in the active connection table for connections with given local Transport address
 */ 
		while ( closeCurrent != NULL ) {
			if ( memcmp( closeCurrent->transportAddress, 
				localTAddress, addressSize)  == 0 ) {
/*
 *	if matching address found and more then one connection then decrease one
 */ 
				if ((closeCurrent->numberOfOpens) > 1 ) 
				{
					(closeCurrent->numberOfOpens)-- ;
					return OK ;
				} else {
					TransClose( transportDomain, closeCurrent->handle ) ;
					if (closeCurrent == CLListenTable )
						CLListenTable = CLListenTable->next;
					else 	previous->next = closeCurrent->next ;
					free (closeCurrent) ; /* memory leak? taddresses? */	
				} 	/* endif */

				break ;

			} else {
				previous = closeCurrent ; 
				closeCurrent = closeCurrent->next ;
			} 	/* endif */

		} 	/*end while current */

		if ( CLListenTable == NULL )
			transportDomainTable[ transportDomain ] = FALSE ;

		return OK ; 

	}	/* end case UDPDOMAIN */
	case TCPDOMAIN :
	case CONSDOMAIN :  {
		int			i ;
		listenTable		*current = NULL, *previous = NULL ;
		activeConnectionsTable	*currentConnection = NULL ,
					*previousConnection = NULL ;

		current = previous = COListenTable ;
/*
 *	look in the active connection table for connections with given local Transport address
 */ 
		while ( current != NULL ) {
			if ( memcmp( current->transportAddress, localTAddress, addressSize)  == 0 ) {
				if (current->numberOfOpens > 1 ) {
					(current->numberOfOpens)-- ;
					return OK  ;
				} else {
/*
 *	look for active connections from same listenport 
 */ 
					currentConnection = previousConnection = activeConnections ;

					while ( currentConnection != NULL ) {
						if (memcmp( currentConnection->localTAddress, 
							localTAddress, addressSize)  == 0 ) 
						{ 
/*
 *							and close the connection
 */ 
							TransClose( transportDomain, currentConnection->handle ) ;
							if (currentConnection == activeConnections )
								activeConnections = activeConnections->next ;
							else {
								previousConnection->next  = currentConnection->next ;
							}
							free ( currentConnection ) ;
							currentConnection = previousConnection ;
						}	/* endif */
						currentConnection =  currentConnection->next ;
					} 	/* end while */

					TransClose( transportDomain, current->handle) ;
					if (current == COListenTable )
						COListenTable = COListenTable->next ;
					else {
						previous->next  = current->next ;
					}	/* endif */ 
					free ( current ) ;
					break ;
				} 	/* endif */
			} else {
				previous = current ;
				current = current->next ;
			}	/* endif memcmp */
		}	/* end while */
/*
 *	If the table is empty "unactivate" entry in transportDomainTable
 */ 
		if ( COListenTable == NULL ) 
			transportDomainTable[transportDomain] = FALSE ;

		return OK ; 
	}	/* end case TCPDOMAIN */

	default: 

	} 	/* end switch */

	return FAILED ;

} /* end of TransportCloseConnection */

/* ------------------------ end of file spmTransport.c -------------------------*/
