/* -----------------------------------------------------------------------
 *
 * Project;	UT-SNMP
 *
 * Filename:    spmTrans.c
 *
 * Date:        May 1994
 *
 * Purpose:	This file is an abstract mapping between the transport interface
 * 		and the real interface e.g. socket interface.  
 *
 * Author(s);   Vincent G.M. Berkhout, Frederic P. H. van Hengstum 
 *
 * -----------------------------------------------------------------------
 */

#include <spmTIM.h>



/*------------------------------------------------------------------------------*/
STATUS TransCheckInputs( transportDomain, allHandles, activeHandles ) 
TDOMAIN	transportDomain; 	/* in, type of transport domain */
THANDLE	*allHandles,		/* in, all handles that might be active */
	*activeHandles ; 	/* out, all active handles */
/*
 *Purpose:	
 *	maps the transport abstract check input on real function SocketCheckInputs
 *Remarks:	
 *	the transport domains are filled in for clarity
 */ 
{
	switch( transportDomain )
	{
		case UDPDOMAIN:
		case TCPDOMAIN:
			return  (SocketCheckInputs( (int *) allHandles, (int *) activeHandles ) ) ; 
		case CLNSDOMAIN :
		case CONSDOMAIN :
		case DDPDOMAIN :
		case IPXDOMAIN :
	}
	return FAILED ;
}	/* end of function TransCheckInput */


/*------------------------------------------------------------------------------*/
STATUS TransClose( transportDomain, handle)
TDOMAIN		transportDomain; 	/* in, type of transport domain */
THANDLE		handle ;		/* in, handle to close */
/*
 * Purpose:	
 *	This functions maps the abstract transport close on real transport close
 * Remarks:	
 *	handle is not a pointer, but int!
 */ 
{
	switch( transportDomain )
	{
		case UDPDOMAIN:
		case TCPDOMAIN:
			return SocketClose( (int) handle ) ;
		case CLNSDOMAIN :
		case CONSDOMAIN :
		case DDPDOMAIN :
		case IPXDOMAIN :
	}
	return FAILED ;
}	/* end of function TransClose */


/*------------------------------------------------------------------------------*/
STATUS TransPassiveOpen( transportDomain, localTAddress, connectHandle ) 
TDOMAIN		transportDomain; 	/* in, type of transport domain */
TADDRESS	*localTAddress ;	/* in, local address containing the portnumber */
THANDLE		*connectHandle ;	/* out, a listen handle for new connections */
/*
 * Purpose:	
 *	This functions maps the abstract transport passive open on the real transport 
 * 	passive open.
 * Remarks:
 *		
 */ 
{
	switch( transportDomain )
	{
		case TCPDOMAIN:
			return SocketStreamPassiveOpen ( (char *) localTAddress, 
							  (int *) connectHandle )  ; 
		case CONSDOMAIN :
	}
	return FAILED ;

} 	/* end of function TransPassiveOpen */


/*------------------------------------------------------------------------------*/
STATUS TransActiveOpen( transportDomain, remoteTAddress, dataHandle )  
TDOMAIN		transportDomain; 	/* in, type of transport domain */
TADDRESS	*remoteTAddress ;	/* in, address of remote entity */
THANDLE		*dataHandle ; 		/* out, the handle of the new connection */
/*
 * Purpose:	
 *	This functions maps the abstract transport active open on the real transport 
 * 	active open.
 * Remarks:
 * 	
 */ 
{
	switch( transportDomain )
	{
		case TCPDOMAIN:
			return SocketStreamActiveOpen (	(char *) remoteTAddress, 
							(int *) dataHandle )  ; 
		case CONSDOMAIN :
	}
	return FAILED ;

} 	/* end of function TransActiveOpen */


/*------------------------------------------------------------------------------*/
STATUS TransAcceptConnection ( transportDomain, connectHandle, remoteTAddress, dataHandle )  
TDOMAIN		transportDomain; 	/* in, type of transport domain */
THANDLE		*connectHandle,		/* in, the listen handle */
		*dataHandle ;		/* out, the handle referencing to new connection */
TADDRESS 	*remoteTAddress ;	/* out, the address of the remote peer */
/*
 * Purpose:	
 *	This functions maps the abstract transport acceptance of a connection  
 * 	on the real transport acceptance of a connection.
 * Remarks:	
 */ 
{
	switch( transportDomain )
	{
		case TCPDOMAIN:
			return SocketStreamAcceptConnection( (int *) connectHandle,
				(int *) dataHandle, (char *) remoteTAddress )  ; 
		case CONSDOMAIN :
	}
	return FAILED ;

} 	/* end of function TransAcceptConnection */
 

/*------------------------------------------------------------------------------*/
STATUS TransStreamSend( transportDomain, dataHandle, message )  
TDOMAIN		transportDomain; 	/* in, type of transport domain */
THANDLE		*dataHandle ;		/* in, data handle on which to send the message */
TMESSAGE	*message ;		/* out, the message to be send */
/*
 * Purpose:	
 *	This functions maps the abstract transport send on the real transport send.
 * Remarks:
 *		
 */ 
{
	switch( transportDomain )
	{
		case TCPDOMAIN:
			return SocketStreamSend ( (int *) dataHandle, (char *) message ) ; 
		case CONSDOMAIN :
	}
	return FAILED ;

} 	/* end of function TransStreamSend */


/*------------------------------------------------------------------------------*/
STATUS TransStreamReceive( transportDomain, dataHandle, message )  
TDOMAIN	transportDomain; 	/* in, type of transport domain */
int	*dataHandle ;		/* in, data handle on which to receive the message */
char	*message ;		/* out, the message that has been received */
/*
 * Purpose:	
 *	This functions maps the abstract transport receive on the real transport receive.
 * Remarks: -	
 */ 
{
	switch( transportDomain )
	{
		case TCPDOMAIN:
			return SocketStreamReceive ( (int *) dataHandle, (char *) message ) ; 
		case CONSDOMAIN :
	}
	return FAILED ;

} 	/* end of function TransStreamReceive */


/*------------------------------------------------------------------------------*/
STATUS TransPacketOpen( transportDomain, localTAddress, dataHandle )  
TDOMAIN		transportDomain; 	/* in, type of transport domain */
TADDRESS	*localTAddress ;	/* in, address containing the port on which to listen */
THANDLE		*dataHandle ;		/* out, the handle referencing to the new connection */
/*
 * Purpose:	
 *	This functions maps the abstract transport packet open on the real 
 * 	transport packet open.
 * Remarks:	
 */ 
{
	switch( transportDomain )
	{
		case UDPDOMAIN:
			return SocketPacketOpen ( (char *)localTAddress, 
						(int *) dataHandle ) ; 
		case CLNSDOMAIN :
		case DDPDOMAIN :
		case IPXDOMAIN :
	}
	return FAILED ;

} 	/* end of function TransPacketOpen */


/*------------------------------------------------------------------------------*/
STATUS TransPacketSend( transportDomain, dataHandle, remoteTAddress, message )  
TDOMAIN		transportDomain; 	/* in, type of transport domain */
THANDLE		*dataHandle ;		/* in, handle on which to send the message */
TADDRESS	*remoteTAddress ;	/* in, address of remote entity */
TMESSAGE	*message ;		/* in, the message to be send  */
/*
 * Purpose:	
 *	This functions maps the abstract transport packet send on the real transport send.
 * Remarks:	
 */ 
{
	switch( transportDomain )
	{
		case UDPDOMAIN:
			return SocketPacketSend ( (int *) dataHandle, 
						(char *) remoteTAddress, 
						(char *) message )  ; 
		case CLNSDOMAIN :
		case DDPDOMAIN :
		case IPXDOMAIN :
	}
	return FAILED ;
} 	/* end of function TransPacketSend */


/*------------------------------------------------------------------------------*/
STATUS TransPacketReceive( transportDomain, dataHandle, remoteTAddress, message )  
TDOMAIN		transportDomain; 	/* in, type of transport domain */
THANDLE		*dataHandle ;		/* in, handle (port) on which to receive the message */
TADDRESS	*remoteTAddress ;	/* out, address of remote entity */
TMESSAGE	*message ;		/* out, the message that has been received */
/*
 * Purpose:	
 *	This functions maps the abstract transport packet receive on the real 
 * 	transport receive.
 * Remarks:	
 */ 
{
	switch( transportDomain )
	{
		case UDPDOMAIN:
			return SocketPacketReceive ( (int *) dataHandle, 
						(char *) remoteTAddress, 
						(char *) message )  ; 
		case CLNSDOMAIN :
		case DDPDOMAIN :
		case IPXDOMAIN :
	}
	return FAILED ;

} 	/* end of function  TransPacketReceive */

/*------------------------ end of file spmTrans.c ------------------------*/


