/* -----------------------------------------------------------------------
 *
 * Project:     UT-SNMP
 * Filename:    MSG.c
 * Date:        September 1994
 * Purpose:     This file implements the Interface functions, which 
 *				are needed for the accessability of IPC_MSG(message)s.
 * Author(s):   Vincent Berkhout, Frederic P. H. van Hengstum
 *
 * -----------------------------------------------------------------------
 */

#include <apiMSG.h>
#include <snmpTypes.h>

/* 
 * -----------------------------------------------------------------------
 *
 * Function: 	msgOpen
 * Arguments:	Handle, reference to messagebuffer (OUT).
 * Returns:		OK on success and ERROR otherwise.
 * Purpose:		This function allocates space for an IPC_MSG message.
 * Remarks:		none.
 *
 * -----------------------------------------------------------------------
 */
	STATUS
msgOpen( AsnInt *handle )
{
	IPC_MSG *msg;

	msg = (IPC_MSG *) malloc( sizeof( IPC_MSG ) );
	if( ! msg ) return ERROR;
	msg->size = sizeof( IPC_MSG ) ;
	*handle = (AsnInt) msg;

	return( OK );
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	msgPut
 * Arguments:	handle, reference of message (IN).
 *				...
 *				vars, reference to variables (IN).
 * Returns:		OK on success and ERROR otherwise.
 * Purpose:		This function fills the referenced messagebuffer with
 *				the provided parameterinformation.
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS
msgPut( AsnInt *handle,
		AsnInt *id, AsnInt *type, AsnOcts *taddress, AsnInt *qos, 
		AsnInt *etype, AsnInt *eindex, AsnOid *lb, AsnOid *ub,
		AsnOcts *cle, AsnOid *clt, AsnVbl *vars )
{
	IPC_MSG *msg;

	if( handle == (AsnInt *) NULL )
		return ERROR ;
	msg = (IPC_MSG *) *handle;
	
	if( id != (AsnInt *) NULL )
		msg->id = *id;
	if( type != (AsnInt *) NULL )
		msg->type = *type;
	if( taddress != (AsnOcts *) NULL )
	{
		msg->taddressLen = taddress->octetLen;
		memcpy( msg->taddress, taddress->octs, taddress->octetLen );
	}
	if( qos != (AsnInt *) NULL )
		msg->qos = *qos;
	if( etype != (AsnInt *) NULL )
		msg->etype = *etype;
	if( eindex != (AsnInt *) NULL )
		msg->eindex = *eindex;
	if( lb != (AsnOid *) NULL ) 
	{
		msg->lbLen = lb->octetLen;
		memcpy( msg->lb, lb->octs, lb->octetLen );
	}
	if( ub != (AsnOid *) NULL ) 
	{
		msg->ubLen = ub->octetLen;
		memcpy( msg->ub, ub->octs, ub->octetLen );
	}
	if( cle != (AsnOcts *) NULL ) 
	{
		msg->cleLen = cle->octetLen;
		memcpy( msg->cle, cle->octs, cle->octetLen );
	}
	if( clt != (AsnOid *) NULL ) 
	{
		msg->cltLen = clt->octetLen;
		memcpy( msg->clt, clt->octs, clt->octetLen );
	}
	if( vars != (AsnVbl *) NULL ) 
		DynToStatVbl( vars, msg->vars );
	
	return OK ;
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	msgGet
 * Arguments: 	mType, type of message (IN).
 *				mHandle, reference to messagebuffer (OUT).
 *				mMsg, reference to message (OUT).
 * Returns:		OK on success and ERROR otherwise.
 * Purpose:		This fucntion searches for message with type mType. If found,
 *				the parameters will be filled with information contained in
 *				the message.
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS
msgGet( AsnInt *handle,
        AsnInt *id, AsnInt *type, AsnOcts *taddress, AsnInt *qos,
		AsnInt *etype, AsnInt *eindex, AsnOid *lb, AsnOid *ub,
        AsnOcts *cle, AsnOid *clt, AsnVbl *vars )

{
    IPC_MSG *msg;

    if( handle == (AsnInt *) NULL )
        return ERROR ;
    msg = (IPC_MSG *) *handle;

    if( id != (AsnInt *) NULL )
        *id = msg->id;
    if( type != (AsnInt *) NULL )
        *type = msg->type;
    if( taddress != (AsnOcts *) NULL )
    {
        taddress->octetLen = msg->taddressLen;
        memcpy( taddress->octs, msg->taddress, taddress->octetLen );
    }
    if( qos != (AsnInt *) NULL )
        *qos = msg->qos;
    if( etype != (AsnInt *) NULL )
        *etype = msg->etype;
    if( eindex != (AsnInt *) NULL )
        *eindex = msg->eindex;
    if( lb != (AsnOid *) NULL )
    {
        lb->octetLen = msg->lbLen;
        memcpy( lb->octs, msg->lb, lb->octetLen );
    }
    if( ub != (AsnOid *) NULL )
    {
        ub->octetLen = msg->ubLen;
        memcpy( ub->octs, msg->ub, ub->octetLen );
    }
    if( cle != (AsnOcts *) NULL )
    {
        cle->octetLen = msg->cleLen;
        memcpy( cle->octs, msg->cle, cle->octetLen );
    }
    if( clt != (AsnOid *) NULL )
    {
        clt->octetLen = msg->cltLen;
        memcpy( clt->octs, msg->clt, clt->octetLen );
    }
    if( vars != (AsnVbl *) NULL )
		StatToDynVbl( msg->vars, vars );

	return( OK );
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	msgClose
 * Arguments:	handle, reference to messagebuffer.
 * Returns:		OK on success or otherwise ERROR.
 * Purpose:		Function clears messagebuffer.
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	STATUS
msgClose( AsnInt handle )
{
	IPC_MSG *msg = (IPC_MSG *) handle;

	if( msg != (IPC_MSG *) NULL )
	{
		free( (char *) handle );
		return( OK );
	}
	return( ERROR );
}



/* -----------------------------------------------------------------------
 */
