/* -----------------------------------------------------------------------
 *
 * project:     UT-SNMP
 *
 * Filename:    apiTYPES.c
 *
 * Date:        June 1994
 *
 * Purpose:     Additional functions to fool around with variable bindings
 *				in a more convenient manner.
 *
 * Author(s):   Vincent Berkhout, Frederic P. H. van Hengstum
 *
 * -----------------------------------------------------------------------
 */

#include <snmpTypes.h>



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mOcNew
 *
 * Arguments:	l, length of string (IN).
 *
 * Return:	On success, reference to created variable, or NULL otherwise.
 *
 * Purpose:    	Allocate space for variable.
 *
 * Remarks:	The allocation of the string field is an desired 
 *		side-effect !
 *
 * -----------------------------------------------------------------------
 */
	AsnOcts *
mOcNew( unsigned int l )
{
	AsnOcts *p;
	p = (AsnOcts *) Asn1Alloc( sizeof( AsnOcts ) );
	p->octs = (unsigned char *) Asn1Alloc( p->octetLen = l );
	return( p );
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mOcPut
 *
 * Arguments:	p, reference to variable (IN).
 *		s, contence of string (IN).
 *		l, length of string (IN).
 *
 * Return:	Reference to variable.
 *
 * Purpose:    	Fill variable with string and length.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	AsnOcts *
mOcPut( AsnOcts *p, unsigned char *s, unsigned int l )
{
	memcpy( p->octs, s, p->octetLen = l );
	return( p );
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mOcGet
 *
 * Arguments:	p, reference to variable (IN).
 *		s, contence of string (OUT).
 *		l, length of string (OUT).
 *
 * Return:	Reference to variable.
 *
 * Purpose: 	Fill string and length with variable values.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	AsnOcts *
mOcGet( AsnOcts *p, unsigned char *s, unsigned int *l )
{
	memcpy( s, p->octs, *l = p->octetLen );
	return( p );
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mBiNew
 *
 * Arguments:	l, length of string (bits) (IN).
 *
 * Return:	On success, reference to created variable, or NULL otherwise.
 *
 * Purpose:    	Allocate space for variable.
 *
 * Remarks:	The allocation of the (bit)string filed is an desired 
 *		side-effect !
 *
 * -----------------------------------------------------------------------
 */
	AsnBits *
mBiNew( int l )
{
	AsnBits *p;
	p = (AsnBits *) Asn1Alloc( sizeof( AsnBits ) );
	p->bits = (char *) Asn1Alloc( p->bitLen = l );
	return( p );
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mBiPut
 *
 * Arguments:	p, reference to variable (IN).
 *		s, contence of string (bits) (IN).
 *		l, length of string (IN).
 *
 * Return:	Reference to variable.
 *
 * Purpose:    	Fill variable with string and length.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	AsnBits *
mBiPut( AsnBits *p, char *s, int l )
{
	memcpy( p->bits, s, p->bitLen = l );
	return( p );
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mBiGet
 *
 * Arguments:	p, reference to variable (IN).
 *		s, contence of string (OUT).
 *		l, length of string (OUT).
 *
 * Return:	Reference to variable. 
 *
 * Purpose:    	Fill variable with string (bits) and length.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	AsnBits *
mBiGet( AsnBits *p, char *s, int *l )
{
	memcpy( s, p->bits, *l = p->bitLen );
	return( p );
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mObNew
 *
 * Arguments:	none.	
 *
 * Return:	On success, reference to created variable, or NULL otherwise.
 *
 * Purpose:    	Allocate space for variable.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	ObjectSyntax *
mObNew( )
{
	return (ObjectSyntax *) Asn1Alloc( sizeof( ObjectSyntax ) );
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mObPut
 *
 * Arguments:	p, reference to variable (IN).
 *		i, variable type (IN).
 *		v, variable value (IN).
 *
 * Return:	Reference to variable.
 *
 * Purpose:    	Fill variable with type and value.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	ObjectSyntax *
mObPut( ObjectSyntax *p, ObjectSyntaxChoiceId i, AsnInt v )
{
	p->choiceId = i;
	p->a.integer_value = v;
	return p;
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mObGet
 *
 * Arguments:	p, reference to variable (IN).
 *		i, variable type (IN).
 *
 * Return:	Reference to variable value.
 *
 * Purpose:    	Fill type and value with varibale values.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	AsnInt 
mObGet( ObjectSyntax *p, ObjectSyntaxChoiceId *i) 
{
	*i = p->choiceId;
	return p->a.integer_value;
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mVcPut
 *
 * Arguments:	p, reference to variable (IN).
 *		i, variable type (IN).
 *		v, variable value (IN).
 *
 * Return:	Reference to variable.
 *
 * Purpose:    	Fill variable with type and value.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	VarBindChoice *
mVcPut( VarBindChoice *p, VarBindChoiceChoiceId i, ObjectSyntax *v )
{
	p->choiceId = i;
	p->a.value = v;
	return p;
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mVcGet
 *
 * Arguments:	p, reference to variable (IN).
 *		i, varible type (OUT)
 *
 * Return:	Reference to variable value.
 *
 * Purpose:    	Fill type with variable type.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	ObjectSyntax *
mVcGet( VarBindChoice *p, VarBindChoiceChoiceId *i ) 
{
	*i = p->choiceId;
	return p->a.value;
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mVcNew
 *
 * Arguments:	l, length of string (IN)
 *
 * Return:	On success, reference to created variable, or NULL otherwise.
 *
 * Purpose:    	Allocate space for variable.
 *
 * Remarks:	The allocation of the string filed is an desired side-
 *		effect !
 *
 * -----------------------------------------------------------------------
 */
	VarBind *
mVbNew( unsigned int l )
{
	VarBind *p;
	p = (VarBind *) Asn1Alloc( sizeof( VarBind ) );
	p->name.octs = (unsigned char *) Asn1Alloc( p->name.octetLen=l );
	return p;
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mVbPut
 *
 * Arguments:	p, reference to variable (IN).
 *		n, name of variable (IN).
 *		v, value of variable (IN).
 *
 * Return:	Reference to variable.
 *
 * Purpose:   	Fill variable with name and value.	 
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	VarBind *
mVbPut( VarBind *p, ObjectName *n, VarBindChoice *v )
{
	memcpy( p->name.octs, n->octs, p->name.octetLen=n->octetLen );
	p->varBindChoice = v;
	return p;
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mVbGet
 *
 * Arguments:	p, reference to variable (IN).
 *		n, name of variable (OUT).
 *
 * Return:	Reference to variable value.
 *
 * Purpose:    	Fill name with variable name.
 *
 * Remarks:
 *
 * -----------------------------------------------------------------------
 */
	VarBindChoice *
mVbGet( VarBind *p, ObjectName *n )
{
	memcpy( n->octs, p->name.octs, n->octetLen=p->name.octetLen );
	return p->varBindChoice;
}

/* -----------------------------------------------------------------------
 */

