/*
\begin{verbatim}
*/

/*********************************************************************/
/*                                                                   */
/* SNMP-DPI API - SNMP Distributed Protocol Interface                */
/*                Application Programming Interface                  */
/*                                                                   */
/* Feb 03, 1994 - Version 0.13i                                      */
/*                                                                   */
/* Copyright    - (C) International Business Machines Corp. 1994     */
/*                                                                   */
/*   Permission to use, copy, modify, and distribute this software   */
/*   and its documentation for any lawful purpose and without fee is */
/*   hereby granted, provided that this notice be retained unaltered,*/
/*   and that the names of IBM and all other contributors shall not  */
/*   be used in advertising or publicity pertaining to distribution  */
/*   of the software without specific written prior permission.      */
/*   No contributor makes any representations about the suitability  */
/*   of this software for any purpose.  It is provided "as is"       */
/*   without express or implied warranty.                            */
/*                                                                   */
/*   IBM AND ALL OTHER CONTRIBUTORS DISCLAIM ALL WARRANTIES WITH     */
/*   REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF    */
/*   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, TITLE,    */
/*   AND NON-INFRINGEMENT.                                           */
/*                                                                   */
/*   IN NO EVENT SHALL IBM OR ANY OTHER CONTRIBUTOR BE LIABLE FOR    */
/*   ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, WHETHER IN      */
/*   CONTRACT, TORT, OR OTHER FORM OF ACTION, ARISING OUT OF OR IN   */
/*   CONNECTION WITH, THE USE OR PERFORMANCE OF THIS SOFTWARE.       */
/*                                                                   */
/* snmp_lDPI.c     - system specific code for DPI related functions. */
/*                 - this sample code works for OS/2 and AIX 3.2     */
/*                   and AIX/ESA. It probably also works on other    */
/*                   UNIX systems that support the BSD 4.3 socket    */
/*                   API. Had to do some #ifdef stuff though, see    */
/*                   notes below.                                    */
/*                                                                   */
/*-------------------------------------------------------------------*/
/* Note(s):                                                          */
/* - When compiling this source, use flags to set ANSI C checking.   */
/*   Further, use these extra flags for specific Operating Systems:  */
/*   - OS/2 (C-set/2 compiler): -DOS2 and/or -DOS2_SELECT            */
/*     If you only specify -DOS2, then we use -DBSD_SELECT on OS/2   */
/*   - AIX/ESA (MetaWare High C compiler) does not need special flags*/
/*   - AIX 3.x (ANSI C-compiler) does not need special flags         */
/*-------------------------------------------------------------------*/

#ifdef RCS_ID
static char rcsid[] =
 "$Id: snmp_lDPI.c,v 1.3 1994/02/04 00:38:36 bert Exp $";
#endif /* RCS_ID */

/* Change activitity:
 *
 * $Log: snmp_lDPI.c,v $
 * Revision 1.3  1994/02/04  00:38:36  bert
 * Fixes after DPI 2.0 review
 *
 * Revision 1.2  1994/01/24  17:05:57  bert
 * dpiPortForXXX names
 *
 * Revision 1.1  1994/01/03  22:24:55  bert
 * Initial revision
 *
 */

/*********************************************************************/
/* Here is some OS/2 specific stuff. Compile with flags:             */
/* -DOS2           - will cause OS/2 specific stuff, but it will use */
/*                   BSD-style select(). So it defines BSD_SELECT.   */
/* -DOS2_SELECT    - will cause OS/2 specific stuff and will also    */
/*                   OS/2 specific form of select().                 */
/* See TCP/IP 2.0 for OS/2 Programmers Reference for details on the  */
/* 2 forms of select(0 processing.                                   */
/*********************************************************************/
#ifdef  OS2_SELECT               /* If you want to use OS2_SELECT    */
#ifndef OS2                      /* then that is fine, and I will    */
#define OS2                      /* make sure that we us OS/2        */
#endif  /* OS2 */                /* methods for socket functions     */
#else   /* BSD_SELECT */         /* This is my default as on UNIX    */
#ifdef  OS2                      /* So if we are on OS/2, then I     */
#define BSD_SELECT               /* must define this for BSD_SELECT  */
#endif  /* OS2 */                /* in oder to use BSD-like methods  */
#endif  /* OS2_SELECT */         /* Sorry... What can I say.         */

#ifdef  OS2                      /* If on OS/2, sockets are special  */
#define CLOSE soclose            /* -  close()    is soclose()       */
#define PERROR psock_errno       /* -  perror()   is pscokerrno()    */
#include <types.h>               /* -  types.h is not in sys on OS/2 */
#include <sys/select.h>
#else   /* not OS2 */            /* If not OS/2 (which is default)   */
#define CLOSE close              /* we use standard BSD 4.3 style    */
#define PERROR perror            /* also for this one                */
#endif  /* OS2 */                /* end OS2 ed on OS/2               */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netdb.h>

#include <snmp_lDPI.h>

/*********************************************************************/
/* These are definitions for values used throughout this program     */
/*********************************************************************/
#define SNMP_QUERY_PORT_TRIES 3  /* max tries for SNMP GET dpiPort.x */
#define RESPONSE_TIMEOUT      3  /* default timeout (seconds)        */
#define SNMP_PORT           161  /* standard SNMP (agent) port       */
#define SNMP_TRAP_PORT      162  /* SNMP TRAP port being used        */
#define SNMP_COMMUNITY  "public" /* Commonly used public community   */
#define TRAP_COMMUNITY  "public" /* Commonly used public community   */
#define IP_LOOPBACK  "localhost" /* Common IP loopback address       */

static int mustread(int fd, char *buf, int len, int type);
/*static int read_dpi_packet_on_fd(int fd, int type);                 */               
static void dump_bfr(CONST void *buf, int len, CONST void *hdr);
/*#define MAX_CONNECTIONS 3 changed into 32, see snmp_lDPI.h          */

int   connection_fds[MAX_CONNECTIONS+1];   /* file descriptors  */
						  /* +1 for listenSock */
int   connection_types[MAX_CONNECTIONS+1]; /* socket types      */

/* defines below are placed in snmp_lDPI.h                             */
/*             #define UDP_connection 1           /* one of these      */
/*             #define TCP_connection 2                                */
/*             #define NMQ_connection 3                                */
/*             #define SNA_connection 4                                */

struct sockaddr_in 						
               connection_socks[MAX_CONNECTIONS]; /* socket addresses  */
char *connection_qnames[MAX_CONNECTIONS];  /* queue names       */
int   init_not_yet_done = 1;

char  dpi_buffer[SNMP_DPI_BUFSIZE];        /* buffer for received   */
int   dpi_buflen;                          /* (DPI) packets; length */

/*********************************************************************/
/* common routine to create a socket, bind it and then connect if it */
/* is a SOCK_STREAM socket.                                          */
/* It returns -1 if we fail;                                         */
/* returns handle(index) into connections array if success.          */
/*********************************************************************/
int setup_socket(                  /* setup/connect a socket  */
  int                  socket_type,       /* for this type           */
  char                *hostname_p,        /* destination host name   */
  long int             port)              /* destination port        */
{
   int                 fd;                /* for file descriptor     */
   int                 rc;                /* for return code         */
   int                 i;                 /* for loop control        */
   struct sockaddr_in  s_sock,            /* source socket structure */
                       d_sock;            /* destination socket str. */
   unsigned long       host_addr;         /* for IP address          */
 
   if (init_not_yet_done) {
   /********* i was origineel 0 **************/
      for (i=1; i < MAX_CONNECTIONS; i++) /* set all entries free    */
      {                                   /* when we get here for    */
          connection_fds[i] = -1;         /* the first time          */
      } /* endfor */
#ifdef OS2
      rc = sock_init();                   /* OS/2 here we come       */
      if (rc) {                           /* what?? you failed me!!  */
         DPI_DEBUG(1,PERROR(              /* if debug active, then   */
                     "sock_init():"));    /* print error msg and     */
         return(-1);                      /* -1 means failure        */
      } /* endif */
#endif /* OS2 */
      init_not_yet_done = 0;              /* reset first time flag   */
   } /* endif */

   host_addr = lookup_host(hostname_p);   /* find target IP address  */
   if (host_addr == 0) {                  /* if unknown, that's it   */
      DPI_DEBUG(1,DPI_DM(                 /* if debug_lvl >= 1 tell  */
        DPI_M_NO_IPADDR_FOR_HOST,         /* that we could not find  */
        hostname_p));                     /* IP address              */
      return(-1);                         /* -1 means failure        */
   } /* endif */

   fd = socket(AF_INET,socket_type,0);    /* create a socket         */
   if (fd < 0) {                          /* failure to create socket*/
      DPI_DEBUG(1,PERROR("socket():"));   /* print error msg and     */
      return(-1);                         /* -1 means failure        */
   } /* endif */
 
   memset(&s_sock,0,sizeof(s_sock));      /* clear source structure  */
   s_sock.sin_family      = AF_INET;      /* set AF_INET family      */
   s_sock.sin_port        = 0;            /* give us any port        */
   s_sock.sin_addr.s_addr =               /* and we are hapy with    */
                       htonl(INADDR_ANY); /* any local IP address    */
   rc = bind(fd,                          /* bind our socket (fd)    */
             (struct sockaddr *)&s_sock,  /* to this address         */
             sizeof(s_sock));             /* length of the structure */
   if (rc < 0) {                          /* if we failed, then      */
      DPI_DEBUG(1,PERROR("bind():"));     /* print error msg and     */
      close(fd);                          /* close the socket, and   */
      return(-1);                         /* -1 means failure        */
   } /* endif */
/****************** i was origineel 0 ***************/
   for (i=1; i < MAX_CONNECTIONS; i++) {  /* find free netry in our  */
      if (connection_fds[i] == -1) break; /* table of connections    */
   } /* endfor */
   if (i >= MAX_CONNECTIONS) {            /* no free entry           */
      DPI_DEBUG(1,DPI_DM(                 /* if debug_lvl >= 1 tell  */
          DPI_M_ALL_CONNECTIONS_USED,     /* user about it and what  */
          MAX_CONNECTIONS));              /* MAX number is           */
      return(-1);                         /* -1 means failure        */
   } /* endif */

   memset(&d_sock,0,sizeof(d_sock));      /* clear destination sock  */
   d_sock.sin_family = AF_INET;           /* set AF_INET family      */
   d_sock.sin_port   = htons(port);       /* set requested port      */
   d_sock.sin_addr.s_addr = host_addr;    /* destination IP address  */

   switch (socket_type) {                 /* do by socket type       */
   case SOCK_DGRAM:                       /* for a UDP socket        */
     connection_types[i] = UDP_connection;
     break;                               /* we are all done         */
   case SOCK_STREAM:                      /* for a TCP socket        */
     connection_types[i] = TCP_connection;
     rc = connect(fd,                     /* connect to target       */
              (struct sockaddr *)&d_sock, /* this is address/port    */
              sizeof(d_sock));            /* size of structure       */
     if (rc < 0) {                        /* if we failed, then      */
         DPI_DEBUG(1,PERROR("connect():"));  /* print error msg and  */
         close(fd);                       /* close the socket, and   */
         return(-1);                      /* -1 means failure        */
     } /* endif */
     break;
   default:                               /* never expect to be here */
     close(fd);                           /* close the socket, and   */
     return(-1);                          /* -1 means failure        */
     /* break; */                         /* NEVER REACHED           */
   } /* endswitch */

   if (fd >=0 ) {                         /* if valid socket (fd)    */
      connection_fds[i]   = fd;           /* save the fd             */
      memcpy(&connection_socks[i],        /* copy destination socket */
          &d_sock,                        /* address structure       */
          sizeof(struct sockaddr_in));    /* size of such structure  */
   } /* endif */

   DPI_DEBUG(1,DPI_DM(                    /* if any debug active,    */
       DPI_M_CONNECTION_SETUP,            /* print information msg   */
       (socket_type == SOCK_STREAM) ?     /* and pass TCP or UDP     */
       "TCP" : "UDP", fd));               /* and file descriptor     */
   return(i);                             /* return index/handle     */
}

/*********************************************************************/
/* Close and reset a connection as specified by handle into our array*/
/* we assume handle is valid, we are static function, so checking is */
/* done elsewere in thos source file before calling us               */
/*********************************************************************/
static void close_socket_by_handle(    /* close socket by handle     */
                          int handle)  /* handle into our array      */
{
  if ((handle<0)                ||     /* ensure valid handle        */
      (handle>=MAX_CONNECTIONS) ||     /* ...                        */
      (connection_fds[handle] == -1))  /* and active connection      */
  {
     return;                           /* if not: nothing to do      */
  } /* endif */

  DPI_DEBUG(1,DPI_DM(                  /* if debug_lvl >=1 tell user */
      DPI_M_CLOSING_FD,                /* that we are closing this   */
      connection_fds[handle]));        /* file descriptor            */

  close(connection_fds[handle]);       /* close file descriptor      */
  connection_fds[handle] = -1;         /* reset entry in array       */
}

/*********************************************************************/
/* Close and reset a connection as specified by file descriptor      */
/*********************************************************************/
static void close_socket_by_fd(int fd)    /* close socket connection */
{
   int i;                                 /* for loop control        */

   close(fd);                             /* always close socket     */
   for (i=0; i <= MAX_CONNECTIONS; i++) { /* find the entry in our   */
      if (connection_fds[i] == fd) break; /* table of connections    */
   } /* endfor */
   if (i >= MAX_CONNECTIONS) return;      /* not in our table.       */

   DPI_DEBUG(1,DPI_DM(                    /* if debug_lvl >=1 tell   */
      DPI_M_CLOSING_FD,                   /* that we are closing     */
      connection_fds[i]));                /* this file descriptor    */

   connection_fds[i] = -1;                /* reset entry to free     */
}

/*********************************************************************/
/* Function to wait for data (or anything) on a filedescriptor fd    */
/* The time_out argument is the maximum number of seconds to wait.   */
/* A time_out value of zero means: do not wait, just check if data   */
/* is available. A time_out value of -1 means wait forever untill    */
/* data (or EOF) arrives.                                            */
/*********************************************************************/
int wait_for_data(                 /* wait for data to arrive */
  int                       fd,           /* on this file descriptor */
  int                       time_out)     /* max wait in useconds     */
{
   struct timeval    timeout;             /* for timeout spec        */
   int               rc;                  /* for select return code  */

   timeout.tv_sec  = time_out;            /* timeout value (seconds) */
   timeout.tv_usec = 0;                   /* zero micro seconds      */
 
#ifndef OS2_SELECT
{
   struct fd_set readfds;                 /* for BSD-type select()   */
   struct timeval *t_p;                   /* ptr to time value       */
   if (time_out == -1) {                  /* minus 1 means wait      */
      t_p = (struct timeval *)0;          /* forever, which must be  */
   } else {                               /* passed as a NULL ptr    */
      t_p = &timeout;                     /* else just point to the  */
   } /* endif */                          /* timeval structure       */
   FD_ZERO(&readfds);                     /* zero out bit mask       */
   FD_SET(fd, &readfds);                  /* set fd to wait for      */
   rc = select(fd+1, &readfds, 0, 0, t_p);/* await response which is */
}                                         /* expected within timeout */
#else /* TCP/IP OS/2 SELECT */
{
printf("In de else van wait_for_data\n");
   int socks[2];                          /* only for OS/2 select    */
   if (time_out && time_out != -1) {      /* special timeout values  */
      timeout.tv_sec *= 1000L;            /* must not be multiplied  */
   } /* endif */                          /* for OS/2 either         */
   socks[0] = fd;                         /* set fd to wait for      */
   rc = select(socks,1,0,0,               /* await response which is */
               timeout.tv_sec);           /* expected within timeout */
printf("In wait_for_data rc=%d",rc);
}
#endif /* OS2_SELECT */

   if (rc >= 1) {                         /* This means fd is ready  */
      rc = DPI_RC_OK;                     /* so set RC OK            */
   } else if (rc == 0) {                  /* zero means timeout      */
      DPI_DEBUG(1,DPI_DM(                 /* if debug_lvl >=1 tell   */
        DPI_M_TIMEOUT_ON_FD,time_out,fd));/* user about timeout      */
      rc = DPI_RC_TIMEOUT;                /* and pass RC to caller   */
   } else {                               /* anything else is error  */
      DPI_DEBUG(1,PERROR("select():"));   /* say so if debug is on   */
      rc = DPI_RC_IO_ERROR;               /* set retcode             */
   } /* endif */
   return(rc);                            /* pass retcode to caller  */
}

/*********************************************************************/
/* Function lookup_host() returns the IpAddress (network byte order) */
/* of the specified host. The first IpAddress found is returned.     */
/* The host can be specified by name or in standard dot notation.    */
/*********************************************************************/
unsigned long lookup_host(char *hostname_p) /* lookup IpAddress      */
{
   register unsigned long  ret_addr;        /* receives the address  */
   struct hostent         *host_p;          /* ptr to hostent struct */
   struct in_addr         *addr_p;          /* ptr to in_addr struct */

   if (isdigit(*hostname_p)) {              /* Starts with digit,    */
      ret_addr = inet_addr(hostname_p);     /* try if dot notation   */
      if (ret_addr) return(ret_addr);       /* if found, then return */
   } /* endif */                            /* the address           */

   host_p = gethostbyname(hostname_p);      /* Try if known by name  */
   if (host_p) {                            /* if found, then point  */
      addr_p = (struct in_addr *)           /* to first address in   */
               (host_p->h_addr_list[0]);    /* list of addresses     */
      ret_addr = addr_p->s_addr;            /* pick up that addess   */
   } else {                                 /* hostname not found    */
      ret_addr = 0;                         /* set address to zero   */
   } /* endif */

   return(ret_addr);                        /* return the address    */
}

/*********************************************************************/
/* Query the DPI port as known by the SNMP agent at specified host.  */
/* A community name must be passed so that we can use it on the SNMP */
/* GET request that we send to the agent.                            */
/* - A NULL ptr for the hostname is interpreted as "localhost"       */
/* - A NULL ptr for the community name is interpreted as "public"    */
/*********************************************************************/
long int              qDPIport(        /* Query (GET) SNMP_DPI port  */
  char                 *hostname_p,    /* target hostname/IPaddress  */
  char                 *community_p,   /* communityname for GET      */
  int                   instance)      /* port instance:             */
                                       /*   dpiPortForTCP (=1)       */
                                       /*   dpiPortForUDP (=2)       */
{
   unsigned char      *packet_p;          /* for ptr to SNMP packet  */
   int                 packet_len;        /* for length of packet    */
   int                 community_len;     /* for length of community */
   int                 handle;            /* for connection          */
   long int            rc;                /* for return code/dpiPort */
   int                 tries;             /* for count of tries      */
   struct sockaddr_in  fsock;
   int                 fsock_len;

   community_len = strlen(community_p);   /* get length of community */
 
   packet_p = mkDPIqport_packet(          /* make an SNMP GET packet */
                1,                        /* for SNMP version 1      */
                instance,                 /* for dpiPort.instance.0  */
                community_p,              /* this community name     */
                community_len,            /* this community length   */
                &packet_len);             /* receives packet length  */

   if (packet_p) {                        /* if packet got created   */
      handle = setup_socket(SOCK_DGRAM,   /* setup socket for UDP    */
                     hostname_p,          /* communication to this   */
                     SNMP_PORT);          /* host and the SNMP port  */
      if (handle < 0) return(-1L);        /* no socket available     */
   } else {                               /* couldn't create packet  */
      return(-1L);                        /* so return -1, meaning   */
   } /* endif */                          /* can't find DPIport      */
 
   tries = 0;                             /* zero tries done sofar   */
   while (tries++ < SNMP_QUERY_PORT_TRIES)/* try max somany times    */
   {                                      /* to send an SNMP packet  */
     rc = sendto(connection_fds[handle],  /* to agent via this fd    */
              (char *)packet_p,packet_len,/* ptr to packet; length   */
              0,                          /* flags = 0               */
              (struct sockaddr *)         /* ptr to sock_addr struct */
              &connection_socks[handle],  /* is this one, but it     */
              sizeof(struct sockaddr_in));/* has this size           */
     rc = wait_for_data(                  /* wait for a response     */
                   connection_fds[handle],/* on this file descriptor */
                   RESPONSE_TIMEOUT);     /* but max this long       */
     if (rc != DPI_RC_OK) continue;       /* if not, try again       */
     fsock_len = sizeof(fsock);
     dpi_buflen = recvfrom(               /* now receive a packet    */
            connection_fds[handle],       /* on this file descriptor */
            dpi_buffer,                   /* into DPI buffer (one    */
            sizeof(dpi_buffer),           /* time use) max this size */
            0,                            /* flags = 0, and a NULL   */
            (struct sockaddr *)&fsock,    /* ptr for address info    */
            &fsock_len);                  /* length for addr info    */
     if (dpi_buflen <= 0) {               /* if error, return -1     */
        if (dpi_buflen < 0) {             /* error on recvfrom()     */
            DPI_DEBUG(1,PERROR(           /* if debug active, then   */
                      "recvfrom():"));    /* display the error       */
        } else {                          /* else received EOF       */
            DPI_DEBUG(1,DPI_DM(           /* if debuge active        */
                DPI_M_EOF_RECEIVED,"UDP", /* display this message    */
                connection_fds[handle]));
        } /* endif */
        close_socket_by_handle(handle);   /* always close socket     */
        return(-1L);                      /* error, return -1        */
     } /* endif */
     close_socket_by_handle(handle);      /* no error. close socket  */
     rc = extract_DPI_port(               /* extract the DPIport from*/
                  dpi_buffer,             /* this buffer which is    */
                  dpi_buflen);            /* the SNMP response PDU   */
     return(rc);                          /* and pass it to caller   */
   } /* endwhile */

   close_socket_by_handle(handle);        /* an error. close socket  */
   return(-1L);                           /* so close and return -1  */
}

/*********************************************************************/
/*                                                                   */
/*********************************************************************/
int DPIconnect_to_agent_TCP(           /* Connect to DPI TCP port    */
  char                 *hostname_p,    /* target hostname/IPaddress  */
  char                 *community_p)   /* communityname              */
{
   long int    port;                   /* receives DPI port number   */
   int         handle;                 /* receives connection handle */

   port = qDPIport(hostname_p,         /* find DPI port at this host */
              community_p,             /* with this community name   */
              dpiPortForTCP);          /* and for the TCP port       */
   if (port <= 0) {                    /* can't find port number     */
      return(DPI_RC_NO_PORT);          /* so return this error       */
   } /* endif */

   handle = setup_socket(SOCK_STREAM,  /* setup socket with TCP      */
                  hostname_p,          /* connection to this host    */
                  port);               /* and with the DPI TCP port  */

   if (handle >= 0) return(handle);    /* success, return handle     */
   return(DPI_RC_IO_ERROR);            /* else return this error     */
}

/*********************************************************************/
/*                                                                   */
/*********************************************************************/
int DPIconnect_to_agent_UDP(           /* Connect to DPI UDP port    */
  char                 *hostname_p,    /* target hostname/IPaddress  */
  char                 *community_p)   /* communityname              */
{
   long int    port;                   /* receives DPI port number   */
   int         handle;                 /* receives connection handle */

   port = qDPIport(hostname_p,         /* find DPI port at this host */
              community_p,             /* with this community name   */
              dpiPortForUDP);          /* and for the UDP port       */
   if (port <= 0) {                    /* can't find port number     */
      return(DPI_RC_NO_PORT);          /* so return this error       */
   } /* endif */

   handle = setup_socket(SOCK_DGRAM,   /* setup socket with TCP      */
                  hostname_p,          /* connection to this host    */
                  port);               /* and with the DPI TCP port  */

   if (handle >= 0) return(handle);    /* success, eturn handle      */
   return(DPI_RC_IO_ERROR);            /* else return this error     */
}

/*********************************************************************/
/*                                                                   */
/*********************************************************************/
int DPIconnect_to_agent_NMQ(           /* Connect to DPI Named Queue */
  char                 *t_qname_p,     /* target (agent) queue name  */
  char                 *s_qname_p)     /* source (subagent) q name   */
{
   int i;
   /* guess you want to setup a connection and save info in the array*/
   /* of connections, similar to how we do it for TCP and UDP.       */
   /* then you return the handle into that array. An error NO_PORT   */
   /* can be interpreted as NO_NAMED_QUEUE, or maybe you want to add */
   /* some DPI_RC_xxxx codes for names queues to the snmp_dpi.h file */

   for (i=0; i < MAX_CONNECTIONS; i++) {  /* find free entry in our  */
      if (connection_fds[i] == -1) break; /* table of connections    */
   } /* endfor */
   if (i >= MAX_CONNECTIONS) {            /* no free entry           */
      DPI_DEBUG(1,DPI_DM(                 /* if debug_lvl >= 1 tell  */
          DPI_M_ALL_CONNECTIONS_USED,     /* user about it and what  */
          MAX_CONNECTIONS));              /* MAX number is           */
      return(-1);                         /* -1 means failure        */
   } /* endif */

   connection_qnames[i] = (char *)        /* allocate memeory for    */
              malloc(strlen(t_qname_p));  /* copy of target q name   */
   if (connection_qnames[i]) {            /* if success with malloc  */
      memcpy(connection_qnames[i],        /* then copy the q name    */
           t_qname_p, strlen(t_qname_p)); /* to our own area         */
   } else {                               /* if failure on malloc    */
      DPI_DEBUG(1,DPI_DM(                 /* if debug_lvl >=1 print  */
        DPI_M_OUT_OF_MEMORY));            /* this message            */
      return(DPI_RC_NO_PORT);             /* and return this error   */
   } /* endif */

   /* guess you want to setup a connection and save more in array    */
   /* of connections, similar to how we do it for TCP and UDP.       */
   /* then you return the handle into that array. An error NO_PORT   */
   /* can be interpreted as NO_NAMED_QUEUE, or maybe you want to add */
   /* some DPI_RC_xxxx codes for names queues to the snmp_dpi.h file */

   return(DPI_RC_NO_PORT);             /* return this error          */
}

/*********************************************************************/
/*                                                                   */
/*********************************************************************/
void DPIdisconnect_from_agent(         /* disconnect from DPI (agent)*/
  int                   handle)        /* close this connection      */
{

  if ((handle<0)                ||     /* ensure valid handle        */
      (handle>=MAX_CONNECTIONS) ||     /* ...                        */
      (connection_fds[handle] == -1))  /* and active connection      */
  {
     return;                           /* if not: nothing to do      */
  } /* endif */

  switch (connection_types[handle]) {
  case UDP_connection:
  case TCP_connection:
    close_socket_by_handle(handle);    /* close the connections      */
    break;
  case NMQ_connection:
  case SNA_connection:
    /* do whatever needs to be done.. close connection..  etc.       */
    if (connection_qnames[handle]) {      /* if we allocated memory  */
       free(connection_qnames[handle]);   /* free it now             */
    } /* endif */
    connection_fds[handle] = -1;          /* reset entry to free     */
    break;
  default:
    break;
  } /* endswitch */
}

/*********************************************************************/
/* Await a DPI packet from the SNMP agent.                           */
/* - stores ptr to buffer in callers message_p argument              */
/*   (NULL ptr if we fail to receive a packet)                       */
/* - stores packet length in callers length argument                 */
/*   (zero if we fail to receive a packet                            */
/* Possible return codes:                                            */
/* - DPI_RC_OK             - all OK, ptr and length stored           */
/* - DPI_RC_NOK            - undefined error occured.  ptr/length 0  */
/* - DPI_RC_INVALID_HANDLE - an invalid handle passed. ptr/length 0  */
/* - DPI_RC_TIMEOUT        - a timeout occured.        ptr/length 0  */
/* - others as defined in include file snmp_dpi.h                    */
/*********************************************************************/
int DPIawait_packet_from_agent(        /* await a DPI packet         */
  int                   handle,        /* on this connection         */
  int                   timeout,       /* timeout in seconds.        */
  unsigned char       **message_p,     /* receives ptr to data       */
  unsigned long        *length)        /* receives length of data    */
{
  int           rc;                    /* for return code            */

  *message_p = (unsigned char *)0;     /* no data yet, so NULL ptr   */
  *length    = 0;                      /* and also zero length       */

  if ((handle<0)                ||     /* ensure valid handle        */
      (handle>=MAX_CONNECTIONS) ||     /* ...                        */
      (connection_fds[handle] == -1))  /* and active connection      */
  {
     DPI_DEBUG(1,DPI_DM(               /* if debug_lvl >= 1 tell     */
         DPI_M_INVALID_HANDLE,handle));/* this message               */
     return(DPI_RC_INVALID_HANDLE);    /* if not return this error   */
  } /* endif */

  switch (connection_types[handle]) {  /* handle by connection type  */
  case UDP_connection:                    /* UDP connection,         */
  case TCP_connection:
    rc = wait_for_data(                   /* wait till there is data */
              connection_fds[handle],     /* on this fd              */
              timeout);                   /* max wait in seconds     */
    if (rc != DPI_RC_OK) return(rc);      /* return if error/timeout */
    rc = read_dpi_packet_on_fd(           /* read a DPI packet       */
              connection_fds[handle],     /* on this file descriptor */
              connection_types[handle]);  /* this type of descriptor */
    if (rc > 0) {                         /* we got data if rc>0     */
       *message_p = (unsigned char *)     /* cast to unsigned char   */
                    dpi_buffer;           /* so set ptr to buffer    */
       *length    = dpi_buflen;           /* set length of packet    */
       rc = DPI_RC_OK;                    /* show that allo is OK    */
    } /* endif */
    break;
  case NMQ_connection:
  case SNA_connection:
    /* Do whatever is needed to send to named queue. You have a      */
    /* handle that is an index into the array of connections that    */
    /* you keep a ptr to the qname and a handle. The handle might be */
    /* a ptr to help you go fast, like we use the handle to keep     */
    /* the fd for UDP and TCP domains.                               */
    /* in the case of SNA you have a ptr to luname and also a handle */
    rc = DPI_RC_NOK;                      /* untill we do something  */
    break;
  default:                                /* should never occur, if  */
    DPI_DEBUG(1,DPI_DM(                   /* it does, code is out of */
        DPI_M_CODE_OUT_OF_SYNC,           /* sync in this source file*/
        __FILE__, __LINE__));             /* at this place           */
    rc = DPI_RC_NOK;                      /* set this retcode        */
    break;
  } /* endswitch */

  return(rc);                             /* pass retcode to caller  */
}

/*********************************************************************/
/*                                                                   */
/*********************************************************************/
int DPIsend_packet_to_agent(           /* send a DPI packet          */
  int                   handle,        /* on this connection         */
  CONST unsigned char  *packet_p,      /* the packet to send         */
  unsigned long         length)        /* length of packet           */
{
  int           rc;                    /* for return code            */

  if ((handle<0)                ||     /* ensure valid handle        */
      (handle>=MAX_CONNECTIONS) ||     /* ...                        */
      (connection_fds[handle] == -1))  /* and active connection      */
  {
     DPI_DEBUG(1,DPI_DM(               /* if debug_lvl >= 1 tell     */
         DPI_M_INVALID_HANDLE,handle));/* this message               */
     return(DPI_RC_INVALID_HANDLE);    /* if not return this error   */
  } /* endif */

  switch (connection_types[handle]) {
  case UDP_connection:
    rc = sendto(connection_fds[handle],   /* send packet on this fd  */
               (char *)packet_p,          /* ptr to DPI packet       */
               length, 0,                 /* packet length, flags=0  */
               (struct sockaddr *)        /* ptr to sock_addr struct */
               &connection_socks[handle], /* is this one, but it     */
              sizeof(struct sockaddr_in));/* has this size           */
    if (rc == length) {                   /* All bytes sent          */
       DPI_DEBUG(1,DPI_DM(                /* if debug_lvl>=1, print  */
         DPI_M_SENT_PACKET_ON_UDP_FD,
         length, connection_fds[handle], 
         inet_ntoa(connection_socks[handle].sin_addr),
         ntohs(connection_socks[handle].sin_port)));
       DPI_DEBUG(2,dump_bfr(packet_p,     /* if debug_lvl>=2, then   */
         length, DPI_M_OUT_PACKET));      /* also dump DPI packet    */
       rc = DPI_RC_OK;                    /* so all is OK            */
    } else if (rc<0) {                    /* error on sendto         */
       DPI_DEBUG(1,PERROR("sendto():"));  /* inform user             */
       rc = DPI_RC_IO_ERROR;              /* set this error code     */
    } else {                              /* else incomplete send    */
       DPI_DEBUG(1,DPI_DM(                /* if debug_lvl>=1, print  */
           DPI_M_NOT_ALL_BYTES_SENT,      /* that we could only      */
           rc, length));                  /* sent aprtial data       */
       rc = DPI_RC_IO_ERROR;              /* error on sendto         */
    } /* endif */
    break;
  case TCP_connection:
    rc = send(connection_fds[handle],     /* send packet on this fd  */
              (char *)packet_p,           /* ptr to DPI packet       */
              length, 0);                 /* packet length, flags=0  */
    if (rc == length) {                   /* All bytes sent          */
       DPI_DEBUG(1,DPI_DM(                /* if debug_lvl >=1, print */
         DPI_M_SENT_PACKET_ON_TCP_FD,     /* message that we did     */
         length, connection_fds[handle]));/* send the packet.        */
       DPI_DEBUG(2,dump_bfr(packet_p,     /* if debug_lvl>=2, then   */
         length, DPI_M_OUT_PACKET));      /* also dump DPI packet    */
       rc = DPI_RC_OK;                    /* so all is OK            */
    } else if (rc<0) {                    /* error on send           */
       DPI_DEBUG(1,PERROR("sendto():"));  /* inform user             */
       rc = DPI_RC_IO_ERROR;              /* set this error code     */
    } else {                              /* else incomplete send    */
       DPI_DEBUG(1,DPI_DM(                /* if debug_lvl>=1, print  */
           DPI_M_NOT_ALL_BYTES_SENT,      /* that we could only      */
           rc, length));                  /* sent aprtial data       */
       rc = DPI_RC_IO_ERROR;              /* error on send           */
    } /* endif */
    break;
  case NMQ_connection:
  case SNA_connection:
    /* Do whatever is needed to send to named queue. You have a      */
    /* handle that is an index into the array of connections that    */
    /* you keep a ptr to the qname and a handle. The handle might be */
    /* a ptr to help you go fast, like we use the handle to keep     */
    /* the fd for UDP and TCP domains.                               */
    /* in the case of SNA you have a ptr to luname and also a handle */
    rc = DPI_RC_NOK;                      /* untill we do something  */
    break;
  default:                                /* should never occur, if  */
    DPI_DEBUG(1,DPI_DM(                   /* it does, code is out of */
        DPI_M_CODE_OUT_OF_SYNC,           /* sync in this source file*/
        __FILE__, __LINE__));             /* at this place           */
    rc = DPI_RC_NOK;                      /* set this retcode        */
    break;
  } /* endswitch */

  return(rc);                             /* pass retcode to caller  */
}

/*********************************************************************/
/* Common routine to read a specified number of bytes from a TCP or  */
/* UDP socket. Keeps reading till number of bytes received or EOF    */
/* or till an error occurs. Then displays debug info and returns     */
/* negative DPI_RC_xxx retcode. If success, return postibe number of */
/* bytes received.                                                   */
/*********************************************************************/
static int mustread(int fd, char *buf, int len, int type)
{
  int rc;                                 /* for retcode or bytes    */
  int bytes_read = 0;                     /* count of bytes read     */
  struct sockaddr_in fsock;
  int                fsock_len;

  while (bytes_read < len) {
/*    rc = wait_for_data(fd, 3);          /* ensure there is data    */
    rc = wait_for_data(fd, WAITING_TIME);           /* wait_time 10 usec       */
    if (rc != DPI_RC_OK) return(rc);      /* return if error         */
    if (type == TCP_connection) {         /* recv(from) by type      */
       rc = recv(fd, buf+bytes_read,      /* read (more) data bytes  */
                 len-bytes_read, 0);      /* max length we want      */
    } else {                              /* assume UDP_connection   */
       fsock_len = sizeof(fsock);
       rc = recvfrom(fd,buf+bytes_read,   /* read (more) data bytes  */
               len-bytes_read,0,          /* max length we want      */
               (struct sockaddr *)&fsock, /* ptr for address info    */
               &fsock_len);               /* length for addr info    */
    } /* endif */                         /* from address stored     */
    if (rc > 0) {                         /* if we got any bytes     */
       bytes_read += rc;                  /* add them to count       */
       continue;                          /* while loop if needed    */
    } else if (rc == 0) {                 /* zero bytes means EOF    */
       DPI_DEBUG(1,DPI_DM(                /* if debug_lvl>=1 then    */
         DPI_M_EOF_RECEIVED,              /* tell user what we got   */
         (type == TCP_connection) ?       /* pass correct socket     */
         "TCP" : "UDP", fd));             /* type and fd number      */
       close_socket_by_fd(fd);            /* then close the socket   */
       return(DPI_RC_EOF);                /* and return error        */
    } else if (rc < 0) {                  /* negative means error    */
       DPI_DEBUG(1,PERROR(                /* if debug_lvl >= 1       */
         (type == TCP_connection) ?       /* tell error and pass     */
         "recv()" : "recvfrom()"));       /* correct function name   */
       close_socket_by_fd(fd);            /* then close the socket   */
       return(DPI_RC_IO_ERROR);           /* and return error        */
    } /* endif */
  } /* endwhile */

  return(bytes_read);                     /* return what we got      */
}

/*********************************************************************/
/* common routine to read a DPI packet from a TCP or UDP connection  */
/* - Returns positive number of bytes if read is successful.         */
/* - Returns negative DPI_RC_xxx return code if error                */
/*                                                                   */
/* Note(s):                                                          */
/* - depends on global scope of dpi_buffer and dpi_buflen            */
/*********************************************************************/
int read_dpi_packet_on_fd(int fd,  /* read DPI packet on fd   */
                                 int type)/* of this type connection */
{
   int     rc;                            /* retcode or bytes read   */

   rc = mustread(fd, dpi_buffer, 5420, type);/* always block read  */
   if (rc != 5420) {                         /* only OK if read 5420 bytes */
      DPI_DEBUG(1,DPI_DM(                 /* if debug_lvl >=1, print */
        DPI_M_INCOMPLETE_IN_PACKET,       /* this message and pass   */
        (type == TCP_connection) ?        /* correct socket type     */
        "TCP" : "UDP", fd));              /* and fd number           */
      return(DPI_RC_IO_ERROR);            /* return this retcode     */
   } /* endif */
   else return(rc);

   dpi_buflen = (256 *                    /* now calculate remaining */
          ((unsigned char)dpi_buffer[0])  /* length; length bytes    */
          + (unsigned char)dpi_buffer[1]);/* in network byte order   */
   if (dpi_buflen>(sizeof(dpi_buffer)-2)) /* ensure data fits in     */
   {                                      /* our dpi_buffer. If not, */
      DPI_DEBUG(1,DPI_DM(                 /* if debug_lvl > 1        */
        DPI_M_PACKET_DROP,                /* give user a message     */
        dpi_buflen+2,sizeof(dpi_buffer)));/* pass packet & buf size  */
      while (dpi_buflen>0) {              /* keep reading for length;*/
        rc = mustread(fd, dpi_buffer,     /* that actually cuases a  */
                sizeof(dpi_buffer),type); /* packet drop             */
        if (rc <= 0) return(rc);          /* return if error         */
        dpi_buflen -=rc;                  /* deduct length           */
      } /* endwhile */
      return(DPI_RC_PACKET_TOO_LARGE);    /* return this to caller   */
   } /* endif */

   rc = mustread(fd, dpi_buffer+2,        /* read the DPI packet of  */
                    dpi_buflen, type);    /* this size on this fd    */

   if (rc<=0) return(rc);                 /* if we have an error     */

   DPI_DEBUG(1,printf(                    /* if debug_lvl>=1, print  */
       DPI_M_RECEIVED_PACKET_ON_FD,       /* this message and        */
       dpi_buflen + 2,                    /* pass length             */
       (type == TCP_connection) ?         /* pass correct socket     */
       "TCP" : "UDP", fd));               /* type and fd number      */
   DPI_DEBUG(2,dump_bfr(                  /* if debug_lvl>=2 then    */
       dpi_buffer, dpi_buflen + 2,        /* also dump packet buffer */
       DPI_M_IN_PACKET));

   return(rc);                            /* pass retcode or bytes   */
}

/*********************************************************************/
/* function to make a hex dump of a buffer to stdout                 */
/*********************************************************************/
static void dump_bfr(CONST void *buf, int len, CONST void *hdr)
{
   register int   i;
   unsigned char *ucp;

   DPI_DM(DPI_M_DUMP_OF_BYTES, len, hdr);
   ucp = (unsigned char *) buf;
   for (i=0; i<len; i++) {
       printf("%2.2x%s",*ucp,((i&15) == 15) ? "\n  " : " ");
       ucp++;
   }
   printf("\n");
}

/*
\end{verbatim}
*/
