/* 
\begin{verbatim}
*/

/*********************************************************************/
/*                                                                   */
/* SNMP-DPI API - SNMP Distributed Protocol Interface                */
/*                Application Programming Interface                  */
/*                                                                   */
/* Feb 15, 1994 - Version 0.13i                                      */
/*                                                                   */
/* Copyright    - (C) International Business Machines Corp. 1994     */
/*                                                                   */
/*   Permission to use, copy, modify, and distribute this software   */
/*   and its documentation for any lawful purpose and without fee is */
/*   hereby granted, provided that this notice be retained unaltered,*/
/*   and that the names of IBM and all other contributors shall not  */
/*   be used in advertising or publicity pertaining to distribution  */
/*   of the software without specific written prior permission.      */
/*   No contributor makes any representations about the suitability  */
/*   of this software for any purpose.  It is provided "as is"       */
/*   without express or implied warranty.                            */
/*                                                                   */
/*   IBM AND ALL OTHER CONTRIBUTORS DISCLAIM ALL WARRANTIES WITH     */
/*   REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF    */
/*   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, TITLE,    */
/*   AND NON-INFRINGEMENT.                                           */
/*                                                                   */
/*   IN NO EVENT SHALL IBM OR ANY OTHER CONTRIBUTOR BE LIABLE FOR    */
/*   ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, WHETHER IN      */
/*   CONTRACT, TORT, OR OTHER FORM OF ACTION, ARISING OUT OF OR IN   */
/*   CONNECTION WITH, THE USE OR PERFORMANCE OF THIS SOFTWARE.       */
/*                                                                   */
/* snmp_mkDPI.c  -  DPI packet parse and DPI packet builder          */
/*                                                                   */
/*********************************************************************/

#ifdef RCS_ID
static char rcsid[] =
 "$Id: snmp_mDPI.c,v 1.7 1994/02/15 17:21:08 bert Exp $";
#endif /* RCS_ID */

/* Change activitity:
 *
 * $Log: snmp_mDPI.c,v $
 * Revision 1.7  1994/02/15  17:21:08  bert
 * Add revision log and RCSID
 *
 * Revision 1.6  1994/02/15  13:38:36  bert
 * Fix to put selected character set into DPI OPEN packet
 * This is in the mkDPIopen() function
 *
 * Revision 1.5  1994/02/04  13:28:16  bert
 * Ensure correct DPI error text
 *
 * Revision 1.4  1994/02/08  10:38:36  bert
 * add version info
 *
 * Revision 1.3  1994/02/04  00:38:36  bert
 * Fixes after DPI 2.0 review
 *
 * Revision 1.2  1994/01/24  17:05:57  bert
 * dpiPortForXXX names
 *
 * Revision 1.1  1994/01/03  22:24:55  bert
 * Initial revision
 *
 */


#define SNMP_TYPE_DPI_1_INTERNET (128|5)     /* for DPI 1.x support  */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <snmp_lDPI.h>

static snmp_dpi_open_packet *pDPIopen(       /* parse DPI OPEN packet*/
           snmp_dpi_hdr  *hdr_p,             /* ptr to a parse hdr   */
           unsigned char *packet_p,          /* ptr to the packet    */
           int            len);              /* length of the packet */

static snmp_dpi_close_packet *pDPIclose(     /* parse CLOSE packet   */
           snmp_dpi_hdr  *hdr_p,             /* ptr to a parse hdr   */
           unsigned char *packet_p,          /* ptr to the packet    */
           int            len);              /* length of the packet */

static snmp_dpi_reg_packet *pDPIreg(         /* parse REGISTER packet*/
           snmp_dpi_hdr  *hdr_p,             /* ptr to a parse hdr   */
           unsigned char *packet_p,          /* ptr to the packet    */
           int            len);              /* length of the packet */

static snmp_dpi_ureg_packet *pDPIureg(       /* parse UNREGISTER     */
           snmp_dpi_hdr  *hdr_p,             /* ptr to a parse hdr   */
           unsigned char *packet_p,          /* ptr to the packet    */
           int            len);              /* length of the packet */

static snmp_dpi_get_packet  *pDPIget(        /* parse DPI GET packet */
           snmp_dpi_hdr  *hdr_p,             /* ptr to a parse hdr   */
           unsigned char *packet_p,          /* ptr to the packet    */
           int            len);              /* length of the packet */
 
static snmp_dpi_next_packet *pDPInext(       /* parse GETNEXT packet */
           snmp_dpi_hdr  *hdr_p,             /* ptr to a parse hdr   */
           unsigned char *packet_p,          /* ptr to the packet    */
           int            len);              /* length of the packet */
 
static snmp_dpi_set_packet  *pDPIset(        /* parse DPI SET packet */
           snmp_dpi_hdr  *hdr_p,             /* ptr to a parse hdr   */
           unsigned char *packet_p,          /* ptr to the packet    */
           int            len);              /* length of the packet */
 
static snmp_dpi_trap_packet *pDPItrap(       /* parse DPI TRAP packet*/
           snmp_dpi_hdr  *hdr_p,             /* ptr to a parse hdr   */
           unsigned char *packet_p,          /* ptr to the packet    */
           int            len);              /* length of the packet */
 
static snmp_dpi_resp_packet *pDPIresp(       /* parse RESPONSE packet*/
           snmp_dpi_hdr  *hdr_p,             /* ptr to a parse hdr   */
           unsigned char *packet_p,          /* ptr to the packet    */
           int            len);              /* length of the packet */
 
static void fDPIopen_packet(           /* free open packet           */
   snmp_dpi_open_packet  *packet_p);   /* ptr to packet parse tree   */

static void fDPIclose_packet(          /* free close packet          */
   snmp_dpi_close_packet *packet_p);   /* ptr to packet parse tree   */

static void fDPIget_packet(            /* free get packet            */
   snmp_dpi_get_packet   *packet_p);   /* ptr to packet parse tree   */

static void fDPIreg_packet(            /* free register packet       */
   snmp_dpi_reg_packet   *packet_p);   /* ptr to packet parse tree   */

static void fDPIureg_packet(           /* free unregister packet     */
   snmp_dpi_ureg_packet  *packet_p);   /* ptr to packet parse tree   */

static void fDPInext_packet(           /* free a DPI getnext packet  */
   snmp_dpi_next_packet  *packet_p);   /* ptr to packet parse tree   */

extern void fDPIset_packet(            /* free a DPI set packet      */
   snmp_dpi_set_packet   *packet_p);   /* ptr to packet parse tree   */

static void fDPItrap_packet(           /* free a DPI trap packet     */
   snmp_dpi_trap_packet  *packet_p);   /* ptr to packet parse tree   */

static void fDPIresp_packet(           /* free a DPI response packet */
   snmp_dpi_resp_packet  *packet_p);   /* ptr to packet parse tree   */
 
static int cDPIopen(                   /* create OPEN packet         */
   snmp_dpi_open_packet  *packet_p);   /* ptr to OPEN parse tree     */
static int cDPIclose(                  /* create CLOSE packet        */
   snmp_dpi_close_packet *packet_p);   /* ptr to CLOSE parse tree    */
static int cDPIreg(                    /* create REGISTER packet     */
   snmp_dpi_reg_packet   *packet_p);   /* ptr to GET parse tree      */
static int cDPIureg(                   /* create GET packet          */
   snmp_dpi_ureg_packet  *packet_p);   /* ptr to GET parse tree      */
static int cDPIget(                    /* create GET packet          */
   snmp_dpi_get_packet   *packet_p);   /* ptr to GET parse tree      */
static int cDPInext(                   /* create GETNEXT packet      */
   snmp_dpi_next_packet  *packet_p);   /* ptr to GETNEXT parse tree  */
static int cDPIset(                    /* create SET packet          */
   snmp_dpi_set_packet   *packet_p);   /* ptr to SET parse tree      */
static int cDPItrap(                   /* create TRAP packet         */
   snmp_dpi_trap_packet  *packet_p);   /* ptr to TRAP parse tree     */
static int cDPIresp(                   /* create RESPONSE packet     */
   snmp_dpi_resp_packet  *packet_p);   /* ptr to RESPONSE parse tree */
 
static long int_val(                   /* convert 4 network ordered  */
       unsigned char *bytes);          /* bytes to a long integer    */
static short s_int_val(                /* convert 2 network ordered  */
       unsigned char *bytes);          /* bytes to a short integer   */
static char *copy(                     /* copy null terminated string*/
              char *cp,                /* from cp into dynamic memory*/
              int *len);               /* return ptr, size in len    */
static char *concat(                   /* concatenate null terminated*/
              char *p1,                /* string (p2) to base (p2) in*/
              char *p2);               /* new memory and return ptr  */

static unsigned char  packet_buf[SNMP_DPI_BUFSIZE];
static int            packet_len;
static unsigned short packet_id;
int                   dpi_debug;       /* Also used by qDPI and lDPI */
#include <dpi_version.h>               /* include the version ID     */

#define PACKET_VERSION  packet_buf[3]         /* this is DPI version */
#define PACKET_RELEASE  packet_buf[4]         /* this is DPI release */

#define PACKET_COPY(a,b)                      /* copy data into DPI */\
   if ((packet_len + (b))<sizeof(packet_buf)) /* buffer, but only if*/\
   {                                          /* it fits in buffer  */\
      memcpy(&packet_buf[packet_len],(a),(b));/* copy the data      */\
      packet_len += (b);                      /* add new length     */\
   } else {                                   /* if it does not fit */\
      DPI_DEBUG(1,DPI_DM(                     /* if debug active    */\
                DPI_M_DPI_BUFFER_OVERFLOW));  /* display message    */\
      return(-2);                             /* and return error   */\
   } /* endif */                              /* to caller          */

#define PACKET_UCOPY(a,b)                     /* copy data to packet*/\
   if ((packet_len + (b))<sizeof(packet_buf)) /* buffer, but only if*/\
   {                                          /* it fits in buffer  */\
      memcpy(&packet_buf[packet_len],(a),(b));/* copy the data      */\
      packet_len += (b);                      /* add new length     */\
   } else {                                   /* if it does not fit */\
      DPI_DEBUG(1,DPI_DM(                     /* if debug active    */\
                DPI_M_DPI_BUFFER_OVERFLOW));  /* display message    */\
      return((unsigned char *)0);             /* return a NULL ptr  */\
   } /* endif */                              /* to caller          */

#ifdef DPI_DEBUG_CODE                     /* lots of code that is    */
                                          /* only needed for DEBUG   */
static char *dpi_type_text[] = {          /* names DPI packet types  */
         "Unknown",
         "SNMP_DPI_GET",
         "SNMP_DPI_GETNEXT",
         "SNMP_DPI_SET",
         "SNMP_DPI_TRAP",
         "SNMP_DPI_RESPONSE",
         "SNMP_DPI_REGISTER",
         "SNMP_DPI_UNREGISTER",
         "SNMP_DPI_OPEN",
         "SNMP_DPI_CLOSE",
         "SNMP_DPI_COMMIT",
         "SNMP_DPI_UNDO",
         "SNMP_DPI_GETBULK",
         "SNMP_DPI_TRAPV2",
         "SNMP_DPI_INFORM",
         "SNMP_DPI_ARE_YOU_THERE"
};
#define SNMP_DPI_MAXTYPE ((sizeof(dpi_type_text)-1)/sizeof(char *))

static char *snmp_type_text[] = {         /* names SNMP_DPI types    */
         "Unknown",
         "Integer32",
         "OCTET_STRING",
         "OBJECT_IDENTIFIER",
         "NULL",
         "IpAddress",
         "Counter32",
         "Gauge32",
         "TimeTicks",
         "DisplayString",
         "BIT_STRING",
         "NsapAddress",
         "UInteger32",
         "Counter64",
         "Opaque",
         "noSuchObject",
         "noSuchInstance",
         "endOfMibView"
};
#define SNMP_TYPE_MAXTYPE ((sizeof(snmp_type_text)-1)/sizeof(char *))

static char *snmp_error_text[] = {    /* names SNMP error codes   */
   /*  0 */ "noError",
   /*  1 */ "tooBig",
   /*  2 */ "noSuchName",
   /*  3 */ "badValue",
   /*  4 */ "readOnly",
   /*  5 */ "genErr",
   /*  6 */ "noAccess",
   /*  7 */ "wrongType",
   /*  8 */ "wrongLength",
   /*  9 */ "wrongEncoding",
   /* 10 */ "wrongValue",
   /* 11 */ "noCreation",
   /* 12 */ "inconsistentValue",
   /* 13 */ "resourceUnavailable",
   /* 14 */ "commitFailed",
   /* 15 */ "undoFailed",
   /* 16 */ "authorizationError",
   /* 17 */ "notWritable",
   /* 18 */ "inconsistentName",
   /* 19 */ "*** Unknown error-status ***"/* keep as last entry      */
};
#define SNMP_ERROR_MAX ((sizeof(snmp_error_text)-1)/sizeof(char *))

static char *dpi_error_text[] = {         /* names DPI error codes   */
#define DPI_ERROR_FIRST      101          /* first DPI error code    */
        "DPI_otherError",
        "DPI_notFound",
        "DPI_alreadyRegistered",
        "DPI_higherPriorityRegistered",
        "DPI_mustOpenFirst",
        "DPI_notAuthorized",
        "DPI_viewSelectionNotSupported",
        "DPI_getBulkSelectionNotSupported",
        "DPI_duplicateSubAgentIdentifier",
        "DPI_invalidDisplayString",
        "DPI_characterSetSelectionNotSupported"
};
#define DPI_ERROR_MAX ((sizeof(dpi_error_text)-1)/sizeof(char *)) + \
        DPI_ERROR_FIRST

static char *close_text[] = {             /* close reason codes      */
        "unknown",
        "otherReason",
        "goingDown",
        "unsupportedVersion",
        "protocolError",
        "authenticationFailure",
        "byManager",
        "timeout",
        "openError"
};
#define CLOSE_TEXT_MAX ((sizeof(close_text)-1)/sizeof(char *))

static char *ureg_text[] = {              /* unregister reason codes */
        "unknown",
        "otherReason",
        "goingDown",
        "justUnregister",
        "newRegistration",
        "higherPriorityRegistered",
        "byManager",
        "timeout"
};
#define UREG_TEXT_MAX ((sizeof(ureg_text)-1)/sizeof(char *))

/*********************************************************************/
/* Function to print a hdr info when in debug mode                   */
/*********************************************************************/
static void dpi_debug_print_hdr(          /* print dpi hdr info      */
                char         *cp,         /* ptr to function name    */
                snmp_dpi_hdr *hdr_p)      /* ptr to hdr data         */
{
   int i = hdr_p->packet_type;            /* pick up packet type     */

   if (i > SNMP_DPI_MAXTYPE) i = 0;       /* ensure valid index      */

   DPI_DM(DPI_M_VERSION1,                 /* display this message    */
          cp,
          hdr_p->proto_major,
          hdr_p->proto_version,
          hdr_p->proto_release,
          hdr_p->packet_id,
          dpi_type_text[i]);
}
#ifdef SNMP_DPI_VIEW_SELECTION 
static void dpi_debug_print_hdr2(         /* print dpi hdr2 info     */
                char         *cp,         /* ptr to function name    */
                snmp_dpi_hdr *hdr_p)      /* ptr to hdr data         */
{
   switch (hdr_p->packet_type) {
   case SNMP_DPI_GET:
   case SNMP_DPI_GETNEXT:
   case SNMP_DPI_SET:
   case SNMP_DPI_COMMIT:
   case SNMP_DPI_UNDO:
     DPI_DM(DPI_M_VERSION2,               /* display this message    */
          strlen(cp),"",
          hdr_p->community_len ? hdr_p->community_len :
                                 strlen(DPI_M_NONE),
          hdr_p->community_len ? (char *)hdr_p->community_p :
                                 DPI_M_NONE);
      break;
   } /* endswitch */
}
#endif

/*********************************************************************/
/* Function to print a dpi_open packet when in debug mode            */
/*********************************************************************/
static void dpi_debug_print_open(         /* print DPI open info     */
          char                 *cp,       /* ptr to function name    */
          snmp_dpi_open_packet *pack_p)   /* ptr to packet data      */
{
   DPI_DM(DPI_M_OPEN_PART_1,
          cp,
          pack_p->oid_p         ? pack_p->oid_p         : DPI_M_NONE);
   DPI_DM(DPI_M_OPEN_PART_2,
          strlen(cp),"",
          pack_p->description_p ? pack_p->description_p : DPI_M_NONE);
   DPI_DM(DPI_M_OPEN_PART_3,
          strlen(cp),"",
          pack_p->character_set,
          pack_p->character_set ? DPI_M_CSET_ASCII : 
                                  DPI_M_CSET_NATIVE);
   DPI_DM(DPI_M_OPEN_PART_4,
          strlen(cp),"",
          pack_p->timeout, pack_p->max_varBinds,
          pack_p->password_len ? pack_p->password_len : 
                                 strlen(DPI_M_NONE),
          pack_p->password_len ? (char *)pack_p->password_p :
                                 DPI_M_NONE);
}

/*********************************************************************/
/* Function to print a dpi_close packet when in debug mode           */
/*********************************************************************/
static void dpi_debug_print_close(        /* print DPI close info    */
          char                  *cp,      /* ptr to function name    */
          snmp_dpi_close_packet *pack_p)  /* ptr to packet data      */
{
   unsigned int i = pack_p->reason_code;

   if (i>CLOSE_TEXT_MAX) i = 0;

   DPI_DM(DPI_M_CLOSE,
          cp, pack_p->reason_code, close_text[i]);
}

/*********************************************************************/
/* Function to print a dpi_reg   packet when in debug mode           */
/*********************************************************************/
static void dpi_debug_print_reg(          /* print DPI get   info    */
          char                  *cp,      /* ptr to function name    */
          snmp_dpi_reg_packet   *pack_p)  /* ptr to packet data      */
{
   DPI_DM(DPI_M_REGISTER1,
           cp,                           
           pack_p->group_p ? pack_p->group_p : DPI_M_NONE,
           pack_p->priority, pack_p->timeout);
#ifdef SNMP_DPI_VIEW_SELECTION 
   DPI_DM(DPI_M_REGISTER2,
           strlen(cp),"",
           pack_p->view ? DPI_M_YES : DPI_M_NO);
#endif
#ifdef SNMP_DPI_BULK_SELECTION
   DPI_DM(DPI_M_REGISTER3,
           strlen(cp),"",
           pack_p->bulk ? DPI_M_YES : DPI_M_NO);
#endif
#ifdef SNMP_DPI_REGISTER_WITH_COMMUNITY
   DPI_DM(DPI_M_REGISTER4,
           strlen(cp),"",
           pack_p->community_len ? pack_p->community_len :
                                   strlen(DPI_M_NONE),
           pack_p->community_len ? (char *)pack_p->community :
                                   DPI_M_NONE);
#endif
}

/*********************************************************************/
/* Function to print a dpi_ureg  packet when in debug mode           */
/*********************************************************************/
static void dpi_debug_print_ureg(         /* print DPI get   info    */
          char                  *cp,      /* ptr to function name    */
          snmp_dpi_ureg_packet  *pack_p)  /* ptr to packet data      */
{
   unsigned int i = pack_p->reason_code;

   if (i>UREG_TEXT_MAX) i = 0;

   DPI_DM(DPI_M_UNREGISTER,
          cp,                            
          pack_p->group_p ? pack_p->group_p : DPI_M_NONE,
          pack_p->reason_code, ureg_text[i]);
}

/*********************************************************************/
/* Function to print a dpi_get packet when in debug mode             */
/*********************************************************************/
static void dpi_debug_print_get(          /* print DPI get   info    */
          char                  *cp,      /* ptr to function name    */
          snmp_dpi_get_packet   *pack_p)  /* ptr to packet data      */
{
   DPI_DM(DPI_M_OBJECT_PART_1,
          cp,                            
          pack_p->group_p    ? pack_p->group_p    : DPI_M_NONE,
          pack_p->instance_p ? pack_p->instance_p : DPI_M_NONE);
   DPI_DM(DPI_M_OBJECT_PART_2,
          strlen(cp),"",
          pack_p->object_p   ? pack_p->object_p   : DPI_M_NONE);
}

/*********************************************************************/
/* Function to print a dpi_next  packet when in debug mode           */
/*********************************************************************/
static void dpi_debug_print_next(         /* print DPI get   info    */
          char                  *cp,      /* ptr to function name    */
          snmp_dpi_next_packet  *pack_p)  /* ptr to packet data      */
{
   DPI_DM(DPI_M_OBJECT_PART_1,
          cp,                            
          pack_p->group_p    ? pack_p->group_p    : DPI_M_NONE,
          pack_p->instance_p ? pack_p->instance_p : DPI_M_NONE);
   DPI_DM(DPI_M_OBJECT_PART_2,
          strlen(cp),"",
          pack_p->object_p   ? pack_p->object_p   : DPI_M_NONE);
}

/*********************************************************************/
/* Function to print a dpi_set packet when in debug mode             */
/*********************************************************************/
static void dpi_debug_print_set(          /* print set structure     */
            char                *cp,      /* ptr to function name    */
            snmp_dpi_set_packet *pack_p)  /* ptr to the packet data  */
{
   int cp_len = strlen(cp);
   int line_len;
   int i = pack_p->value_type & SNMP_TYPE_MASK; /* type without flag */

   if (i > SNMP_TYPE_MAXTYPE) i = 0;      /* ensure within our table */

   DPI_DM(DPI_M_OBJECT_PART_1,
          cp,                             /* print function name     */
          pack_p->group_p    ? pack_p->group_p    : DPI_M_NONE,
          pack_p->instance_p ? pack_p->instance_p : DPI_M_NONE);
   DPI_DM(DPI_M_OBJECT_PART_2,
          cp_len,"",
          pack_p->object_p   ? pack_p->object_p   : DPI_M_NONE);
   DPI_DM(DPI_M_VALUE_PART_1,
          cp_len,"",
          snmp_type_text[i],              /* name of type            */
          pack_p->value_type,             /* hex value of type       */
          pack_p->value_len);             /* length of value         */
   DPI_DM(DPI_M_VALUE_PART_2,
          cp_len,"");                     /* value line              */

   switch (pack_p->value_type) {          /* process by value type   */
   case SNMP_TYPE_Integer32:              /* a number (signed int)   */
     DPI_DM(DPI_M_VALUE_INTEGER32,
            *((long *)pack_p->value_p),
            *((long *)pack_p->value_p));
     break;
   case SNMP_TYPE_DPI_1_INTERNET:         /* Internet (DPI 1.x style)*/
   case SNMP_TYPE_IpAddress:              /* Internet IP address     */
     DPI_DM(DPI_M_VALUE_IPADDRESS, 
            ((unsigned char *)pack_p->value_p)[0],
            ((unsigned char *)pack_p->value_p)[1],
            ((unsigned char *)pack_p->value_p)[2],
            ((unsigned char *)pack_p->value_p)[3]);
     break;
   case SNMP_TYPE_Counter32:              /* These are unsigned      */
   case SNMP_TYPE_Gauge32:                /* long integers           */
   case SNMP_TYPE_TimeTicks:
   case SNMP_TYPE_UInteger32:
     DPI_DM(DPI_M_VALUE_U_INTEGER32,
            *((unsigned long*)pack_p->value_p),
            *((unsigned long*)pack_p->value_p));
     break;
   case SNMP_TYPE_Counter64:              /* 64 bit integer          */
   { long double d,dt;
   /*d  = *((unsigned long*)pack_p->value_p) * pow(2.0,32.0); */
     dt = (double)(1<<31);
     d  = *((unsigned long*)pack_p->value_p) * dt;
     d += *((unsigned long*)pack_p->value_p+4);
     DPI_DM(DPI_M_VALUE_COUNTER64, d,
            ((snmp_dpi_u64 *)pack_p->value_p)->high,
            ((snmp_dpi_u64 *)pack_p->value_p)->low);
   } /* end case SNMP_TYPE_Counter64; */
   break;
   case SNMP_TYPE_NULL:                   /* These ones have no value*/
   case SNMP_TYPE_noSuchObject:
   case SNMP_TYPE_noSuchInstance:
   case SNMP_TYPE_endOfMibView:
     DPI_DM(DPI_M_VALUE_NULL);
     break;
   case SNMP_TYPE_OBJECT_IDENTIFIER:      /* OBJECT IDENTIFIER       */
     DPI_DM(DPI_M_VALUE_OID,
            pack_p->value_len,            /* use length, just incase */
            (char *)pack_p->value_p);     /* not NULL terminated     */
     break;
   case SNMP_TYPE_DisplayString:          /* DisplayString           */
     DPI_DM(DPI_M_VALUE_DISPLAY_STRING,
            pack_p->value_len,            /* use length, just incase */
            (char *)pack_p->value_p);     /* not NULL terminated     */
     break;
   case SNMP_TYPE_BIT_STRING:             /* Handled as OCTET STRING */
   case SNMP_TYPE_NsapAddress:            /* These can be printed    */
   case SNMP_TYPE_Opaque:                 /* as ...                  */
   case SNMP_TYPE_OCTET_STRING:           /* OCTECT STRING           */
   default:                               /* just in case...         */
     line_len=cp_len + 9;
     DPI_DM(DPI_M_VALUE_OCTETS_PART_1);
     for (i=0; i < pack_p->value_len; i++)/* display byte by byte    */
     {                                    /* in hex                  */
         DPI_DM(DPI_M_VALUE_OCTETS_PART_2,
                ((unsigned char *)pack_p->value_p)[i]);
         line_len +=2;
         if (line_len >= 77) {
            DPI_DM(DPI_M_VALUE_OCTETS_PART_3, cp_len+9);
            line_len = cp_len+9;
         } /* endif */
     } /* endfor */
     DPI_DM(DPI_M_VALUE_OCTETS_PART_4);
     break;
   } /* endswitch */
}

/*********************************************************************/
/* Function to print a dpi_resp packet when in debug mode            */
/*********************************************************************/
static void dpi_debug_print_resp(         /* print resp structure    */
            char                 *cp,     /* ptr to function name    */
            snmp_dpi_resp_packet *pack_p) /* ptr to the packet data  */
{
   unsigned int  i = pack_p->error_code;
   char         *text_p;

   if (i < SNMP_ERROR_MAX) {              /* ensure valid index      */
      text_p = snmp_error_text[i];
   } else if ((i>=DPI_ERROR_FIRST) && (i<=DPI_ERROR_MAX)) {
      text_p = dpi_error_text[i-DPI_ERROR_FIRST];
   } else {
      text_p = snmp_error_text[SNMP_ERROR_MAX];
   } /* endif */

   DPI_DM(DPI_M_RESPONSE,
        cp,
        pack_p->error_code, pack_p->error_code,
        text_p, pack_p->error_index);
}

/*********************************************************************/
/* Function to print a dpi_trap packet when in debug mode            */
/*********************************************************************/
static void dpi_debug_print_trap(         /* print trap structure    */
            char                 *cp,     /* ptr to function name    */
            snmp_dpi_trap_packet *pack_p) /* ptr to the packet data  */
{
 DPI_DM(DPI_M_TRAP,
        cp,
        pack_p->generic, pack_p->generic,    
        pack_p->specific,pack_p->specific,
        pack_p->enterprise_p ? pack_p->enterprise_p : DPI_M_NONE);
}
#endif /* def DPI_DEBUG_CODE */

/*********************************************************************/
/* Function to turn DPI debug mode on or off.                        */
/*    If the argument is zero, then DPI debug is turned off          */
/*    Any other value turns DPI debug on.                            */
/*********************************************************************/
void DPIdebug(int level)                  /* turn DPI debug on/off   */
{
   if (dpi_debug != level) {              /* If debug level changes, */
      DPI_DM(DPI_M_DEBUG_ONOFF, level ?   /* tell user is debug is   */
        DPI_M_DEBUG_ON : DPI_M_DEBUG_ON,  /* ON or OFF (as is the    */
        level);                           /* case) and also level    */
      printf("%s\n", dpi_version);        /* Display version also    */
      fflush(stdout);
   } /* endif */
   dpi_debug = level;                     /* set what caller wants   */
}

/*********************************************************************/
/* Function to parse a DPI packet into a DPI parse tree.             */
/* - input is a ptr to the DPI packet                                */
/* - if success, we return a ptr to a DPI parse tree (snmp_dpi_hdr)  */
/*   which has been dynamically allocated. It is the callers         */
/*   responsibility to free the parse tree when no longer needed.    */
/*   To do so, the one can call the fDPIhdr() function.              */
/* - if we fail to parse the packet, then we return a NULL ptr.      */
/*********************************************************************/
snmp_dpi_hdr *pDPIpacket(                 /* parse a DPI packet      */
               unsigned char *packet_p)   /* ptr to the packet       */
{
   snmp_dpi_hdr  *hdr_p;                  /* ptr to a dpi hdr        */
   int            len,                    /* length remaining packet */
                  offset,                 /* offset into packet      */
                  level_unsupported;      /* flag                    */
   unsigned char *cp;                     /* ptr into packet         */

   if (packet_p == (unsigned char*)0) {   /* ensure we have a ptr to */
      DPI_DEBUG(1,DPI_DM(DPI_M_NULL_PTR,  /* a packet; error if not  */
                "pDPIpacket"));           /* pass function name      */
      return(snmp_dpi_hdr_NULL_p);        /* then return NULL ptr    */
   } /* endif */
   hdr_p = (snmp_dpi_hdr *) calloc(1,     /* allocate a DPI hdr      */
                    sizeof(snmp_dpi_hdr));/* initialized to zero     */

   if (hdr_p == snmp_dpi_hdr_NULL_p) {    /* a NULL ptr means error  */
      DPI_DEBUG(1,DPI_DM(                 /* if debug_lvl >= 1       */
          DPI_M_OUT_OF_MEMORY));          /* then give this message  */
      return(snmp_dpi_hdr_NULL_p);        /* return NULL to caller   */
   } /* endif */

   /* first 2 bytes of a packet define the remaining packet length.  */
   /* they are in network byte order (MSB to LSB)                    */
   len = s_int_val(packet_p);             /* pick up packet length   */
   len += 2;                              /* add length bytes        */
   offset = 2;                            /* data starts after that  */
   hdr_p->proto_major   = packet_p[offset++]; /* pick up protocol    */
   hdr_p->proto_version = packet_p[offset++]; /* protocol version    */
   hdr_p->proto_release = packet_p[offset++]; /* protocol release    */

   level_unsupported = 0;                 /* untill proven wrong     */
   switch (hdr_p->proto_version) {        /* check by DPI version    */
   case 1:                                /* DPI Version 1.x         */
      switch (hdr_p->proto_release) {     /* switch by Release       */
      case 0:                             /* Release 0 is supported  */
      case 1:                             /* Release 1 is supported  */
        break;                            /* All OK sofar            */
      default:                            /* any other release is    */
        level_unsupported = 1;            /* not supported           */
      } /* endswitch */
      break;
   case 2:                                /* DPI Version 2.x         */
      switch (hdr_p->proto_release) {     /* switch by Release       */
      case 0:                             /* Release 0 is supported  */
        hdr_p->packet_id =                /* get packetID; 2 bytes   */
            s_int_val(&packet_p[offset]); /* in network byte order   */
        offset +=2;                       /* point after it          */
        break;                            /* All OK sofar            */
      default:                            /* any other release is    */
        level_unsupported = 1;            /* not supported           */
      } /* endswitch */
      break;
   default:                               /* Any other versions of   */
      level_unsupported = 1;              /* DPI are not supported   */
   } /* endswitch (hdr_p->proto_version) */

   hdr_p->packet_type = packet_p[offset++];  /* get DPI packet type  */

   DPI_DEBUG(1,dpi_debug_print_hdr(       /* if in debug mode, then  */
                "pDPIpacket", hdr_p));    /* print hdrinfo on stdout */

   if (level_unsupported || (hdr_p->proto_major != SNMP_DPI_PROTOCOL)){
      DPI_DEBUG(1,DPI_DM(                 /* if debug active         */
         DPI_M_INVALID_DPI_LEVEL,         /* tell this message       */
         "pDPIpacket"));
      free(hdr_p);                        /* free hdr we allocared   */
      return(snmp_dpi_hdr_NULL_p);        /* return NULL ptr         */
   } /* endif */

   len  -= offset;                        /* set remeaining length   */
   cp    = packet_p + offset;             /* ptr to current position */

#ifdef SNMP_DPI_VIEW_SELECTION
   if (hdr_p->proto_version >= 2) {       /* handle View Selection   */
      switch (hdr_p->packet_type) {       /* but only for these      */
      case SNMP_DPI_GET:                  /* packet types            */
      case SNMP_DPI_GETNEXT:
      case SNMP_DPI_SET:
      case SNMP_DPI_COMMIT:
      case SNMP_DPI_UNDO:
        hdr_p->community_len =            /* pickup community length */
               s_int_val(cp);             /* which is 2 bytes        */
        if (hdr_p->community_len) {       /* if community present    */
           hdr_p->community_p =
                (unsigned char *)malloc(
                    hdr_p->community_len);
           if (hdr_p->community_p) {
              memcpy(hdr_p->community_p, cp+2,
                     hdr_p->community_len);
           } else {
              DPI_DEBUG(1,DPI_DM(            /* if debug_lvl >= 1    */
                  DPI_M_OUT_OF_MEMORY));     /* give this message    */
              free(hdr_p);                   /* free allocated memory*/
              return(snmp_dpi_hdr_NULL_p);   /* return NULL ptr      */
           } /* endif */
        } /* endif */    
        len  -= (hdr_p->community_len+2); /* set remeaining length   */
        cp   += (hdr_p->community_len+2); /* ptr to current position */
        break;
      } /* endswitch */
   } /* endif */
   DPI_DEBUG(1,dpi_debug_print_hdr2(      /* if in debug mode, then  */
                "pDPIpacket", hdr_p));    /* print hdrinfo on stdout */

#endif /* def SNMP_DPI_VIEW_SELECTION */

   switch (hdr_p->packet_type) {          /* go by DPI packet type   */
   case SNMP_DPI_GET:
     hdr_p->data_u.get_p =                /* parse remaining packet  */
            pDPIget(hdr_p, cp, len);      /* as if GET packet type   */
     break;
   case SNMP_DPI_GETNEXT:
     hdr_p->data_u.next_p =               /* parse remaining packet  */
            pDPInext(hdr_p, cp, len);     /* as a GETNEXT packet     */
     break;
   case SNMP_DPI_RESPONSE:
     hdr_p->data_u.resp_p =               /* parse remaining packet  */
            pDPIresp(hdr_p, cp, len);     /* as a RESPONSE packet    */
     break;
   case SNMP_DPI_SET:
   case SNMP_DPI_COMMIT:
   case SNMP_DPI_UNDO:
     hdr_p->data_u.set_p =                /* parse remaining packet  */
            pDPIset(hdr_p, cp, len);      /* as a SET packet         */
     break;
   case SNMP_DPI_TRAP:
     hdr_p->data_u.trap_p =               /* parse remaining packet  */
            pDPItrap(hdr_p, cp, len);     /* as a TRAP packet        */
     break;
   case SNMP_DPI_REGISTER:
     hdr_p->data_u.reg_p =                /* parse remaining packet  */
            pDPIreg(hdr_p, cp, len);      /* as a REGISTER packet    */
     break;
   case SNMP_DPI_UNREGISTER:
     hdr_p->data_u.ureg_p =               /* parse remaining packet  */
            pDPIureg(hdr_p,cp,len);       /* as an UNREGISTER packet */
     break;
   case SNMP_DPI_OPEN:
     hdr_p->data_u.open_p =               /* parse remaining packet  */
            pDPIopen(hdr_p, cp, len);     /* as an OPEN packet       */
     break;
   case SNMP_DPI_CLOSE:
     hdr_p->data_u.close_p =              /* parse remaining packet  */
            pDPIclose(hdr_p, cp, len);    /* as a CLOSE packet       */
     break;
   case SNMP_DPI_ARE_YOU_THERE:
     break;                               /* has no data             */
   case SNMP_DPI_TRAPV2:                  /* place holders for       */
   case SNMP_DPI_INFORM:                  /* SNMPv2 support          */
   case SNMP_DPI_GETBULK:                 /* TODO: implement.        */
     DPI_DEBUG(1,DPI_DM(                  /* if debug active, tell   */
               DPI_M_INVALID_DPI_TYPE));  /* untill we implement it  */
     break;
   default:                               /* should not occur        */
     DPI_DEBUG(1,DPI_DM(                  /* if debug active, tell   */
               DPI_M_INVALID_DPI_TYPE));  /* anything else is error  */
     fDPIparse(hdr_p);                    /* free allocated hdr      */
     hdr_p = snmp_dpi_hdr_NULL_p;         /* set NULL pointer        */
     break;
   } /* endswitch hdr_p->packet_type) */

   if ((hdr_p->data_u.any_p) ||           /* if pDPIxxx returned ptr */
       (hdr_p->packet_type ==             /* or if no data expected  */
        SNMP_DPI_ARE_YOU_THERE))
   {
      return(hdr_p);                      /* return ptr to caller    */
   } /* endif */                          /* else error in pDPIxxx   */

   /* if we get here we had an error parsing the specific packet     */
   fDPIparse(hdr_p);                      /* return the parse tree   */
   return(snmp_dpi_hdr_NULL_p);           /* return NULL ptr         */
}

/*********************************************************************/
/* Function to parse a DPI REGISTER packet                           */
/* - if success, return ptr to dpi_reg_packet structure              */
/* - if failure, return NULL ptr                                     */
/*********************************************************************/
static snmp_dpi_reg_packet *pDPIreg(         /* parse a DPI RESGITER */
                snmp_dpi_hdr  *hdr_p,        /* ptr to hdr structure */
                unsigned char *packet_p,     /* ptr to the packet    */
                int            len)          /* length of packet     */
{
   snmp_dpi_reg_packet  *new_p,              /* ptr to REG structure */
                        *first_p,            /* ptr to first (if any)*/
                        *next_p;             /* ptr to next (if any) */
   int                   l;                  /* length of string     */
   char                 *cp;                 /* ptr to current byte  */

   if (len == 0) {                           /* nothing to parse     */
      DPI_DEBUG(1,DPI_DM(DPI_M_ZERO_LENGTH));/* tell if debug active */
      return(snmp_dpi_reg_packet_NULL_p);    /* return NULL ptr      */
   } /* endif */

   new_p   = snmp_dpi_reg_packet_NULL_p;     /* No new   ptr sofar   */
   first_p = snmp_dpi_reg_packet_NULL_p;     /* No first ptr sofar   */
   next_p  = snmp_dpi_reg_packet_NULL_p;     /* No next ptr sofar    */
   cp      = (char *)packet_p;               /* point to first byte  */
 
   while (len) {                             /* process whole packet */
    new_p = (snmp_dpi_reg_packet *)calloc(1, /* one REG structure    */
            sizeof(snmp_dpi_reg_packet));    /* initialized to zero  */

    if (new_p == snmp_dpi_reg_packet_NULL_p) /* NULL ptr means error */
    {
       DPI_DEBUG(1,DPI_DM(                   /* if debug_lvl >= 1    */
          DPI_M_OUT_OF_MEMORY));             /* give this message    */
       if (first_p) fDPIreg_packet(first_p); /* free if  allocated   */
       return(snmp_dpi_reg_packet_NULL_p);   /* return NULL ptr      */
    } /* endif */

    if (first_p) next_p->next_p = new_p;     /* Add at end of chain  */
    else first_p = new_p;                    /* unless this is first */
    next_p = new_p;                          /* save for next round  */
 
    switch (hdr_p->proto_version) {
    case 1:                                  /* DPI Version 1.x      */
      new_p->group_p = copy(cp, &l);         /* consider it groupOID */
      cp   += l;                             /* set ptr to next byte */
      len  -= l;                             /* set remaining length */
      break;
    case 2:                                  /* DPI Version 2.x      */
    default:
      if (len >= 6) {                        /* ensure timeout+prior */
         new_p->priority =                   /* set requested        */
             int_val((unsigned char *)cp);   /* priority 4-bytes     */
         cp  += 4;
         len -= 4;
         new_p->timeout  =                   /* pick up timeout      */
             s_int_val((unsigned char *)cp); /* which is 2 bytes     */
         cp  += 2;
         len -= 2;
      } else {
         DPI_DEBUG(1,DPI_DM(                 /* msg about missing    */
             DPI_M_MISSING_PRIORITY));       /* things               */
         fDPIreg_packet(new_p);              /* free allocated memory*/
         return(snmp_dpi_reg_packet_NULL_p); /* return NULL ptr      */
      } /* endif */

#ifdef SNMP_DPI_VIEW_SELECTION
      if (len >= 1) {                        /* ensure view select   */
         new_p->view = *cp++;
         len--;
      } else {
         DPI_DEBUG(1,DPI_DM(                 /* msg about missing    */
             DPI_M_MISSING_VIEW_SELECT));    /* things               */
         fDPIreg_packet(new_p);              /* free allocated memory*/
         return(snmp_dpi_reg_packet_NULL_p); /* return NULL ptr      */
      } /* endif */
#endif /* def SNMP_DPI_VIEW_SELECTION */
#ifdef SNMP_DPI_BULK_SELECTION
      if (len >= 1) {                        /* ensure view select   */
         new_p->bulk = *cp++;
         len--;
      } else {
         DPI_DEBUG(1,DPI_DM(                 /* msg about missing    */
             DPI_M_MISSING_BULK_SELECT));    /* things               */
         fDPIreg_packet(new_p);              /* free allocated memory*/
         return(snmp_dpi_reg_packet_NULL_p); /* return NULL ptr      */
      } /* endif */
#endif /* def SNMP_DPI_BULK_SELECTION */
#ifdef SNMP_DPI_REGISTER_WITH_COMMUNITY
      if (len >= 2) {                        /* ensure view select   */
         new_p->community_len =              /* pickup length        */
             s_int_val((unsigned char *)cp); /* which is 2 bytes     */
         if (new_p->community_len) {         /* if there is one      */
            new_p->community_p =             /* passed, then alloc   */
                   (unsigned char *)malloc(  /* memory for it        */
                       new_p->community_len);
            if (new_p->community_p) {        /* success, so copy the */
               memcpy(new_p->community_p, cp+2, /* data to our area  */
                      new_p->community_len);
            } else {
               DPI_DEBUG(1,DPI_DM(           /* if debug_lvl >= 1    */
                   DPI_M_OUT_OF_MEMORY));    /* give this message    */
               if (first_p) fDPIreg_packet(first_p); /* free if  any */
               return(snmp_dpi_reg_packet_NULL_p);   /* return NULL  */
            } /* endif */
         } /* endif */
         cp  += new_p->community_len + 2;    /* point to next byte   */
         len -= (new_p->community_len + 2);  /* set remaining length */
      } else {
         DPI_DEBUG(1,DPI_DM(                 /* msg about missing    */
             DPI_M_MISSING_COMMUNITY));      /* things               */
         fDPIreg_packet(new_p);              /* free allocated memory*/
         return(snmp_dpi_reg_packet_NULL_p); /* return NULL ptr      */
      } /* endif */
#endif /* def SNMP_DPI_REGISTER_WITH_COMMUNITY */

      new_p->group_p = copy(cp, &l);         /* consider it groupOID */
      cp   += l;                             /* set ptr to next byte */
      len  -= l;                             /* set remaining length */
      break;
    } /* endswitch */
                         
    if (new_p->group_p == (char *)0) {       /* failure, so          */
       DPI_DEBUG(1,DPI_DM(                   /* tell if debug active */
           DPI_M_MISSING_GROUP_ID));
       fDPIreg_packet(new_p);                /* free allocated memory*/
       return(snmp_dpi_reg_packet_NULL_p);   /* return NULL ptr      */
    } /* endif */
    DPI_DEBUG(1,dpi_debug_print_reg(         /* if debug: print line */
                          "pDPIreg",new_p)); /* showing packet       */
   } /* endwhile */

   return(new_p);                            /* return the ptr       */
}

/*********************************************************************/
/* Function to parse a DPI UNREGISTER packet                         */
/* - if success, return ptr to dpi_ureg_packet structure             */
/* - if failure, return NULL ptr                                     */
/*********************************************************************/
static snmp_dpi_ureg_packet *pDPIureg(       /* parse DPI UNREGISTER */
                snmp_dpi_hdr  *hdr_p,        /* ptr to hdr structure */
                unsigned char *packet_p,     /* ptr to the packet    */
                int            len)          /* length of packet     */
{
   snmp_dpi_ureg_packet *new_p,              /* ptr to UREG structure*/
                        *first_p,            /* ptr to first (if any)*/
                        *next_p;             /* ptr to next (if any) */
   int                   l;                  /* length of string     */
   char                 *cp;                 /* ptr to current byte  */

   if (len == 0) {                           /* nothing to parse     */
      DPI_DEBUG(1,DPI_DM(                    /* tell if debug active */
             DPI_M_ZERO_LENGTH));
      return(snmp_dpi_ureg_packet_NULL_p);   /* return NULL ptr      */
   } /* endif */

   new_p   = snmp_dpi_ureg_packet_NULL_p;    /* No new   ptr sofar   */
   first_p = snmp_dpi_ureg_packet_NULL_p;    /* No first ptr sofar   */
   next_p  = snmp_dpi_ureg_packet_NULL_p;    /* No next ptr sofar    */
   cp      = (char *)packet_p;               /* point to first byte  */
 
   while (len) {                             /* process whole packet */
    new_p = (snmp_dpi_ureg_packet *)calloc(1,/* one UREG structure   */
            sizeof(snmp_dpi_ureg_packet));   /* initialized to zero  */

    if (new_p == snmp_dpi_ureg_packet_NULL_p)/* NULL ptr means error */
    {
       DPI_DEBUG(1,DPI_DM(                   /* if debug_lvl >= 1    */
          DPI_M_OUT_OF_MEMORY));             /* give this message    */
       if (first_p) fDPIureg_packet(first_p);/* free if  allocated   */
       return(snmp_dpi_ureg_packet_NULL_p);  /* return NULL ptr      */
    } /* endif */

    if (first_p) next_p->next_p = new_p;     /* Add at end of chain  */
    else first_p = new_p;                    /* unless this is first */
    next_p = new_p;                          /* save for next round  */
 
    switch (hdr_p->proto_version) {
    case 1:   /* DPI 1.x does not have it */ /* DPI Version 1.x      */
    case 2:                                  /* DPI Version 2.x      */
    default:
      if (len >= 1) {
         new_p->reason_code = *cp++;         /* pick up reason code  */
         len--;                              /* set remaining length */
      } else {
         DPI_DEBUG(1,DPI_DM(                 /* tell if debug active */
                DPI_M_MISSING_REASON_CODE));
         fDPIureg_packet(new_p);             /* free allocated memory*/
         return(snmp_dpi_ureg_packet_NULL_p);/* return NULL ptr      */
      } /* endif */
      new_p->group_p = copy(cp, &l);         /* consider it groupOID */
      if (new_p->group_p) {                  /* success, so          */
         cp   += l;                          /* set ptr to next byte */
         len  -= l;                          /* set remaining length */
      } else {                               /* failure, so          */
         fDPIureg_packet(new_p);             /* free allocated memory*/
         return(snmp_dpi_ureg_packet_NULL_p);/* return NULL ptr      */
      } /* endif */
      break;
    } /* endswitch */
    DPI_DEBUG(1,dpi_debug_print_ureg(        /* if debug: print line */
                          "pDPIureg",new_p));/* showing packet       */
   } /* endwhile */

   return(new_p);                            /* return the ptr       */
}

/*********************************************************************/
/* Function to parse a DPI GET packet                                */
/* - if success, return ptr to snmp_dpi_get_packet structure         */
/* - if failure, return NULL ptr                                     */
/*********************************************************************/
static snmp_dpi_get_packet *pDPIget(      /* parse a DPI get packet  */
                snmp_dpi_hdr   *hdr_p,    /* ptr to hdr structure    */
                unsigned char  *packet_p, /* ptr to packet           */
                int             len)      /* length remaining packet */
{
   snmp_dpi_get_packet  *new_p,           /* ptr to a GET structure  */
                        *first_p,         /* ptr to first (if any)   */
                        *next_p;          /* ptr to next (if any)    */
   int                   l;               /* length of string        */
   char                 *cp;              /* current ptr into packet */

   if (len == 0) {                        /* nothing to parse        */
      DPI_DEBUG(1,DPI_DM(                 /* tell if debug active    */
             DPI_M_ZERO_LENGTH));
      return(snmp_dpi_get_packet_NULL_p); /* return NULL ptr         */
   } /* endif */

   first_p = snmp_dpi_get_packet_NULL_p;     /* No first ptr sofar   */
   next_p  = snmp_dpi_get_packet_NULL_p;     /* No next ptr sofar    */
   cp      = (char *)packet_p;               /* point to first byte  */

   while (len > 0) {                         /* process whole packet */
    new_p = (snmp_dpi_get_packet *)calloc(1, /* one GET structure    */
            sizeof(snmp_dpi_get_packet));    /* initialized to zero  */

    if (new_p == snmp_dpi_get_packet_NULL_p) /* NULL ptr means error */
    {
       DPI_DEBUG(1,DPI_DM(                   /* if debug_lvl >= 1    */
          DPI_M_OUT_OF_MEMORY));             /* give this message    */
       if (first_p) fDPIget_packet(first_p); /* free if  allocated   */
       return(snmp_dpi_get_packet_NULL_p);   /* return NULL ptr      */
    } /* endif */

    if (first_p) next_p->next_p = new_p;     /* Add at end of chain  */
    else first_p = new_p;                    /* unless this is first */
    next_p = new_p;                          /* save for next round  */
 
    switch (hdr_p->proto_version) {
    case 1:                                  /* DPI version 1.x      */
      new_p->group_p = copy(cp, &l);         /* consider it groupOID */
      if (new_p->group_p) {                  /* if success, then     */
          new_p->object_p = copy(cp, &l);    /* copy it as an object */
      } /* endif */                          /* OID too.             */
      cp   += l;                             /* set ptr to next byte */
      len  -= l;                             /* set remaining length */
      break;                                 /* from switch          */
    case 2:                                  /* DPI version 2.x      */
    default:                                 /* and up               */
      if (len && *cp) {
         new_p->group_p = copy(cp, &l);      /* copy group OID       */
         cp   += l;                          /* set ptr to next byte */
         len  -= l;                          /* set remaining length */
      } /* endif */
      if (len && *cp) {
         new_p->instance_p = copy(cp, &l);   /* copy instance OID    */
         cp   += l;                          /* set ptr to next byte */
         len  -= l;                          /* set remaining length */
      } else if (len) {                      /* must be null string  */
         cp++;                               /* point to next byte   */
         len--;                              /* set remaining length */
      } /* endif */
      new_p->object_p = concat(              /* OID is concatenation */
                       new_p->group_p,       /* of groupID and       */
                       new_p->instance_p);   /* instanceID           */
      break;
    } /* endswitch */

    if (new_p->object_p) {                   /* If we have object ID */
       l = strlen(new_p->object_p)-1;        /* Possible trailing dot*/
       if (new_p->object_p[l] =='.') {       /* If object ID has a   */
           new_p->object_p[l] = '\0';        /* trailing dot, then   */
       } /* endif */                         /* remove it            */
    } else {                                 /* failed to get memory */
       if (first_p) fDPIget_packet(first_p); /* free if allocated    */
       return(snmp_dpi_get_packet_NULL_p);   /* return NULL ptr      */
    } /* endif */

    DPI_DEBUG(1,dpi_debug_print_get(         /* if debug: print line */
                          "pDPIget",new_p)); /* showing packet       */
   } /* endwhile */

   return(first_p);                          /* return ptr to caller */
}

/*********************************************************************/
/* Function to parse a DPI GETNEXT packet                            */
/* - if success, return ptr to dpi_next_packet structure             */
/* - if failure, return NULL ptr                                     */
/*********************************************************************/
static snmp_dpi_next_packet *pDPInext(    /* parse a DPI GETNEXT     */
                snmp_dpi_hdr  *hdr_p,     /* ptr to hdr structure    */
                unsigned char *packet_p,  /* ptr to the packet       */
                int            len)       /* length of packet        */
{
   snmp_dpi_next_packet *new_p,           /* ptr to a GETNEXT struct */
                        *first_p,         /* ptr to first (if any)   */
                        *next_p;          /* ptr to next (if any)    */
   int                   l, lg;           /* length of string        */
   char                 *cp;              /* current ptr into packet */

   if (len == 0) {                        /* nothing to parse        */
      DPI_DEBUG(1,DPI_DM(                 /* tell if debug active    */
             DPI_M_ZERO_LENGTH));
      return(snmp_dpi_next_packet_NULL_p);/* return NULL ptr         */
   } /* endif */

   first_p = snmp_dpi_next_packet_NULL_p;    /* No first ptr sofar   */
   next_p  = snmp_dpi_next_packet_NULL_p;    /* No next ptr sofar    */
   cp      = (char *)packet_p;               /* point to first byte  */

   while (len > 0) {                         /* process whole packet */
    new_p = (snmp_dpi_next_packet *)calloc(1,/* one getnext structure*/
             sizeof(snmp_dpi_next_packet));  /* initialized to zero  */

    if (new_p == snmp_dpi_next_packet_NULL_p)/* error if NULL ptr    */
    {
       DPI_DEBUG(1,DPI_DM(                   /* if debug_lvl >= 1    */
          DPI_M_OUT_OF_MEMORY));             /* give this message    */
       if (first_p) fDPInext_packet(first_p);/* free if  allocated   */
       return(snmp_dpi_next_packet_NULL_p);  /* return NULL ptr      */
    } /* endif */

    if (first_p) next_p->next_p = new_p;     /* Add at end of chain  */
    else first_p = new_p;                    /* unless this is first */
    next_p = new_p;                          /* save for next loop   */

    switch (hdr_p->proto_version) {
    case 1:                                  /* DPI version 1.x      */
      if (len && *cp) {
         new_p->object_p = copy(cp, &l);     /* copy object OID      */
         cp   += l;                          /* set ptr to next byte */
         len  -= l;                          /* set remaining length */
      } /* endif */
      if (new_p->object_p && len && *cp) {   /* success, make group  */
         new_p->group_p = copy(cp, &lg);     /* copy group OID       */
         cp   += lg;                         /* set ptr to next byte */
         len  -= lg;                         /* set remaining length */
      } /* endif */
      if (new_p->group_p && (lg < l)) {      /* if instance present  */
         new_p->instance_p = copy(           /* then copy it in a    */
               &new_p->object_p[l-lg], &l);  /* separate string      */
         if (new_p->instance_p == (char *)0)    /* NUL ptr returned  */
         {                                      /* means error       */
          if (first_p) fDPInext_packet(first_p);/* free if allocated */
          return(snmp_dpi_next_packet_NULL_p);  /* return NULL ptr   */
         } /* endif */
      } /* endif */
      break;                                 /* from switch          */
    case 2:                                  /* DPI version 2.x      */
    default:                                 /* and up               */
      if (len && *cp) {
         new_p->group_p = copy(cp, &l);      /* copy group OID       */
         cp   += l;                          /* set ptr to next byte */
         len  -= l;                          /* set remaining length */
      } /* endif */
      if (len && *cp) {
         new_p->instance_p = copy(cp, &l);   /* copy instance OID    */
         cp   += l;                          /* set ptr to next byte */
         len  -= l;                          /* set remaining length */
      } else if (len) {                      /* must be null string  */
         cp++;                               /* point to next byte   */
         len--;                              /* set remaining length */
      } /* endif */
      new_p->object_p = concat(              /* OID is concatenation */
                       new_p->group_p,       /* of groupID and       */
                       new_p->instance_p);   /* instanceID           */
      break;
    } /* endswitch */

    if (new_p->object_p) {                   /* If we have object ID */
       l = strlen(new_p->object_p)-1;        /* Possible trailing dot*/
       if (new_p->object_p[l] =='.') {       /* If object ID has a   */
           new_p->object_p[l] = '\0';        /* trailing dot, then   */
       } /* endif */                         /* remove it            */
    } else {                                 /* failed to get memory */
       if (first_p) fDPInext_packet(first_p);/* free if allocated    */
       return(snmp_dpi_next_packet_NULL_p);  /* return NULL ptr      */
    } /* endif */

    DPI_DEBUG(1,dpi_debug_print_next(        /* if debug: print line */
                          "pDPInext",new_p));/* showing packet       */
   } /* endwhile */

   return(first_p);                          /* return ptr to caller */
}

/*********************************************************************/
/* Function to parse a DPI SET packet                                */
/* - if success, return ptr to dpi_set_packet structure              */
/* - if failure, return NULL ptr                                     */
/*********************************************************************/
static snmp_dpi_set_packet *pDPIset(      /* parse a DPI set packet  */
                snmp_dpi_hdr  *hdr_p,     /* ptr to hdr structure    */
                unsigned char *packet_p,  /* ptr to packet data      */
                int            len)       /* length remaining packet */
{
   snmp_dpi_set_packet  *new_p,           /* ptr to a SET structure  */
                        *first_p,         /* ptr to first one        */
                        *next_p;          /* ptr to next (if any)    */
   int                   l;               /* length of string        */
   char                 *cp;              /* current ptr into packet */
   long int              li;              /* receives a long int     */

   if (len == 0) {                        /* nothing to parse        */
      DPI_DEBUG(1,DPI_DM(                 /* tell if debug active    */
             DPI_M_ZERO_LENGTH));
      return(snmp_dpi_set_packet_NULL_p); /* return NULL ptr         */
   } /* endif */

   first_p = snmp_dpi_set_packet_NULL_p;     /* No first ptr sofar   */
   next_p  = snmp_dpi_set_packet_NULL_p;     /* No next ptr sofar    */
   cp      = (char *)packet_p;               /* point to first byte  */

   while (len > 0) {                         /* process whole packet */
    new_p = (snmp_dpi_set_packet *)calloc(1, /* memory for one SET   */
             sizeof(snmp_dpi_set_packet));   /* initialized to zero  */

    if (new_p == snmp_dpi_set_packet_NULL_p) /* error if NULL ptr    */
    {
       DPI_DEBUG(1,DPI_DM(                   /* if debug_lvl >= 1    */
          DPI_M_OUT_OF_MEMORY));             /* give this message    */
       if (first_p) fDPIset_packet(first_p); /* free if  allocated   */
       return(snmp_dpi_set_packet_NULL_p);   /* return NULL ptr      */
    } /* endif */

    if (first_p) next_p->next_p = new_p;     /* Add at end of chain  */
    else first_p = new_p;                    /* unless this is first */
    next_p = new_p;                          /* save for next round  */

    switch (hdr_p->proto_version) {
    case 1:                                  /* DPI version 1.x      */
      new_p->group_p = copy(cp, &l);         /* consider it groupOID */
      if (new_p->group_p) {                  /* if success, then     */
          new_p->object_p = copy(cp, &l);    /* copy it as an object */
      } /* endif */                          /* OID too.             */
      cp   += l;                             /* set ptr to next byte */
      len  -= l;                             /* set remaining length */
      break;                                 /* from switch          */
    case 2:                                  /* DPI version 2.x      */
    default:                                 /* and up               */
      if (len && *cp) {
         new_p->group_p = copy(cp, &l);      /* copy group OID       */
         cp   += l;                          /* set ptr to next byte */
         len  -= l;                          /* set remaining length */
      } /* endif */
      if (len && *cp) {
         new_p->instance_p = copy(cp, &l);   /* copy instance OID    */
         cp   += l;                          /* set ptr to next byte */
         len  -= l;                          /* set remaining length */
      } else if (len) {                      /* must be null string  */
         cp++;                               /* point to next byte   */
         len--;                              /* set remaining length */
      } /* endif */
      new_p->object_p = concat(              /* OID is concatenation */
                       new_p->group_p,       /* of groupID and       */
                       new_p->instance_p);   /* instanceID           */
      break;
    } /* endswitch */

    if (new_p->object_p) {                   /* If we have object ID */
       l = strlen(new_p->object_p)-1;        /* Possible trailing dot*/
       if (new_p->object_p[l] =='.') {       /* If object ID has a   */
           new_p->object_p[l] = '\0';        /* trailing dot, then   */
       } /* endif */                         /* remove it            */
    } else {                                 /* failed to get memory */
       if (first_p) fDPIset_packet(first_p); /* free if allocated    */
       return(snmp_dpi_set_packet_NULL_p);   /* return NULL ptr      */
    } /* endif */

    if (len<3) {                             /* remaining length < 3 */
       DPI_DEBUG(1,DPI_DM(                   /* tell if debug active */
              DPI_M_MISSING_VALUE));
       fDPIset_packet(first_p);              /* free allocated mem.  */
       return(snmp_dpi_set_packet_NULL_p);   /* return NULL ptr      */
    } /* endif */
 
    new_p->value_type = *(cp++);             /* pick up  value type  */
    l  = s_int_val((unsigned char *)cp);     /* and value length,    */
    cp  += 2;                                /* which is 2 bytes     */
    len -= 3;                                /* set remaining length */
 
    if (len<l) {                             /* remaining length<len */
       DPI_DEBUG(1,DPI_DM(                   /* tell if debug active */
              DPI_M_INVALID_VAL_LENGTH));
       fDPIset_packet(first_p);              /* free allocated mem.  */
       return(snmp_dpi_set_packet_NULL_p);   /* return NULL ptr      */
    } /* endif */
 
    new_p->value_len = l;                    /* set value length and */
    if (l) {                                 /* if length not zero   */
       new_p->value_p = (char *) malloc(l);  /* get memory for value */
       if (new_p->value_p == (char *)0) {    /* NULL ptr means error */
          DPI_DEBUG(1,DPI_DM(                /* if debug_lvl >= 1    */
              DPI_M_OUT_OF_MEMORY));         /* give this message    */
          fDPIset_packet(first_p);           /* free allocated mem.  */
          return(snmp_dpi_set_packet_NULL_p);/* return NULL ptr      */
       } /* endif */
    } /* endif */
 
    if (new_p->value_type & ~SNMP_TYPE_MASK) /* an integer type      */
    {
       if (new_p->value_len != 4) {          /* must have length 4   */
          DPI_DEBUG(1,DPI_DM(                /* show error if debug  */
            DPI_M_INVALID_INT_LENGTH,
            "pDPIset", new_p->value_len));
          fDPIset_packet(first_p);           /* free memory if any   */
          return(snmp_dpi_set_packet_NULL_p);/* return NULL ptr      */
       } /* endif */

       if (new_p->value_type ==              /* INTERNET IP address  */
           SNMP_TYPE_DPI_1_INTERNET)         /* DPI 1.x format       */
       {                                     /* By definition it is  */
          memcpy(new_p->value_p, cp,         /* in network byte order*/
                 new_p->value_len);          /* copy to value area   */
       } else {                              /* else: integer type   */
          li = int_val((unsigned char *)cp); /* convert to long int  */
          memcpy(new_p->value_p, &li,        /* then copy it to      */
                 new_p->value_len);          /* value area           */
       } /* endif */
    } else if (new_p->value_type ==          /* a type of Counter 64 */
                SNMP_TYPE_Counter64)         /* needs extra handling */
    {                                        /* which is done here   */
       snmp_dpi_u64 *u64;
       if (new_p->value_len != 8) {          /* must have length 8   */
          DPI_DEBUG(1,DPI_DM(                /* show error if debug  */
            DPI_M_INVALID_I64_LENGTH,
            "pDPIset", new_p->value_len));
          fDPIset_packet(first_p);           /* free memory if any   */
          return(snmp_dpi_set_packet_NULL_p);/* return NULL ptr      */
       } /* endif */
       if (sizeof(snmp_dpi_u64)>             /* if allocated size is */
            new_p->value_len)                /* too small for struct */
       {                                     /* (not expected, but?) */
          free(new_p->value_p);              /* then free old area   */
          new_p->value_p = (char *) malloc(  /* and obtain new one   */
                       sizeof(snmp_dpi_u64));/* for the structure    */
          if (new_p->value_p == (char *)0) { /* NULL ptr means error */
             DPI_DEBUG(1,DPI_DM(             /* if debug_lvl >= 1    */
                 DPI_M_OUT_OF_MEMORY));      /* give this message    */
             fDPIset_packet(first_p);        /* free allocated mem.  */
             return(snmp_dpi_set_packet_NULL_p);/* return NULL ptr   */
          } /* endif */
       } /* endif */
       u64 = (snmp_dpi_u64 *)new_p->value_p;    /* point to structure*/
       u64->high=int_val((unsigned char *)cp);  /* convert high order*/
       u64->low=int_val((unsigned char *)cp+4); /* convert low order */
    } else if (l) {                             /* An octet string   */
       memcpy(new_p->value_p, cp, l);           /* copy the value    */
       /* BIT_STRING handled as OCTET_STRING */
    } /* endif (new_p->type & ~SNMP_TYPE_MASK) */

    cp  += l;                                /* point to next SET    */
    len -= l;                                /* set remaining length */
 
    DPI_DEBUG(1,dpi_debug_print_set(         /* if debug: print line */
                          "pDPIset",new_p)); /* showing packet       */
   } /* while (len > 0) */

   return(first_p);                          /* return ptr to caller */
}

/*********************************************************************/
/* Function to parse a DPI TRAP packet                               */
/* - if success, return ptr to dpi_trap_packet structure             */
/* - if failure, return NULL ptr                                     */
/*********************************************************************/
static snmp_dpi_trap_packet *pDPItrap(    /* parse a DPI trap packet */
            snmp_dpi_hdr  *hdr_p,         /* ptr to hdr structure    */
            unsigned char *packet_p,      /* ptr to packet data      */
            int            len)           /* length packet data      */
{
   snmp_dpi_trap_packet *new_p;           /* ptr to a TRAP structure */
   int                   l;               /* length of string        */
   char                 *cp;              /* current ptr into packet */

   if (len == 0) {                        /* nothing to parse        */
      DPI_DEBUG(1,DPI_DM(                 /* tell if debug active    */
             DPI_M_ZERO_LENGTH));
      return(snmp_dpi_trap_packet_NULL_p);/* return NULL ptr         */
   } /* endif */

   cp = (char *)packet_p;                 /* point to first byte     */

   new_p = (snmp_dpi_trap_packet *)calloc(1, /* allocate one TRAP    */
           sizeof(snmp_dpi_trap_packet));    /* structure, init zero */

   if (new_p == snmp_dpi_trap_packet_NULL_p) /* NULL means error     */
   {
      DPI_DEBUG(1,DPI_DM(                    /* if debug_lvl >= 1    */
              DPI_M_OUT_OF_MEMORY));         /* give this message    */
      return(snmp_dpi_trap_packet_NULL_p);   /* return a NULL ptr    */
   }

   switch (hdr_p->proto_version) {        /* handle by DPI version   */
   case 1:                                /* Version 1.x             */
     switch (hdr_p->proto_release) {      /* handle by Release       */
     case 0:                              /* DPI Version 1.0         */
       new_p->generic  = *packet_p;       /* set generic trap type   */
       new_p->specific = *(packet_p + 1); /* set specific trap type  */
       cp  += 2;                          /* point to varBind data   */
       len -= 2;                          /* set remeaining length   */
       break;
     case 1:                              /* DPI Version 1.1         */
     default:
       new_p->generic  = int_val(packet_p);  /* get 4-byte generic   */
       new_p->specific = int_val(packet_p+4);/* and 4-byte specific  */
       cp  += 8;                             /* point to Enterprise  */
       len -= 8;                             /* set remaining length */
       if (len && *cp) {                     /* Enterprise specified */
          new_p->enterprise_p = copy(cp, &l);/* so copy it           */
          if (new_p->enterprise_p) {         /* means success        */
             cp  += l;                       /* point to varBind data*/
             len -= l;                       /* set remaining length */
          } else {                           /* NULL ptr means error */
             free(new_p);                    /* free alocated memory */
             return(snmp_dpi_trap_packet_NULL_p); /* return NULL ptr */
          } /* endif */
       } else if (len) {
          cp++;                              /* point to varBind data*/
          len--;                             /* set remaining length */
       } /* endif */
       break; 
     } /* end switch (hdr_p->proto_release) */
     break;
   case 2:                                   /* DPI Version 2.x      */
     switch (hdr_p->proto_release) {         /* handle by Release    */
     case 0:                                 /* DPI Version 2.0      */
     default:                                /* DPI Version 2.x      */
       new_p->generic  = int_val(packet_p);  /* get 4-byte generic   */
       new_p->specific = int_val(packet_p+4);/* and 4-byte specific  */
       cp  += 8;                             /* point to Enterprise  */
       len -= 8;                             /* set remaining length */
       if (len && *cp) {                     /* Enterprise specified */
          new_p->enterprise_p = copy(cp, &l);/* so copy it           */
          if (new_p->enterprise_p) {         /* means success        */
             cp  += l;                       /* point to varBind data*/
             len -= l;                       /* set remaining length */
          } else {                           /* NULL ptr means error */
             free(new_p);                    /* free alocated memory */
             return(snmp_dpi_trap_packet_NULL_p); /* return NULL ptr */
          } /* endif */
       } else if (len) {                     /* NULL as enterpriseID */
          cp++;                              /* account for the NULL */
          len--;                             /* byte in the packet   */
       } /* endif */
       break; 
     } /* end switch (hdr_p->proto_release) */
     break;
   default:
     DPI_DEBUG(1,DPI_DM(                    /* tell if debug active  */
            DPI_M_INVALID_DPI_VERSION));
     fDPItrap_packet(new_p);                /* free allocated memory */
     return(snmp_dpi_trap_packet_NULL_p);   /* return NULL ptr       */
     /* break; */                           /* NEVER REACHED         */
   } /* end switch (hdr->proto_minor) */

   DPI_DEBUG(1,dpi_debug_print_trap(        /* if debug: print line  */
                         "pDPItrap",new_p));/* showing packet        */

   if (len) {
      new_p->varBind_p = pDPIset(hdr_p,     /* process varBind data  */
                     (unsigned char *)cp,   /* which is SET packet   */
                     len);                  /* remaining length      */
      if (new_p->varBind_p ==
          snmp_dpi_set_packet_NULL_p)       /* NULL means error      */
      {                                     /* so we must            */
        fDPItrap_packet(new_p);             /* free memory and       */
        return(snmp_dpi_trap_packet_NULL_p);/* return NULL ptr       */
      } /* endif */
   } /* endif */

   return(new_p);                           /* return ptr to caller  */
}

/*********************************************************************/
/* Function to parse a DPI RESPONSE packet                           */
/* - if success, return ptr to dpi_resp_packet structure             */
/* - if failure, return NULL ptr                                     */
/*********************************************************************/
static snmp_dpi_resp_packet *pDPIresp(      /* parse RESPONSE packet */
                snmp_dpi_hdr  *hdr_p,       /* ptr to hdr structure  */
                unsigned char *packet_p,    /* ptr to packet data    */
                int            len)         /* remaining length      */
{
   snmp_dpi_resp_packet *new_p;             /* ptr to RESP structure */
   char                 *cp;                /* ptr to current char   */

   if (len == 0) {                          /* nothing to parse      */
      DPI_DEBUG(1,DPI_DM(                   /* tell if debug active  */
             DPI_M_ZERO_LENGTH));
      return(snmp_dpi_resp_packet_NULL_p);  /* return NULL ptr       */
   } /* endif */

   cp    = (char *)packet_p;                /* point to first byte   */

   new_p = (snmp_dpi_resp_packet *)calloc(1,/* allocate 1 RESPONSE   */
            sizeof(snmp_dpi_resp_packet));  /* structure, init zero  */

   if (new_p == snmp_dpi_resp_packet_NULL_p)/* NULL ptr means error  */
   {
      DPI_DEBUG(1,DPI_DM(                    /* if debug_lvl >= 1    */
              DPI_M_OUT_OF_MEMORY));         /* give this message    */
      return(snmp_dpi_resp_packet_NULL_p);  /* so return a NULL ptr  */
   } /* endif */

   switch (hdr_p->proto_version) {
   case 1:                                  /* DPI Version 1.x       */
     new_p->error_code  = *cp;              /* has a 1-byte retcode  */
     if (new_p->error_code) {               /* if error, then set    */
        new_p->error_index = 1;             /* index to 1, DPI 1.x   */
     } /* endif */                          /* always had 1 varBind  */
     cp++;                                  /* point to varBind data */
     len--;                                 /* set remeaining length */
     break;
   case 2:                                  /* DPI Version 2.x       */
   default:                                 /* and up                */
     new_p->error_code  = *cp++;            /* has 1-byte retcode    */
     new_p->error_index = int_val(          /* and 4-byte index      */
                (unsigned char *)cp);       /* convert to host order */
     cp  += 4;                              /* point to varBind data */
     len -= 5;                              /* set remaining length  */
     break;
   } /* endswitch */

   DPI_DEBUG(1,dpi_debug_print_resp(        /* print contents of the */
                     "pDPIresp", new_p));   /* packet if debug is on */

   if (len) {                               /* if varBinds present   */
      new_p->varBind_p = pDPIset(hdr_p,     /* then call pDPIset     */
                 (unsigned char *)cp, len); /* to parse them         */
      if (new_p->varBind_p ==
          snmp_dpi_set_packet_NULL_p)       /* NULL means error      */
      {                                     /* so we must            */
        fDPIresp_packet(new_p);             /* free allocated memory */
        return(snmp_dpi_resp_packet_NULL_p);/* and return NULL ptr   */
      } /* endif */
   } /* endif */

   return(new_p);                           /* return ptr to caller  */
}

/*********************************************************************/
/* Function to parse a DPI OPEN packet                               */
/* - if success, return ptr to dpi_open_packet structure             */
/* - if failure, return NULL ptr                                     */
/*********************************************************************/
static snmp_dpi_open_packet *pDPIopen(      /* parse OPEN packet     */
                snmp_dpi_hdr  *hdr_p,       /* ptr to hdr structure  */
                unsigned char *packet_p,    /* ptr to packet data    */
                int            len)         /* remaining length      */
{
   snmp_dpi_open_packet *new_p;             /* ptr to OPEN structure */
   char                 *cp;                /* ptr to current char   */
   int                   l;

   if (len == 0) {                          /* nothing to parse      */
      DPI_DEBUG(1,DPI_DM(                   /* tell if debug active  */
             DPI_M_ZERO_LENGTH));
      return(snmp_dpi_open_packet_NULL_p);  /* return NULL ptr       */
   } /* endif */

   cp    = (char *)packet_p;                /* point to first byte   */

   new_p = (snmp_dpi_open_packet *)calloc(1,/* allocate 1 OPEN       */
            sizeof(snmp_dpi_open_packet));  /* structure, init zero  */

   if (new_p == snmp_dpi_open_packet_NULL_p)/* NULL ptr means error  */
   {
      DPI_DEBUG(1,DPI_DM(                   /* if debug_lvl >= 1     */
              DPI_M_OUT_OF_MEMORY));        /* give this message     */
      return(snmp_dpi_open_packet_NULL_p);  /* so return a NULL ptr  */
   } /* endif */

   switch (hdr_p->proto_version) {
   case 1:   /* Does not really have OPEN *//* DPI Version 1.x       */
   case 2:                                  /* DPI Version 2.x       */
   default:                                 /* and up                */
     new_p->timeout      = s_int_val(packet_p);   /* 2-byte timeout  */
     new_p->max_varBinds = s_int_val(packet_p+2); /* 2-byte max vB   */
     cp  += 4;                              /* point to next byte    */
     len -= 4;                              /* set remaining length  */
     if (len) {                             /* pick up Selected      */
        new_p->character_set = *cp++;       /* character set         */
        len--;                              /* set remaining length  */
     } else {
        DPI_DEBUG(1,DPI_DM(                 /* if debug_lvl >= 1     */
                DPI_M_MISSING_CSET));       /* give this message     */
        return(snmp_dpi_open_packet_NULL_p);/* so return a NULL ptr  */
     } /* endif */
     if (len && *cp) {                      /* Identifier specified  */
        new_p->oid_p = copy(cp, &l);        /* so copy it for us     */
        if (new_p->oid_p) {                 /* success, so now       */
           cp  += l;                        /* point to next byte    */
           len -= l;                        /* set remaining length  */
        } else {                            /* NULL ptr means error  */
           fDPIopen_packet(new_p);          /* free alocated memory  */
           return(snmp_dpi_open_packet_NULL_p);/* return NULL ptr    */
        } /* endif */
     } else {
        DPI_DEBUG(1,DPI_DM(                 /* tell if debug active  */
               DPI_M_MISSING_SA_OID));
        fDPIopen_packet(new_p);             /* free allocated memory */
        return(snmp_dpi_open_packet_NULL_p);/* return NULL ptr       */
     } /* endif */
     if (len && *cp) {                      /* Description specified */
        new_p->description_p = copy(cp, &l);/* so copy it for us     */
        if (new_p->description_p) {         /* if success, then      */
           cp  += l;                        /* point to next byte    */
           len -= l;                        /* set remaining length  */
        } else {                            /* NULL ptr means error  */
           fDPIopen_packet(new_p);          /* free alocated memory  */
           return(snmp_dpi_open_packet_NULL_p);/* return NULL ptr    */
        } /* endif */
     } else {
        DPI_DEBUG(1,DPI_DM(                 /* tell if debug active  */
             DPI_M_MISSING_SA_DESCR));
        fDPIopen_packet(new_p);             /* free allocated memory */
        return(snmp_dpi_open_packet_NULL_p);/* return NULL ptr       */
     } /* endif */
     new_p->password_len  =                 /* password len; 2 bytes */
            s_int_val((unsigned char *)cp); /* in network byte order */
     cp  += 2;                              /* point to next byte    */
     len -= 2;                              /* set remaining length  */
     if (new_p->password_len) {             /* if password present   */
        new_p->password_p = (unsigned char *)  /* allocate memory    */
               malloc(new_p->password_len); /* for it now.           */
        if (new_p->password_p) {            /* if success, then      */
           memcpy(new_p->password_p, cp,    /* copy the password     */
                  new_p->password_len);     /* to our new area       */
           cp += new_p->password_len;       /* point after password  */
        } else {                            /* else error            */
           DPI_DEBUG(1,DPI_DM(              /* if debug_lvl >= 1     */
              DPI_M_OUT_OF_MEMORY));        /* give this message     */
           fDPIopen_packet(new_p);          /* free alocated memory  */
           return(snmp_dpi_open_packet_NULL_p);/* return NULL ptr    */
        } /* endif */
     } /* endif */
     break;
   } /* endswitch */

   DPI_DEBUG(1,dpi_debug_print_open(        /* if debug is on, then  */
                      "pDPIopen", new_p));  /* print packet content  */

   return(new_p);                           /* return ptr to caller  */
}

/*********************************************************************/
/* Function to parse a DPI CLOSE packet                              */
/* - if success, return ptr to dpi_close_packet structure            */
/* - if failure, return NULL ptr                                     */
/*********************************************************************/
static snmp_dpi_close_packet *pDPIclose(    /* parse CLOSE packet    */
                snmp_dpi_hdr  *hdr_p,       /* ptr to hdr structure  */
                unsigned char *packet_p,    /* ptr to packet data    */
                int            len)         /* remaining length      */
{
   snmp_dpi_close_packet *new_p;            /* ptr to CLOSE structure*/
   char                  *cp;               /* ptr to current char   */

   if (len == 0) {                          /* nothing to parse      */
      DPI_DEBUG(1,DPI_DM(                   /* tell if debug active  */
             DPI_M_ZERO_LENGTH));
      return(snmp_dpi_close_packet_NULL_p); /* return NULL ptr       */
   } /* endif */

   cp    = (char *)packet_p;                /* point to first byte   */

   new_p = (snmp_dpi_close_packet *)calloc( /* allocate 1 CLOSE      */
           1,sizeof(snmp_dpi_close_packet));/* structure, init zero  */

   if (new_p == snmp_dpi_close_packet_NULL_p)  /* NULL  means error  */
   {
      DPI_DEBUG(1,DPI_DM(                   /* if debug_lvl >= 1     */
              DPI_M_OUT_OF_MEMORY));        /* give this message     */
      return(snmp_dpi_close_packet_NULL_p); /* so return a NULL ptr  */
   } /* endif */

   switch (hdr_p->proto_version) {
   case 1:   /* does not really have CLOSE*//* DPI Version 1.x       */
   case 2:                                  /* DPI Version 2.x       */
   default:                                 /* and up                */
     new_p->reason_code = *cp;              /* has 1-byte reason code*/
     cp++;                                  /* point to varBind data */
     len--;                                 /* set remaining length  */
     break;
   } /* endswitch */

   DPI_DEBUG(1,dpi_debug_print_close(       /* if debug is on, then  */
                      "pDPIclose", new_p)); /* print packet content  */

   return(new_p);                           /* return ptr to caller  */
}

/*********************************************************************/
/* Function to free complete DPI parse tree                          */
/*********************************************************************/
void fDPIparse(snmp_dpi_hdr *hdr_p)       /* free a DPI parse tree   */
{
   if (hdr_p) {
      switch (hdr_p->packet_type) {
      case SNMP_DPI_GET:
           fDPIget_packet(hdr_p->data_u.get_p);
           break;
      case SNMP_DPI_GETNEXT:
           fDPInext_packet(hdr_p->data_u.next_p);
           break;
      case SNMP_DPI_RESPONSE:
           fDPIresp_packet(hdr_p->data_u.resp_p);
           break;
      case SNMP_DPI_SET:
      case SNMP_DPI_COMMIT:
      case SNMP_DPI_UNDO:
           fDPIset_packet(hdr_p->data_u.set_p);
           break;
      case SNMP_DPI_TRAP:
           fDPItrap_packet(hdr_p->data_u.trap_p);
           break;
      case SNMP_DPI_REGISTER:
           fDPIreg_packet(hdr_p->data_u.reg_p);
           break;
      case SNMP_DPI_UNREGISTER:
           fDPIureg_packet(hdr_p->data_u.ureg_p);
           break;
      case SNMP_DPI_OPEN:
           fDPIopen_packet(hdr_p->data_u.open_p);
           break;
      case SNMP_DPI_CLOSE:
           fDPIclose_packet(hdr_p->data_u.close_p);
           break;
      case SNMP_DPI_ARE_YOU_THERE:   
           break;                         /* no packet data          */
      case SNMP_DPI_GETBULK:              /* TODO: implement         */
      case SNMP_DPI_TRAPV2:
      case SNMP_DPI_INFORM:
      default:                            /* should not occur        */
           DPI_DEBUG(1,DPI_DM(            /* tell if debug active    */
               DPI_M_INVALID_DPI_TYPE));
           break;
      } /* endswitch */
#ifdef SNMP_DPI_VIEW_SELECTION
      if (hdr_p->community_p) free(hdr_p->community_p);
#endif /* def SNMP_DPI_VIEW_SELECTION */
      free(hdr_p);
   }
}

/*********************************************************************/
/* Function to free DPI GET parse tree                               */
/*********************************************************************/
static void fDPIget_packet(               /* free a GET structure    */
             snmp_dpi_get_packet *pack_p) /* ptr to structure        */
{
   snmp_dpi_get_packet *ret_p, *next_p;   /* ptr to GET structure    */

   next_p = pack_p;                       /* point to structure      */

   while (next_p) {                       /* do all GET structures   */
      ret_p = next_p;                     /* point to the structure  */
      if (ret_p->object_p)                /* free OIDstring if any   */
         free(ret_p->object_p);
      if (ret_p->group_p)                 /* free OIDstring if any   */
         free(ret_p->group_p);
      if (ret_p->instance_p)              /* free INSTANCE  if any   */
         free(ret_p->instance_p);
      next_p = ret_p->next_p;             /* pick up ptr to next one */
      free(ret_p);                        /* free GET structure      */
   } /* endwhile */
}

/*********************************************************************/
/* Function to free DPI GETNEXT parse tree                           */
/*********************************************************************/
static void fDPInext_packet(              /* free GETNEXT structure  */
             snmp_dpi_next_packet *pack_p)/* ptr to structure        */
{
   snmp_dpi_next_packet *ret_p, *next_p;  /* ptr to GETNEXT structure*/

   next_p = pack_p;                       /* point to structure      */

   while (next_p) {                       /* do all GETNEXT structs  */
      ret_p = next_p;                     /* point to the structure  */
      if (ret_p->object_p)                /* free OIDstring if any   */
         free(ret_p->object_p);
      if (ret_p->group_p)                 /* free groupOID if any    */
         free(ret_p->group_p);
      if (ret_p->instance_p)              /* free INSTANCE  if any   */
         free(ret_p->instance_p);
      next_p = ret_p->next_p;             /* pick up ptr to next one */
      free(ret_p);                        /* free GETNEXT structure  */
   } /* endwhile */
}

/*********************************************************************/
/* Function to free DPI SET parse tree                               */
/*********************************************************************/
extern void fDPIset_packet(               /* free a SET structure    */
             snmp_dpi_set_packet *pack_p) /* ptr to SET structure    */
{
   snmp_dpi_set_packet *ret_p, *next_p;   /* ptr to SET structure    */

   next_p = pack_p;                       /* pick up ptr to SET      */

   while (next_p) {                       /* do all SET structures   */
      ret_p = next_p;                     /* point to the structure  */
      if (ret_p->object_p)                /* free OIDstring if any   */
         free(ret_p->object_p);
      if (ret_p->group_p)                 /* free OIDstring if any   */
         free(ret_p->group_p);
      if (ret_p->instance_p)              /* free INSTANCE  if any   */
         free(ret_p->instance_p);
      if (ret_p->value_p)                 /* free value if any       */
         free(ret_p->value_p);
      next_p = ret_p->next_p;             /* pick up ptr to next one */
      free(ret_p);                        /* free SET structure      */
   } /* endwhile */
}

/*********************************************************************/
/* Function to free DPI TRAP parse tree                              */
/*********************************************************************/
static void fDPItrap_packet(              /* free DPI trap structure */
             snmp_dpi_trap_packet *pack_p)/* ptr to trap structure   */
{
   if (pack_p) {                          /* if valid ptr            */
      if (pack_p->enterprise_p)           /* free enterprise if any  */
          free(pack_p->enterprise_p);
      if (pack_p->varBind_p)              /* free varBind if any     */
         fDPIset_packet(pack_p->varBind_p);
      free(pack_p);                       /* free trap structure     */
   } /* endif */
}

/*********************************************************************/
/* Function to free DPI RESPONSE parse tree                          */
/*********************************************************************/
static void fDPIresp_packet(snmp_dpi_resp_packet *pack_p)
{
   if (pack_p) {
      if (pack_p->varBind_p)              /* free varBind if any     */
         fDPIset_packet(pack_p->varBind_p);
      free(pack_p);                       /* free response structure */
   } /* endif */
}

/*********************************************************************/
/* Function to free DPI OPEN parse tree                              */
/*********************************************************************/
static void fDPIopen_packet(snmp_dpi_open_packet *pack_p)
{
   if (pack_p) {
      if (pack_p->oid_p)                  /* free oid if any         */
          free(pack_p->oid_p);
      if (pack_p->description_p)          /* free description if any */
          free(pack_p->description_p);
      if (pack_p->password_p)             /* free password if any    */
          free(pack_p->password_p);
      free(pack_p);                       /* free open structure     */
   } /* endif */
}

/*********************************************************************/
/* Function to free DPI CLOSE parse tree                             */
/*********************************************************************/
static void fDPIclose_packet(snmp_dpi_close_packet *pack_p)
{
   if (pack_p) {
      free(pack_p);                       /* free close structure    */
   } /* endif */
}

/*********************************************************************/
/* Function to free DPI REGISTER parse tree                          */
/*********************************************************************/
static void fDPIreg_packet(snmp_dpi_reg_packet *pack_p)
{
   snmp_dpi_reg_packet *ret_p, *next_p;   /* ptr to REG structure    */

   next_p = pack_p;                       /* pick up ptr to REG      */

   while (next_p) {                       /* do all REG structures   */
      ret_p = next_p;                     /* point to the structure  */
      if (ret_p->group_p)                 /* free OIDstring if any   */
         free(ret_p->group_p);
      next_p = ret_p->next_p;             /* pick up ptr to next one */
      free(ret_p);                        /* free SET structure      */
   } /* endwhile */
}

/*********************************************************************/
/* Function to free DPI UNREGISTER parse tree                        */
/*********************************************************************/
static void fDPIureg_packet(snmp_dpi_ureg_packet *pack_p)
{
   snmp_dpi_ureg_packet *ret_p, *next_p;  /* ptr to UREG structure   */

   next_p = pack_p;                       /* pick up ptr to REG      */

   while (next_p) {                       /* do all REG structures   */
      ret_p = next_p;                     /* point to the structure  */
      if (ret_p->group_p)                 /* free OIDstring if any   */
         free(ret_p->group_p);
      next_p = ret_p->next_p;             /* pick up ptr to next one */
      free(ret_p);                        /* free SET structure      */
   } /* endwhile */
}

/*********************************************************************/
/* Function to create a complete DPI packet from DPI parse tree      */
/*********************************************************************/
unsigned char *cDPIpacket(                /* create a DPI packet     */
                  snmp_dpi_hdr *hdr_p)    /* ptr to parse tree       */
{
   int rc, len;                           /* for retcode and length  */

   if (hdr_p == snmp_dpi_hdr_NULL_p) {    /* if NULL ptr passed,     */
      return((unsigned char *)0);         /* return NULL ptr also    */
   } /* endif */

   DPI_DEBUG(1,dpi_debug_print_hdr(       /* if debug is on, then    */
                 "cDPIpacket", hdr_p));   /* print packet content    */

   packet_len = 2;                        /* always 2 length bytes   */
   packet_buf[packet_len++] = hdr_p->proto_major;
   packet_buf[packet_len++] = hdr_p->proto_version;
   packet_buf[packet_len++] = hdr_p->proto_release;
   if (hdr_p->proto_version >= 2) {
       packet_buf[packet_len++] = (hdr_p->packet_id >> 8)& 0xff;
       packet_buf[packet_len++] = (hdr_p->packet_id     )& 0xff;
   } /* endif */
   packet_buf[packet_len++] = hdr_p->packet_type;

#ifdef SNMP_DPI_VIEW_SELECTION
   if (hdr_p->proto_version >= 2) {       /* handle View Selection   */
      switch (hdr_p->packet_type) {       /* but only for these      */
      case SNMP_DPI_GET:                  /* packet types            */
      case SNMP_DPI_GETNEXT:
      case SNMP_DPI_SET:
      case SNMP_DPI_COMMIT:
      case SNMP_DPI_UNDO:
        packet_buf[packet_len++] = (hdr_p->community_len >> 8)& 0xff;
        packet_buf[packet_len++] = (hdr_p->community_len     )& 0xff;
        if (hdr_p->community_len) {       /* if community present    */
           if (hdr_p->community_p) {
              PACKET_UCOPY(hdr_p->community_p, hdr_p->community_len);
           } else {
              DPI_DEBUG(1,DPI_DM(            /* if debug_lvl >= 1    */
                  DPI_M_MISSING_COMMUNITY)); /* give this message    */
           } /* endif */
        } /* endif */    
        break;
      } /* endswitch */
   } /* endif */
#endif

   len = packet_len;
   switch (hdr_p->packet_type) {
   case SNMP_DPI_GET:
        rc = cDPIget(hdr_p->data_u.get_p);
        break;
   case SNMP_DPI_GETNEXT:
        rc = cDPInext(hdr_p->data_u.next_p);
        break;
   case SNMP_DPI_SET:
   case SNMP_DPI_COMMIT:
   case SNMP_DPI_UNDO:
        rc = cDPIset(hdr_p->data_u.set_p);
        break;
   case SNMP_DPI_TRAP:
        rc = cDPItrap(hdr_p->data_u.trap_p);
        break;
   case SNMP_DPI_RESPONSE:
        rc = cDPIresp(hdr_p->data_u.resp_p);
        break;
   case SNMP_DPI_REGISTER:
        rc = cDPIreg(hdr_p->data_u.reg_p);
        break;
   case SNMP_DPI_UNREGISTER:
        rc = cDPIureg(hdr_p->data_u.ureg_p);
        break;
   case SNMP_DPI_OPEN:
        rc = cDPIopen(hdr_p->data_u.open_p);
        break;
   case SNMP_DPI_CLOSE:
        rc = cDPIclose(hdr_p->data_u.close_p);
        break;
   case SNMP_DPI_ARE_YOU_THERE:           /* place holders for       */
        rc = 0;                           /* no packet data          */
        break;     
   case SNMP_DPI_TRAPV2:                  /* place holders for       */
   case SNMP_DPI_INFORM:                  /* SNMPv2 support          */
   case SNMP_DPI_GETBULK:                 /* TODO: implement         */
   default:                               /* should not occur        */
        DPI_DEBUG(1,DPI_DM(               /* tell if debug active    */
               DPI_M_INVALID_DPI_TYPE));
        rc = -3;
        break;
   } /* endswitch */

   if (rc < 0) {                          /* error from cDPIxxx      */
      if (rc == -1) {
         DPI_DEBUG(1,DPI_DM(              /* tell if debug active    */
             DPI_M_DPI_ENCODING_ERROR));
      } else if (rc == -2) {
         DPI_DEBUG(1,DPI_DM(              /* tell if debug active    */
             DPI_M_DPI_BUFFER_OVERFLOW));
      } /* endif */
      return((unsigned char *)0);         /* return a NULL ptr       */
   }

   if ((len == packet_len) &&             /* anything besides hdr ?  */
       (hdr_p->packet_type !=             /* for any packet other    */
        SNMP_DPI_ARE_YOU_THERE))          /* than this one it is     */
   {                                      /* an error                */
      return((unsigned char *)0);         /* return NULL ptr if not  */
   } /* endif */

   len             = packet_len - 2;      /* Length of data after    */
   packet_buf[1]   = len & 0xff;          /* first 2 length bytes    */
   len           >>= 8;                   /* goes into first 2 bytes */
   packet_buf[0]   = len & 0xff;          /* in network byte order   */

   return(packet_buf);                    /* ptr to static buffer    */
}

/*********************************************************************/
/* Function to create a DPI OPEN packet from DPI OPEN tree           */
/*********************************************************************/
static int cDPIopen(                         /* serialize OPEN       */
            snmp_dpi_open_packet *pack_p)    /* ptr to OPEN strcuture*/
{
 int           len;                          /* receives length      */
 unsigned char data[4];

 if (pack_p) {                               /* only if we have a ptr*/
    DPI_DEBUG(1,dpi_debug_print_open(        /* if debug is on, then */
                      "cDPIopen", pack_p));  /* print packet content */
   data[0] = (pack_p->timeout>>8) & 0xff;    /* timeout in network   */
   data[1] = (pack_p->timeout   ) & 0xff;    /* byte order please    */
   data[2] = (pack_p->max_varBinds>>8)&0xff; /* same for the         */
   data[3] = (pack_p->max_varBinds   )&0xff; /* max_varBinds pls     */
   PACKET_COPY(data,4);                      /* copy to packet       */
   PACKET_COPY(&pack_p->character_set,1);    /* copy select cset     */
   if (pack_p->oid_p) {                      /* if Identifier exists,*/
      len = strlen(pack_p->oid_p)+1;         /* then get its length  */
      PACKET_COPY(pack_p->oid_p, len);       /* copy it into packet  */
   } else {
      DPI_DEBUG(1,DPI_DM(                    /* tell if debug active */
             DPI_M_MISSING_SA_OID));
      return(-1);                            /* return error         */
   } /* endif */
   if (pack_p->description_p) {              /* if Description exists*/
      len = strlen(pack_p->description_p)+1; /* then get its length  */
      PACKET_COPY(pack_p->description_p,len);/* copy it into packet  */
   } else {
      PACKET_COPY("",1);                     /* indicate none there  */
   } /* endif */
   if (pack_p->password_len &&               /* if password present  */
       pack_p->password_p)                   /* then put length in   */
   {                                         /* network byte order   */
     data[0] = (pack_p->password_len>>8)&0xff;
     data[1] = (pack_p->password_len>>8)&0xff;
     PACKET_COPY(data,2);                    /* copy length to packet*/
     PACKET_COPY(pack_p->password_p,         /* then copy password   */
                 pack_p->password_len);      /* into packet          */
   } else {
     data[0] = 0;                            /* set zero length in   */
     data[1] = 0;                            /* network byte order   */
     PACKET_COPY(data,2);                    /* copy length to packet*/
   } /* endif */
 } /* endif */
 
 return(0);                                  /* return rc=0, all OK  */
}

/*********************************************************************/
/* Function to create a DPI CLOSE packet from DPI CLOSE tree         */
/*********************************************************************/
static int cDPIclose(                        /* serialize CLOSE      */
            snmp_dpi_close_packet *pack_p)   /* ptr to CLOSEstrcuture*/
{
  if (pack_p) {                              /* only if we have a ptr*/
    DPI_DEBUG(1,dpi_debug_print_close(       /* if debug is on, then */
                      "cDPIclose", pack_p)); /* print packet content */
    PACKET_COPY(&pack_p->reason_code,1);     /* copy to packet       */
  } /* endif */

  return(0);                                 /* return rc=0, all OK  */
}

/*********************************************************************/
/* Function to create a DPI REGISTER packet from DPI REG tree        */
/*********************************************************************/
static int cDPIreg(                          /* serialize REGISTER   */
            snmp_dpi_reg_packet *pack_p)     /* ptr to REG strcuture */
{
 int           len;                          /* receives length      */
 unsigned char data[4];
 
 while (pack_p) {                            /* do all groupIDs      */
   DPI_DEBUG(1,dpi_debug_print_reg(          /* if debug is on, then */
                      "cDPIreg", pack_p));   /* print packet content */
   if (pack_p->group_p) {                    /* if group ID exits    */
      len = strlen(pack_p->group_p)+1;       /* then get its length  */
      if (pack_p->group_p[len-2] != '.') {   /* must be a dot        */
         DPI_DEBUG(1,DPI_DM(                 /* tell if debug active */
             DPI_M_MISSING_TRAILING_DOT));
         return(-1);                         /* return with error    */
      } /* endif */
   } else {
      DPI_DEBUG(1,DPI_DM(                    /* tell if debug active */
             DPI_M_MISSING_GROUP_ID));
      return(-1);                            /* return error         */
   } /* endif */

   switch (PACKET_VERSION) {
   case 1:                                   /* DPI version 1.x      */
     PACKET_COPY(pack_p->group_p, len);      /* copy group to packet */
     break;
   case 2:
     data[0] = (pack_p->priority>>24) & 0xff;/* convert priority to  */
     data[1] = (pack_p->priority>>16) & 0xff;/* network byte order   */
     data[2] = (pack_p->priority>> 8) & 0xff;/* ....                 */
     data[3] = (pack_p->priority    ) & 0xff;/* ....                 */
     PACKET_COPY(data,4);                    /* copy into packet     */
     data[0] = (pack_p->timeout >> 8) & 0xff;/* convert timeout  to  */
     data[1] = (pack_p->timeout     ) & 0xff;/* network byte order   */
     PACKET_COPY(data,2);                    /* copy into packet     */
#ifdef SNMP_DPI_VIEW_SELECTION
     PACKET_COPY(&pack_p->view,1);
#endif /* def SNMP_DPI_VIEW_SELECTION */
#ifdef SNMP_DPI_BULK_SELECTION
     PACKET_COPY(&pack_p->bulk,1);
#endif /* def SNMP_DPI_BULK_SELECTION */
#ifdef SNMP_DPI_REGISTER_WITH_COMMUNITY
     data[0] = (pack_p->community_len >> 8) & 0xff;
     data[1] = (pack_p->community_len     ) & 0xff;
     PACKET_COPY(data,2);                    /* copy into packet     */
     if (pack_p->community_len) {
        PACKET_COPY(pack_p->community_p,
                    pack_p->community_len);
     } /* endif */
#endif /* SNMP_DPI_REGISTER_WITH_COMMUNITY */
     PACKET_COPY(pack_p->group_p, len);      /* copy group to packet */
     break;
   default:                                  /* code out of sync     */
     DPI_DEBUG(1,DPI_DM(                     /* tell if debug active */
            DPI_M_INVALID_DPI_VERSION));
     return(-1);                             /* return error         */
     /* break; */                            /* NEVER REACHED        */
   } /* endswitch */
   pack_p = pack_p->next_p;                  /* pick up next one     */
 } /* endwhile */

 return(0);                                  /* return rc=0, all OK  */
}

/*********************************************************************/
/* Function to create a DPI UNREGISTER packet from DPI UREG tree     */
/*********************************************************************/
static int cDPIureg(                         /* serialize UNREGISTER */
            snmp_dpi_ureg_packet *pack_p)    /* ptr to UREG strcuture*/
{
 int           len;                          /* receives length      */
 
 while (pack_p) {                            /* do all groupIDs      */
   DPI_DEBUG(1,dpi_debug_print_ureg(         /* if debug is on, then */
                      "cDPIureg", pack_p));  /* print packet content */
   if (pack_p->group_p) {                    /* if group ID exits    */
      len = strlen(pack_p->group_p)+1;       /* then get its length  */
      if (pack_p->group_p[len-2] != '.') {   /* must be a dot        */
         DPI_DEBUG(1,DPI_DM(                 /* tell if debug active */
             DPI_M_MISSING_TRAILING_DOT));
         return(-1);                         /* return with error    */
      } /* endif */
   } else {
      DPI_DEBUG(1,DPI_DM(                    /* tell if debug active */
             DPI_M_MISSING_GROUP_ID));
      return(-1);                            /* return error         */
   } /* endif */

   PACKET_COPY(&pack_p->reason_code,1);      /* copy reason code     */
   PACKET_COPY(pack_p->group_p, len);        /* copy group to packet */
   pack_p = pack_p->next_p;                  /* pick up next one     */
 } /* endwhile */

 return(0);                                  /* return rc=0, all OK  */
}

/*********************************************************************/
/* Function to create a DPI GET packet from DPI GET tree             */
/*********************************************************************/
static int cDPIget(                          /* serialize GET        */
            snmp_dpi_get_packet *pack_p)     /* ptr to GET strcuture */
{
   int len;                                  /* receives length      */


   while (pack_p) {                          /* do all variables     */
     DPI_DEBUG(1,dpi_debug_print_get(        /* if debug is on, then */
                      "cDPIget", pack_p));   /* print packet content */
     if (pack_p->group_p) {                  /* if group ID exits    */
        len = strlen(pack_p->group_p)+1;     /* then get its length  */
        PACKET_COPY(pack_p->group_p, len);   /* copy it into packet  */
     } else {
        DPI_DEBUG(1,DPI_DM(                  /* tell if debug active */
               DPI_M_MISSING_GROUP_ID));
        return(-1);                          /* return error         */
     } /* endif */
     if (pack_p->instance_p) {               /* if instance ID exists*/
        if (PACKET_VERSION < 2) packet_len--;/* DPI 1.x needs one ID */
        len = strlen(pack_p->instance_p)+1;  /* then get its length  */
        PACKET_COPY(pack_p->instance_p,len); /* copy it into packet  */
     } else {
        if (PACKET_VERSION < 2)              /* DPI 1.x needs one ID */
            packet_len -=2;                  /* without trailing dot */
        PACKET_COPY("",1);                   /* indicate no instance */
     } /* endif */
     pack_p = pack_p->next_p;                /* pick up next one     */
   } /* endwhile */

   return(0);                                /* return rc=0, all OK  */
}

/*********************************************************************/
/* Function to create a DPI GETNEXT packet from DPI GETNEXT tree     */
/*********************************************************************/
static int cDPInext(snmp_dpi_next_packet *pack_p)
{
 int len;                                    /* receives length      */

 while (pack_p) {                            /* do all variables     */
   DPI_DEBUG(1,dpi_debug_print_next(         /* if debug is on, then */
                      "cDPInext", pack_p));  /* print packet content */
   switch (PACKET_VERSION) {
   case 1:                                   /* for DPI version 1.x  */
     if (pack_p->group_p) {                  /* if group ID exits    */
        len = strlen(pack_p->group_p);       /* then get its length  */
        PACKET_COPY(pack_p->group_p, len);   /* copy it into packet  */
     } else {
        DPI_DEBUG(1,DPI_DM(                  /* tell if debug active */
               DPI_M_MISSING_GROUP_ID));
        return(-1);                          /* return error         */
     } /* endif */
     if (pack_p->instance_p) {               /* if instance ID exists*/
        len = strlen(pack_p->instance_p)+1;  /* then get its length  */
        PACKET_COPY(pack_p->instance_p,len); /* copy it into packet  */
     } else {                                /* if no instance ID,   */
        packet_len -= 1;                     /* remove trailing dot  */
        PACKET_COPY("",1);                   /* sets NULL at end     */
     } /* endif */
     len = strlen(pack_p->group_p)+1;        /* add group ID at after*/
     PACKET_COPY(pack_p->group_p, len);      /* the objectID         */
     break;
   case 2:                                   /* for DPI version 2.x  */
     if (pack_p->group_p) {                  /* if group ID exits    */
        len = strlen(pack_p->group_p)+1;     /* then get its length  */
        PACKET_COPY(pack_p->group_p, len);   /* copy it into packet  */
     } else {
        DPI_DEBUG(1,DPI_DM(                  /* tell if debug active */
               DPI_M_MISSING_GROUP_ID));
        return(-1);                          /* return error         */
     } /* endif */
     if (pack_p->instance_p) {               /* if instance ID exists*/
        len = strlen(pack_p->instance_p)+1;  /* then get its length  */
        PACKET_COPY(pack_p->instance_p,len); /* copy it into packet  */
     } else {
        PACKET_COPY("",1);                   /* indicate no instance */
     } /* endif */
     break;
   default:                                  /* code out of sync     */
     DPI_DEBUG(1,DPI_DM(                     /* tell if debug active */
            DPI_M_INVALID_DPI_VERSION));
     return(-1);
     /* break */                             /* NEVER REACHED        */
   } /* endswitch */                         
   pack_p = pack_p->next_p;                  /* pick up next one     */
 } /* endwhile */

 return(0);                                  /* return rc=0, all OK  */
}

/*********************************************************************/
/* Function to create a DPI SET packet from DPI SET tree             */
/*********************************************************************/
static int cDPIset(snmp_dpi_set_packet *pack_p)
{
 int            len;
 long int      *i_p;
 snmp_dpi_u64  *u64;
 unsigned char  data[4];

 while (pack_p) {
    DPI_DEBUG(1,dpi_debug_print_set(         /* if debug is on, then */
                      "cDPIset", pack_p));   /* print packet content */
    if (pack_p->group_p) {                   /* if group ID exits    */
       len = strlen(pack_p->group_p)+1;      /* then get its length  */
       PACKET_COPY(pack_p->group_p, len);    /* copy it into packet  */
    } else {
       DPI_DEBUG(1,DPI_DM(                   /* tell if debug active */
               DPI_M_MISSING_GROUP_ID));
       return(-1);                           /* return error         */
    } /* endif */
    if (pack_p->instance_p) {                /* if instance ID exists*/
       if (PACKET_VERSION < 2) packet_len--; /* DPI 1.x needs one ID */
       len = strlen(pack_p->instance_p)+1;   /* then get its length  */
       PACKET_COPY(pack_p->instance_p,len);  /* copy it into packet  */
    } else {
       if (PACKET_VERSION < 2) {             /* DPI 1.x needs one ID */
          if (packet_buf[packet_len-2]=='.'){
              packet_len -=2;                /* without trailing dot */
          } else {
              packet_len -=1;                /* else just at end     */
          } /* endif */
       } /* endif */
       PACKET_COPY("",1);                    /* indicate no instance */
    } /* endif */

    if ((pack_p->value_len) && 
        (pack_p->value_p))
    {                                        /* we have a value      */
       if (PACKET_VERSION >= 2) {            /* for version 2 all OK */
          data[0] = pack_p->value_type;      /* pick up value type   */
       } else {                              /* for version 1.x      */
          data[0] = pack_p->value_type;      /* pick up value type   */
          if (data[0] == SNMP_TYPE_IpAddress) { /* convert this type */
            data[0] = SNMP_TYPE_DPI_1_INTERNET; /* to one known by   */
          } /* endif */                         /* DPI 1.x           */
       } /* endif */
       if (pack_p->value_type ==
           SNMP_TYPE_Counter64)              /* 64-bit Counter       */
       {
          data[1] = 0;                       /* length is 8 bytes    */
          data[2] = 8;                       /* network byte order   */
       } else {                              /* other types          */
          len = pack_p->value_len >> 8;      /* and its length in    */
          data[1] = len & 0xff;              /* network byte order   */
          data[2] = pack_p->value_len & 0xff;/* please. Then copy    */
       } /* endif */
       PACKET_COPY(data,3);                  /* type and length      */
       if ((pack_p->value_type & 
            ~SNMP_TYPE_MASK)                 /* integer type value   */
           && (pack_p->value_len != 4))      /* must have length 4   */
       {                                     /* if not,              */
         DPI_DEBUG(1,DPI_DM(                 /* show if debug on     */
           DPI_M_INVALID_INT_LENGTH,         /* this message         */
           "pDPIset", pack_p->value_len));   /* these arguments      */
         return(-1);
       } /* endif */
       switch (pack_p->value_type) {
       case SNMP_TYPE_Integer32:             /* 32-bit INTEGER       */
       case SNMP_TYPE_Counter32:             /* 32-bit Counter       */
       case SNMP_TYPE_Gauge32:               /* 32-bit Gauge         */
       case SNMP_TYPE_TimeTicks:             /* 32-bit TimeTicks     */
       case SNMP_TYPE_UInteger32:            /* 32-bit INTEGER       */
         i_p = (long int *)pack_p->value_p;  /* point long int value */
         data[0] = (*i_p >> 24) & 0xff;
         data[1] = (*i_p >> 16) & 0xff;
         data[2] = (*i_p >>  8) & 0xff;
         data[3] = (*i_p      ) & 0xff;
         PACKET_COPY(data,4);                /* copy it to packet    */
         break;
       case SNMP_TYPE_DPI_1_INTERNET:        /* IP address DPI 1.x   */
       case SNMP_TYPE_OCTET_STRING:          /* OCTET STRING         */
       case SNMP_TYPE_DisplayString:         /* IMPLICIT OCTET STR   */
       case SNMP_TYPE_OBJECT_IDENTIFIER:     /* OBJECT IDENTIFIER    */
       case SNMP_TYPE_NULL:                  /* NULL value           */
       case SNMP_TYPE_IpAddress:             /* IMPLICIT OCTET STR(4)*/
       case SNMP_TYPE_NsapAddress:           /* IMPLICIT OCTET STR   */
       case SNMP_TYPE_Opaque:                /* IMPLICIT OCTET STR   */
       case SNMP_TYPE_noSuchObject:          /* IMPLICIT NULL        */
       case SNMP_TYPE_noSuchInstance:        /* IMPLICIT NULL        */
       case SNMP_TYPE_endOfMibView:          /* IMPLICIT NULL        */
       case SNMP_TYPE_BIT_STRING:            /* BIT STRING           */
         PACKET_COPY(pack_p->value_p,        /* plain copy of value  */
                pack_p->value_len);          /* into the packet      */
         break;
       case SNMP_TYPE_Counter64:             /* 64-bit Counter       */
         u64 = (snmp_dpi_u64 *)              /* point to the 64bit   */
                pack_p->value_p;             /* structure            */
         data[0] = (u64->high >> 24) & 0xff; /* pick up high order   */
         data[1] = (u64->high >> 16) & 0xff; /* bytes                */
         data[2] = (u64->high >>  8) & 0xff;
         data[3] = (u64->high      ) & 0xff;
         PACKET_COPY(data,4);                /* copy it to packet    */
         data[0] = (u64->low  >> 24) & 0xff; /* pick up low order    */
         data[1] = (u64->low  >> 16) & 0xff; /* bytes                */
         data[2] = (u64->low  >>  8) & 0xff;
         data[3] = (u64->low       ) & 0xff;
         PACKET_COPY(data,4);                /* copy it to packet    */
         break;
       default:                              /* code out of sync ??  */
         DPI_DEBUG(1,DPI_DM(                 /* tell if debug active */
             DPI_M_INVALID_SNMP_TYPE,
             "cDPIset_packet", 
             pack_p->value_type));
         return(-1);
         /* break; */                        /* NEVER REACHED        */
       } /* endswitch */
    } else if (pack_p->value_len == 0) {     /* valid if NULL value  */
       switch (pack_p->value_type) {
       case SNMP_TYPE_NULL:                  /* NULL value           */
       case SNMP_TYPE_noSuchObject:          /* IMPLICIT NULL        */
       case SNMP_TYPE_noSuchInstance:        /* IMPLICIT NULL        */
       case SNMP_TYPE_endOfMibView:          /* IMPLICIT NULL        */
       case SNMP_TYPE_BIT_STRING:            /* NULL value           */
         data[0] = pack_p->value_type;       /* pick up value type   */
         len = pack_p->value_len >> 8;       /* and its length in    */
         data[1] = len & 0xff;               /* network byte order   */
         data[2] = pack_p->value_len & 0xff; /* please. Then copy    */
         PACKET_COPY(data,3);                /* type and length      */
         break;
       default:
         DPI_DEBUG(1,DPI_DM(                 /* tell if debug active */
              DPI_M_MISSING_VALUE));
         return(-1);                         /* return error         */
         /* break */                         /* NEVER REACHED        */
       } /* endswitch */
    } else {                                 /* no value             */
       DPI_DEBUG(1,DPI_DM(                   /* tell if debug active */
              DPI_M_MISSING_VALUE));
       return(-1);                           /* return error         */
    } /* endif */
    pack_p = pack_p->next_p;
 } /* endwhile */

 return(0);
}

/*********************************************************************/
/* Function to create a DPI RESPONSE packet from DPI RESPONSE tree   */
/*********************************************************************/
static int cDPIresp(snmp_dpi_resp_packet *pack_p)
{
  int rc = -1;                               /* untill proven wrong  */
  unsigned char data[4];

  if (pack_p) {                              /* if valid packet      */
   DPI_DEBUG(1,dpi_debug_print_resp(         /* if debug is on, then */
                      "cDPIresp", pack_p));  /* print packet content */
   switch (PACKET_VERSION) {
   case 1:
     PACKET_COPY(&pack_p->error_code,1)      /* copy to buffer       */
     if (pack_p->varBind_p) {                /* if any varBinds,     */
        rc = cDPIset(pack_p->varBind_p);     /* handle SET structure */
     } else rc = 0;                          /* no varBinds, all OK  */
     break;
   case 2:
     PACKET_COPY(&pack_p->error_code,1)         /* copy to buffer    */
     data[0] = (pack_p->error_index>>24)& 0xff; /* convert the error */
     data[1] = (pack_p->error_index>>16)& 0xff; /* index into network*/
     data[2] = (pack_p->error_index>> 8)& 0xff; /* byte order first  */
     data[3] = (pack_p->error_index    )& 0xff; /* ....              */
     PACKET_COPY(data,4);                       /* copy into packet  */
     if (pack_p->varBind_p) {                /* if any varBinds,     */
        rc = cDPIset(pack_p->varBind_p);     /* handle SET structure */
     } else rc = 0;                          /* no varBinds, all OK  */
     break;
   default:                                  /* code out of sync ??  */
     DPI_DEBUG(1,DPI_DM(                     /* tell if debug active */
         DPI_M_INVALID_DPI_VERSION));
     rc = -1;
     break;
   } /* endswitch */
  } /* endif */

  return(rc);                                /* return rc to caller  */
}

/*********************************************************************/
/* Function to create a DPI TRAP packet from DPI TRAP tree           */
/*********************************************************************/
static int cDPItrap(snmp_dpi_trap_packet *pack_p)
{
 int            rc = -1;                     /* untill proven wrong  */
 unsigned char  data[4];

 if (pack_p) {
   DPI_DEBUG(1,dpi_debug_print_trap(         /* if debug is on, then */
                      "cDPItrap", pack_p));  /* print packet content */
   switch (PACKET_VERSION) {                 /* DPI version 1.x      */
   case 1:
     switch (PACKET_RELEASE) {               /* DPI version 1.1      */
     case 0:
       data[0] = pack_p->generic  & 0xff;    /* isolate generic code */
       data[1] = pack_p->specific & 0xff;    /* isolate specific code*/
       PACKET_COPY(data,2)                   /* copy them to packet  */
       if (pack_p->varBind_p) {              /* handle varBinds      */
          rc = cDPIset(pack_p->varBind_p);   /* if any present       */
       } else rc = 0;                        /* else all done and OK */
       break;
     case 1:
       data[0] = (pack_p->generic  >> 24) & 0xff; /* convert generic */
       data[1] = (pack_p->generic  >> 16) & 0xff; /* trap code into  */
       data[2] = (pack_p->generic  >>  8) & 0xff; /* network byte    */
       data[3] = (pack_p->generic       ) & 0xff; /* order, then     */
       PACKET_COPY(data,4)                        /* copy to packet  */
       data[0] = (pack_p->specific >> 24) & 0xff; /* convert specific*/
       data[1] = (pack_p->specific >> 16) & 0xff; /* trap code into  */
       data[2] = (pack_p->specific >>  8) & 0xff; /* network byte    */
       data[3] = (pack_p->specific      ) & 0xff; /* order, then     */
       PACKET_COPY(data,4)                        /* copy to packet  */
       if (pack_p->enterprise_p) {                /* copy enterprise */
          PACKET_COPY(pack_p->enterprise_p,       /* to packet if it */
                 strlen(pack_p->enterprise_p)+1); /* specified, else */
       } else {                                   /* copy a '\0' to  */
          PACKET_COPY("",1);                      /* indicate none   */
       } /* endif */                              /* present         */
       if (pack_p->varBind_p) {                   /* handle varBinds */
          rc = cDPIset(pack_p->varBind_p);        /* if any present  */
       } else rc = 0;                             /* else all OK     */
       break;
     default:                                     /* code out of sync*/
       DPI_DEBUG(1,DPI_DM(                        /* if debug active */
           DPI_M_INVALID_DPI_RELEASE));
       rc = -1;
       break;
     } /* endswitch (PACKET_RELEASE) */
     break;                                       /* PACKET_VERSION  */
   case 2:
       data[0] = (pack_p->generic  >> 24) & 0xff; /* convert generic */
       data[1] = (pack_p->generic  >> 16) & 0xff; /* trap code into  */
       data[2] = (pack_p->generic  >>  8) & 0xff; /* network byte    */
       data[3] = (pack_p->generic       ) & 0xff; /* order, then     */
       PACKET_COPY(data,4)                        /* copy to packet  */
       data[0] = (pack_p->specific >> 24) & 0xff; /* convert specific*/
       data[1] = (pack_p->specific >> 16) & 0xff; /* trap code into  */
       data[2] = (pack_p->specific >>  8) & 0xff; /* network byte    */
       data[3] = (pack_p->specific      ) & 0xff; /* order, then     */
       PACKET_COPY(data,4)                        /* copy to packet  */
       if (pack_p->enterprise_p) {                /* copy enterprise */
          PACKET_COPY(pack_p->enterprise_p,       /* to packet if it */
                 strlen(pack_p->enterprise_p)+1); /* specified, else */
       } else {                                   /* copy a '\0' to  */
          PACKET_COPY("",1);                      /* indicate none   */
       } /* endif */                              /* present         */
       if (pack_p->varBind_p) {                   /* handle varBinds */
          rc = cDPIset(pack_p->varBind_p);        /* if any present  */
       } else rc = 0;                             /* else all OK     */
     break;
   default:                                       /* code out of sync*/
     DPI_DEBUG(1,DPI_DM(                          /* if debug active */
         DPI_M_INVALID_DPI_VERSION));
     rc = -1;
     break;
   } /* endswitch */
 } /* endif */

 return(rc);                                      /* return to caller*/
}

/*********************************************************************/
/* Make a DPI packet for any DPI packet type                         */
/* - Returns ptr to static buffer containing DPI packet if success   */
/* - Returns a NULL ptr if unsuccessful.                             */
/* Note(s):                                                          */
/* - The callers parse tree (pointed to by the hdr_p argument) will  */
/*   always be freed. So the caller can/should no longer reference   */
/*   that snmp_dpi_hdr structure after a mkDPIpacket call.           */
/*********************************************************************/
unsigned char *mkDPIpacket(            /* Make a DPIrequest packet   */
    snmp_dpi_hdr              *hdr_p)  /* ptr to a parse tree        */
{
   unsigned char *cp;                  /* receives ptr to DPI packet */

   cp = cDPIpacket(hdr_p);             /* static routine             */
   fDPIparse(hdr_p);                   /* free the whole parse tree  */
   return(cp);                         /* return ptr or NULL         */
}

/*********************************************************************/
/* Make a DPI packet for a DPI OPEN packet type.                     */
/* - Returns ptr to static buffer containing DPI packet if success   */
/* - Returns a NULL ptr if unsuccessful.                             */
/* Note(s):                                                          */
/* - Callers ptrs are always still valid upon return.                */
/*********************************************************************/
unsigned char *mkDPIopen(              /* make a DPI open packet     */
  char                 *oid_p,         /* my subagent OID            */
  char                 *description_p, /* my descriptive name        */
  unsigned long         timeout,       /* requested default timeout  */
  unsigned long         max_varBinds,  /* max varBinds I can handle  */
  char                  character_set, /* selected character set     */
  unsigned long         password_len,  /* length of pasword (if any) */
  unsigned char        *password_p)    /* ptr to password (if any)   */
{
   snmp_dpi_hdr         *hdr_p;              /* ptr to hdr structure */
   snmp_dpi_open_packet *o_p;                /* ptr to OPEN structure*/
   unsigned char *cp = (unsigned char *)0;   /* ptr to packet buffer */

   hdr_p = mkDPIhdr(SNMP_DPI_OPEN);          /* make an OPEN hdr     */

   if (hdr_p) {                              /* if success, then     */
      o_p = (snmp_dpi_open_packet *)calloc(  /* allocate an OPEN     */
            1, sizeof(snmp_dpi_open_packet));/* structure            */
      if (o_p) {                             /* if success, then     */
         hdr_p->data_u.open_p = o_p;         /* set ptr in hdr       */
         o_p->oid_p          = oid_p;        /* point to subagent OID*/
         o_p->description_p  = description_p;/* to its description,  */
         o_p->timeout        = timeout;      /* set timeout,         */
         o_p->max_varBinds   = max_varBinds; /* set max varBinds     */
         o_p->character_set  = character_set;/* set selected cset    */
         o_p->password_len   = password_len; /* set password length  */
         o_p->password_p     = password_p;   /* and ptr to password  */
         cp = cDPIpacket(hdr_p);             /* serialize packet     */
         o_p->oid_p         = (char *)0;     /* reset ptrs to NULL,  */
         o_p->description_p = (char *)0;     /* we only borrowed     */
         o_p->password_p  = (unsigned char *)0; /* them from caller  */
      } /* endif */
      fDPIparse(hdr_p);                      /* free the parse tree  */
   } /* endif */                             /* so free it           */

   return(cp);                               /* return ptr or NULL   */
}

/*********************************************************************/
/* Make a DPI packet for a DPI CLOSE packet type.                    */
/* - Returns ptr to static buffer containing DPI packet if success   */
/* - Returns a NULL ptr if unsuccessful.                             */
/*********************************************************************/
unsigned char *mkDPIclose(             /* make a DPI close packet    */
  char                  reason_code)   /* the reason for closing     */
{
   snmp_dpi_hdr          *hdr_p;             /* ptr to hdr structure */
   snmp_dpi_close_packet *c_p;               /* ptr to CLOSE struct  */
   unsigned char *cp = (unsigned char *)0;   /* ptr to packet buffer */

   hdr_p = mkDPIhdr(SNMP_DPI_CLOSE);         /* make a CLOSE hdr     */

   if (hdr_p) {                              /* if success, then     */
      c_p = (snmp_dpi_close_packet *)calloc( /* allocate an CLOSE    */
            1,sizeof(snmp_dpi_close_packet));/* structure            */
      if (c_p) {                             /* if success, then     */
         hdr_p->data_u.close_p = c_p;        /* set ptr in hdr       */
         c_p->reason_code  = reason_code;    /* set reason code      */
         cp = cDPIpacket(hdr_p);             /* serialize packet     */
      } /* endif */                          /* which returns a ptr  */
      fDPIparse(hdr_p);                      /* free the parse tree  */
   } /* endif */                             /* so free it           */

   return(cp);                               /* return ptr or NULL   */
}

/*********************************************************************/
/* Make a DPI packet for a DPI ARE_YOU_THERE packet type.            */
/* - Returns ptr to static buffer containing DPI packet if success   */
/* - Returns a NULL ptr if unsuccessful.                             */
/*********************************************************************/
unsigned char *mkDPIayt(void)                /* make a DPI AYT packet*/
{
   snmp_dpi_hdr          *hdr_p;             /* ptr to hdr structure */
   unsigned char *cp = (unsigned char *)0;   /* ptr to packet buffer */

   hdr_p = mkDPIhdr(SNMP_DPI_ARE_YOU_THERE); /* make a AYT hdr       */

   if (hdr_p) {                              /* if success, then     */
      cp = cDPIpacket(hdr_p);                /* serialize packet     */
      fDPIparse(hdr_p);                      /* free the parse tree  */
   } /* endif */                             /* so free it           */

   return(cp);                               /* return ptr or NULL   */
}

/*********************************************************************/
/* Make a DPI packet for a DPI REGISTER packet type                  */
/* - Returns ptr to static buffer containing DPI packet if success   */
/* - Returns a NULL ptr if unsuccessful.                             */
/* Note(s):                                                          */
/* - Callers ptrs are always still valid upon return.                */
/*********************************************************************/
unsigned char *mkDPIreg(               /* make a DPI register packet */
  unsigned short        timeout,       /* requested timeout value    */
  long int              priority,      /* requested priority         */
  char                 *group_p,       /* ptr to grou ID (subtree)   */
  char                  bulk_select,   /* Bulk selection yes(1)/no(0)*/
  char                  view_select)   /* View selection yes(1)/no(0)*/
                                       /* (view_select for BNR DPI   */
                                       /*  1.2 compatibility only)   */
{
   snmp_dpi_hdr  *hdr_p;                     /* ptr to hdr structure */
   unsigned char *cp = (unsigned char *)0;   /* ptr to packet buffer */

   hdr_p = mkDPIhdr(SNMP_DPI_REGISTER);      /* make a REGISTER hdr  */

   if (hdr_p) {                              /* if success, then     */
      hdr_p->data_u.reg_p =                  /* make REG structure   */
             mkDPIreg_packet(                /* (first in chain) and */
               snmp_dpi_reg_packet_NULL_p,   /* based on callers     */
               timeout, priority,group_p,    /* arguments abd if     */
               view_select, bulk_select);
      if (hdr_p->data_u.reg_p) {             /* that also succeeds   */
          cp = cDPIpacket(hdr_p);            /* serialize packet     */
      } /* endif */                          /* which returns ptr    */
      fDPIparse(hdr_p);                      /* no longer need hdr   */
   } /* endif */                             /* so free it           */

   return(cp);                               /* return ptr or NULL   */
}

/*********************************************************************/
/* Make a DPI packet for a DPI UNREGISTER packet type                */
/* - Returns ptr to static buffer containing DPI packet if success   */
/* - Returns a NULL ptr if unsuccessful.                             */
/* Note(s):                                                          */
/* - Callers ptrs are always still valid upon return.                */
/*********************************************************************/
unsigned char *mkDPIureg(              /* make DPI unregister packet */
    char                  reason_code, /* reason for unregister      */
    char                 *group_p)     /* ptr to group ID (subtree)  */
{
   snmp_dpi_hdr  *hdr_p;                     /* ptr to hdr structure */
   unsigned char *cp = (unsigned char *)0;   /* ptr to packet buffer */

   hdr_p = mkDPIhdr(SNMP_DPI_UNREGISTER);    /* make UNREGISTER hdr  */

   if (hdr_p) {                              /* if success, then     */
      hdr_p->data_u.ureg_p =                 /* make UREG structure  */
             mkDPIureg_packet(               /* (first in chain) and */
               snmp_dpi_ureg_packet_NULL_p,  /* based on callers     */
               reason_code, group_p);        /* arguments abd if     */
      if (hdr_p->data_u.ureg_p) {            /* that also succeeds   */
         cp = cDPIpacket(hdr_p);             /* then serialize packet*/
      } /* endif */                          /* which returns ptr    */
      fDPIparse(hdr_p);                      /* no longer need hdr   */
   } /* endif */                             /* so free it           */

   return(cp);                               /* return ptr or NULL   */
}

/*********************************************************************/
/* Make a DPI packet for a DPI RESPONSE packet type                  */
/* - Returns ptr to static buffer containing DPI packet if success   */
/* - Returns a NULL ptr if unsuccessful.                             */
/* Note(s):                                                          */
/* - The callers parse tree (pointed to by the pack_p argument) will */
/*   always be freed. So the caller can/should no longer reference   */
/*   that snmp_dpi_set_packet structure after a mkDPIresp call.      */
/*********************************************************************/
unsigned char *mkDPIresp(              /* make a DPI RESPONSE packet */
  snmp_dpi_hdr         *ihdr_p,        /* ptr to packet to respond to*/
  long int              error_code,    /* return or error code       */
  long int              error_index,   /* index to varBind in error  */
  snmp_dpi_set_packet  *pack_p)        /* ptr to chain of varBinds   */
{
   snmp_dpi_hdr         *hdr_p;              /* ptr to hdr structure */
   snmp_dpi_resp_packet *r_p;                /* ptr to REG struct    */
   unsigned char *cp = (unsigned char *)0;   /* ptr to packet buffer */

   hdr_p = mkDPIhdr(SNMP_DPI_RESPONSE);      /* make a RESPONSE hdr  */

   if (hdr_p) {                              /* if success, then     */
      if (ihdr_p) {
         hdr_p->proto_version =              /* copy these fields    */
                ihdr_p->proto_version;       /* from parse tree      */
         hdr_p->proto_release =              /* that we are          */
                ihdr_p->proto_release;       /* responding to        */
         hdr_p->packet_id = 
                ihdr_p->packet_id;           /* ...                  */
      } /* endif */
      r_p = (snmp_dpi_resp_packet *)calloc(  /* allocate a response  */
            1, sizeof(snmp_dpi_resp_packet));/* structure            */
      if (r_p) {                             /* if success, then     */
         hdr_p->data_u.resp_p = r_p;         /* set ptr in hdr       */
         r_p->varBind_p       = pack_p;      /* point to varBinds    */
         r_p->error_code      = error_code;  /* set return code      */
         r_p->error_index     = error_index; /* set index            */
         cp = cDPIpacket(hdr_p);             /* serialize packet     */
      } /* endif */                          /* which returns a ptr  */
      fDPIparse(hdr_p);                      /* free the parse tree  */
   } else {                                  /* no hdr allocated     */
      fDPIset_packet(pack_p);                /* we claim to always   */
   } /* endif */                             /* free this, so do it  */

   return(cp);                               /* return ptr or NULL   */
}

/*********************************************************************/
/* Make a DPI packet for a DPI TRAP packet type.                     */
/* - Returns ptr to static buffer containing DPI packet if success   */
/* - Returns a NULL ptr if unsuccessful.                             */
/* Note(s):                                                          */
/* - The callers parse tree (pointed to by the pack_p argument) will */
/*   always be freed. So the caller can/should no longer reference   */
/*   that snmp_dpi_set_packet structure after a mkDPItrap(e) call.   */
/*   The enterprise OID is left untouched, so callers enterprise_p   */
/*   ptr is always still valid.                                      */
/*********************************************************************/
unsigned char *mkDPItrape(             /* Make a DPI TRAP packet     */
    long int             generic,      /* generic trap type          */
    long int             specific,     /* specific trap type         */
    snmp_dpi_set_packet *pack_p,       /* ptr to chain of varBinds   */
    char                *enterprise_p) /* ptr to enterprise OID      */
{
   snmp_dpi_hdr         *hdr_p;              /* ptr to hdr structure */
   snmp_dpi_trap_packet *t_p;                /* ptr to TRAP structure*/
   unsigned char *cp = (unsigned char *)0;   /* ptr to packet buffer */

   hdr_p = mkDPIhdr(SNMP_DPI_TRAP);          /* make a TRAP hdr      */

   if (hdr_p) {                              /* if success, then     */
      t_p = (snmp_dpi_trap_packet *)calloc(  /* alocate a trap       */
            1, sizeof(snmp_dpi_trap_packet));/* structure            */
      if (t_p) {                             /* if success, then     */
         hdr_p->data_u.trap_p = t_p;         /* set ptr in hdr       */
         t_p->varBind_p       = pack_p;      /* point to varBinds    */
         t_p->generic         = generic;     /* set generic type     */
         t_p->specific        = specific;    /* set specific type    */
         t_p->enterprise_p    = enterprise_p;/* borrow caller ptr    */
         cp = cDPIpacket(hdr_p);             /* serialize packet     */
         t_p->enterprise_p    = (char *)0;   /* set NULL; so callers */
      } /* endif */                          /* ptr stays valid      */
      fDPIparse(hdr_p);                      /* free the parse tree  */
   } else {                                  /* no hdr allocated     */
      fDPIset_packet(pack_p);                /* we claim to always   */
   } /* endif */                             /* free this, so do it  */

   return(cp);                               /* return ptr or NULL   */
}

/*********************************************************************/
/* Function to allocate a snmp_dpi_hdr structure and fill out the    */
/* specified packet type and also the default fields, like the       */
/* major protocol (SNMP_DPI) and the current/default DPI Version     */
/* Release. Also a new packet_id is assigned and set in the hdr.     */
/* - Returns ptr to dynamically allocated snmp_dpi_hdr structure     */
/*   if successful.                                                  */
/* - Returns a NULL ptr if unsuccessful.                             */
/*********************************************************************/
snmp_dpi_hdr         *mkDPIhdr_version(/* Make DPI hdr for parse tree*/
  int                   type,          /* packet type; SNMP_DPI_xxxx */
  char                  version,       /* for this specific version  */
  char                  release)       /* for this specific release  */
{
  snmp_dpi_hdr   *hdr_p;                     /* receives ptr to hdr  */

  switch (version) {                         /* ensure caller passed */
  case 1:                                    /* a supported set of   */
   switch (release) {                        /* version and release  */
   case 0:                                  
   case 1:
     break;                                  /* from switch (release)*/
   default: 
     DPI_DEBUG(1,DPI_DM(                     /* tell if debug active */
         DPI_M_INVALID_DPI_RELEASE));
     return(snmp_dpi_hdr_NULL_p);            /* return NULL ptr      */
     /* break; */                            /* NEVER REACHED        */
   } /* endswitch */
   break;                                    /* from swicth (version)*/
  case 2:
   switch (release) {
   case 0:                                   
     break;                                  /* from switch (release)*/
   default: 
     DPI_DEBUG(1,DPI_DM(                     /* tell if debug active */
         DPI_M_INVALID_DPI_RELEASE));
     return(snmp_dpi_hdr_NULL_p);            /* return NULL ptr      */
     /* break; */                            /* NEVER REACHED        */
   } /* endswitch */
   break;                                    /* from swicth (version)*/
  default:
   DPI_DEBUG(1,DPI_DM(                       /* tell if debug active */
       DPI_M_INVALID_DPI_VERSION));
   return(snmp_dpi_hdr_NULL_p);              /* return NULL ptr      */
   /* break; */                              /* NEVER REACHED        */
  } /* endswitch */

  hdr_p = (snmp_dpi_hdr *)                   /* allocate storage     */
           calloc(1, sizeof(snmp_dpi_hdr));  /* init at zero         */

  if (hdr_p) {                               /* if success, then     */
    hdr_p->proto_major   = SNMP_DPI_PROTOCOL;/* set protocol         */
    hdr_p->proto_version = version;          /*     version          */
    hdr_p->proto_release = release;          /*     release          */
    hdr_p->packet_type   = type;             /*     packet type      */
    if (version >= 2) {
       hdr_p->packet_id  = packet_id++;      /*     packet ID        */
    } /* endif */
  } else {                                   /* if failure           */
    DPI_DEBUG(1,DPI_DM(                     /* if debug_lvl >= 1     */
              DPI_M_OUT_OF_MEMORY));        /* give this message     */
  } /* endif */

  return(hdr_p);                             /* return ptr or NULL   */
}

/*********************************************************************/
/* function to Create or Extend a snmp_dpi_reg_packet structure.     */
/* - If success we return a ptr to an snmp_dpi_reg_packet structure. */
/*   - A new packet structure is dynamically allocated.              */
/*   - If caller passed an existing chain of packet structures via   */
/*     the packet_p argument, then we have added the new structure   */
/*     at the end of the chain and we return the ptr packet_p.       */
/*   - If caller passed a NULL ptr for packet_p, then we return a    */
/*     ptr to the one (and thus first) packet that we just created.  */
/*   - the callers other arguments are copied to newly allocated     */
/*     memory, so upon return callers ptrs to those are still valid. */
/* - If failure, we return a NULL ptr.                               */
/*********************************************************************/
snmp_dpi_reg_packet  *mkDPIreg_packet( /* Make DPI reg_packet tree   */
  snmp_dpi_reg_packet  *packet_p,      /* ptr to REGISTER structure  */
  unsigned short        timeout,       /* in seconds (16-bit)        */
  long int              priority,      /* requested priority         */
  char                 *group_p,       /* ptr to group ID (subtree)  */
  char                  bulk_select,   /* Bulk selection yes(1)/no(0)*/
  char                  view_select)   /* View selection yes(1)/no(0)*/
                                       /* (view_select for BNR DPI   */
                                       /*  1.2 compatibility only)   */
{
   snmp_dpi_reg_packet  *new_p, *next_p;  /* ptr to REG structure    */
   int                   l;               /* for length calculations */

   new_p = (snmp_dpi_reg_packet *)calloc( /* allocate a new REG      */
           1,sizeof(snmp_dpi_reg_packet));/* structure for new packet*/

   if (new_p) {                           /* if success, then copy   */
      new_p->group_p = copy(group_p,&l);  /* group OID (subtree)     */
      if (new_p->group_p) {               /* if copy was successful  */
         new_p->timeout   = timeout;      /* set timeout             */
         new_p->priority  = priority;     /* set priority            */
#ifdef SNMP_DPI_VIEW_SELECTION 
         new_p->view = view_select;       /* set view selection      */
#endif
#ifdef SNMP_DPI_BULK_SELECTION
         new_p->bulk = bulk_select;       /* set bulk selection      */
#endif
         if (packet_p) {                  /* caller passed existing  */
             next_p = packet_p;           /* chain, then find last   */
             while (next_p->next_p) {     /* entry in the chain      */
                next_p = next_p->next_p;  /* so we can add the new   */
             } /* endwhile */             /* structure at the end    */
             next_p->next_p = new_p;      /* of the chain.           */
             return(packet_p);            /* return ptr to first one */
         } /* endif */                    /* in the chain            */
         return(new_p);                   /* else new one IS the 1st */
      } /* endif */                       /* one, so return its ptr  */
      fDPIreg_packet(new_p);              /* we failed on concat, so */
      return(snmp_dpi_reg_packet_NULL_p); /* free allocated memory   */
   } /* endif */                          /* and return NULL ptr     */

   return(new_p);                         /* calloc failed, NULL ptr */
}

/*********************************************************************/
/* function to Create or Extend a snmp_dpi_ureg_packet structure.    */
/* - If success we return a ptr to an snmp_dpi_ureg_packet structure.*/
/*   - A new packet structure is dynamically allocated.              */
/*   - If caller passed an existing chain of packet structures via   */
/*     the packet_p argument, then we have added the new structure   */
/*     at the end of the chain and we return the ptr packet_p.       */
/*   - If caller passed a NULL ptr for packet_p, then we return a    */
/*     ptr to the one (and thus first) packet that we just created.  */
/*   - the callers other arguments are copied to newly allocated     */
/*     memory, so upon return callers ptrs to those are still valid. */
/* - If failure, we return a NULL ptr.                               */
/*********************************************************************/
snmp_dpi_ureg_packet *mkDPIureg_packet(/* Make DPI ureg packet tree  */
  snmp_dpi_ureg_packet *packet_p,      /* ptr to UNREGISTER structure*/
  char                  reason_code,   /* subagent specific code     */
  char                 *group_p)       /* ptr to group ID (subtree)  */

{
   snmp_dpi_ureg_packet *new_p, *next_p;  /* ptr to UREG structure   */
   int                   l;               /* for length calculations */

   new_p = (snmp_dpi_ureg_packet *)calloc(/* allocate a new UREG     */
          1,sizeof(snmp_dpi_ureg_packet));/* structure for new packet*/

   if (new_p) {                           /* if success, then copy   */
      new_p->group_p = copy(group_p,&l);  /* group OID (subtree)     */
      if (new_p->group_p) {               /* if copy was successful  */
         new_p->reason_code = reason_code;/* set reason code         */
         if (packet_p) {                  /* caller passed existing  */
             next_p = packet_p;           /* chain, then find last   */
             while (next_p->next_p) {     /* entry in the chain      */
                next_p = next_p->next_p;  /* so we can add the new   */
             } /* endwhile */             /* structure at the end    */
             next_p->next_p = new_p;      /* of the chain.           */
             return(packet_p);            /* return ptr to first one */
         } /* endif */                    /* in the chain            */
         return(new_p);                   /* else new one IS the 1st */
      } /* endif */                       /* one, so return its ptr  */
      fDPIureg_packet(new_p);             /* we failed on concat, so */
      return(snmp_dpi_ureg_packet_NULL_p);/* free allocated memory   */
   } /* endif */                          /* and return NULL ptr     */

   return(new_p);                         /* calloc failed, NULL ptr */
}

/*********************************************************************/
/* function to Create or Extend a snmp_dpi_get_packet structure.     */
/* - If success we return a ptr to an snmp_dpi_get_packet structure. */
/*   - A new packet structure is dynamically allocated.              */
/*   - If caller passed an existing chain of packet structures via   */
/*     the packet_p argument, then we have added the new structure   */
/*     at the end of the chain and we return the ptr packet_p.       */
/*   - If caller passed a NULL ptr for packet_p, then we return a    */
/*     ptr to the one (and thus first) packet that we just created.  */
/*   - the callers other arguments are copied to newly allocated     */
/*     memory, so upon return callers ptrs to those are still valid. */
/* - If failure, we return a NULL ptr.                               */
/*********************************************************************/
snmp_dpi_get_packet *mkDPIget_packet(  /* Make/extend DPIget packet  */
     snmp_dpi_get_packet *packet_p,    /* ptr to existing chain ifany*/
     char                *group_p,     /* ptr to group ID (subtree)  */
     char                *instance_p)  /* ptr to instance ID         */
{
   snmp_dpi_get_packet  *new_p, *next_p;  /* ptr to GET structure    */
   int                   l;               /* for length calculations */

   new_p = (snmp_dpi_get_packet *)calloc( /* allocate a new GET      */
           1,sizeof(snmp_dpi_get_packet));/* structure for new packet*/

   if (new_p) {                           /* if success, then copy   */
      new_p->group_p = copy(group_p,&l);  /* callers groupID/subtree */
      if (instance_p && *instance_p) {    /* if more than a NULL,    */
         new_p->instance_p =              /* then also copy callers  */
                     copy(instance_p,&l); /* instance OID            */
      } /* endif */
      if (new_p->group_p &&               /* if both copies were     */
          new_p->instance_p)              /* successful then we      */
      {                                   /* can concatenate the     */
         new_p->object_p = concat(        /* group ID and instance ID*/
                    new_p->group_p,       /* into an object ID as    */
                    new_p->instance_p);   /* expected by DPI 1.x     */
      } else if (new_p->group_p) {        /* if no instance present  */
         new_p->object_p = copy(          /* then use the group ID   */
                           group_p,&l);   /* as the object ID also   */
      } /* endif */

      if (new_p->object_p) {              /* if still successful and */
         l = strlen(new_p->object_p)-1;   /* find possible trailing  */
         if (new_p->object_p[l] =='.') {  /* dot. If object ID has   */
            new_p->object_p[l] = '\0';    /* trailing dot, then      */
         } /* endif */                    /* remove it               */
         if (packet_p) {                  /* caller passed existing  */
             next_p = packet_p;           /* chain, then find last   */
             while (next_p->next_p) {     /* entry in the chain      */
                next_p = next_p->next_p;  /* so we can add the new   */
             } /* endwhile */             /* structure at the end    */
             next_p->next_p = new_p;      /* of the chain.           */
             return(packet_p);            /* return ptr to first one */
         } /* endif */                    /* in the chain            */
         return(new_p);                   /* else new one IS the 1st */
      } /* endif */                       /* one, so return its ptr  */
      fDPIget_packet(new_p);              /* we failed on concat, so */
      return(snmp_dpi_get_packet_NULL_p); /* free allocated memory   */
   } /* endif */                          /* and return NULL ptr     */

   return(new_p);                         /* calloc failed, NULL ptr */
}

/*********************************************************************/
/* function to Create or Extend a snmp_dpi_next_packet structure.    */
/* - If success we return a ptr to an snmp_dpi_get_packet structure. */
/*   - A new packet structure is dynamically allocated.              */
/*   - If caller passed an existing chain of packet structures via   */
/*     the packet_p argument, then we have added the new structure   */
/*     at the end of the chain and we return the ptr packet_p.       */
/*   - If caller passed a NULL ptr for packet_p, then we return a    */
/*     ptr to the one (and thus first) packet that we just created.  */
/*   - the callers other arguments are copied to newly allocated     */
/*     memory, so upon return callers ptrs to those are still valid. */
/* - If failure, we return a NULL ptr.                               */
/*********************************************************************/
snmp_dpi_next_packet *mkDPInext_packet(/* Make DPI next packet tree  */
  snmp_dpi_next_packet *packet_p,      /* ptr to NEXT structure      */
  char                 *group_p,       /* ptr to group ID (subtree)  */
  char                 *instance_p)    /* ptr to instance OID string */
{
   snmp_dpi_next_packet *new_p, *next_p;  /* ptr to NEXT structure   */
   int                   l;               /* for length calculations */

   new_p = (snmp_dpi_next_packet *)calloc(   /* allocate a new NEXT  */
           1,sizeof(snmp_dpi_next_packet));  /* structure new packet */

   if (new_p) {                           /* if success              */
      new_p->group_p = copy(group_p,&l);  /* callers groupID/subtree */
      if (instance_p && *instance_p) {    /* if more than a NULL,    */
         new_p->instance_p =              /* then also copy callers  */
                     copy(instance_p,&l); /* instance OID            */
      } /* endif */
      if (new_p->group_p &&               /* if both copies were     */
          new_p->instance_p)              /* successful then we      */
      {                                   /* can concatenate the     */
         new_p->object_p = concat(        /* group ID and instance ID*/
                    new_p->group_p,       /* into an object ID as    */
                    new_p->instance_p);   /* expected by DPI 1.x     */
      } else if (new_p->group_p) {        /* if no instance present  */
         new_p->object_p = copy(          /* then use the group ID   */
                           group_p,&l);   /* as the object ID also   */
      } /* endif */

      if (new_p->object_p) {              /* if still successful and */
         l = strlen(new_p->object_p)-1;   /* find possible trailing  */
         if (new_p->object_p[l] =='.') {  /* dot. If object ID has   */
            new_p->object_p[l] = '\0';    /* trailing dot, then      */
         } /* endif */                    /* remove it               */
         if (packet_p) {                  /* caller passed existing  */
             next_p = packet_p;           /* chain, then find last   */
             while (next_p->next_p) {     /* entry in the chain      */
                next_p = next_p->next_p;  /* so we can add the new   */
             } /* endwhile */             /* structure at the end    */
             next_p->next_p = new_p;      /* of the chain.           */
             return(packet_p);            /* return ptr to first one */
         } /* endif */                    /* in the chain            */
         return(new_p);                   /* else new one IS the 1st */
      } /* endif */                       /* one, so return its ptr  */
      fDPInext_packet(new_p);             /* we failed on concat, so */
      return(snmp_dpi_next_packet_NULL_p);/* free allocated memory   */
   } /* endif */                          /* and return NULL ptr     */

   return(new_p);                         /* calloc failed, NULL ptr */
}

/*********************************************************************/
/* function to Create or Extend a snmp_dpi_set_packet structure      */
/*********************************************************************/
snmp_dpi_set_packet  *mkDPIset_packet( /* Make DPI set packet tree   */
  snmp_dpi_set_packet  *packet_p,      /* ptr to SET structure       */
  char                 *group_p,       /* ptr to group ID (subtree)  */
  char                 *instance_p,    /* ptr to instance OID string */
  int                   value_type,    /* value type (SNMP_TYPE_xxx) */
  int                   value_len,     /* length of value            */
  void                 *value_p)       /* ptr to value               */
{
   snmp_dpi_set_packet  *new_p, *next_p;
   int                   l;

   switch (value_type) {               /* validate the value type    */
   case SNMP_TYPE_Integer32:           /* 32-bit INTEGER             */
   case SNMP_TYPE_OCTET_STRING:        /* OCTET STRING               */
   case SNMP_TYPE_OBJECT_IDENTIFIER:   /* OBJECT IDENTIFIER          */
   case SNMP_TYPE_NULL:                /* NULL value                 */
   case SNMP_TYPE_DPI_1_INTERNET:      /* Old DPI 1.x type           */
   case SNMP_TYPE_Counter32:           /* 32-bit Counter (unsigned)  */
   case SNMP_TYPE_Gauge32:             /* 32-bit Gauge   (unsigned)  */
   case SNMP_TYPE_TimeTicks:           /* 32-bit TimeTicks (unsigned)*/
   case SNMP_TYPE_DisplayString:       /* Textual Convention         */
   case SNMP_TYPE_IpAddress:           /* IMPLICIT OCTET STRING (4)  */
   case SNMP_TYPE_BIT_STRING:          /* BIT STRING                 */
   case SNMP_TYPE_NsapAddress:         /* IMPLICIT OCTET STRING      */
   case SNMP_TYPE_UInteger32:          /* 32-bit INTEGER (unsigned)  */
   case SNMP_TYPE_Counter64:           /* 64-bit Counter (unsigned)  */
   case SNMP_TYPE_Opaque:              /* IMPLICIT OCTET STRING      */
   case SNMP_TYPE_noSuchObject:        /* IMPLICIT NULL              */
   case SNMP_TYPE_noSuchInstance:      /* IMPLICIT NULL              */
   case SNMP_TYPE_endOfMibView:        /* IMPLICIT NULL              */
     break;
   default:
     DPI_DEBUG(1,DPI_DM(                  /* If debug active,        */
         DPI_M_INVALID_SNMP_TYPE,         /* tell this message       */
         "mkDPIset_packet", value_type));
     return(snmp_dpi_set_packet_NULL_p);  /* return a NULLptr        */
     /* break; */                         /* NEVER REACHED           */
   } /* endswitch */

   new_p = (snmp_dpi_set_packet *)calloc( /* allocate a new SET      */
           1,sizeof(snmp_dpi_set_packet));/* structure for new packet*/

   if (new_p) {                           /* if success              */
      new_p->group_p = copy(group_p,&l);  /* callers groupID/subtree */
      if (instance_p && *instance_p) {    /* if more than a NULL,    */
         new_p->instance_p =              /* then also copy callers  */
                     copy(instance_p,&l); /* instance OID            */
      } /* endif */
      if (new_p->group_p &&               /* if both copies were     */
          new_p->instance_p)              /* successful then we      */
      {                                   /* can concatenate the     */
         new_p->object_p = concat(        /* group ID and instance ID*/
                    new_p->group_p,       /* into an object ID as    */
                    new_p->instance_p);   /* expected by DPI 1.x     */
      } else if (new_p->group_p) {        /* if no instance present  */
         new_p->object_p = copy(          /* then use the group ID   */
                           group_p,&l);   /* as the object ID also   */
      } /* endif */

      if (new_p->object_p) {              /* if still successful     */
         l = strlen(new_p->object_p)-1;   /* find possible trailing  */
         if (new_p->object_p[l] =='.') {  /* dot. If object ID has   */
            new_p->object_p[l] = '\0';    /* trailing dot, then      */
         } /* endif */                    /* remove it               */
         new_p->value_type = value_type;  /* set type in packet      */
         new_p->value_len  = value_len;   /* set length also         */
         if (value_len) {                 /* if value has a length   */
            new_p->value_p = (char *)     /* allocate memory for     */
                       malloc(value_len); /* the value itself        */
            if (new_p->value_p) {         /* if success,             */
                memcpy(new_p->value_p,    /* then copy the           */
                       value_p,value_len);/* value from caller       */
            } else {
               DPI_DEBUG(1,DPI_DM(        /* if debug active then    */
                    DPI_M_OUT_OF_MEMORY));/* give this message       */
               fDPIset_packet(new_p);     /* free allocated memory   */
               return(snmp_dpi_set_packet_NULL_p);/* return NULL ptr */
            } /* endif */
         } /* endif */
         if (packet_p) {                  /* caller passed existing  */
            next_p = packet_p;            /* chain, then find last   */
            while (next_p->next_p) {      /* entry in the chain      */
                 next_p = next_p->next_p; /* so we can add the new   */
            } /* endwhile */              /* structure at the end    */
            next_p->next_p = new_p;       /* of the chain.           */
            return(packet_p);             /* return ptr to first one */
         } /* endif */                    /* in the chain            */
         return(new_p);                   /* else new one IS the 1st */
      } /* endif */                       /* we failed on concat or  */
   } /* endif */                          /* allocating memory       */

   DPI_DEBUG(1,DPI_DM(                    /* if debug active then    */
              DPI_M_OUT_OF_MEMORY));      /* give this message       */
   if (new_p) fDPIset_packet(new_p);      /* free memory (if any)    */
   return(snmp_dpi_set_packet_NULL_p);    /* and return NULL ptr     */
}

/*********************************************************************/
/*  The following code effectively performs the following statement: */
/*      sum = ntohl((unsigned long)*bytes);                          */
/*  We want to be system independent, so we use this piece of code.  */
/*********************************************************************/
static long int_val(                   /* convert 4 bytes from       */
  unsigned char     *bytes)            /* network to host byte order */
{
   register int           i;
   register unsigned long sum;

   sum = 0;
   for (i=0; i<4; i++) {
       sum <<= 8;
       sum |= *(bytes++);
   } /* endfor */
   return(sum);
}

/*********************************************************************/
/*  The following code effectively performs the following statement  */
/*      sum = ntohs((unsigned short)*bytes);                         */
/*  We want to be system independent, so we use this piece of code.  */
/*********************************************************************/
static short s_int_val(                /* convert 2 bytes from       */
  unsigned char     *bytes)            /* network to host byte order */
{
   register int            i;
   register unsigned short sum;

   sum = 0;
   for (i=0; i<2; i++) {
       sum <<= 8;
       sum |= *(bytes++);
   } /* endfor */
   return(sum);
}

/*********************************************************************/
/*  The following code allocates memory for a null terminated string */
/*  copies the string and returns a ptr to the newly allocated area. */
/*  The callers length argument is set to the length of the string,  */
/*  including the terminating null character.                        */
/*  If a failure occurs, then a NULL ptr is returned, the callers    */
/*  length argument will still be set to the length of the string.   */
/*********************************************************************/
static char *copy(char *cp, int *len)
{
   char *new_p;

   if (cp) {
      *len   = strlen(cp) + 1;               /* get length of string */
      new_p  = (char *) malloc(*len);        /* get memory for it    */
      if (new_p) {                           /* success, so copy it  */
         strcpy(new_p, cp);                  /* including '\0' char  */
      } else {                               /* else error           */
         DPI_DEBUG(1,DPI_DM(                 /* if debug_lvl >= 1    */
              DPI_M_OUT_OF_MEMORY));         /* give this message    */
      } /* endif */
   } else {
      *len  = 0;
      new_p = (char *)0;
   } /* endif */
   return(new_p);                            /* return ptr to caller */
}

/*********************************************************************/
/*  The following code allocates memory for a null terminated string */
/*  that will be built by concatenating the 2 null terminated strings*/
/*  beings passed (p1 and p2). Returns a ptr to newly allocated area.*/
/*  If a failure occurs, then a NULL ptr is returned.                */
/*********************************************************************/
static char *concat(char *p1, char *p2)
{
   char *new_p;                              /* ptr to new memory    */
   int   len;

   if (p1) {                                 /* if ptr1 is valid     */
      len   = strlen(p1) + 1;                /* get length of string */
   } else {
      return((char *)0);                     /* return NULL ptr      */
   } /* endif */

   if (p2) len += strlen(p2) + 1;            /* add length of string */

   new_p  = (char *) malloc(len);            /* get memory for it    */

   if (new_p) {                              /* success, so now      */
      strcpy(new_p, p1);                     /* copy first string    */
      if (p2) strcat(new_p, p2);             /* append second string */
   } else {                                  /* else error           */
      DPI_DEBUG(1,DPI_DM(                    /* if debug_lvl >= 1    */
              DPI_M_OUT_OF_MEMORY));         /* give this message    */
   } /* endif */
   return(new_p);                            /* return ptr to caller */
}

/*
\end{verbatim}
*/

